
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include <math.h>
#define	ANSI
/*
	Adaptive quadrature routine.
	Called with
		lft:  left endpoint
		rt:   right endpoint
		f:    pointer to function to integrate
		err:  pointer to allowable error
		p1:   function value at left endpoint
		p2:   function value at center of interval
		p3:   function value at right endpoint
	Returns
		adaptive2   evaluated integral
		err         pointer to upper limit of actual error
*/
#if	defined(ANSI)
static double adaptive2(double lft, double rt, double (*f)(double x),
		double *err, double p1, double p3, double p5)
#else
static double adaptive2(lft, rt, f, err, p1, p3, p5)
double lft, rt, (*f)(), *err, p1, p3, p5;
#endif
{
	double h;	/* width of this interval */
	double p2, p4;	/* function values at pts 2 and 4 */
	double error;	/* actual error on this interval */
	double err1, err2;	/* errors on the sub-intervals */
	double int3, int5;	/* integral with Simpson's 3- and 5-point appox. */
	h = rt - lft;	/* get the width of the interval */

	/* get the 2 approximations to the integral */

	p2 = (*f)(lft + h/4.0);
	p4 = (*f)(rt - h/4.0);

	int3 = h*(p1 + 4.0*p3 + p5)/6.0;
	int5 = h*(p1 + 4.0*p2 + 2.0*p3 + 4.0*p4 + p5)/12.0;

	error = fabs(int3 - int5)/15.0;	/* find the actual error */
	if (error < *err) {	/* if within tolerance, */
		*err = error;	/* copy the actual error to caller, */
		return int5;	/* return the more accurate approximation */
	} else {	/* otherwise, */
		err1 = err2 = *err/2.0;	/* integrate the two half-regions, */
		int5 = adaptive2(lft, lft+h/2.0, f, &err1, p1, p2, p3);
		int5 += adaptive2(lft+h/2.0, rt, f, &err2, p3, p4, p5);
		*err = err1 + err2;	/* sum the real errors, */
		return int5;	/* and return the integral. */
	}
}
/*
	Adaptive quadrature interface.
	Called with
		lft:  left endpoint
		rt:   right endpoint
		f:    pointer to function to integrate
		err:  pointer to allowable error
	Returns
		adaptive    evaluated integral
		err         pointer to upper limit of actual error
*/
#if	defined(ANSI)
double adaptive(double lft, double rt, double (*f)(double x), double *err)
#else
double adaptive(lft, rt, f, err)
double lft, rt, (*f)(), *err;
#endif
{
	double p1, p2, p3;	/* initial function evaluations */
	double h;	/* width of this interval */

	h = rt - lft;	/* get the width of the interval */

	p1 = (*f)(lft);	/* pre-compute the initial function values */
	p2 = (*f)(lft + h/2.0);
	p3 = (*f)(rt);

	/* call adaptive2 to do the real work */
	return adaptive2(lft, rt, f, err, p1, p2, p3);
}
