
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <time.h>	/* defines tm structure, declares time and clock */
#if	defined(BSD)	/* BSD-UNIX */
#include <sys/types.h>	/* defines time_t */
#include <sys/file.h>	/* lseek() flags */
#endif
#if	defined(SYSV)	/* UNIX System V */
#include <sys/types.h>	/* defines time_t */
#include <unistd.h>	/* lseek() flags */
#endif
#if	defined(BSD)	/* BSD-UNIX */
#include <strings.h>	/* string header */
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros, errno declared */
#include <stdlib.h>	/* errno declared */
#else	/* not ANSI-C */
#define remove unlink	/* UNIX unlink is equivalent to ANSI remove */
extern time_t time();
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
/*----------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*----------------------------------------------------------------------*/
#if	!defined(SEEK_SET)	/* fseek() flags */
#define	SEEK_SET	0
#endif
#if	!defined(SEEK_CUR)
#define	SEEK_CUR	1
#endif
#if	!defined(SEEK_END)
#define	SEEK_END	2
#endif
/*----------------------------------------------------------------------*/
#define	LARGE	10000000L	/* large number */
#define	NFILE	65	/* maximum number of open image files */
#define	PATHSIZE	128	/* path buffer size */	
/*----------------------------------------------------------------------*/
/*
	define maximum length of each acceptable file name
	under MS-DOS and UNIX, and define directory slash
*/
#if	defined(BSD) || defined(SYSV)	/* UNIX-C */
#define	MAXLEN	14
#define	DSLASH	'/'
#else	/* MS-DOS */
#define	MAXLEN	12
#define	DSLASH	'\\'
#endif
/*----------------------------------------------------------------------*/
typedef struct {	/* image structure */
	char *ifile;	/* path name of image file */
	long	pixel;	/* image pointer in pixel units */
	long	*start;	/* pointer table to first pixel in each line */
	long	psize;	/* image size in pixel units */
	int	bsize;	/* image size in block units */
	FILE	*stream;	/* file stream */
	int	rw;	/* read/write flag */
	int	upd;	/* last update flag: 0 = read, 1 = write */
	int	ptype;	/* pixel type = 1, 2, 3, 4, 5 */
	int	plen;	/* 1 pixel = 2**plen bytes */
	int	blen;	/* 1 block = 2**blen pixels */
	int	xmax;	/* number of pixels in one line */
	int	ymax;	/* number of lines in image */
	int	stx;	/* x-value for upper left corner */
	int	sty;	/* y-value for upper left corner */
	int	extx;	/* x-dim. of window */
	int	exty;	/* y-dim. of window */
} IMAGE;
/*----------------------------------------------------------------------*/
static IMAGE *itab[NFILE] = {(IMAGE *)0};	/* image table */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static double TimeCode(
struct tm *tp)
#else	/* old style */
static double TimeCode(tp)	/* calculate a double from time structure */
struct tm *tp;
#endif
{
	double tcode = 0.0;
	tcode += (double)tp->tm_year;
	tcode *= 12.0;
	tcode += (double)(tp->tm_mon % 12);
	tcode *= 32.0;
	tcode += (double)(tp->tm_mday % 32);
	tcode *= 24.0;
	tcode += (double)(tp->tm_hour % 24);
	tcode *= 60.0;
	tcode += (double)(tp->tm_min % 60);
	tcode *= 60.0;
	tcode += (double)(tp->tm_sec % 60);
	tcode /= 1000.0;
	return tcode;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static long FileLength(
const char *path)
#else	/* old style */
static long FileLength(path)	/* file length in bytes */
const char *path;	/* path name of existing file */
#endif
{
	FILE *stream;
	long length;
	if ((stream = fopen(path, O_RO)) == (FILE *)0)
		return(-1L);
	fseek(stream, 0L, SEEK_END);
	length = ftell(stream);
	fclose(stream);
	return(length);
}
/*----------------------------------------------------------------------*/
static char ifile[PATHSIZE] = {(char)0};	/* image file path */
static char ipath[PATHSIZE] = {(char)0};	/* image directory path */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void setipath(
const int i)
#else	/* old style */
void setipath(i)	/* assign image path from environment */
const int i;	/* select i'th path name */
#endif
{
	char **p;
	int n;
	if ((p = getpath("IMGPATH", &n)) == (char **)0)
		strcpy(ipath, ".");	/* copy current directory path */
	else if (i < n)
		strcpy(ipath, p[i]);	/* copy i'th environment path */
	else
		strcpy(ipath, ".");	/* copy current directory path */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void AssignImageFile(
const char *file)
#else	/* old style */
static void AssignImageFile(file)	/* cat image file to path name */
const char *file;	/* image file */
#endif
{
	strcpy(ifile, ipath);	/* image directory path */
	if (strlen(ifile) > 0)
		strcat(ifile, "/");	/* append "/" */
	strncat(ifile, file, MAXLEN);	/* append image file */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void NewFile(
char *file)
#else	/* old style */
static void NewFile(file)	/* create new file name */
char *file;	/* the new file name: TimeCode(tp) */
#endif
{
	FILE *stream;
	struct tm *tp;
	time_t t;
	double tcode;
	time(&t);
	if ((tp = gmtime(&t)) == (struct tm *)0)
		tp = localtime(&t);
	tcode = TimeCode(tp);
	sprintf(file, "%.3f", tcode);
	AssignImageFile(file);
	while ((stream = fopen(ifile, O_RO)) != (FILE *)0) {
		fclose(stream);
		tcode += 0.001;
		sprintf(file, "%.3f", tcode);
		AssignImageFile(file);
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int Simple(
const int type)
#else	/* old style */
static int Simple(type)	/* is it a simple FITS image? */
const int type;	/* image type */
#endif
{
	int simple;
	switch(type) {	/* assign 'T' or 'F' */
		case 1:	/* unsigned char */
		case 2:	/* short int */
		case 3:	/* float */
		case 4:	/* LONG int */
			simple = 1;
			break;
		case 5:	/* complex */
		default:	/* undefined */
			simple = 0;
			break;
	}
	return(simple);	/* returns 1 or 0 */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int SetImageStruc(
FILE * stream,
const int rw,
const int type,
const int xmax,
const int ymax)
#else	/* old style */
static int SetImageStruc(stream, rw, type, xmax, ymax)
FILE * stream;	/* file stream */
const int rw;	/* read/write flag */
const int type;	/* image type */
const int xmax;	/* number of pixels in x-direction */
const int ymax;	/* number of pixels in y-direction */
#endif
{
	IMAGE *iptr;	/* pointer to image structure */
	int id, y;
	for(id = 0; id < NFILE; id++)	/* search for free slot */
		if (itab[id] == (IMAGE *)0)
			break;	/* found free slot */
	if (id >= NFILE) {
		fprintf(stderr,
		"SetImageStruc: Image table is full\n");
		return(-1);
	}
	/* allocate space for image structure */
	if ((iptr = (IMAGE *)malloc(sizeof(IMAGE))) == (IMAGE *)0) {
		fprintf(stderr,
		"SetImageStruc: Allocation of image structure failed\n");
		return(-1);
	}
	/* allocate space for string containing image file name */
	/* the static string ifile MUST contain the image file name */
	if ((iptr->ifile = (char *)malloc(strlen(ifile)+1)) == (char *)0) {
		fprintf(stderr,
		"SetImageStruc: Allocation of image file name failed\n");
		free(iptr);	/* deallocate image structure */
		return(-1);
	}
	/* allocate space for table containing pixel numbers
		of the first pixel in each line */
	iptr->start = (long *)malloc(ymax * sizeof(long));
	if (iptr->start == (long *)0) {
		fprintf(stderr,
		"SetImageStruc: Allocation of row offset table failed\n");
		free(iptr->ifile);	/* deallocate image file name */
		free(iptr);	/* deallocate image structure */
		return(-1);
	}
	strcpy(iptr->ifile, ifile);	/* copy image file name */
	iptr->stream = stream;	/* file stream */
	iptr->rw = rw;	/* read/write flag */
	iptr->upd = -1;	/* invoke fseek() before any read or write */
	iptr->ptype = type;	/* pixel type */
	iptr->plen = msb(bytepix(type));	/* 1 pixel = 2**plen bytes */
	iptr->blen = msb(buffsize())-iptr->plen;	/* block = 2**blen pixels */
	iptr->xmax = xmax;	/* no of pixels in x-dimension */
	iptr->ymax = ymax;	/* no of pixels in y-dimension */
	iptr->stx  = 0;
	iptr->sty  = 0;
	iptr->extx = iptr->xmax;
	iptr->exty = iptr->ymax;
	iptr->pixel = 0L;
	iptr->start[0] = 0L;
	for(y = 1; y < iptr->ymax; y++)
		iptr->start[y] = iptr->start[y - 1] + iptr->xmax;
	iptr->psize = iptr->start[iptr->ymax - 1] + iptr->xmax;
	iptr->bsize = (int)(((iptr->psize - 1) >> iptr->blen) + 1);
	itab[id] = iptr;
	return(id);	/* return image number */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static FILE *OpenImageFile(
const char *name,
const int rw)
#else	/* old style */
static FILE *OpenImageFile(name, rw)	/* open existing image file */
const char *name;	/* image name */
const int rw;	/* read/write flag */
#endif
{
	char *file;
	FILE *stream;
	long nbytes;
	if ((file = imagefile(name)) == (char *)0) {	/* find image file */
		fprintf(stderr,
		"OpenImageFile: %s: Image file not found\n", name);
		return((FILE *)0);
	}
	AssignImageFile(file);	/* assign image file */
	if ((nbytes = FileLength(ifile)) < 0L) {	/* size of image file */
		fprintf(stderr,
		"OpenImageFile: %s: Image file %s: ", name, file);
		perror("");
		return((FILE *)0);
	}
	if (imagesize(name) != nbytes) {	/* check image size */
		fprintf(stderr,
		"OpenImageFile: %s: Size error in %s\n", name, file);
		return((FILE *)0);
	}
	switch(rw) {	/* open image file */
	case 0:	/* read only */
		if ((stream = fopen(ifile, O_RO)) == (FILE *)0) {
			fprintf(stderr,
			"OpenImageFile: %s: Image file %s: ", name, file);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
			"OpenImageFile: %s: setvbuf() failed for image file %s\n",
				name, file);
			fclose(stream);
			stream = (FILE *)0;
#endif
		}
		break;
	case 2:	/* read and write */
		if ((stream = fopen(ifile, O_RW)) == (FILE *)0) {
			fprintf(stderr,
			"OpenImageFile: %s: Image file %s: ", name, file);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
			"OpenImageFile: %s: setvbuf() failed for image file %s\n",
				name, file);
			fclose(stream);
			stream = (FILE *)0;
#endif
		}
		break;
	default:
		fprintf(stderr,
		"OpenImageFile: %s: Illegal access mode: %d\n", name, rw);
		stream = (FILE *)0;
		break;
	}
	return (stream);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static FILE *CreatImageFile(
const char *name,
const int rw)
#else	/* old style */
static FILE *CreatImageFile(name, rw)	/* create image file */
const char *name;	/* image name */
const int rw;	/* read/write flag */
#endif
{
	char *file;
	FILE *stream;
	char chr = 0;
	long nbytes;
	if ((file = imagefile(name)) == (char *)0) {	/* find image file */
		fprintf(stderr,
		"CreatImageFile: %s: Image file not found\n", name);
		return((FILE *)0);
	}
	if ((nbytes = imagesize(name)) == 0L) {	/* check image size */
		fprintf(stderr,
		"CreatImageFile: %s: Image has zero length\n", name);
		return((FILE *)0);
	}
	AssignImageFile(file);	/* assign image file */
	switch(rw) {	/* create new image file */
	case 1:	/* write only */
		if ((stream = fopen(ifile, C_WO)) == (FILE*)0) {
			fprintf(stderr,
			"CreatImageFile: %s: Image file %s: ", name, file);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
			"CreatImageFile: %s: setvbuf() failed for image file %s\n",
				name, file);
			fclose(stream);
			stream = (FILE *)0;
#endif
		} else {
			fseek(stream, nbytes - 1L, SEEK_SET);
			fwrite(&chr, sizeof(char), 1, stream);	/* extend file */
			if (ferror(stream)) {	/* write error */
				fprintf(stderr,
				"CreatImageFile: %s: Image file %s: ", name, file);
				perror("");
				fclose(stream);
				stream = (FILE *)0;
			} else
				fseek(stream, 0L, SEEK_SET);	/* rewind file */
		}
		break;
	case 2:	/* read and write */
		if ((stream = fopen(ifile, C_RW)) == (FILE *)0) {
			fprintf(stderr,
			"CreatImageFile: %s: Image file %s: ", name, file);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
			"CreatImageFile: %s: setvbuf() failed for image file %s\n",
				name, file);
			fclose(stream);
			stream = (FILE *)0;
#endif
		} else {
			fseek(stream, nbytes - 1L, SEEK_SET);
			fwrite(&chr, sizeof(char), 1, stream);	/* extend file */
			if (ferror(stream)) {	/* write error */
				fprintf(stderr,
				"CreatImageFile: %s: Image file %s: ", name, file);
				perror("");
				fclose(stream);
				stream = (FILE *)0;
			} else
				fseek(stream, 0L, SEEK_SET);	/* rewind file */
		}
		break;
	default:
		fprintf(stderr,
		"CreatImageFile: %s: Illegal access mode: %d\n", name, rw);
		stream = (FILE *)0;
		break;
	}
	return(stream);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int CopyImageFile(
const char *oname,
const char *nname)
#else	/* old style */
static int CopyImageFile(oname, nname)	/* copy image file */
const char *oname;	/* old image name */
const char *nname;	/* new image name */
#endif
{
	FILE *ostream, *nstream;
	char *file, *buf;
	long nbytes;
	unsigned n, bsize = buffsize();	/* buffer size */
	if ((file = imagefile(oname)) == (char *)0) {	/* find image file */
		fprintf(stderr,
		"CopyImageFile: %s: Image file not found\n", oname);
		return(-1);
	}
	AssignImageFile(file);	/* assign image file */
	if ((nbytes = FileLength(ifile)) < 0L) {	/* size of image file */
		fprintf(stderr,
		"CopyImageFile: %s: Image file %s: ", oname, file);
		perror("");
		return(-1);
	}
	if (imagesize(oname) != nbytes) {	/* check image size */
		fprintf(stderr,
		"CopyImageFile: %s: Size error in %s\n", oname, file);
		return(-1);
	}
	if ((ostream = fopen(ifile, O_RO)) == (FILE *)0) {	/* open image file */
		fprintf(stderr,
		"CopyImageFile: %s: Image file %s: ", oname, file);
		perror("");
		return(-1);
#if	BUF_SIZ	/* set stream buffer size */
	} else if (setvbuf(ostream, NULL, _IOFBF, bsize)) {
		fprintf(stderr,
		"CopyImageFile: %s: setvbuf() failed for image file %s\n",
			oname, file);
		fclose(ostream);
		return(-1);
#endif
	}
	if ((file = imagefile(nname)) == (char *)0) {	/* find image file */
		fprintf(stderr,
		"CopyImageFile: %s: Image file not found\n", nname);
		fclose(ostream);
		return(-1);
	}
	if (imagesize(nname) != nbytes) {	/* check image size */
		fprintf(stderr,
		"CopyImageFile: %s: Size in conflict with %s\n", nname, oname);
		fclose(ostream);
		return(-1);
	}
	AssignImageFile(file);	/* assign image file */
	if ((nstream = fopen(ifile, C_WO)) == (FILE *)0) {	/* create image file */
		fprintf(stderr,
		"CopyImageFile: %s: Image file %s: ", nname, file);
		perror("");
		fclose(ostream);
		return(-1);
#if	BUF_SIZ	/* set stream buffer size */
	} else if (setvbuf(nstream, NULL, _IOFBF, bsize)) {
		fprintf(stderr,
		"CopyImageFile: %s: setvbuf() failed for image file %s\n",
			nname, file);
		fclose(nstream);
		fclose(ostream);
		return(-1);
#endif
	}
	if ((buf = (char *)malloc(bsize)) == (char *)0) {
		fprintf(stderr,
		"CopyImageFile: %s: Buffer allocation failed\n", oname);
		fclose(nstream);
		fclose(ostream);
		return(-1);
	}
	while ((n = fread(buf, sizeof (char), bsize, ostream)) > 0) {
		if (ferror(ostream)) {	/* read error */
			clearerr(ostream);
			fprintf(stderr, "CopyImageFile: Read error cleared\n");
		}
		fwrite(buf, sizeof (char), n, nstream);
		if (ferror(nstream)) {	/* write error */
			fprintf(stderr,
			"CopyImageFile: %s: Error writing image file %s: ",
				nname, file);
			perror("");
			free(buf);	/* remember to deallocate the buffer */
			fclose(nstream);
			fclose(ostream);
			return(-1);
		}
	}
	free(buf);	/* remember to deallocate the buffer */
	fclose(nstream);
	fclose(ostream);
	if (FileLength(ifile) != nbytes) {	/* has image file been copied? */
		fprintf(stderr,
		"CopyImageFile: %s: Size error in %s\n", nname, file);
		return(-1);
	}
	return(0);	/* returns 0 if string is successfully executed */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int idel(
const char *name)
#else	/* old style */
int idel(name)	/* delete image header and image file */
const char *name;	/* image name */
#endif
{
	char *file;
	long nbytes;
   if (!xist(name))  /* image does not exist */
      return 0;
	if ((file = imagefile(name)) == (char *)0) {	/* find image file */
		fprintf(stderr,
			"idel: %s: Image file not found\n", name);
		return(-1);
	}
	AssignImageFile(file);	/* assign image file */
	if ((nbytes = FileLength(ifile)) < 0L) {	/* size of image file */
		fprintf(stderr, "idel: %s: Image file %s: ", name, file);
		perror("");
		return(-1);
	}
	if (imagesize(name) != nbytes) {	/* check image size */
		fprintf(stderr,
			"idel: %s: Size error in %s\n", name, file);
		return(-1);
	}
	if (remove(ifile) != 0) {	/* remove image file */
		fprintf(stderr,
			"idel: %s: Image file %s not deleted: ", name, file);
		perror("");
		return(-1);
	}
	rmi(name);	/* remove image header */
	return(0);	/* image file and image header both deleted */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int iremove(
const char *name)
#else	/* old style */
int iremove(name)	/* fast remove of image header and image file */
const char *name;	/* image name */
#endif
{
	char *file;
   if (!xist(name))  /* image does not exist */
      return 0;
	if ((file = imagefile(name)) == (char *)0) {	/* find image file */
		fprintf(stderr,
			"iremove: %s: Image file not found\n", name);
		return(-1);
	}
	AssignImageFile(file);	/* assign image file */
	if (remove(ifile) != 0) {	/* remove image file */
		fprintf(stderr,
			"iremove: %s: Image file %s not deleted: ", name, file);
		perror("");
		return(-1);
	}
	rmi(name);	/* remove image header */
	return(0);	/* image file and image header both removed */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int iopen(
const char *name,
const int rw,
int *type,
int *npix,
int *nlin)
#else	/* old style */
int iopen(name, rw, type, npix, nlin)	/* open existing image file */
const char *name;	/* image name */
const int rw;	/* 0: read only, 2: read and write */
int *type;	/* 1: uchar, 2: short, 3: float, 4: LONG, 5: complex */
int *npix, *nlin;	/* x-dimension and y-dimension */
#endif
{
	int id;
	FILE *stream;
	if (!xist(name)) {
		fprintf(stderr,
		"iopen: %s: Image not found\n", name);
		return(-1);
	}
	if (naxis(name) != 2) {
		fprintf(stderr,
		"iopen: %s: Image not 2-dimensional\n", name);
		return(-1);
	}
	if ((*npix = naxis1(name)) == 0) {
		fprintf(stderr,
		"iopen: %s: No pixels along x-axis\n", name);
		return(-1);
	}
	if ((*nlin = naxis2(name)) == 0) {
		fprintf(stderr,
		"iopen: %s: No pixels along y-axis\n", name);
		return(-1);
	}
	*type = pixtype(name);
	if (bytepix(*type) == 0) {	/* undefined type */
		fprintf(stderr,
		"iopen: %s: Undefined pixel type: %d\n", name, *type);
		return(-1);
	}
	if (imagesize(name) <= 0L) {
		fprintf(stderr,
		"iopen: %s: Image has zero length\n", name);
		return(-1);
	}
	if ((stream = OpenImageFile(name, rw)) == (FILE *)0) {
		fprintf(stderr,
		"iopen: %s: Image file not opened\n", name);
		return(-1);
	}
	if ((id = SetImageStruc(stream, rw, *type, *npix, *nlin)) < 0) {
		fprintf(stderr,
		"iopen: %s: Image structure not created\n", name);
		fclose(stream);	/* close image file */
		return(-1);
	}
	return(id);	/* return image number */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int icreat(
const char *name,
const int rw,
const int type,
const int npix,
const int nlin)
#else	/* old style */
int icreat(name, rw, type, npix, nlin)	/* create new image file */
const char *name;	/* image name */
const int rw;	/* 1: write only, 2: read and write */
const int type;	/* 1: uchar, 2: short, 3: float, 4: LONG, 5: complex */
const int npix, nlin;	/* x-dimension and y-dymension */
#endif
{
	char *str;
	long nbytes;
	int bpix, bit, id;
	FILE *stream;
	char file[MAXLEN+1];	/* image file name */
	int simp = Simple(type);	/* FITS type */
	if (xist(name))	/* existing image header */
		if ((str = imagefile(name)) != (char *)0)
			if (strlen(str) > 0) {	/* existing image file */
			fprintf(stderr,
			"icreat: %s: Existing image header and file\n",
			name);
			return(-1);
			}
	if ((bpix = bytepix(type)) == 0) {	/* bytes per pixel */
		fprintf(stderr,
		"icreat: %s: Undefined pixel type: %d\n",
		name, type);
		return(-1);
	}
	nbytes = bpix;
	nbytes *= npix;
	nbytes *= nlin;	/* image size in bytes */
	if (nbytes <= 0L) {
		fprintf(stderr,
		"icreat: %s: Image has zero length\n",
		name);
		return(-1);
	}
	NewFile(file);	/* create new file name using time() */
	if (setimagefile(name, file) == 0) {	/* enter FILE keyword */
		fprintf(stderr,
		"icreat: %s: FILE not inserted\n",
		name);
		rmi(name);	/* remove image header */
		return(-1);
	}
	if (setsimple(name, simp) == 0) {
		fprintf(stderr,
		"icreat: %s: SIMPLE not inserted\n",
		name);
		rmi(name);	/* remove image header */
		return(-1);
	}
	switch(type) {	/* assign BITPIX value */
		case 1:	bit =   8;	break;	/* unsigned char */
		case 2:	bit =  16;	break;	/* short int */
		case 3:	bit = -32;	break;	/* float */
		case 4:	bit =  32;	break;	/* LONG int */
		case 5:	bit = -64;	break;	/* complex */
		default:	bit =   0;	break;	/* undefined */
	}
	if (setbitpix(name, bit) == 0) {
		fprintf(stderr,
		"icreat: %s: BITPIX not inserted\n",
		name);
		rmi(name);	/* remove image header */
		return(-1);
	}
	if (setnaxis(name, 2) == 0) {
		fprintf(stderr,
		"icreat: %s: NAXIS not inserted\n",
		name);
		rmi(name);
		return(-1);
	}
	if (setnaxis1(name, npix) == 0) {
		fprintf(stderr,
		"icreat: %s: NAXIS1 not inserted\n",
		name);
		rmi(name);
		return(-1);
	}
	if (setnaxis2(name, nlin) == 0) {
		fprintf(stderr,
		"icreat: %s: NAXIS2 not inserted\n",
		name);
		rmi(name);
		return(-1);
	}
	if (setpixtype(name, type) == 0) {
		fprintf(stderr,
		"icreat: %s: PIXTYPE not inserted\n",
		name);
		rmi(name);
		return(-1);
	}
	switch (type) {	/* for keywords depending on pixel type */
		case 1:	/* unsigned char */
			setdatamin(name, (double)0);
			setdatamax(name, (double)UCHAR_Max);
         rmk(name, "BLANK");
			break;
		case 2:	/* short int */
			setdatamin(name, (double)((LONG)(SHORT_Min + 1)));
			setdatamax(name, (double)((LONG)SHORT_Max));
			setblank(name, (double)((LONG)SHORT_Min));
			break;
		case 3:	/* float */
			setdatamin(name, -0.9*FLT_BLANK);
			setdatamax(name, 0.9*FLT_BLANK);
			setblank(name, FLT_BLANK);
			break;
		case 4:	/* LONG int */
			setdatamin(name, (double)(LONG_Min + 1));
			setdatamax(name, (double)LONG_Max);
			setblank(name, (double)LONG_Min);
			break;
		case 5:	/* COMPLEX */
			setdatamin(name, -0.9*FLT_BLANK);
			setdatamax(name, 0.9*FLT_BLANK);
			setblank(name, FLT_BLANK);
			break;
		default:
			break;
	}
	setdate(name);	/* enter current date */
	if ((stream = CreatImageFile(name, rw)) == (FILE *)0) {
		fprintf(stderr,
		"icreat: %s: Image file not created\n",
		name);
		rmi(name);	/* remove image header */
		return(-1);
	}
	if ((id = SetImageStruc(stream, rw, type, npix, nlin)) < 0) {
		fprintf(stderr,
		"icreat: %s: Image structure not created\n",
		name);
		fclose(stream);	/* close image file */
		idel(name);	/* remove image file, and image header */
		return(-1);
	}
	return(id);	/* return image number */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int icopy(
const char *oname,
const char *nname,
const int rw,
const int cpf)
#else	/* old style */
int icopy(oname, nname, rw, cpf)	/* copy new image from old image */
const char *oname;	/* old image name */
const char *nname;	/* new image name */
const int rw;	/* 0: read only, 1: write only, 2: read and write */
const int cpf;	/* -1: create and extend, 0: copy only, 1: copy and open */
#endif
{
	int npix, nlin, type, id;
	FILE *stream;
	char file[MAXLEN+1];	/* image file name */
	if (!xist(oname)) {
		fprintf(stderr,
		"icopy: %s: Image not found\n",
		oname);
		return(-1);
	} else if (strcmp(oname, nname) == 0) {
		if (cpf >= 0) {
			fprintf(stderr,
			"icopy: cpf = %d: Illegal when newname = oldname\n",
			cpf);
			return(-1);
		}
	} else if (xist(nname)) {
		fprintf(stderr,
		"icopy: %s: Existing image\n",
		nname);
		return(-1);
	} else if (cpi(oname, nname) < 1) {	/* copy image header */
		fprintf(stderr,
		"icopy: cpi(%s, %s) has failed\n",
		oname, nname);
		return(-1);
	}
	if (naxis(nname) != 2) {
		fprintf(stderr,
		"icopy: %s: Image not 2-dimensional\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	if ((npix = naxis1(nname)) == 0) {
		fprintf(stderr,
		"icopy: %s: No pixels along x-axis\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	if ((nlin = naxis2(nname)) == 0) {
		fprintf(stderr,
		"icopy: %s: No pixels along y-axis\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	type = pixtype(nname);
	if (bytepix(type) == 0) {	/* undefined pixel type */
		fprintf(stderr,
		"icopy: %s: Undefined pixel type: %d\n",
		nname, type);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	if (imagesize(nname) <= 0L) {
		fprintf(stderr,
		"icopy: %s: Image has zero length\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	NewFile(file);	/* create new file name using time() */
	if (setimagefile(nname, file) == 0) {	/* enter image file */
		fprintf(stderr,
		"icopy: %s: FILE not inserted\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	}
	setdate(nname);	/* enter current date */
	if (cpf < 0) {	/* create and extend image file */
		if ((stream = CreatImageFile(nname, rw)) == (FILE *)0) {
			fprintf(stderr,
			"icopy: %s: Image file not created\n",
			nname);
			rmi(nname);	/* remove image header */
			return(-1);
		}
	} else if (CopyImageFile(oname, nname) < 0) {	/* copy image file */
		fprintf(stderr,
		"icopy: %s: Image file not copied\n",
		nname);
		rmi(nname);	/* remove image header */
		return(-1);
	} else if (cpf == 0)	/* image file successfully copied */
			return(0);	/* image file copied, but not opened */
	else if ((stream = OpenImageFile(nname, rw)) == (FILE *)0) {
		fprintf(stderr,
		"icopy: %s: Image file not opened\n",
		nname);
		idel(nname);	/* remove image file, and image header */
		return(-1);
	}
	if ((id = SetImageStruc(stream, rw, type, npix, nlin)) < 0) {
		fprintf(stderr,
		"icopy: %s: Image structure not created\n",
		nname);
		fclose(stream);	/* close image file */
		idel(nname);	/* remove image file, and image header */
		return(-1);
	}
	return(id);	/* return image number */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int window(
const int id,
const int stx,
const int sty,
const int extx,
const int exty)
#else	/* old style */
int window(id, stx, sty, extx, exty)	/* specifies a window */
const int id;	/* image No. as returned by iopen/icreat/icopy */
const int stx;	/* start x */
const int sty;	/* start y */
const int extx;	/* extend x */
const int exty;	/* extend y */
#endif
{
	IMAGE *p = itab[id];
	int y;
	if (extx <= 0 || exty <= 0)
		return(0);	/* window dimensions must be positive */
	if (stx < 0 || sty < 0)
		return(0);	/* extends above or to the left of image */
	if (stx + extx > p->xmax || sty + exty > p->ymax)
		return(0);	/* extends below or to the right of image */
	p->stx = stx;
	p->sty = sty;
	p->extx = extx;
	p->exty = exty;
	p->pixel = p->stx;
	for(y = 0; y < p->sty; y++)
		p->pixel += p->xmax;
	p->start[0] = p->pixel;
	for(y = 1; y < p->exty; y++)
		p->start[y] = p->start[y - 1] + p->xmax;
	for(y = p->exty; y < p->ymax; y++)
		p->start[y] = LARGE;
	p->upd = -1;	/* invoke fseek() before any read or write */
	return(1);	/* window is inside image */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int getline(
const int id,
const int y,
void *line)
#else	/* old style */
int getline(id, y, line)	/* read a line as defined by window */
const int id;	/* image No. as returned by iopen/icreate/icopy */
const int y;	/* line No. in window */
void *line;	/* line pointer */
#endif
{
	IMAGE *p = itab[id];
	int npix;
	long pixel;
	if (y < 0 || y >= p->exty)	/* line number out of range */
		return(0);
	pixel = p->start[y];
	if (p->upd != 0 || p->pixel != pixel) { /* update pixel pointer */
		p->upd = 0;
		p->pixel = pixel;
		fseek(p->stream, p->pixel << p->plen, SEEK_SET);
	}
	npix = fread((char *)line, 1 << p->plen, p->extx, p->stream);
	if (npix > 0)
		p->pixel += npix;
	if (ferror(p->stream)) {	/* read error */
		clearerr(p->stream);
		fprintf(stderr, "getline: Read error cleared\n");
		p->upd = -1;
	}
	return(npix);	/* number of pixels read */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int putline(
const int id,
const int y,
void *line)
#else	/* old style */
int putline(id, y, line)	/* write a line as defined by window */
const int id;	/* image No. as returned by iopen/icreate/icopy */
const int y;	/* line No. in window */
void *line;	/* line pointer */
#endif
{
	IMAGE *p = itab[id];
	int npix;
	long pixel;
	if (y < 0 || y >= p->exty)	/* line number out of range */
		return(0);
	pixel = p->start[y];
	if (p->upd != 1 || p->pixel != pixel) {	/* update pixel pointer */
		p->upd = 1;
		p->pixel = pixel;
		fseek(p->stream, p->pixel << p->plen, SEEK_SET);
	}
	npix = fwrite((char *)line, 1 << p->plen, p->extx, p->stream);
	if (npix > 0)
		p->pixel += npix;
	if (ferror(p->stream)) {	/* write error */
		clearerr(p->stream);
		fprintf(stderr, "putline: Write error cleared\n");
		p->upd = -1;
	}
	return(npix);	/* number of pixels written */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int bread(
const int id,
const int block,
void *buffer)
#else	/* old style */
int bread(id, block, buffer)	/* read one block from image */
const int id;	/* image No. as returned by iopen/icreate/icopy */
const int block;	/* block number */
void *buffer;	/* buffer containing one block of data */
#endif
{
	IMAGE *p = itab[id];
	int npix;
	long pixel;
	if (block < 0 || block >= p->bsize)	/* block out of range */
		return(0);
	pixel = (long)block << p->blen;
	if (p->upd != 0 || p->pixel != pixel) {	/* update pixel pointer */
		p->upd = 0;
		p->pixel = pixel;
		fseek(p->stream, p->pixel << p->plen, SEEK_SET);
	}
	if (block == p->bsize - 1)	/* check if block is the last one */
		npix = (int)(p->psize - p->pixel);
	else
		npix = 1 << p->blen;	/* block length in pixels */
	npix = fread((char *)buffer, 1 << p->plen, npix, p->stream);
	if (npix > 0)
		p->pixel += npix;
	if (ferror(p->stream)) {	/* read error */
		clearerr(p->stream);
		fprintf(stderr, "bread: Read error cleared\n");
		p->upd = -1;
	}
	return(npix);	/* number of pixels read */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int bwrite(
const int id,
const int block,
void *buffer)
#else	/* old style */
int bwrite(id, block, buffer)	/* write one block to image */
const int id;	/* image No. as returned by iopen/icreate/icopy */
const int block;	/* block number */
void *buffer;	/* buffer containing one block of data */
#endif
{
	IMAGE *p = itab[id];
	int npix;
	long pixel;
	if (block < 0 || block >= p->bsize)	/* block out of range */
		return(0);
	pixel = (long)block << p->blen;
	if (p->upd != 1 || p->pixel != pixel) {	/* update pixel pointer */
		p->upd = 1;
		p->pixel = pixel;
		fseek(p->stream, p->pixel << p->plen, SEEK_SET);
	}
	if (block == p->bsize - 1)	/* check if block is the last one */
		npix = (int)(p->psize - p->pixel);
	else
		npix = 1 << p->blen;	/* block length in pixels */
	npix = fwrite((char *)buffer, 1 << p->plen, npix, p->stream);
	if (npix > 0)
		p->pixel += npix;
	if (ferror(p->stream)) {	/* write error */
		clearerr(p->stream);
		fprintf(stderr, "bwrite: Write error cleared\n");
		p->upd = -1;
	}
	return(npix);	/* number of pixels written */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void iflush(
const int id)
#else	/* old style */
void iflush(id)	/* flush image file */
const int id;	/* image No. as returned by iopen/icreate/icopy */
#endif
{
	IMAGE *p;
	if (id >= 0 && (p = itab[id]) != (IMAGE *)0)
		fflush(p->stream);	/* flush image file */
   update();   /* update header files */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void iclose(
const int id)
#else	/* old style */
void iclose(id)	/* close image file */
const int id;	/* image No. as returned by iopen/icreate/icopy */
#endif
{
	IMAGE *p;
	if (id >= 0 && (p = itab[id]) != (IMAGE *)0) {
		fclose(p->stream);	/* close image file */
		free(p->start);	/* deallocate line-start table */
		free(p->ifile);	/* deallocate image file name */
		free(p);	/* deallocate image structure */
		itab[id] = (IMAGE *)0;	/* this entry is now empty */
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int chiomod(
const int id,
const int rw)
#else	/* old style */
int chiomod(id, rw)	/* change read/update mode */
const int id;	/* image No. as returned by iopen/icreate/icopy */
const int rw;	/* 0: read only, 1: Illegal, 2: update */
#endif
{
	IMAGE *p;
	if ((p = itab[id]) == (IMAGE *)0)	/* incorrect image number */
		return(0);
	if (p->rw == rw)	/* already opened with this IO mode */
		return(1);
	switch(rw) {	/* reopen image file */
	case 0:	/* read only */
		if ((p->stream = freopen(p->ifile,O_RO,p->stream)) == (FILE *)0){
			fprintf(stderr, "chiomod: Image file %s: ", p->ifile);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(p->stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
				"chiomod: setvbuf() failed for %s\n", p->ifile);
			fclose(p->stream);
			p->stream = (FILE *)0;
#endif
		}
		break;
	case 2:	/* update (read and write) */
		if ((p->stream = freopen(p->ifile,O_RW,p->stream)) == (FILE *)0){
			fprintf(stderr, "chiomod: Image file %s: ", p->ifile);
			perror("");
#if	BUF_SIZ	/* set stream buffer size */
		} else if (setvbuf(p->stream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
				"chiomod: setvbuf() failed for %s\n", p->ifile);
			fclose(p->stream);
			p->stream = (FILE *)0;
#endif
		}
		break;
	default:
		fprintf(stderr, "chiomod: Illegal access mode: %d\n", rw);
		fclose(p->stream);	/* close image file */
		p->stream = (FILE *)0;
		break;
	}
	if (p->stream == (FILE *)0) {	/* image file was not re-opened */
		free(p->start);	/* deallocate line-start table */
		free(p->ifile);	/* deallocate image file name */
		free(p);	/* deallocate image structure */
		itab[id] = (IMAGE *)0;	/* this entry is now empty */
		return (0);	/* chiomod failed */
	}
	p->rw = rw;	/* read/write flag */
	p->upd = -1;	/* invoke fseek() before any read or write */
	return (1);	/* chiomod has successfully re-opened image file */
}
/*----------------------------------------------------------------------*/
