
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <ctype.h>	/* ctype macros */
#include <time.h>	/* declares time functions */
#if	defined(BSD)	/* BSD-UNIX */
#include <sys/types.h>	/* defines time_t */
#include <strings.h>	/* string header */
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros, errno declared */
#include <stdlib.h>	/* errno declared */
#else	/* not ANSI-C */
extern time_t time();
#endif
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
/*----------------------------------------------------------------------*/
#define	TEXTSIZE	256	/* text buffer size */
#define	FIELDSIZ	256	/* field buffer size */
/*----------------------------------------------------------------------*/
static char FieldBuf[FIELDSIZ] = {(char)0};	/* field buffer */
/* the field buffer contains the current keyword field */
/* all keyword string functions return a pointer to this buffer */
/* calling a new string function will overwrite the buffer */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Field(const char *text)
#else	/* old style */
static char *Field(text)	/* extract value field from text */
const char *text;
#endif
{
	const char *p = text;
	char *q = FieldBuf;
	for (; isspace(*p); p++)
		;	/* skip leading spaces */
	while (*p != '\0' && *p != '\\')	/* copy until '\0' or '\\' */
		*q++ = *p++;
	for (--q; (q >= FieldBuf) && isspace(*q); q--)
		;
	*++q = '\0';	/* truncate trailing spaces */
	if (*FieldBuf == '\0')
		return((char *)0);	/* empty value field */
	else
		return(FieldBuf);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *imagefile(const char *name)
#else	/* old style */
char *imagefile(name)	/* get image file to FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "FILE")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "FILE" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setimagefile(const char *name, const char *file)
#else	/* old style */
int setimagefile(name, file)	/* set image file */
const char *name;	/* image name */
const char *file;	/* image file */
#endif
{
	if(putk(name, "FILE", file) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "FILE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int naxis(const char *name)
#else	/* old style */
int naxis(name)	/* get number of dimensions */
const char *name;	/* image name */
#endif
{
	char *s;
	int r;
	if((s = getk(name, "NAXIS")) == (char *)0)	/* get keyword */
		return 0;	/* can't find "NAXIS" */
	sscanf(s, "%d", &r);	/* convert to integer value */
	return r;
}
/*----------------------------------------------------------------------*/
static char TextBuf[TEXTSIZE] = {(char)0};	/* text buffer */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setnaxis(const char *name, const int n)
#else	/* old style */
int setnaxis(name, n)	/* set number of dimensions */
const char *name;	/* image name */
const int n;	/* number of dimensions */
#endif
{
	sprintf(TextBuf, "%d", n);	/* convert to string */
	if(putk(name, "NAXIS", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "NAXIS" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int naxis1(const char *name)
#else	/* old style */
int naxis1(name)	/* pixels in the x dimension */
const char *name;	/* image name */
#endif
{
	char *s;
	int r;
	if((s = getk(name, "NAXIS1")) == (char *)0)	/* get keyword */
		return 0;	/* can't find "NAXIS1" */
	sscanf(s, "%d", &r);	/* convert to integer value */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setnaxis1(const char *name, const int nx)
#else	/* old style */
int setnaxis1(name, nx)	/* set no. of pixels along x-axis */
const char *name;	/* image name */
const int nx;	/* number of pixels in x-dimension */
#endif
{
	sprintf(TextBuf, "%d", nx);	/* convert to string */
	if(putk(name, "NAXIS1", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "NAXIS1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int naxis2(const char *name)
#else	/* old style */
int naxis2(name)	/* lines in the y dimension */
const char *name;	/* image name */
#endif
{
	char *s;
	int r;
	if((s = getk(name, "NAXIS2")) == (char *)0)	/* get keyword */
		return 0;	/* can't find "NAXIS2" */
	sscanf(s, "%d", &r);	/* convert to integer value */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setnaxis2(const char *name, const int ny)
#else	/* old style */
int setnaxis2(name, ny)	/* set no. of pixels along y-axis */
const char *name;	/* image name */
const int ny;	/* number of pixels in y-dimension */
#endif
{
	sprintf(TextBuf, "%d", ny);	/* convert to string */
	if(putk(name, "NAXIS2", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "NAXIS2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int simple(const char *name)
#else	/* old style */
int simple(name)	/* is it a simple FITS image */
const char *name;	/* image name */
#endif
{
	char *s;
	if((s = getk(name, "SIMPLE")) == (char *)0)	/* get keyword */
		return(0);	/* can't find "SIMPLE" */
	for (; isspace(*s); s++)
		;	/* skip leading spaces */
	if (*s != 'T')
		return(0);	/* not a simple FITS image */
	else
		return(1);	/* it is a simple FITS image */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setsimple(const char *name, const int c)
#else	/* old style */
int setsimple(name, c)	/* set "SIMPLE" to 'T' or 'F' */
const char *name;	/* image name */
const int c;	/* 1 or 0 */
#endif
{
	if (c)
		strcpy(TextBuf, "T");	/* simple */
	else
		strcpy(TextBuf, "F");	/* not simple */
	if(putk(name, "SIMPLE", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "SIMPLE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int bitpix(const char *name)
#else	/* old style */
int bitpix(name)	/* get number of bits per pixel */
const char *name;	/* image name */
#endif
{
	char *s;
	int r;
	if((s = getk(name, "BITPIX")) == (char *)0)	/* get keyword */
		return 0;	/* can't find "BITPIX" */
	sscanf(s, "%d", &r);	/* convert to integer value */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setbitpix(const char *name, const int bit)
#else	/* old style */
int setbitpix(name, bit)	/* set number of bits per pixels */
const char *name;	/* image name */
const int bit;	/* number of bits per pixel */
#endif
{
	sprintf(TextBuf, "%d", bit);	/* convert to string */
	if(putk(name, "BITPIX", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "BITPIX" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int pixtype(const char *name)
#else	/* old style */
int pixtype(name)	/* get pixel data type */
const char *name;	/* image name */
#endif
{
	char *s;
	int r;
	if((s = getk(name, "PIXTYPE")) == (char *)0) {	/* PIXTYPE not given */
		switch(bitpix(name)) {	/* assign pixel data type from BITPIX */
			case   8:	r = 1;	break;	/* unsigned char */
			case  16:	r = 2;	break;	/* short int */
			case -32:	r = 3;	break;	/* float */
			case  32:	r = 4;	break;	/* LONG int */
			case -64:	r = 5;	break;	/* complex */
			default :	r = 0;	break;	/* undefined */
		}
		setpixtype(name, r);	/* set missing PIXTYPE value */
	} else	/* PIXTYPE is given */
		sscanf(s, "%d", &r);	/* convert to integer value */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setpixtype(const char *name, const int type)
#else	/* old style */
int setpixtype(name, type)	/* set pixel data type */
const char *name;	/* image name */
const int type;	/* pixel type number */
#endif
{
	sprintf(TextBuf, "%d", type);	/* convert to string */
	if(putk(name, "PIXTYPE", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "PIXTYPE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int bytepix(const int type)
#else	/* old style */
int bytepix(type)	/* number of bytes per pixel */
const int type;	/* image type */
#endif
{
	int bpix;
	switch(type) {	/* assign bytes per pixel */
		case 1:	bpix = sizeof(UCHAR);	break;	/* unsigned char */
		case 2:	bpix = sizeof(SHORT);	break;	/* short int */
		case 3:	bpix = sizeof(FLOAT);	break;	/* float */
		case 4:	bpix = sizeof (LONG);	break;	/* LONG int */
		case 5:	bpix = sizeof(COMPLEX);	break;	/* complex */
		default:	bpix = 0;	break;	/* undefined */
	}
	return bpix;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int pixsize(const char *name)
#else	/* old style */
int pixsize(name)	/* pixel size in bytes */
const char *name;
#endif
{
	return bytepix(pixtype(name));
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int bpixels(const char *name)
#else	/* old style */
int bpixels(name)	/* returns the block size in pixels */
const char *name;
#endif
{
	return (int)(buffsize()/pixsize(name));
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
long imagesize(const char *name)
#else	/* old style */
long imagesize(name)	/* image size in bytes */
const char *name;
#endif
{
	int npix, nlin;
	long nbytes;
	npix = naxis1(name);	/* number of pixels */
	nlin = naxis2(name);	/* number of lines */
	nbytes = pixsize(name);	/* number of bytes per pixel */
	nbytes *= npix;
	nbytes *= nlin;
	return nbytes;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double bscale(const char *name)
#else	/* old style */
double bscale(name)	/* get brightness scale factor */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "BSCALE")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "BSCALE" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setbscale(const char *name, const double scale)
#else	/* old style */
int setbscale(name, scale)	/* set brightness scale factor */
const char *name;	/* image name */
const double scale;	/* scale factor */
#endif
{
	sprintf(TextBuf, "%.10G", scale);	/* convert to string */
	if(putk(name, "BSCALE", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "BSCALE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double bbzero(const char *name)
#else	/* old style */
double bbzero(name)	/* get brightness offset value */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "BZERO")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "BZERO" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setbzero(const char *name, const double zero)
#else	/* old style */
int setbzero(name, zero)	/* set brightness offset value */
const char *name;	/* image name */
const double zero;	/* offset value */
#endif
{
	sprintf(TextBuf, "%.10G", zero);	/* convert to string */
	if(putk(name, "BZERO", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "BZERO" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *bunit(const char *name)
#else	/* old style */
char *bunit(name)	/* get brightness unit into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "BUNIT")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "BUNIT" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setbunit(const char *name, const char *unit)
#else	/* old style */
int setbunit(name, unit)	/* set brightness unit */
const char *name;	/* image name */
const char *unit;	/* brightness unit */
#endif
{
	if(putk(name, "BUNIT", unit) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "BUNIT" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *object(const char *name)
#else	/* old style */
char *object(name)	/* get object name into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "OBJECT")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "OBJECT" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setobject(const char *name, const char *obj)
#else	/* old style */
int setobject(name, obj)	/* set object name */
const char *name;	/* image name */
const char *obj;	/* object name */
#endif
{
	if(putk(name, "OBJECT", obj) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "OBJECT" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *filter(const char *name)
#else	/* old style */
char *filter(name)	/* get filter name into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "FILTER")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "FILTER" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setfilter(const char *name, const char *fil)
#else	/* old style */
int setfilter(name, fil)	/* set filter name */
const char *name;	/* image name */
const char *fil;	/* filter name */
#endif
{
	if(putk(name, "FILTER", fil) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "FILTER" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *grism(const char *name)
#else	/* old style */
char *grism(name)	/* get name of grism into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "GRISM")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "GRISM" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setgrism(const char *name, const char *gri)
#else	/* old style */
int setgrism(name, gri)	/* set grism name */
const char *name;	/* image name */
const char *gri;	/* name of grism */
#endif
{
	if(putk(name, "GRISM", gri) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "GRISM" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *origin(const char *name)
#else	/* old style */
char *origin(name)	/* get institution name into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "ORIGIN")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "ORIGIN" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setorigin(const char *name, const char *org)
#else	/* old style */
int setorigin(name, org)	/* set institution name */
const char *name;	/* image name */
const char *org;	/* institution name */
#endif
{
	if(putk(name, "ORIGIN", org) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "ORIGIN" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *instrume(const char *name)
#else	/* old style */
char *instrume(name)	/* get acquisition instrument into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "INSTRUME")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "INSTRUME" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setinstrume(const char *name, const char *instr)
#else	/* old style */
int setinstrume(name, instr)	/* set acquisition instrument */
const char *name;	/* image name */
const char *instr;	/* instrument */
#endif
{
	if(putk(name, "INSTRUME", instr) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "INSTRUME" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *telescop(const char *name)
#else	/* old style */
char *telescop(name)	/* get telescope into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "TELESCOP")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "TELESCOP" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int settelescop(const char *name, const char *tel)
#else	/* old style */
int settelescop(name, tel)	/* set telescope */
const char *name;	/* image name */
const char *tel;	/* telescope */
#endif
{
	if(putk(name, "TELESCOP", tel) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "TELESCOP" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *observer(const char *name)
#else	/* old style */
char *observer(name)	/* get observer name into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "OBSERVER")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "OBSERVER" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setobserver(const char *name, const char *obs)
#else	/* old style */
int setobserver(name, obs)	/* set observer name */
const char *name;	/* image name */
const char *obs;	/* observer */
#endif
{
	if(putk(name, "OBSERVER", obs) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "OBSERVER" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crval1(const char *name)
#else	/* old style */
double crval1(name)	/* get physical coordinate at reference pixel */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CRVAL1")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CRVAL1" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crval2(const char *name)
#else	/* old style */
double crval2(name)	/* get physical coordinate at reference pixel */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CRVAL2")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CRVAL2" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrval1(const char *name, const double crval)
#else	/* old style */
int setcrval1(name, crval)	/* set physical coordinate at ref. pixel */
const char *name;	/* image name */
const double crval;	/* physical coordinate */
#endif
{
	sprintf(TextBuf, "%.10G", crval);	/* convert to string */
	if(putk(name, "CRVAL1", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CRVAL1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrval2(const char *name, const double crval)
#else	/* old style */
int setcrval2(name, crval)	/* set physical coordinate at ref. pixel */
const char *name;	/* image name */
const double crval;	/* physical coordinate */
#endif
{
	sprintf(TextBuf, "%.10G", crval);	/* convert to string */
	if(putk(name, "CRVAL2", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CRVAL2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crpix1(const char *name)
#else	/* old style */
double crpix1(name)	/* get array location of reference pixel */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CRPIX1")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CRPIX1" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crpix2(const char *name)
#else	/* old style */
double crpix2(name)	/* get array location of reference pixel */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CRPIX2")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CRPIX2" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrpix1(const char *name, const double crpix)
#else	/* old style */
int setcrpix1(name, crpix)	/* set array location of reference pixel */
const char *name;	/* image name */
const double crpix;	/*  array location of reference pixel */
#endif
{
	sprintf(TextBuf, "%.10G", crpix);	/* convert to string */
	if(putk(name, "CRPIX1", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CRPIX1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrpix2(const char *name, const double crpix)
#else	/* old style */
int setcrpix2(name, crpix)	/* set array location of reference pixel */
const char *name;	/* image name */
const double crpix;	/*  array location of reference pixel */
#endif
{
	sprintf(TextBuf, "%.10G", crpix);	/* convert to string */
	if(putk(name, "CRPIX2", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CRPIX2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double cdelt1(const char *name)
#else	/* old style */
double cdelt1(name)	/* get increment in physical coordinate */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CDELT1")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CDELT1" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double cdelt2(const char *name)
#else	/* old style */
double cdelt2(name)	/* get increment in physical coordinate */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CDELT2")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CDELT2" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcdelt1(const char *name, const double cdelt)
#else	/* old style */
int setcdelt1(name, cdelt)	/* set increment in physical coordinate */
const char *name;	/* image name */
const double cdelt;	/* increment in physical coordinate */
#endif
{
	sprintf(TextBuf, "%.10G", cdelt);	/* convert to string */
	if(putk(name, "CDELT1", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CDELT1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcdelt2(const char *name, const double cdelt)
#else	/* old style */
int setcdelt2(name, cdelt)	/* set increment in physical coordinate */
const char *name;	/* image name */
const double cdelt;	/* increment in physical coordinate */
#endif
{
	sprintf(TextBuf, "%.10G", cdelt);	/* convert to string */
	if(putk(name, "CDELT2", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CDELT2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crota1(const char *name)
#else	/* old style */
double crota1(name)	/* get rotation angle of actual axis */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CROTA1")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CROTA1" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double crota2(const char *name)
#else	/* old style */
double crota2(name)	/* get rotation angle of actual axis */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "CROTA2")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "CROTA2" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrota1(const char *name, const double crota)
#else	/* old style */
int setcrota1(name, crota)	/* set rotation angle of actual axis */
const char *name;	/* image name */
const double crota;	/* rotation angle of actual axis */
#endif
{
	sprintf(TextBuf, "%.10G", crota);	/* convert to string */
	if(putk(name, "CROTA1", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CROTA1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setcrota2(const char *name, const double crota)
#else	/* old style */
int setcrota2(name, crota)	/* set rotation angle of actual axis */
const char *name;	/* image name */
const double crota;	/* rotation angle of actual axis */
#endif
{
	sprintf(TextBuf, "%.10G", crota);	/* convert to string */
	if(putk(name, "CROTA2", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CROTA2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *ctype1(const char *name)
#else	/* old style */
char *ctype1(name)	/* get type of physical coordinate into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "CTYPE1")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "CTYPE1" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *ctype2(const char *name)
#else	/* old style */
char *ctype2(name)	/* get type of physical coordinate into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "CTYPE2")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "CTYPE2" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setctype1(const char *name, const char *ctype)
#else	/* old style */
int setctype1(name, ctype)	/* set type of physical coordinate on axis */
const char *name;	/* image name */
const char *ctype;	/* coordinate type */
#endif
{
	if(putk(name, "CTYPE1", ctype) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CTYPE1" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setctype2(const char *name, const char *ctype)
#else	/* old style */
int setctype2(name, ctype)	/* set type of physical coordinate on axis */
const char *name;	/* image name */
const char *ctype;	/* coordinate type */
#endif
{
	if(putk(name, "CTYPE2", ctype) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "CTYPE2" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double datamax(const char *name)
#else	/* old style */
double datamax(name)	/* get the maximum pixel value */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "DATAMAX")) == (char *)0) {	/* DATAMAX not given */
		switch (pixtype(name)) {	/* set DATAMAX value */
			case 1:	/* unsigned char */
				r = (double) UCHAR_Max;
				setdatamax(name, r);
				break;
			case 2:	/* short int */
				r = (double) ((LONG) SHORT_Max);
				setdatamax(name, r);
				break;
			case 3:	/* float */
				r = 0.9*FLT_BLANK;
				setdatamax(name, r);
				break;
			case 4:	/* LONG int */
				r = (double) LONG_Max;
				setdatamax(name, r);
				break;
			case 5:	/* COMPLEX */
				r = 0.9*FLT_BLANK;
				setdatamax(name, r);
				break;
			default:
				r = 0.0;
				break;
		}
	} else	/* DATAMAX is given */
		sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setdatamax(const char *name, const double max)
#else	/* old style */
int setdatamax(name, max)	/* set maximum pixel value */
const char *name;	/* image name */
const double max;	/* maximum pixel value */
#endif
{
	sprintf(TextBuf, "%.10G", max);	/* convert to string */
	if(putk(name, "DATAMAX", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "DATAMAX" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double datamin(const char *name)
#else	/* old style */
double datamin(name)	/* get the minimum pixel value */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "DATAMIN")) == (char *)0) {	/* DATAMIN not given */
		switch (pixtype(name)) {	/* set DATAMIN value */
			case 1:	/* unsigned char */
				r = (double) 0;
				setdatamin(name, r);
				break;
			case 2:	/* short int */
				r = (double) ((LONG) (SHORT_Min + 1));
				setdatamin(name, r);
				break;
			case 3:	/* float */
				r = -0.9*FLT_BLANK;
				setdatamin(name, r);
				break;
			case 4:	/* LONG int */
				r = (double) (LONG_Min + 1);
				setdatamin(name, r);
				break;
			case 5:	/* COMPLEX */
				r = -0.9*FLT_BLANK;
				setdatamin(name, r);
				break;
			default:
				r = 0.0;
				break;
		}
	} else	/* DATAMIN is given */
		sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setdatamin(const char *name, const double min)
#else	/* old style */
int setdatamin(name, min)	/* set minimum pixel value */
const char *name;	/* image name */
const double min;	/* minimum pixel value */
#endif
{
	sprintf(TextBuf, "%.10G", min);	/* convert to string */
	if(putk(name, "DATAMIN", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "DATAMIN" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double epoch(const char *name)
#else	/* old style */
double epoch(name)	/* get epoch of coordinate system */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "EPOCH")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "EPOCH" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setepoch(const char *name, const double ep)
#else	/* old style */
int setepoch(name, ep)	/* set epoch of coordinate system */
const char *name;	/* image name */
const double ep;	/* epoch */
#endif
{
	sprintf(TextBuf, "%.10G", ep);	/* convert to string */
	if(putk(name, "EPOCH", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "EPOCH" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double blank(const char *name)
#else	/* old style */
double blank(name)	/* value assigned to indefinite pixels */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if ((s = getk(name, "BLANK")) == (char *)0) {	/* BLANK not given */
		switch (pixtype(name)) {	/* set BLANK value */
			case 2:	/* short int */
				r = (double) ((LONG) SHORT_Min);
				setblank(name, r);
				break;
			case 3:	/* float */
				r = FLT_BLANK;
				setblank(name, r);
				break;
			case 4:	/* LONG int */
				r = (double) LONG_Min;
				setblank(name, r);
				break;
			case 5:	/* COMPLEX */
				r = FLT_BLANK;
				setblank(name, r);
				break;
			default:
				r = 0.0;
				break;
		}
	} else	/* BLANK is given */
		sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setblank(const char *name, const double blank)
#else	/* old style */
int setblank(name, blank)	/* set blank value */
const char *name;	/* image name */
const double blank;	/* blank pixel value */
#endif
{
	sprintf(TextBuf, "%.10G", blank);	/* convert to string */
	if(putk(name, "BLANK", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "BLANK" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *date(const char *name)
#else	/* old style */
char *date(name)	/* get date file was written into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "DATE")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "DATE" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *dateobs(const char *name)
#else	/* old style */
char *dateobs(name)	/* get date of data acquisition into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "DATE-OBS")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "DATE-OBS" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setdate(const char *name)
#else	/* old style */
int setdate(name)	/* set "DATE" keyword with current date */
const char *name;	/* image name */
#endif
{
	time_t t;
	struct tm *p;
	time(&t);
	if ((p = gmtime(&t)) == (struct tm *)0)
		p = localtime(&t);
	sprintf(TextBuf, "%02d/%02d/%02d %02d:%02d:%02d",
		p->tm_mday, p->tm_mon+1, p->tm_year,
		p->tm_hour, p->tm_min, p->tm_sec);
	if(putk(name, "DATE", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "DATE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setdateobs(const char *name)
#else	/* old style */
int setdateobs(name)	/* set "DATE-OBS" keyword with current date */
const char *name;	/* image name */
#endif
{
	time_t t;
	struct tm *p;
	time(&t);
	if ((p = gmtime(&t)) == (struct tm *)0)
		p = localtime(&t);
	sprintf(TextBuf, "%02d/%02d/%02d %02d:%02d:%02d",
		p->tm_mday, p->tm_mon+1, p->tm_year,
		p->tm_hour, p->tm_min, p->tm_sec);
	if(putk(name, "DATE-OBS", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "DATE-OBS" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double jdstart(const char *name)
#else	/* old style */
double jdstart(name)	/* get JD of exposure start */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "JD-START")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "JD-START" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setjdstart(const char *name, const double start)
#else	/* old style */
int setjdstart(name, start)	/* set JD of exposure start */
const char *name;	/* image name */
const double start;	/* JD */
#endif
{
	sprintf(TextBuf, "%.10G", start);	/* convert to string */
	if(putk(name, "JD-START", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "JD-START" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double jdstop(const char *name)
#else	/* old style */
double jdstop(name)	/* get JD of exposure stop */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "JD-STOP")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "JD-STOP" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setjdstop(const char *name, const double stop)
#else	/* old style */
int setjdstop(name, stop)	/* set JD of exposure stop */
const char *name;	/* image name */
const double stop;	/* JD */
#endif
{
	sprintf(TextBuf, "%.10G", stop);	/* convert to string */
	if(putk(name, "JD-STOP", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "JD-STOP" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *history(const char *name)
#else	/* old style */
char *history(name)	/* get history into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "HISTORY")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "HISTORY" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int sethistory(const char *name, const char *upd)
#else	/* old style */
int sethistory(name, upd)	/* update "HISTORY" */
const char *name;	/* image name */
const char *upd;	/* update to existing history */
#endif
{
	char *s;
	strcpy(TextBuf, upd);	/* copy update string */
	if((s = getk(name, "HISTORY")) != (char *)0) {	/* get history */
		strcat(TextBuf, ";");	/* separator */
		strcat(TextBuf, s);	/* append existing history */
	}
	if(putk(name, "HISTORY", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "HISTORY" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *iname(const char *name)
#else	/* old style */
char *iname(name)	/* get image name into FieldBuf */
const char *name;	/* image name */
#endif
{
	char *s, *t;
	if((s = getk(name, "IMAGE")) == (char *)0)	/* get keyword */
		return((char *)0);	/* can't find "IMAGE" */
	else if((t = Field(s)) == (char *)0)	/* value field */
		return((char *)0);	/* can't find value field */
	else
		return(t);	/* return value field from text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setiname(const char *name)
#else	/* old style */
int setiname(name)	/* set "IMAGE" keyword with image name */
const char *name;	/* image name */
#endif
{
	if(putk(name, "IMAGE", name) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "IMAGE" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int hdrcp(
const char *old,
const char *new)
#else	/* old style */
int hdrcp(old, new)	/* copy old header to new header */
const char *old;	/* old header */
const char *new;	/* new header */
#endif
{
   if (xist(old) == 0) {
      return -1;
	} else if (xist(new) > 0) {
      return -1;
	} else if (cpi(old, new) < 1) {	/* copy header */
      return -1;
	} else if (setimagefile(new, "") == 0) {	/* erase FILE value */
		rmi(new);	/* remove header */
      return -1;
	}
   return 0;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double exptime(const char *name)
#else	/* old style */
double exptime(name)	/* get exposure time */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "EXPTIME")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "EXPTIME" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setexptime(const char *name, const double et)
#else	/* old style */
int setexptime(name, et)	/* set exposure time */
const char *name;	/* image name */
const double et;	/* exposure time */
#endif
{
	sprintf(TextBuf, "%.10G", et);	/* convert to string */
	if(putk(name, "EXPTIME", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "EXPTIME" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double darktime(const char *name)
#else	/* old style */
double darktime(name)	/* get dark time */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "DARKTIME")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "DARKTIME" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int setdarktime(const char *name, const double dt)
#else	/* old style */
int setdarktime(name, dt)	/* set dark time */
const char *name;	/* image name */
const double dt;	/* dark time */
#endif
{
	sprintf(TextBuf, "%.10G", dt);	/* convert to string */
	if(putk(name, "DARKTIME", TextBuf) == (char *)0)	/* put keyword */
		return(0);	/* can't enter "DARKTIME" */
	else
		return(1);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double saturate(const char *name)
#else	/* old style */
double saturate(name)	/* get saturation level */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "SATURATE")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "SATURATE" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double atodgain(const char *name)
#else	/* old style */
double atodgain(name)	/* get A-to-D gain */
const char *name;	/* image name */
#endif
{
	char *s;
	double r;
	if((s = getk(name, "ATODGAIN")) == (char *)0)	/* get keyword */
		return 0.0;	/* can't find "ATODGAIN" */
	sscanf(s, "%lf", &r);	/* convert to double */
	return r;
}
/*----------------------------------------------------------------------*/
