
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#if	defined(BSD)	/* BSD-UNIX */
#include <sys/types.h>	/* size_t defined */
#include <strings.h>
#else	/* not BSD-UNIX */
#include <string.h>
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
#include <math.h>
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "nr.h"	/* function declarations for nr.c */
/*----------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*----------------------------------------------------------------------*/
#define PROD(a,b)	((a).x * (b).x + (a).y * (b).y)
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *nrerror(const char *error_text)
#else	/* old style */
char *nrerror(error_text)
const char *error_text;
#endif
{
	static char *error = "Allocation error in nrerror\n";
	static char *header = "Numerical Recipes run-time error...\n";
	size_t n = strlen(header) + strlen(error_text) + 2;
	char *str = (char *)malloc(n);
	if (str == (char *)0)
		return error;
	strcpy(str, header);
	strcat(str, error_text);
	strcat(str, "\n");
	return str;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int *ivector(int nl, int nh)
#else	/* old style */
int *ivector(nl, nh)
int nl, nh;
#endif
{
	int *v = (int *)calloc(nh - nl + 1, sizeof(int));
	if (v == (int *)0)
		return (int *)0;
	return (v - nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
float *vector(int nl, int nh)
#else	/* old style */
float *vector(nl, nh)
int nl, nh;
#endif
{
	float *v = (float *)calloc(nh - nl + 1, sizeof(float));
	if (v == (float *)0)
		return (float *)0;
	return (v - nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double *dvector(int nl, int nh)
#else	/* old style */
double *dvector(nl, nh)
int nl, nh;
#endif
{
	double *v = (double *)calloc(nh - nl + 1, sizeof(double));
	if (v == (double *)0)
		return (double *)0;
	return (v - nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
COMPLEX *cvector(int nl, int nh)
#else	/* old style */
COMPLEX *cvector(nl, nh)
int nl, nh;
#endif
{
	COMPLEX *v = (COMPLEX *)calloc(nh - nl + 1, sizeof(COMPLEX));
	if (v == (COMPLEX *)0)
		return (COMPLEX *)0;
	return (v - nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int **imatrix(int nrl, int nrh, int ncl, int nch)
#else	/* old style */
int **imatrix(nrl, nrh, ncl, nch)
int nrl, nrh, ncl, nch;
#endif
{
	int i;
	size_t nrows = nrh - nrl + 1;
	size_t ncols = nch - ncl + 1;
	int **m = (int **)calloc(nrows, sizeof(int *));
	if (m == (int **)0)
		return (int **)0;
	m -= nrl;
	m[nrl] = (int *)calloc(nrows, ncols*sizeof(int));
	if (m[nrl] == (int *)0)
		return (int **)0;
	m[nrl] -= ncl;
	for(i = nrl; i < nrh; ++i)
		m[i+1] = m[i] + ncols;
	return m;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
float **matrix(int nrl, int nrh, int ncl, int nch)
#else	/* old style */
float **matrix(nrl, nrh, ncl, nch)
int nrl, nrh, ncl, nch;
#endif
{
	int i;
	size_t nrows = nrh - nrl + 1;
	size_t ncols = nch - ncl + 1;
	float **m = (float **)calloc(nrows, sizeof(float *));
	if (m == (float **)0)
		return (float **)0;
	m -= nrl;
	m[nrl] = (float *)calloc(nrows, ncols*sizeof(float));
	if (m[nrl] == (float *)0)
		return (float **)0;
	m[nrl] -= ncl;
	for(i = nrl; i < nrh; ++i)
		m[i+1] = m[i] + ncols;
	return m;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
float **fmatrix(int nrl, int nrh, int ncl, int nch)
#else	/* old style */
float **fmatrix(nrl, nrh, ncl, nch)
int nrl, nrh, ncl, nch;
#endif
{
	int i;
	size_t nrows = nrh - nrl + 1;
	size_t ncols = nch - ncl + 1;
	char *p = (char *)0;
	void **m = (void **) calloc(
		nrows * sizeof(void *) +       /* for pointer array */
		nrows * ncols * sizeof(float), /* for float elements */
		sizeof(char));

	if (m == (void **)0)
		return (float **)m;
	
	/* link rows to pointers */
	/* first skip pointer array */
	p = (char *) m + nrows * sizeof(void *);
	/* backspace nrl pointers */
	m -= nrl;	/* allocation begins at &m[nrl] */
	/* backspace ncl float words */
	p -= ncl * sizeof(float);
	/* assign pointers to beginning of each row */
	for (i = nrl; i <= nrh; ++i) {
		m[i] = p;
		p += ncols * sizeof(float);
	}

	return (float **) m;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double **dmatrix(int nrl, int nrh, int ncl, int nch)
#else	/* old style */
double **dmatrix(nrl, nrh, ncl, nch)
int nrl, nrh, ncl, nch;
#endif
{
	int i;
	size_t nrows = nrh - nrl + 1;
	size_t ncols = nch - ncl + 1;
	double **m = (double **)calloc(nrows, sizeof(double *));
	if (m == (double **)0)
		return (double **)0;
	m -= nrl;
	m[nrl] = (double *)calloc(nrows, ncols*sizeof(double));
	if (m[nrl] == (double *)0)
		return (double **)0;
	m[nrl] -= ncl;
	for(i = nrl; i < nrh; ++i)
		m[i+1] = m[i] + ncols;
	return m;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
COMPLEX **cmatrix(int nrl, int nrh, int ncl, int nch)
#else	/* old style */
COMPLEX **cmatrix(nrl, nrh, ncl, nch)
int nrl, nrh, ncl, nch;
#endif
{
	int i;
	size_t nrows = nrh - nrl + 1;
	size_t ncols = nch - ncl + 1;
	COMPLEX **m = (COMPLEX **)calloc(nrows, sizeof(COMPLEX *));
	if (m == (COMPLEX **)0)
		return (COMPLEX **)0;
	m -= nrl;
	m[nrl] = (COMPLEX *)calloc(nrows, ncols*sizeof(COMPLEX));
	if (m[nrl] == (COMPLEX *)0)
		return (COMPLEX **)0;
	m[nrl] -= ncl;
	for(i = nrl; i < nrh; ++i)
		m[i+1] = m[i] + ncols;
	return m;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_ivector(int *v, int nl)
#else	/* old style */
void free_ivector(v, nl)
int *v, nl;
#endif
{
	free(v + nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_vector(float *v, int nl)
#else	/* old style */
void free_vector(v, nl)
float *v;
int nl;
#endif
{
	free(v + nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_dvector(double *v, int nl)
#else	/* old style */
void free_dvector(v, nl)
double *v;
int nl;
#endif
{
	free(v + nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_cvector(COMPLEX *v, int nl)
#else	/* old style */
void free_cvector(v, nl)
COMPLEX *v;
int nl;
#endif
{
	free(v + nl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_imatrix(int **m, int nrl, int ncl)
#else	/* old style */
void free_imatrix(m, nrl, ncl)
int **m, nrl, ncl;
#endif
{
	free(m[nrl] + ncl);
	free(m + nrl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_matrix(float **m, int nrl, int ncl)
#else	/* old style */
void free_matrix(m, nrl, ncl)
float **m;
int nrl, ncl;
#endif
{
	free(m[nrl] + ncl);
	free(m + nrl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_dmatrix(double **m, int nrl, int ncl)
#else	/* old style */
void free_dmatrix(m, nrl, ncl)
double **m;
int nrl, ncl;
#endif
{
	free(m[nrl] + ncl);
	free(m + nrl);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void free_cmatrix(COMPLEX **m, int nrl, int ncl)
#else	/* old style */
void free_cmatrix(m, nrl, ncl)
COMPLEX **m;
int nrl, ncl;
#endif
{
	free(m[nrl] + ncl);
	free(m + nrl);
}
/*----------------------------------------------------------------------*/
#define	SWAP(a, b) tempr = (a); (a) = (b); (b) = tempr

#if	defined(ANSI)	/* ANSI-C */
void fourn(float data[], const int nn[], const int ndim, const int isign)
#else	/* old style */
void fourn(data, nn, ndim, isign)
float data[];
const int nn[], ndim, isign;
#endif
{
	int i1, i2, i3, i2rev, i3rev, ip1, ip2, ip3, ifp1, ifp2;
	int ibit, idim, k1, k2, n, nprev, nrem, ntot;
	float tempi, tempr;
	double theta, wi, wpi, wpr, wr, wtemp;

	ntot=1;
	for (idim = 1; idim <= ndim; idim++)
		ntot *= nn[idim];
	nprev = 1;
	for (idim = ndim; idim >= 1; idim--) {
		n = nn[idim];
		nrem = ntot/(n*nprev);
		ip1 = nprev << 1;
		ip2 = ip1*n;
		ip3 = ip2*nrem;
		i2rev = 1;
		for (i2 = 1; i2 <= ip2; i2 += ip1) {
			if (i2 < i2rev) {
				for (i1 = i2; i1 <= i2+ip1-2; i1 += 2) {
					for (i3 = i1; i3 <= ip3; i3 += ip2) {
						i3rev = i2rev + i3 - i2;
						SWAP(data[i3], data[i3rev]);
						SWAP(data[i3+1], data[i3rev+1]);
					}
				}
			}
			ibit = ip2 >> 1;
			while (ibit >= ip1 && i2rev > ibit) {
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1 = ip1;
		while (ifp1 < ip2) {
			ifp2 = ifp1 << 1;
			theta = isign*6.28318530717959/(ifp2/ip1);
			wtemp = sin(0.5*theta);
			wpr = -2.0*wtemp*wtemp;
			wpi = sin(theta);
			wr = 1.0;
			wi = 0.0;
			for (i3 = 1; i3 <= ifp1; i3 += ip1) {
				for (i1 = i3; i1 <= i3+ip1-2; i1 += 2) {
					for (i2 = i1; i2 <= ip3; i2 += ifp2) {
						k1 = i2;
						k2 = k1 + ifp1;
						tempr = (float)(wr*data[k2] - wi*data[k2+1]);
						tempi = (float)(wr*data[k2+1] + wi*data[k2]);
						data[k2] = data[k1] - tempr;
						data[k2+1] = data[k1+1] - tempi;
						data[k1] += tempr;
						data[k1+1] += tempi;
					}
				}
				wr = (wtemp = wr)*wpr - wi*wpi + wr;
				wi = wi*wpr + wtemp*wpi + wi;
			}
			ifp1 = ifp2;
		}
		nprev *= n;
	}
}
#undef	SWAP
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void tridiag(
const double a[],
const double b[],
const double c[],
const double r[],
int n,
double u[])
#else	/* old style */
void tridiag(a, b, c, r, n, u)
const double a[], b[], c[], r[];
double u[];
int n;
#endif
{
	int j;
	double bet, *gam;

	gam = dvector(1, n);
	u[1] = r[1]/(bet = b[1]);
	for (j = 2; j <= n; j++) {
		gam[j] = c[j-1]/bet;
		bet = b[j] - a[j]*gam[j];
		u[j] = (r[j] - a[j]*u[j-1])/bet;
	}
	for (j = (n-1); j >= 1; j--)
		u[j] -= gam[j+1]*u[j+1];
	free_dvector(gam, 1);
}
/*----------------------------------------------------------------------
	nakspl is modified so as to return an error message if n < 5,
	or if some of the points, x[i], are equal or out of sequence.
	Bjarne Thomsen, 31-May-93
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *nakspl(
const double x[],
const double y[],
int n,
double z[])
#else	/* old style */
char *nakspl(x, y, n, z)
const double x[], y[];
int n;
double z[];
#endif
/*
	Given arrays x[1..n] and y[1..n] containing a tabulated function,
	i.e. yi = f(xi), with x1 < x2 < .... < xn. This routine returns an
	array z[1..n] that contains the second derivatives of the interpolating
	function at the tabulated points xi. It uses the not-a-knot boundary
	condition. n must not be less than 5.
*/
{
	int i;
	double *a, *b, *c, *r, p1, q1, pn, qn;

	/* Error check */
	if (n < 5) return nrerror("Too few points in NAKSPL");
	for (i = 1; i < n; ++i) if (x[i+1] <= x[i])
		return nrerror("Points are equal or out of sequence in NAKSPL");

	a = dvector(1, n-2);
	b = dvector(1, n-2);
	c = dvector(1, n-2);
	r = dvector(1, n-2);
	p1 = (x[3] - x[1])/(x[3] - x[2]);
	q1 = (x[1] - x[2])/(x[3] - x[2]);
	pn = (x[n-1] - x[n])/(x[n-1] - x[n-2]);
	qn = (x[n] - x[n-2])/(x[n-1] - x[n-2]);
	/* the lower boundary condition */
	b[1] = (x[3] - x[1])/3.0 + (x[2] - x[1])*p1/6.0;
	c[1] = (x[3] - x[2] + (x[2] - x[1])*q1)/6.0;
	r[1] = (y[3] - y[2])/(x[3] - x[2]) - (y[2] - y[1])/(x[2] - x[1]);
	for (i = 3; i <= (n-2); ++i) {
		a[i-1] = (x[i] - x[i-1])/6.0;
		b[i-1] = (x[i+1] - x[i-1])/3.0;
		c[i-1] = (x[i+1] - x[i])/6.0;
		r[i-1] = (y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
	}
	/* the upper boundary condition */
	a[n-2] = (x[n-1] - x[n-2] + (x[n] - x[n-1])*pn)/6.0;
	b[n-2] = (x[n] - x[n-2])/3.0 + (x[n] - x[n-1])*qn/6.0;
	r[n-2] = (y[n]-y[n-1])/(x[n]-x[n-1]) - (y[n-1]-y[n-2])/(x[n-1]-x[n-2]);
	tridiag(a, b, c, r, n-2, z+1);
	z[1] = p1*z[2] + q1*z[3];
	z[n] = pn*z[n-2] + qn*z[n-1];
	free_dvector(a, 1);
	free_dvector(b, 1);
	free_dvector(c, 1);
	free_dvector(r, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------
	spline is modified so as to return an error message if n < 3,
	or if some of the points, x[i], are equal or out of sequence.
	Bjarne Thomsen, 31-May-93
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *spline(
const double x[],
const double y[],
int n,
double yp1,
double ypn,
double y2[])
#else	/* old style */
char *spline(x, y, n, yp1, ypn, y2)
const double x[], y[];
int n;
double yp1, ypn, y2[];
#endif
/*
	Given arrays x[1..n] and y[1..n] containing a tabulated function,
	i.e. yi = f(xi), with x1 < x2 < .... < xn, and given values yp1 and
	ypn for the first derivative of the interpolating function at points
	1 and n, respectively, this routine returns an array y2[1..n] that
	contains the second derivatives of the interpolating function at the
	tabulated points xi. If yp1 and/or ypn are equal to 1x10^30 or larger,
	the routine is signalled to set the corresponding boundary condition
	for a natural spline, with zero second derivative on the boundary.
	n must not be less than 3.
*/
{
	int i, k;
	double p, qn, sig, un, *u;

	/* Error check */
  if (n < 3) return nrerror("Too few points in SPLINE");
  for (i = 1; i < n; ++i) if (x[i+1] <= x[i])
		return nrerror("Points are equal or out of sequence in SPLINE");

	u = dvector(1, n-1);
	if (yp1 > 0.99e30)	/* the lower boundary condition */
		y2[1] = u[1] = 0.0;	/* is set either to be natural */
	else {	/* or else to have a specified first derivative */
		y2[1] = -0.5;
		u[1] = (3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}
	for (i = 2; i < n; ++i) {
		sig = (x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p = sig*y2[i-1] + 2.0;
/*
		y2[i] = (sig - 1.0)/p;
*/
		u[i] = (y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i] = (6.0*u[i]/(x[i+1]-x[i-1]) - sig*u[i-1])/p;
/*
	Timothy Prince (The C Users Journal, Vol. 10, No. 2, p. 99)
	suggests the assignment of y2[i] be moved to the end of loop.
*/
		y2[i] = (sig - 1.0)/p;
	}
	if (ypn > 0.99e30)	/* the upper boundary condition */
		qn = un = 0.0;	/* is set either to be national */
	else {	/* or else to have a specified first derivative */
		qn = 0.5;
		un = (3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
	}
	y2[n] = (un-qn*u[n-1])/(qn*y2[n-1]+1.0);
	for (k = n-1; k > 0; --k)
		y2[k] = y2[k]*y2[k+1] + u[k];
	free_dvector(u, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------
	It is important to understand that the routine spline is called only
	once to process an entire tabulated function in arrays x[i] and y[i].
	Once this has been done, values of the interpolated function for any
	value of x are obtained by calls (as many as desired) to the separate
	function splint (for "spline interpolation").
	splint is modified in order to return the interpolated value as a
	function value instead of a referenced variable.
	Bjarne Thomsen, 31-May-93
------------------------------------------------------------------------*/
static int Klo = 1;	/* first index of previous interval */
static int Khi = 2;	/* last index of previous interval */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double splint(
const double xa[],
const double ya[],
const double y2a[],
int n,
double x)
#else	/* old style */
double splint(xa, ya, y2a, n, x)
const double xa[], ya[], y2a[];
int n;
double x;
#endif
/*
	Given the arrays xa[1..n] and ya[1..n], which tabulate a function (with
	the xai's in order), and given the array y2a[1..n], which is the output
	from spline above, and given a value of x, this routine returns a
	cubic-spline interpolated value of y.
*/
{
	double h, a, b, y;

	if (Khi > n || x > xa[Khi] || x < xa[Klo]) {
		Klo = 1;
		Khi = n;
		while (Khi-Klo > 1) {
			int k = (Khi+Klo) >> 1;
			if (xa[k] > x) Khi = k; else Klo = k;
		}
	}
	h = xa[Khi] - xa[Klo];
/*
	a = (xa[Khi]-x)/h;
	b = (x-xa[Klo])/h;
	y=a*ya[Klo]+b*ya[Khi]+((a*a*a-a)*y2a[Klo]+(b*b*b-b)*y2a[Khi])*(h*h)/6.0;

	Timothy Prince (The C Users Journal, Vol. 10, No. 2, p. 99)
	suggests the following replacement code:

	b = x - xa[Klo];
	a = xa[Khi] - x;
	y = ((ya[Klo] - y2a[Klo]/6.0*b*(a+h))*a +
			(ya[Khi] - y2a[Khi]/6.0*a*(b+h))*b)/h;

	My own replacement is as follows:
*/
	b = x - xa[Klo];
	a = h - b;
	y = (a*ya[Klo]+b*ya[Khi]-((a+h)*y2a[Klo]+(b+h)*y2a[Khi])*a*b/6.0)/h;

	return y;
}
/*----------------------------------------------------------------------
   Univariate Interpolation (Improved Akima-Method)

   Hiroshi Akima
   Institute for Telecommunication Sciences
   National Telecommunications and Information Administration
   U.S. Department of Commerce
   325 Broadway, Boulder, CO 80303-3328.

   Version of 31-May-93 as modified for C by
   Bjarne Thomsen
   Institute for Physics and Astronomy
   Aarhus University
   Ny Munkegade
   DK-8000 Aarhus C
   Denmark

   The pair of routines, akima and akint, perform univariate interpolation.
   They are based on the improved method developed by Hiroshi Akima:
   "A Method of Univariate Interpolation that has the Accuracy of a
   Third-Degree Polynomial", ACM TOMS, Vol. 17, pp. 341-366, 1991.
   (The equation numbers referred to in the comments below are those
   in the paper).
   In this method, the interpolating is done by a piecewise function
   composed of a set of polynomials applicable to successive intervals
   of the given data points. This method uses third-degree polynomials.
   If x <= x[1] or x >= x[n] linear extrapolation is used.
   An error message is returned if n < 5, or if some of the points,
   x[i], are equal or out of sequence.
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *akima(
const double x[],
const double y[],
int n,
double z[])
#else	/* old style */
char *akima(x, y, n, z)
const double x[], y[];
int n;
double z[];
#endif
/*
   Given arrays x[1..n] and y[1..n] containing a tabulated function,
   i.e. yi = f(xi), with x1 < x2 < .... < xn. This routine returns an
   array z[1..3*n-1] containing the coefficients of the third-degree
   polynomial for each of the sub-intervals: (x[i], x[i+1]), and the
   two slopes used to extrapolate the function outside the end-points.
   n must not be less than 5.
*/
{
   int i;

   /* Error check */
   if (n < 5) return nrerror("Too few points in AKIMA");
   for (i = 1; i < n; ++i) if (x[i+1] <= x[i])
      return nrerror("Points are equal or out of sequence in AKIMA");
/*
   Subcase 1 -- Calculates the slope, z[3*n-2], used to extrapolate
                the function f(x) when x <= x[1].
*/
   {
      double x0 = x[1];
      double x1 = x[2] - x0;
      double x2 = x[3] - x0;
      double x3 = x[4] - x0;
      double y0 = y[1];
      double y1 = y[2] - y0;
      double y2 = y[3] - y0;
      double y3 = y[4] - y0;
      double r1 = x2*x3;
      double r2 = x3*x1;
      double r3 = x1*x2;
      double dlt = (x2-x1)*(x3-x2)*(x3-x1)*x1*r1;
      z[3*n-2] = ((x3-x2)*r1*r1*y1+(x1-x3)*r2*r2*y2+(x2-x1)*r3*r3*y3)/dlt;
   }
/*
   Subcase 2 -- Calculates the slope, z[3*n-1], used to extrapolate
                the function f(x) when x >= x[n].
*/
   {
      double x0 = x[n];
      double x1 = x[n-1] - x0;
      double x2 = x[n-2] - x0;
      double x3 = x[n-3] - x0;
      double y0 = y[n];
      double y1 = y[n-1] - y0;
      double y2 = y[n-2] - y0;
      double y3 = y[n-3] - y0;
      double r1 = x2*x3;
      double r2 = x3*x1;
      double r3 = x1*x2;
      double dlt = (x2-x1)*(x3-x2)*(x3-x1)*x1*r1;
      z[3*n-1] = ((x3-x2)*r1*r1*y1+(x1-x3)*r2*r2*y2+(x2-x1)*r3*r3*y3)/dlt;
   }
/*
   Subcase 3 -- Calculates the coefficients z[i], z[n-1+i], and z[n+n-2+i]
                of the third-degree polynomial used to interpolate the
                function f(x) when x[i] <= x < x[i+1].
*/
   for (i = 1; i < n; ++i) {
      double yp[2];
      int iept;
/*
   The outermost for-loop with respect to the two end-points of
   the interval (x[i], x[i+1]).
*/
      for (iept = 0; iept < 2; ++iept) {
/*
   Calculates the estimate of the first derivative at an endpoint.
   Initial assignments for the calculation.
*/
         int i0 = i + iept;
         double x0 = x[i0];
         double y0 = y[i0];
         double smpef = 0.0;
         double smwtf = 0.0;
         double smpei = 0.0;
         double smwti = 0.0;
         int i1 = i0 - 3;
         int i2 = i0 - 2;
         int i3 = i0 - 1;
         int ipe;
/*
   The innermost for-loop with respect to the four primary
   estimates of the first derivative.
*/
         for (ipe = 0; ipe < 4; ++i1, ++i2, ++i3, ++ipe) {
/*
   Selects point numbers of four consecutive data points for
   calculating the primary estimate of the first derivative.
*/
            double x1, x2, x3, y1, y2, y3, r1, r2, r3, dlt;
            double pe, sx, sy, sxx, sxy, dnm, b0, b1;
            double dy0, dy1, dy2, dy3, vol, epsln;
            if (i1 == i0) ++i1;
            if (i2 == i0) ++i2;
            if (i3 == i0) ++i3;
/*
   Checks if any point number falls outside the legitimate range
   (between 1 and n).  Skips calculation of the primary estimate
   if any does.
*/
            if (i1<1 || i2<1 || i3<1 || i1>n || i2>n || i3>n)
               continue;
/*
   Calculates the primary estimate of the first derivative.
   --  cf. Equation (8)
*/
            x1 = x[i1] - x0;
            x2 = x[i2] - x0;
            x3 = x[i3] - x0;
            y1 = y[i1] - y0;
            y2 = y[i2] - y0;
            y3 = y[i3] - y0;
            r1 = x2*x3;
            r2 = x3*x1;
            r3 = x1*x2;
            dlt = (x2-x1)*(x3-x2)*(x3-x1)*x1*r1;
            pe = ((x3-x2)*r1*r1*y1+(x1-x3)*r2*r2*y2+(x2-x1)*r3*r3*y3)/dlt;
/*
   Calculates the volatility factor, vol, and distance factor,
   sxx, for the primary estimate.  --  cf. Equations (9) and (11)
*/
            sx = x1 + x2 + x3;
            sy = y1 + y2 + y3;
            sxx = x1*x1 + x2*x2 + x3*x3;
            sxy = x1*y1 + x2*y2 + x3*y3;
            dnm = 4.0*sxx - sx*sx;
            b0 = (sxx*sy - sx*sxy)/dnm;
            b1 = (4.0*sxy - sx*sy)/dnm;
            dy0 = -b0;
            dy1 = y1 - (b0 + b1*x1);
            dy2 = y2 - (b0 + b1*x2);
            dy3 = y3 - (b0 + b1*x3);
            vol = dy0*dy0 + dy1*dy1 + dy2*dy2 + dy3*dy3;
/*
   Calculates the epsln value, which is used to decide whether or
   not the volatility factor, vol, is essentially zero.
*/
            r1 = y[i1];
            r2 = y[i2];
            r3 = y[i3];
            epsln = (y0*y0 + r1*r1 + r2*r2 + r3*r3)*1.0e-12;
/*
   Accumulates the weighted primary estimates.
   --  cf. Equations (13) and (14)
*/
            if (vol > epsln) { /* - For finite weight. */
               double wt = 1.0/(vol*sxx);
               smpef += pe*wt;
               smwtf += wt;
            } else { /* - For infinite weight. */
               smpei += pe;
               smwti += 1.0;
            }
         }
/*
   End of the innermost (ipe = 0,3) for-loop.
*/
/*
   Calculates the final estimate of the first derivative.
   --  cf. Equation (14)
   No infinite weights exist. - / - Infinite weights exist.
*/
         yp[iept] = (smwti < 0.5) ? smpef/smwtf : smpei/smwti; 
/*
   End of the calculation of the estimate of the first
   derivative at an end-point.
*/
      }
/*
   End of the outermost (iept = 0,1) for-loop.
*/
/*
   Calculates the coefficients of the third-degree polynomial.
   --  cf. Equation (4)
*/
      {
         double dx = x[i+1] - x[i];
         double dy = y[i+1] - y[i];
         z[i] = yp[0];
         yp[1] -= yp[0];
         yp[0] -= dy/dx;
         z[n-1+i] = -(3.0*yp[0] + yp[1])/dx;
         z[n+n-2+i] =  (2.0*yp[0] + yp[1])/(dx*dx);
      }
   }
   return (char *)0;
}
/*----------------------------------------------------------------------
   The function akima is called only once to process an entire tabulated
   function in arrays x[i] and y[i]. Once this has been done, values of
   the interpolated function for any value of x are obtained by calling
   a separate function akint for "Akima Interpolation".
   Bjarne Thomsen, 31-May-93
------------------------------------------------------------------------*/
static int Index = 0;	/* Index of previous interval */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double akint(
const double xa[],
const double ya[],
const double z[],
int n,
double x)
#else	/* old style */
double akint(xa, ya, z, n, x)
const double xa[], ya[], z[];
int n;
double x;
#endif
/*
   Given the arrays xa[1..n] and ya[1..n], which tabulate a function
   (with the xa[i]'s in order), and given the array z[1..3*n-1] which
   is the output of akima above, and given a value of x, this function
   returns an interpolated value y.
*/
{
   double y;

   /* Locate interval (by binary search). */
   if (x <= xa[1])   /* below first interval */
      Index = 0;
   else if (x >= xa[n])   /* above last interval */
      Index = n;
   else if (Index <= 0 || Index >= n || x < xa[Index] || x >= xa[Index+1]) {
      int index1 = 1;
      int index2 = n;
      while (index2-index1 > 1) {
         int k = (index1+index2) >> 1;
         if (xa[k] > x) index2 = k; else index1 = k;
      }
      Index = index1;
   }

   /* Interpolation or extrapolation in one of the three subcases. */
   if (Index <= 0) {
/*
   Subcase 1 -- Linear extrapolation when the abscissa of the
                desired point is equal to that of the first
                data point or less. Evaluates the y value.
*/
      y = (x - xa[1])*z[n+n+n-2] + ya[1];
   } else if (Index >= n) {
/*
   Subcase 2 -- Linear extrapolation when the abscissa of the
                desired point is equal to that of the last data
                point or greater. Evaluates the y value.
*/
      y = (x - xa[n])*z[n+n+n-1] + ya[n];
   } else {
/*
   Subcase 3 -- Interpolation when the abscissa of the desired
                point is  between those of the first and last
                data points. Evaluate the y value with a third-
                degree polynomial.  --  cf. Equation (3)
*/
      int index1 = Index;
      int index2 = index1 + (n - 1);
      int index3 = index2 + (n - 1);
      double xx = x - xa[index1];
      y = ((z[index3]*xx + z[index2])*xx + z[index1])*xx + ya[index1];
   }

   return y;
}
/*----------------------------------------------------------------------*/
#define SWAP(a,b) {double temp = (a); (a) = (b); (b) = temp;}
#if	defined(ANSI)	/* ANSI-C */
char *gaussj(double *a[], int n, double *b[], int m)
#else	/* old style */
char *gaussj(a, n, b, m)
double *a[], *b[];
int n, m;
#endif
{
	int *indxc, *indxr, *ipiv;
	int i, icol, irow, j, k, l, ll;
	double big, dum, pivinv;
	indxc = ivector(1, n);
	indxr = ivector(1, n);
	ipiv = ivector(1, n);
	for (j = 1; j <= n; j++)
		ipiv[j] = 0;
	for (i = 1; i <= n; i++) {
		big = 0.0;
		for (j = 1; j <= n; j++)
			if (ipiv[j] != 1)
				for (k = 1; k <= n; k++)
					if (ipiv[k] == 0) {
						if (a[j][k] >= big) {
							big = a[j][k];
							irow = j;
							icol = k;
						} else if (-a[j][k] >= big) {
							big = -a[j][k];
							irow = j;
							icol = k;
						}
					} else if (ipiv[k] > 1)
						return nrerror("Singular Matrix in gaussj-1");
		++(ipiv[icol]);
		if (irow != icol) {
			for (l = 1; l <= n; l++)
				SWAP(a[irow][l], a[icol][l])
			for (l = 1; l <= m; l++)
				SWAP(b[irow][l], b[icol][l])
		}
		indxr[i] = irow;
		indxc[i] = icol;
		if (a[icol][icol] == 0.0)
			return nrerror("Singular Matrix in gaussj-2");
		pivinv = 1.0/a[icol][icol];
		a[icol][icol] = 1.0;
		for (l = 1; l <= n; l++)
			a[icol][l] *= pivinv;
		for (l = 1; l <= m; l++)
			b[icol][l] *= pivinv;
		for (ll = 1; ll <= n; ll++)
			if (ll != icol) {
				dum = a[ll][icol];
				a[ll][icol] = 0.0;
				for (l = 1; l <= n; l++)
					a[ll][l] -= a[icol][l]*dum;
				for (l = 1; l <= m; l++)
					b[ll][l] -= b[icol][l]*dum;
			}
	}
	for (l = n; l >= 1; l--)
		if (indxr[l] != indxc[l])
			for (k = 1; k <= n; k++)
				SWAP(a[k][indxr[l]], a[k][indxc[l]]);
	free_ivector(ipiv, 1);
	free_ivector(indxr, 1);
	free_ivector(indxc, 1);
	return (char *)0;
}
#undef SWAP
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void covsrt(
double *covar[],
int ma,
const int lista[],
int mfit)
#else	/* old style */
void covsrt(covar, ma, lista, mfit)
double *covar[];
const int lista[];
int ma, mfit;
#endif
{
	int i, j;
	double swap;
	for (j = 1; j < ma; j++)
		for (i = j+1; i <= ma; i++)
			covar[i][j] = 0.0;
	for (i = 1; i < mfit; i++)
		for (j = i+1; j <= mfit; j++)
			if (lista[j] > lista[i])
				covar[lista[j]][lista[i]] = covar[i][j];
			else
				covar[lista[i]][lista[j]] = covar[i][j];
	swap = covar[1][1];
	for (j = 1; j <= ma; j++) {
		covar[1][j] = covar[j][j];
		covar[j][j] = 0.0;
	}
	covar[lista[1]][lista[1]] = swap;
	for (j = 2; j <= mfit; j++)
		covar[lista[j]][lista[j]] = covar[1][j];
	for (j = 2; j <= ma; j++)
		for (i = 1; i <= j-1; i++)
			covar[i][j] = covar[j][i];
}
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq2cof(
float * const *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int iyc,
int irad,
const double a[],
int ma,
const int lista[],
int mfit,
int *npt,
double *alpha[],
double beta[],
double *chisq,
void (*funcs)(int, int, const double a[], double *z, double dzda[]))
#else	/* old style */
char *mrq2cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
	a, ma, lista, mfit, npt, alpha, beta, chisq, funcs)
float * const *z;
float blank;
const double a[];
double skymean, skyvar, epdu, *alpha[], beta[], *chisq;
const int lista[];
int ixc, iyc, irad, ma, mfit, *npt;
void (*funcs)();
#endif
{
	int y, x, k, j, dr2, dy2, rad2 = irad*irad;
	double zmod, wt, sig2yx, dz, *dzda;
	const float FltMax = (float) (0.9*blank);
	dzda = dvector(1, ma);
	if (dzda == (double *)0)
		return nrerror("Allocation error in mrq2cof");
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= j; k++)
			alpha[j][k] = 0.0;
		beta[j] = 0.0;
	}
	(*chisq) = 0.0;
	(*npt) = 0;
	for (y = (iyc - irad); y <= (iyc + irad); y++) {
		dy2 = y - iyc;
		dy2 *= dy2;
		for (x = (ixc - irad); x <= (ixc + irad); x++) {
			dr2 = x - ixc;
			dr2 *= dr2;
			dr2 += dy2;
			if (dr2 <= rad2) {
				(*funcs)(x, y, a, &zmod, dzda);
				if (z[y][x] <= FltMax) {
					sig2yx = skyvar;
					/* use equal weights if epdu < 0 */
					if (epdu > 0.0 && zmod > 0.0) sig2yx += zmod/epdu;
					sig2yx = 1.0/sig2yx;
					dz = z[y][x] - (zmod + skymean);
				} else {
					sig2yx = 0.0;
					dz = 0.0;
				}
				for (j = 1; j <= mfit; j++) {
					wt = dzda[lista[j]]*sig2yx;
					for (k = 1; k <= j; k++)
						alpha[j][k] += wt*dzda[lista[k]];
					beta[j] += dz*wt;
				}
				(*chisq) += dz*dz*sig2yx;
				(*npt)++;
			}
		}
	}
	for (j = 2; j <= mfit; j++)
		for (k = 1; k <= j - 1; k++)
			alpha[k][j] = alpha[j][k];
	free_dvector(dzda, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------
	Levenberg-Marquardt method, attempting to reduce the value chisq of
	a fit between a set of data points with individual standard deviations
	and a nonlinear function dependent on ma coefficients. The program
	returns current best-fit values for the parameters and chisq.
	On the first call provide an initial guess for the parameters, and
	set alamda < 0 for initialization (which then sets alamda = 0.001).
	If a step succeeds chisq becomes smaller and alamda decreases by a
	factor of 10. If a step fails alamda grows by a factor of 10.
	You must call this routine repeatedly until convergence is achieved.
	Then, make one final call with alamda = 0.0.
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq2min(
float * const *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int iyc,
int irad,
double a[],
int ma,
int lista[],
int mfit,
int *npt,
double *covar[],
double *alpha[],
double *chisq,
void (*funcs)(int, int, const double a[], double *z, double dzda[]),
double *alamda)
#else	/* old style */
char *mrq2min(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
	a, ma, lista, mfit, npt, covar, alpha, chisq, funcs, alamda)
float * const *z;
float blank;
double skymean, skyvar, epdu, a[], *covar[], *alpha[], *chisq, *alamda;
int ixc, iyc, irad, ma, lista[], mfit, *npt;
void (*funcs)();
#endif
{
	char *str;
	int k, kk, j, ihit;
	static double *da, *atry, **oneda, *beta, ochisq;
	if ((*alamda) < 0.0) {	/* initialization */
		if ((oneda = dmatrix(1, mfit, 1, 1)) == (double **)0)
			return nrerror("Allocation error in mrq2min-1");
		if ((atry = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq2min-2");
		if ((da = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq2min-3");
		if ((beta = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq2min-4");
		kk = mfit + 1;
		for (j = 1; j <= ma; j++) {	/* proper permutation? */
			ihit = 0;
			for (k = 1; k <= mfit; k++)
				if (lista[k] == j)
					ihit++;
			if (ihit == 0) {
				lista[kk] = j;
				kk++;
			} else if (ihit > 1)
				return nrerror(
					"Bad lista permutation in mrq2min-1");
		}
		if (kk != ma + 1)
			return nrerror("Bad lista permutation in mrq2min-2");
		(*alamda) = 0.001;
		str = mrq2cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
			a, ma, lista, mfit, npt, alpha, beta, chisq, funcs);
		if (str != (char *)0)
			return str;
		ochisq = (*chisq);
	}
	for (j = 1; j <= mfit; j++) {	/* augment diagonal elements */
		for (k = 1; k <= mfit; k++)
			covar[j][k] = alpha[j][k];
		covar[j][j] = alpha[j][j]*(1.0 + (*alamda));
		oneda[j][1] = beta[j];
	}
	if ((str = gaussj(covar, mfit, oneda, 1)) != (char *)0)
		return str;
	for (j = 1; j <= mfit; j++)
		da[j] = oneda[j][1];
	if ((*alamda) == 0.0) {	/* evaluate covariance matrix */
		covsrt(covar, ma, lista, mfit);
		free_dvector(beta, 1);
		free_dvector(da, 1);
		free_dvector(atry, 1);
		free_dmatrix(oneda, 1, 1);
		return (char *)0;
	}
	for (j = 1; j <= ma; j++)
		atry[j] = a[j];
	for (j = 1; j <= mfit; j++)	/* did the trial succeed? */
		atry[lista[j]] = a[lista[j]] + da[j];
	str = mrq2cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
		atry, ma, lista, mfit, npt, covar, da, chisq, funcs);
	if (str != (char *)0)
		return str;
	if ((*chisq) < ochisq) {	/* success, decrease alamda */
		(*alamda) *= 0.1;
		ochisq = (*chisq);
		for (j = 1; j <= mfit; j++) {	/* accept the new solution */
			for (k = 1; k <= mfit; k++)
				alpha[j][k] = covar[j][k];
			beta[j] = da[j];
			a[lista[j]] = atry[lista[j]];
		}
	} else {	/* failure, increase alamda, keep previous solution */
		(*alamda) *= 10.0;
		(*chisq) = ochisq;
	}
	return (char *)0;
}
/*----------------------------------------------------------------------*/
/*
	Two dimensional Gaussian function, which to the first order takes
	the finite pixel size into account.
*/
#define	SCALE	2.772588722239781	/* 4*ln(2) */
#if	defined(ANSI)	/* ANSI-C */
void gauss2(int x, int y, const double a[], double *z, double dzda[])
#else	/* old style */
void gauss2(x, y, a, z, dzda)
int x, y;
const double a[];
double *z, dzda[];
#endif
{
	double dx, dy, arg, ex, f0, f1, f2, f3, f4, f5, f6;
	dx = (x - a[2])/a[4];
	dy = (y - a[3])/a[4];
	arg = SCALE*(dx*dx + dy*dy);
	ex = exp(-arg);
	f0 = SCALE/6.0;
	f0 /= a[4]*a[4];
	f1 = (arg - 1.0)*f0 + 1.0;
	f2 = f1 - f0;
	f3 = f2 - f0;
	f4 = f1*ex;
	f5 = 2.0*ex*a[1]/a[4];
	f6 = SCALE*f2*f5;
	(*z) = f4*a[1];
	dzda[1] = f4;
	dzda[2] = f6*dx;
	dzda[3] = f6*dy;
	dzda[4] = (f3*arg + f0)*f5;
}
#undef	SCALE
/*----------------------------------------------------------------------*/
#define MA	4
#define IT	3
#define CL	9.0
#if	defined(ANSI)	/* ANSI-C */
char *g2fit(
float * const *image,
float blank,
double epdu,
int irad1,
int irad2,
int irad3,
double *xc,
double *yc,
double *fwhm,
double *dxc,
double *dyc,
double *dfwhm)
#else	/* old style */
char *g2fit(image, blank, epdu, irad1, irad2, irad3, xc, yc, fwhm,
	dxc, dyc, dfwhm)
float * const *image;
float blank;
double epdu, *xc, *yc, *fwhm, *dxc, *dyc, *dfwhm;
int irad1, irad2, irad3;
#endif
{
	char *str;
	float *sky;
	double sum, dat, skymean, skyvar;
	double *a, **covar, **alpha;
	double  alamda, chisq, ochisq;
	const float FltMax = (float) (0.9*blank);
	int *lista, mfit, x, y, i, j, dr2, dy2, nsky, npt, n;
	int ixc = (int) ((*xc) + 0.5);	/* center of fitting area */
	int iyc = (int) ((*yc) + 0.5);
	int r2 = irad2*irad2;
	int r3 = irad3*irad3;
	int skysize = 2*irad3 + 1;
	skysize *= skysize;
	if ((sky = vector(1, skysize)) == (float *)0)
		return nrerror("Allocation error in g2fit-1");
	nsky = 0;
	for (y = (iyc - irad3); y <= (iyc + irad3); y++) {
		dy2 = y - iyc;
		dy2 *= dy2;
		for (x = (ixc - irad3); x <= (ixc + irad3); x++) {
			dr2 = x - ixc;
			dr2 *= dr2;
			dr2 += dy2;
			if ((dr2 > r2) && (dr2 <= r3))
				if (image[y][x] <= FltMax) {
					nsky++;
					sky[nsky] = image[y][x];
				}
		}
	}
	if (nsky <= 1)
		return nrerror("Too few sky points in g2fit-1");
	sum = 0.0;
	for (n = 1; n <= nsky; n++)
		sum += sky[n];
	skymean = sum/nsky;
	sum = 0.0;
	for (n = 1; n <= nsky; n++) {
		dat = sky[n] - skymean;
		dat *= dat;
		sum += dat;
	}
	skyvar = sum/(nsky - 1);
	for (i = 1; i <= IT; i++) {
		double lim = CL*skyvar;
		sum = 0.0;
		npt = 0;
		for(n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			if (dat <= lim) {
				npt++;
				sum += sky[n];
				sky[npt] = sky[n];
			}
		}
		nsky = npt;
		if (nsky <= 1)
			return nrerror("Too few sky points in g2fit-2");
		skymean = sum/nsky;
		sum = 0.0;
		for (n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			sum += dat;
		}
		skyvar = sum/(nsky - 1);
	}
	if ((a = dvector(1, MA)) == (double *)0)
		return nrerror("Allocation error in g2fit-2");
	a[1] = 1.0;
	a[2] = (*xc);
	a[3] = (*yc);
	a[4] = (*fwhm);
	if ((lista = ivector(1, MA)) == (int *)0)
		return nrerror("Allocation error in g2fit-3");
	for (i = 1; i <= MA; i++)
		lista[i] = i;
	if ((covar = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g2fit-4");
	if ((alpha = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g2fit-5");
	mfit = 1;	/* fit Gaussian amplitude */
	alamda = -1.0;	/* initialize mrq2min for 1-parameter solution */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	alamda = 0.0;	/* no iterations, the problem is linear */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	mfit = 3;	/* fit Gaussian amplitude and (x,y) location */
	alamda = -1.0;	/* initialize mrq2min for 3-parameter solution */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 3-parameter problem */
		ochisq = chisq;
		ixc = (int) (a[2] + 0.5);	/* recenter fitting area */
		iyc = (int) (a[3] + 0.5);	/* to nearest pixel position */
		str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss2, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 3-parameter solution */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	mfit = 4;	/* fit amplitude, (x,y) location, and FWHM */
	alamda = -1.0;	/* initialize mrq2min for 4-parameter solution */
	ixc = (int) (a[2] + 0.5);	/* recenter fitting area */
	iyc = (int) (a[3] + 0.5);	/* to nearest pixel position */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 4-parameter problem */
		ochisq = chisq;
		ixc = (int) (a[2] + 0.5);	/* recenter fitting area */
		iyc = (int) (a[3] + 0.5);	/* to nearest pixel position */
		str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss2, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 4-parameter solution */
	str = mrq2min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss2, &alamda);
	if (str != (char *)0) return str;
	if (npt <= mfit)
		return nrerror("Too few fitting points in g2fit");
	(*xc) = a[2];
	(*yc) = a[3];
	(*fwhm) = a[4];
	(*dxc) = sqrt(covar[2][2]*chisq/(npt - mfit));
	(*dyc) = sqrt(covar[3][3]*chisq/(npt - mfit));
	(*dfwhm) = sqrt(covar[4][4]*chisq/(npt - mfit));
	free_dmatrix(alpha, 1, 1);
	free_dmatrix(covar, 1, 1);
	free_ivector(lista, 1);
	free_dvector(a, 1);
	free_vector(sky, 1);
	return (char *)0;
}
#undef MA
#undef IT
#undef CL
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq1cof(
const float *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int irad,
const double a[],
int ma,
const int lista[],
int mfit,
int *npt,
double *alpha[],
double beta[],
double *chisq,
void (*funcs)(int, const double a[], double *z, double dzda[]))
#else	/* old style */
char *mrq1cof(z, blank, skymean, skyvar, epdu, ixc, irad,
	a, ma, lista, mfit, npt, alpha, beta, chisq, funcs)
const float *z;
float blank;
const double a[];
double skymean, skyvar, epdu, *alpha[], beta[], *chisq;
const int lista[];
int ixc, irad, ma, mfit, *npt;
void (*funcs)();
#endif
{
	int x, k, j;
	double zmod, wt, sig2x, dz, *dzda;
	const float FltMax = (float) (0.9*blank);
	dzda = dvector(1, ma);
	if (dzda == (double *)0)
		return nrerror("Allocation error in mrq1cof");
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= j; k++)
			alpha[j][k] = 0.0;
		beta[j] = 0.0;
	}
	(*chisq) = 0.0;
	(*npt) = 0;
	for (x = (ixc - irad); x <= (ixc + irad); x++) {
		(*funcs)(x, a, &zmod, dzda);
		if (z[x] <= FltMax) {
			sig2x = skyvar;
			/* use equal weights if epdu < 0 */
			if (epdu > 0.0 && zmod > 0.0) sig2x += zmod/epdu;
			sig2x = 1.0/sig2x;
			dz = z[x] - (zmod + skymean);
		} else {
			sig2x = 0.0;
			dz = 0.0;
		}
		for (j = 1; j <= mfit; j++) {
			wt = dzda[lista[j]]*sig2x;
			for (k = 1; k <= j; k++)
				alpha[j][k] += wt*dzda[lista[k]];
			beta[j] += dz*wt;
		}
		(*chisq) += dz*dz*sig2x;
		(*npt)++;
	}
	for (j = 2; j <= mfit; j++)
		for (k = 1; k <= j - 1; k++)
			alpha[k][j] = alpha[j][k];
	free_dvector(dzda, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq1min(
const float *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int irad,
double a[],
int ma,
int lista[],
int mfit,
int *npt,
double *covar[],
double *alpha[],
double *chisq,
void (*funcs)(int, const double a[], double *z, double dzda[]),
double *alamda)
#else	/* old style */
char *mrq1min(z, blank, skymean, skyvar, epdu, ixc, irad,
	a, ma, lista, mfit, npt, covar, alpha, chisq, funcs, alamda)
const float *z;
float blank;
double skymean, skyvar, epdu, a[], *covar[], *alpha[], *chisq, *alamda;
int ixc, irad, ma, lista[], mfit, *npt;
void (*funcs)();
#endif
{
	char *str;
	int k, kk, j, ihit;
	static double *da, *atry, **oneda, *beta, ochisq;
	if ((*alamda) < 0.0) {
		if ((oneda = dmatrix(1, mfit, 1, 1)) == (double **)0)
			return nrerror("Allocation error in mrq1min-1");
		if ((atry = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq1min-2");
		if ((da = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq1min-3");
		if ((beta = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq1min-4");
		kk = mfit + 1;
		for (j = 1; j <= ma; j++) {
			ihit = 0;
			for (k = 1; k <= mfit; k++)
				if (lista[k] == j)
					ihit++;
			if (ihit == 0) {
				lista[kk] = j;
				kk++;
			} else if (ihit > 1)
				return nrerror(
					"Bad lista permutation in mrq1min-1");
		}
		if (kk != ma + 1)
			return nrerror("Bad lista permutation in mrq1min-2");
		(*alamda) = 0.001;
		str = mrq1cof(z, blank, skymean, skyvar, epdu, ixc, irad,
			a, ma, lista, mfit, npt, alpha, beta, chisq, funcs);
		if (str != (char *)0)
			return str;
		ochisq = (*chisq);
	}
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= mfit; k++)
			covar[j][k] = alpha[j][k];
		covar[j][j] = alpha[j][j]*(1.0 + (*alamda));
		oneda[j][1] = beta[j];
	}
	if ((str = gaussj(covar, mfit, oneda, 1)) != (char *)0)
		return str;
	for (j = 1; j <= mfit; j++)
		da[j] = oneda[j][1];
	if ((*alamda) == 0.0) {
		covsrt(covar, ma, lista, mfit);
		free_dvector(beta, 1);
		free_dvector(da, 1);
		free_dvector(atry, 1);
		free_dmatrix(oneda, 1, 1);
		return (char *)0;
	}
	for (j = 1; j <= ma; j++)
		atry[j] = a[j];
	for (j = 1; j <= mfit; j++)
		atry[lista[j]] = a[lista[j]] + da[j];
	str = mrq1cof(z, blank, skymean, skyvar, epdu, ixc, irad,
		atry, ma, lista, mfit, npt, covar, da, chisq, funcs);
	if (str != (char *)0)
		return str;
	if ((*chisq) < ochisq) {
		(*alamda) *= 0.1;
		ochisq = (*chisq);
		for (j = 1; j <= mfit; j++) {
			for (k = 1; k <= mfit; k++)
				alpha[j][k] = covar[j][k];
			beta[j] = da[j];
			a[lista[j]] = atry[lista[j]];
		}
	} else {
		(*alamda) *= 10.0;
		(*chisq) = ochisq;
	}
	return (char *)0;
}
/*----------------------------------------------------------------------*/
#define	SCALE	2.772588722239781	/* 4*ln(2) */
#if	defined(ANSI)	/* ANSI-C */
void gauss1(int x, const double a[], double *z, double dzda[])
#else	/* old style */
void gauss1(x, a, z, dzda)
int x;
const double a[];
double *z, dzda[];
#endif
{
	double fun, fac, ex, arg, dx;
	dx = (x - a[2])/a[3];
	arg = (dx*dx)*SCALE;
	ex = exp(-arg);
	fun = a[1]*ex;
	fac = 2.0*SCALE*fun;
	(*z) = fun;
	dzda[1] = ex;
	dzda[2] = fac*dx/a[3];
	dzda[3] = 2.0*arg*fun/a[3];
}
#undef	SCALE
/*----------------------------------------------------------------------*/
#define MA	3
#define IT	3
#define CL	9.0
#if	defined(ANSI)	/* ANSI-C */
char *g1xfit(
float * const *image,
float blank,
double epdu,
int ix1,
int ix2,
int ix3,
double *xc,
double *yc,
double *fwhm,
double *dxc,
double *dfwhm)
#else	/* old style */
char *g1xfit(image, blank, epdu, ix1, ix2, ix3, xc, yc, fwhm, dxc, dfwhm)
float * const *image;
float blank;
double epdu, *xc, *yc, *fwhm, *dxc, *dfwhm;
int ix1, ix2, ix3;
#endif
{
	char *str;
	float *sky, *line;
	double sum, dat, skymean, skyvar;
	double *a, **covar, **alpha;
	double  alamda, chisq, ochisq;
	const float FltMax = (float) (0.9*blank);
	int *lista, mfit, x, y, i, j, nsky, npt, n;
	int ixc = (int) ((*xc) + 0.5);	/* central pixel */
	int iyc = (int) ((*yc) + 0.5);
	int nxl = ixc - ix3;	/* first pixel in row */
	int nxh = ixc + ix3; /* last pixel in row */
	int skysize = ix3 + ix3 + 1;
	skysize *= 2*(ix3 - ix2 + 1);
	if ((sky = vector(1, skysize)) == (float *)0)
		return nrerror("Allocation error in g1xfit-1");
	nsky = 0;
	for (y = (iyc - ix3); y <= (iyc + ix3); y++) {
		for (x = (ixc - ix3); x <= (ixc - ix2); x++)
			if (image[y][x] <= FltMax) {
				nsky++;
				sky[nsky] = image[y][x];
			}
		for (x = (ixc + ix2); x <= (ixc + ix3); x++)
			if (image[y][x] <= FltMax) {
				nsky++;
				sky[nsky] = image[y][x];
			}
	}
	if (nsky <= 1)
		return nrerror("Too few sky points in g1xfit-1");
	sum = 0.0;
	for (n = 1; n <= nsky; n++)
		sum += sky[n];
	skymean = sum/nsky;
	sum = 0.0;
	for (n = 1; n <= nsky; n++) {
		dat = sky[n] - skymean;
		dat *= dat;
		sum += dat;
	}
	skyvar = sum/(nsky - 1);
	for (i = 1; i <= IT; i++) {
		double lim = CL*skyvar;
		sum = 0.0;
		npt = 0;
		for(n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			if (dat <= lim) {
				npt++;
				sum += sky[n];
				sky[npt] = sky[n];
			}
		}
		nsky = npt;
		if (nsky <= 1)
			return nrerror("Too few sky points in g1xfit-2");
		skymean = sum/nsky;
		sum = 0.0;
		for (n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			sum += dat;
		}
		skyvar = sum/(nsky - 1);
	}
	if ((line = vector(nxl, nxh)) == (float *)0)
		return nrerror("Allocation error in g1xfit-2");
	for (x = nxl; x <= nxh; x++) {
		sum = 0.0;
		n = 0;
		for (y = (iyc - ix3); y <= (iyc + ix3); y++)
			if (image[y][x] <= FltMax) {	/* add rows */
				n++;
				sum += image[y][x];
			}
		if (n == (ix3 + ix3 + 1))
			line[x] = (float)sum;
		else
			line[x] = blank;	/* bad pixel in one of the rows */
	}
	skymean *= n;	/* multiply sky mean by nr of rows summed */
	skyvar *= n;	/* multiply sky variance by nr of rows summed */
	if ((a = dvector(1, MA)) == (double *)0)
		return nrerror("Allocation error in g1xfit-3");
	a[1] = 1.0;	/* amplitude */
	a[2] = (*xc);	/* location */
	a[3] = (*fwhm);	/* Full Width at Half Maximum */
	if ((lista = ivector(1, MA)) == (int *)0)
		return nrerror("Allocation error in g1xfit-4");
	for (i = 1; i <= MA; i++)
		lista[i] = i;
	if ((covar = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g1xfit-5");
	if ((alpha = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g1xfit-6");
	mfit = 1;	/* fit Gaussian amplitude, only */
	alamda = -1.0;	/* initialize mrq1min for 1-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	alamda = 0.0;	/* no iterations, the problem is linear */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	mfit = 2;	/* fit Gaussian amplitude and x-location */
	alamda = -1.0;	/* initialize mrq1min for 2-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 2-parameter problem */
		ochisq = chisq;
		ixc = (int) (a[2] + 0.5);	/* recenter fitting range */
		str = mrq1min(line, blank, skymean, skyvar, epdu, ixc,
			ix1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss1, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 2-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	mfit = 3;	/* fit amplitude, x-location, and FWHM */
	alamda = -1.0;	/* initialize mrq1min for 3-parameter solution */
	ixc = (int) (a[2] + 0.5);	/* recenter fitting range */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 3-parameter problem */
		ochisq = chisq;
		ixc = (int) (a[2] + 0.5);	/* recenter fitting range */
		str = mrq1min(line, blank, skymean, skyvar, epdu, ixc,
			ix1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss1, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 3-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, ixc, ix1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	if (npt <= mfit)
		return nrerror("Too few fitting points in g1xfit");
	(*xc) = a[2];
	(*fwhm) = a[3];
	(*dxc) = sqrt(covar[2][2]*chisq/(npt - mfit));
	(*dfwhm) = sqrt(covar[3][3]*chisq/(npt - mfit));
	free_dmatrix(alpha, 1, 1);
	free_dmatrix(covar, 1, 1);
	free_ivector(lista, 1);
	free_dvector(a, 1);
	free_vector(line, nxl);
	free_vector(sky, 1);
	return (char *)0;
}
#undef MA
#undef IT
#undef CL
/*----------------------------------------------------------------------*/
#define MA	3
#define IT	3
#define CL	9.0
#if	defined(ANSI)	/* ANSI-C */
char *g1yfit(
float * const *image,
float blank,
double epdu,
int iy1,
int iy2,
int iy3,
double *xc,
double *yc,
double *fwhm,
double *dyc,
double *dfwhm)
#else	/* old style */
char *g1yfit(image, blank, epdu, iy1, iy2, iy3, xc, yc, fwhm, dyc, dfwhm)
float * const *image;
float blank;
double epdu, *xc, *yc, *fwhm, *dyc, *dfwhm;
int iy1, iy2, iy3;
#endif
{
	char *str;
	float *sky, *line;
	double sum, dat, skymean, skyvar;
	double *a, **covar, **alpha;
	double  alamda, chisq, ochisq;
	const float FltMax = (float) (0.9*blank);
	int *lista, mfit, x, y, i, j, nsky, npt, n;
	int ixc = (int) ((*xc) + 0.5);	/* central pixel */
	int iyc = (int) ((*yc) + 0.5);
	int nyl = iyc - iy3;	/* first pixel in column */
	int nyh = iyc + iy3; /* last pixel in column */
	int skysize = iy3 + iy3 + 1;
	skysize *= 2*(iy3 - iy2 + 1);
	if ((sky = vector(1, skysize)) == (float *)0)
		return nrerror("Allocation error in g1yfit-1");
	nsky = 0;
	for (x = (ixc - iy3); x <= (ixc + iy3); x++) {
		for (y = (iyc - iy3); y <= (iyc - iy2); y++)
			if (image[y][x] <= FltMax) {
				nsky++;
				sky[nsky] = image[y][x];
			}
		for (y = (iyc + iy2); y <= (iyc + iy3); y++)
			if (image[y][x] <= FltMax) {
				nsky++;
				sky[nsky] = image[y][x];
			}
	}
	if (nsky <= 1)
		return nrerror("Too few sky points in g1yfit-1");
	sum = 0.0;
	for (n = 1; n <= nsky; n++)
		sum += sky[n];
	skymean = sum/nsky;
	sum = 0.0;
	for (n = 1; n <= nsky; n++) {
		dat = sky[n] - skymean;
		dat *= dat;
		sum += dat;
	}
	skyvar = sum/(nsky - 1);
	for (i = 1; i <= IT; i++) {
		double lim = CL*skyvar;
		sum = 0.0;
		npt = 0;
		for(n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			if (dat <= lim) {
				npt++;
				sum += sky[n];
				sky[npt] = sky[n];
			}
		}
		nsky = npt;
		if (nsky <= 1)
			return nrerror("Too few sky points in g1yfit-2");
		skymean = sum/nsky;
		sum = 0.0;
		for (n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			sum += dat;
		}
		skyvar = sum/(nsky - 1);
	}
	if ((line = vector(nyl, nyh)) == (float *)0)
		return nrerror("Allocation error in g1yfit-2");
	for (y = nyl; y <= nyh; y++) {
		sum = 0.0;
		n = 0;
		for (x = (ixc - iy3); x <= (ixc + iy3); x++)
			if (image[y][x] <= FltMax) {	/* add columns */
				n++;
				sum += image[y][x];
			}
		if (n == (iy3 + iy3 + 1))
			line[y] = (float)sum;
		else
			line[y] = blank;	/* bad pixel in one of the columns */
	}
	skymean *= n;	/* multiply sky mean by nr of columns summed */
	skyvar *= n;	/* multiply sky variance by nr of columns summed */
	if ((a = dvector(1, MA)) == (double *)0)
		return nrerror("Allocation error in g1yfit-3");
	a[1] = 1.0;	/* amplitude */
	a[2] = (*yc);	/* y-location */
	a[3] = (*fwhm);	/* FWHM */
	if ((lista = ivector(1, MA)) == (int *)0)
		return nrerror("Allocation error in g1yfit-4");
	for (i = 1; i <= MA; i++)
		lista[i] = i;
	if ((covar = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g1yfit-5");
	if ((alpha = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g1yfit-6");
	mfit = 1;	/* fit Gaussian amplitude, only */
	alamda = -1.0;	/* initialize mrq1min for 1-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	alamda = 0.0;	/* no iterations, the problem is linear */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	mfit = 2;	/* fit Gaussian amplitude and y-location */
	alamda = -1.0;	/* initialize mrq1min for 2-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 2-parameter problem */
		ochisq = chisq;
		iyc = (int) (a[2] + 0.5);	/* recenter fitting range */
		str = mrq1min(line, blank, skymean, skyvar, epdu, iyc,
			iy1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss1, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 2-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	mfit = 3;	/* fit amplitude, y-location, and FWHM */
	alamda = -1.0;	/* initialize mrq1min for 3-parameter solution */
	iyc = (int) (a[2] + 0.5);	/* recenter fitting range */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 3-parameter problem */
		ochisq = chisq;
		iyc = (int) (a[2] + 0.5);	/* recenter fitting range */
		str = mrq1min(line, blank, skymean, skyvar, epdu, iyc,
			iy1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss1, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 3-parameter solution */
	str = mrq1min(line, blank, skymean, skyvar, epdu, iyc, iy1, a,
		MA, lista, mfit, &npt, covar, alpha, &chisq, gauss1, &alamda);
	if (str != (char *)0) return str;
	if (npt <= mfit)
		return nrerror("Too few fitting points in g1yfit");
	(*yc) = a[2];
	(*fwhm) = a[3];
	(*dyc) = sqrt(covar[2][2]*chisq/(npt - mfit));
	(*dfwhm) = sqrt(covar[3][3]*chisq/(npt - mfit));
	free_dmatrix(alpha, 1, 1);
	free_dmatrix(covar, 1, 1);
	free_ivector(lista, 1);
	free_dvector(a, 1);
	free_vector(line, nyl);
	free_vector(sky, 1);
	return (char *)0;
}
#undef MA
#undef IT
#undef CL
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq3cof(
float * const *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int iyc,
int irad,
const double a[],
int ma,
const int lista[],
int mfit,
int *npt,
double *alpha[],
double beta[],
double *chisq,
void (*funcs)(int, int, const double a[], double *z, double dzda[], int))
#else	/* old style */
char *mrq3cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
	a, ma, lista, mfit, npt, alpha, beta, chisq, funcs)
float * const *z;
float blank;
const double a[];
double skymean, skyvar, epdu, *alpha[], beta[], *chisq;
const int lista[];
int ixc, iyc, irad, ma, mfit, *npt;
void (*funcs)();
#endif
{
	int y, x, k, j, dr2, dy2, rad2 = irad*irad;
	double zmod, wt, sig2yx, dz, *dzda;
	const float FltMax = (float) (0.9*blank);
	dzda = dvector(1, ma);
	if (dzda == (double *)0)
		return nrerror("Allocation error in mrq3cof");
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= j; k++)
			alpha[j][k] = 0.0;
		beta[j] = 0.0;
	}
	(*chisq) = 0.0;
	(*npt) = 0;
	for (y = (iyc - irad); y <= (iyc + irad); y++) {
		dy2 = y - iyc;
		dy2 *= dy2;
		for (x = (ixc - irad); x <= (ixc + irad); x++) {
			dr2 = x - ixc;
			dr2 *= dr2;
			dr2 += dy2;
			if (dr2 <= rad2) {
				(*funcs)(x, y, a, &zmod, dzda, ma);
				if (z[y][x] <= FltMax) {
					sig2yx = skyvar;
					/* use equal weights if epdu < 0 */
					if (epdu > 0.0 && zmod > 0.0) sig2yx += zmod/epdu;
					sig2yx = 1.0/sig2yx;
					dz = z[y][x] - (zmod + skymean);
				} else {
					sig2yx = 0.0;
					dz = 0.0;
				}
				for (j = 1; j <= mfit; j++) {
					wt = dzda[lista[j]]*sig2yx;
					for (k = 1; k <= j; k++)
						alpha[j][k] += wt*dzda[lista[k]];
					beta[j] += dz*wt;
				}
				(*chisq) += dz*dz*sig2yx;
				(*npt)++;
			}
		}
	}
	for (j = 2; j <= mfit; j++)
		for (k = 1; k <= j - 1; k++)
			alpha[k][j] = alpha[j][k];
	free_dvector(dzda, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq3min(
float * const *z,
float blank,
double skymean,
double skyvar,
double epdu,
int ixc,
int iyc,
int irad,
double a[],
int ma,
int lista[],
int mfit,
int *npt,
double *covar[],
double *alpha[],
double *chisq,
void (*funcs)(int, int, const double a[], double *z, double dzda[], int),
double *alamda)
#else	/* old style */
char *mrq3min(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
	a, ma, lista, mfit, npt, covar, alpha, chisq, funcs, alamda)
float * const *z;
float blank;
double skymean, skyvar, epdu, a[], *covar[], *alpha[], *chisq, *alamda;
int ixc, iyc, irad, ma, lista[], mfit, *npt;
void (*funcs)();
#endif
{
	char *str;
	int k, kk, j, ihit;
	static double *da, *atry, **oneda, *beta, ochisq;
	if ((*alamda) < 0.0) {
		if ((oneda = dmatrix(1, mfit, 1, 1)) == (double **)0)
			return nrerror("Allocation error in mrq3min-1");
		if ((atry = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq3min-2");
		if ((da = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq3min-3");
		if ((beta = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq3min-4");
		kk = mfit + 1;
		for (j = 1; j <= ma; j++) {
			ihit = 0;
			for (k = 1; k <= mfit; k++)
				if (lista[k] == j)
					ihit++;
			if (ihit == 0) {
				lista[kk] = j;
				kk++;
			} else if (ihit > 1)
				return nrerror(
					"Bad lista permutation in mrq3min-1");
		}
		if (kk != ma + 1)
			return nrerror("Bad lista permutation in mrq3min-2");
		(*alamda) = 0.001;
		str = mrq3cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
			a, ma, lista, mfit, npt, alpha, beta, chisq, funcs);
		if (str != (char *)0)
			return str;
		ochisq = (*chisq);
	}
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= mfit; k++)
			covar[j][k] = alpha[j][k];
		covar[j][j] = alpha[j][j]*(1.0 + (*alamda));
		oneda[j][1] = beta[j];
	}
	if ((str = gaussj(covar, mfit, oneda, 1)) != (char *)0)
		return str;
	for (j = 1; j <= mfit; j++)
		da[j] = oneda[j][1];
	if ((*alamda) == 0.0) {
		covsrt(covar, ma, lista, mfit);
		free_dvector(beta, 1);
		free_dvector(da, 1);
		free_dvector(atry, 1);
		free_dmatrix(oneda, 1, 1);
		return (char *)0;
	}
	for (j = 1; j <= ma; j++)
		atry[j] = a[j];
	for (j = 1; j <= mfit; j++)
		atry[lista[j]] = a[lista[j]] + da[j];
	str = mrq3cof(z, blank, skymean, skyvar, epdu, ixc, iyc, irad,
		atry, ma, lista, mfit, npt, covar, da, chisq, funcs);
	if (str != (char *)0)
		return str;
	if ((*chisq) < ochisq) {
		(*alamda) *= 0.1;
		ochisq = (*chisq);
		for (j = 1; j <= mfit; j++) {
			for (k = 1; k <= mfit; k++)
				alpha[j][k] = covar[j][k];
			beta[j] = da[j];
			a[lista[j]] = atry[lista[j]];
		}
	} else {
		(*alamda) *= 10.0;
		(*chisq) = ochisq;
	}
	return (char *)0;
}
/*----------------------------------------------------------------------*/
/*
	A sum of 2-dimensional Gaussians.
*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
#if	defined(ANSI)	/* ANSI-C */
void gauss3(int x, int y, const double a[],
		double *z, double dzda[], int na)
#else	/* old style */
void gauss3(x, y, a, z, dzda, na)
int x, y, na;
const double a[];
double *z, dzda[];
#endif
{
	int i;
	double ar, ex, tm;
	double dx = x - a[na-1];
	double dy = y - a[na];
	double pr = PI*(dx*dx + dy*dy);
	double fr = 0.0;
	double gr = 0.0;
	for (i = 1; i < (na-2); i += 2) {
		ar = pr*a[i+1];
		ex = exp(-ar);
		tm = ex*a[i]*a[i+1];
		fr += tm;
		gr += tm*a[i+1];
		dzda[i] = ex*a[i+1];
		dzda[i+1] = (1.0 - ar)*ex*a[i];
	}
	(*z) = fr;
	gr *= 2.0*PI;
	dzda[na-1] = gr*dx;
	dzda[na] = gr*dy;
}
/*----------------------------------------------------------------------*/
#define MA	6
#define IT	3
#define CL	9.0
#if	defined(ANSI)	/* ANSI-C */
char *g3fit(
float * const *image,
float blank,
double epdu,
int irad1,
int irad2,
int irad3,
double *a1,
double *r1,
double *a2,
double *r2,
double *xc,
double *yc)
#else	/* old style */
char *g3fit(image, blank, epdu, irad1, irad2, irad3,
	a1, r1, a2, r2, xc, yc)
float * const *image;
float blank;
double epdu, *a1, *r1, *a2, *r2, *xc, *yc;
int irad1, irad2, irad3;
#endif
{
	char *str;
	float *sky;
	double sum, dat, skymean, skyvar;
	double *a, **covar, **alpha;
	double  alamda, chisq, ochisq;
	const float FltMax = (float) (0.9*blank);
	int *lista, mfit, x, y, i, j, dr2, dy2, nsky, npt, n;
	int ixc = (int) ((*xc) + 0.5);	/* center of fitting area */
	int iyc = (int) ((*yc) + 0.5);
	int ir2 = irad2*irad2;
	int ir3 = irad3*irad3;
	int skysize = 2*irad3 + 1;
	skysize *= skysize;
	if ((sky = vector(1, skysize)) == (float *)0)
		return nrerror("Allocation error in g3fit-1");
	nsky = 0;
	for (y = (iyc - irad3); y <= (iyc + irad3); y++) {
		dy2 = y - iyc;
		dy2 *= dy2;
		for (x = (ixc - irad3); x <= (ixc + irad3); x++) {
			dr2 = x - ixc;
			dr2 *= dr2;
			dr2 += dy2;
			if ((dr2 > ir2) && (dr2 <= ir3))
				if (image[y][x] <= FltMax) {
					nsky++;
					sky[nsky] = image[y][x];
				}
		}
	}
	if (nsky <= 1)
		return nrerror("Too few sky points in g3fit-1");
	sum = 0.0;
	for (n = 1; n <= nsky; n++)
		sum += sky[n];
	skymean = sum/nsky;
	sum = 0.0;
	for (n = 1; n <= nsky; n++) {
		dat = sky[n] - skymean;
		dat *= dat;
		sum += dat;
	}
	skyvar = sum/(nsky - 1);
	for (i = 1; i <= IT; i++) {
		double lim = CL*skyvar;
		sum = 0.0;
		npt = 0;
		for(n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			if (dat <= lim) {
				npt++;
				sum += sky[n];
				sky[npt] = sky[n];
			}
		}
		nsky = npt;
		if (nsky <= 1)
			return nrerror("Too few sky points in g3fit-2");
		skymean = sum/nsky;
		sum = 0.0;
		for (n = 1; n <= nsky; n++) {
			dat = sky[n] - skymean;
			dat *= dat;
			sum += dat;
		}
		skyvar = sum/(nsky - 1);
	}
	if ((a = dvector(1, MA)) == (double *)0)
		return nrerror("Allocation error in g3fit-2");
	a[1] = (*a1);
	a[2] = 1.0/(*r1);
	a[2] *= a[2];
	a[3] = (*a2);
	a[4] = 1.0/(*r2);
	a[4] *= a[4];
	a[5] = (*xc);
	a[6] = (*yc);
	if ((lista = ivector(1, MA)) == (int *)0)
		return nrerror("Allocation error in g3fit-3");
	for (i = 1; i <= MA; i++)
		lista[i] = i;
	if ((covar = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g3fit-4");
	if ((alpha = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g3fit-5");
	mfit = 1;	/* fit amplitude of first Gaussian */
	alamda = -1.0;	/* initialize mrq3min for a 1-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	alamda = 0.0;	/* no iterations, the problem is linear */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	mfit = 2;	/* fit amplitude and width of first Gaussian */
	alamda = -1.0;	/* initialize mrq3min for a 2-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 2-parameter problem */
		ochisq = chisq;
		str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss3, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 2-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	a[4] = a[2]/4.0;	/* second Gaussian is twice as wide as first one */
	mfit = 3;	/* fit first Gaussian and amplitude of second Gaussian */
	alamda = -1.0;	/* initialize mrq3min for a 3-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 3-parameter problem */
		ochisq = chisq;
		str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss3, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 3-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	mfit = 4;	/* fit amplitudes and widths of the two Gaussians */
	alamda = -1.0;	/* initialize mrq3min for a 4-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 4-parameter problem */
		ochisq = chisq;
		str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss3, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 4-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	mfit = 6;	/* fit the two Gaussians and the (xc, yc) location */
	alamda = -1.0;	/* initialize mrq3min for a 6-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 6-parameter problem */
		ochisq = chisq;
		ixc = (int) (a[5] + 0.5);	/* recenter fitting area */
		iyc = (int) (a[6] + 0.5);	/* to nearest pixel position */
		str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc,
			irad1, a, MA, lista, mfit, &npt, covar, alpha, &chisq,
			gauss3, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 6-parameter solution */
	str = mrq3min(image, blank, skymean, skyvar, epdu, ixc, iyc, irad1,
		a, MA, lista, mfit, &npt, covar, alpha, &chisq, gauss3, &alamda);
	if (str != (char *)0) return str;
	if (npt <= mfit)
		return nrerror("Too few fitting points in g3fit");
	(*a1) = a[1];
	(*r1) = 1.0/sqrt(a[2]);
	(*a2) = a[3];
	(*r2) = 1.0/sqrt(a[4]);
	(*xc) = a[5];
	(*yc) = a[6];
	free_dmatrix(alpha, 1, 1);
	free_dmatrix(covar, 1, 1);
	free_ivector(lista, 1);
	free_dvector(a, 1);
	free_vector(sky, 1);
	return (char *)0;
}
#undef PI
#undef MA
#undef IT
#undef CL
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq4cof(
COMPLEX * const *z,	/* Fourier transform of real PSF */
double blank,
double var,	/* noise variance */
int uc,	/* (uc, vc) is the location of the zero frequency */
int vc,
int rad,	/* the fit is done to points within rad from (uc, vc) */
const double a[],
int ma,
const int lista[],
int mfit,
int *npt,
double *alpha[],
double beta[],
double *chisq,
void (*funcs)(int u, int v, const double a[], COMPLEX *z, COMPLEX dzda[]))
#else	/* old style */
char *mrq4cof(z, blank, var, uc, vc, rad, a, ma, lista, mfit,
	npt, alpha, beta, chisq, funcs)
COMPLEX * const *z;
double blank;
const double a[];
double var, *alpha[], beta[], *chisq;
const int lista[];
int uc, vc, rad, ma, mfit, *npt;
void (*funcs)();
#endif
{
	int u, v, k, j, rad2 = rad*rad;
	COMPLEX *dzda;
	const float FltMax = (float) (0.9*blank);
	const float Zero = (float)0;

	dzda = cvector(1, ma);
	if (dzda == (COMPLEX *)0)
		return nrerror("Allocation error in mrq4cof");
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= j; k++)
			alpha[j][k] = 0.0;
		beta[j] = 0.0;
	}
	(*chisq) = 0.0;
	(*npt) = 0;
	for (v = (vc - rad); v <= (vc + rad); v++) {
		int v2 = v - vc;
		v2 *= v2;
		for (u = uc; u <= (uc + rad); u++) {
			int o2 = u - uc;
			o2 *= o2;
			o2 += v2;
			if ((u > uc) || (v > vc))
				if (o2 <= rad2) {
					COMPLEX zmod, dz, zj;
					double wt;
					(*funcs)(u-uc, v-vc, a, &zmod, dzda);
					if (z[v][u].x<=FltMax && z[v][u].y<=FltMax) {
						wt = 1.0/var;
						dz.x = z[v][u].x - zmod.x;
						dz.y = z[v][u].y - zmod.y;
					} else {
						wt = 0.0;
						dz.x = Zero;
						dz.y = Zero;
					}
					for (j = 1; j <= mfit; j++) {
						zj.x = dzda[lista[j]].x;
						zj.y = dzda[lista[j]].y;
						for (k = 1; k <= j; k++)
							alpha[j][k] += wt*PROD(zj,dzda[lista[k]]);
						beta[j] += wt*PROD(zj,dz);
					}
					(*chisq) += wt*PROD(dz,dz);
					(*npt)++;
				}
		}
	}
	for (j = 2; j <= mfit; j++)
		for (k = 1; k <= j - 1; k++)
			alpha[k][j] = alpha[j][k];
	free_cvector(dzda, 1);
	return (char *)0;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *mrq4min(
COMPLEX * const *z,	/* Fourier transform of real PSF */
double blank,
double var,	/* noise variance */
int uc,	/* (uc, vc) is the location of the zero frequency */
int vc,
int rad,	/* the fit is done to points within rad of (uc, vc) */
double a[],
int ma,
int lista[],
int mfit,
int *npt,
double *covar[],
double *alpha[],
double *chisq,
void (*funcs)(int u, int v, const double a[], COMPLEX *z, COMPLEX dzda[]),
double *alamda)
#else	/* old style */
char *mrq4min(z, blank, var, uc, vc, rad, a, ma, lista, mfit,
	npt, covar, alpha, chisq, funcs, alamda)
COMPLEX * const *z;
double blank;
double var, a[], *covar[], *alpha[], *chisq, *alamda;
int uc, vc, rad, ma, lista[], mfit, *npt;
void (*funcs)();
#endif
{
	char *str;
	int k, kk, j, ihit;
	static double *da, *atry, **oneda, *beta, ochisq;
	if ((*alamda) < 0.0) {
		if ((oneda = dmatrix(1, mfit, 1, 1)) == (double **)0)
			return nrerror("Allocation error in mrq4min-1");
		if ((atry = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq4min-2");
		if ((da = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq4min-3");
		if ((beta = dvector(1, ma)) == (double *)0)
			return nrerror("Allocation error in mrq4min-4");
		kk = mfit + 1;
		for (j = 1; j <= ma; j++) {
			ihit = 0;
			for (k = 1; k <= mfit; k++)
				if (lista[k] == j)
					ihit++;
			if (ihit == 0) {
				lista[kk] = j;
				kk++;
			} else if (ihit > 1)
				return nrerror(
					"Bad lista permutation in mrq4min-1");
		}
		if (kk != ma + 1)
			return nrerror("Bad lista permutation in mrq4min-2");
		(*alamda) = 0.001;
		str = mrq4cof(z, blank, var, uc, vc, rad, a, ma, lista, mfit,
			npt, alpha, beta, chisq, funcs);
		if (str != (char *)0)
			return str;
		ochisq = (*chisq);
	}
	for (j = 1; j <= mfit; j++) {
		for (k = 1; k <= mfit; k++)
			covar[j][k] = alpha[j][k];
		covar[j][j] = alpha[j][j]*(1.0 + (*alamda));
		oneda[j][1] = beta[j];
	}
	if ((str = gaussj(covar, mfit, oneda, 1)) != (char *)0)
		return str;
	for (j = 1; j <= mfit; j++)
		da[j] = oneda[j][1];
	if ((*alamda) == 0.0) {
		covsrt(covar, ma, lista, mfit);
		free_dvector(beta, 1);
		free_dvector(da, 1);
		free_dvector(atry, 1);
		free_dmatrix(oneda, 1, 1);
		return (char *)0;
	}
	for (j = 1; j <= ma; j++)
		atry[j] = a[j];
	for (j = 1; j <= mfit; j++)
		atry[lista[j]] = a[lista[j]] + da[j];
	str = mrq4cof(z, blank, var, uc, vc, rad, atry, ma, lista, mfit,
		npt, covar, da, chisq, funcs);
	if (str != (char *)0)
		return str;
	if ((*chisq) < ochisq) {
		(*alamda) *= 0.1;
		ochisq = (*chisq);
		for (j = 1; j <= mfit; j++) {
			for (k = 1; k <= mfit; k++)
				alpha[j][k] = covar[j][k];
			beta[j] = da[j];
			a[lista[j]] = atry[lista[j]];
		}
	} else {
		(*alamda) *= 10.0;
		(*chisq) = ochisq;
	}
	return (char *)0;
}
/*----------------------------------------------------------------------
	Optical Transfer Function:
	T(u,v) = a[1]*exp(-(o/a[2])^a[5])*cexp(-2*Pi*i*(u*a[6]+v*a[7])),
	   o^2 = u^2 + v^2 + (u^2 - v^2)*a[3] + 2*u*v*a[4]
------------------------------------------------------------------------*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
#if	defined(ANSI)	/* ANSI-C */
void otf2d(int u, int v, const double a[],
	COMPLEX *z, COMPLEX dzda[])
#else	/* old style */
void otf2d(u, v, a, z, dzda)
int u, v;
const double a[];
COMPLEX *z, dzda[];
#endif
{
	double u2 = (double)u*u;
	double v2 = (double)v*v;
	double c2 = u2 - v2;
	double s2 = (2.0*u)*v;
	double o2 = c2*a[3] + s2*a[4] + u2 + v2;
	double pu = 2.0*PI*u;
	double pv = 2.0*PI*v;
	double th = pu*a[6] + pv*a[7];
	double co = cos(th);
	double si = sin(th);
	double x2 = o2/(a[2]*a[2]);
	double n2 = 0.5*a[5];
	double xn = pow(x2, n2);
	double ho = exp(-xn);
	double tf = ho*a[1];
	double tx = tf*xn;
	double to = -tx*n2/o2;
	double t2 = tx*a[5]/a[2];
	double t3 = to*c2;
	double t4 = to*s2;
	double t5 = -0.5*tx*log(x2);
	double t6 = -pu*tf;
	double t7 = -pv*tf;
	z->x = (float) ( tf*co);
	z->y = (float) (-tf*si);
	dzda[1].x = (float) ( ho*co);
	dzda[1].y = (float) (-ho*si);
	dzda[2].x = (float) ( t2*co);
	dzda[2].y = (float) (-t2*si);
	dzda[3].x = (float) ( t3*co);
	dzda[3].y = (float) (-t3*si);
	dzda[4].x = (float) ( t4*co);
	dzda[4].y = (float) (-t4*si);
	dzda[5].x = (float) ( t5*co);
	dzda[5].y = (float) (-t5*si);
	dzda[6].x = (float) (t6*si);
	dzda[6].y = (float) (t6*co);
	dzda[7].x = (float) (t7*si);
	dzda[7].y = (float) (t7*co);
}
/*----------------------------------------------------------------------*/
#define MA	7	/* number of free parameters to fit */
#if	defined(ANSI)	/* ANSI-C */
char *g4fit(
COMPLEX * const *otf,	/* Fourier transform of PSF */
double blank,
int uc,	/* (uc, vc) is the zero frequency */
int vc,
int r1,	/* frequency points inside r1 are used in the fit*/
int r2,	/* the ring between r2 and r3 are used to derive the variance */
int r3,
double *gam,
double *pow,
double *a2,
double *b2,
double *dgam,
double *dpow,
double *da2,
double *db2)
#else	/* old style */
char *g4fit(otf, blank, uc, vc, r1, r2, r3,
	gam, pow, a2, b2, dgam, dpow, da2, db2)
COMPLEX * const *otf;
double blank;
double *gam, *pow, *a2, *b2, *dgam, *dpow, *da2, *db2;
int uc, vc, r1, r2, r3;
#endif
{
	char *str;
	double var;
	double *a, **covar, **alpha;
	double  alamda, chisq, ochisq;
	const float FltMax = (float) (0.9*blank);
	int *lista, mfit, u, v, i, j, npt;
	int rr2 = r2*r2;
	int rr3 = r3*r3;
	npt = 0;
	var = 0.0;
	for (v = (vc - r3); v <= (vc + r3); v++) {
		int v2 = v - vc;
		v2 *= v2;
		for (u = uc; u <= (uc + r3); u++) {
			int o2 = u - uc;
			o2 *= o2;
			o2 += v2;
			if ((u > uc) || (v > vc))
				if ((o2 > rr2) && (o2 <= rr3))
					if (otf[v][u].x<=FltMax && otf[v][u].y<=FltMax) {
						npt++;
						var += PROD(otf[v][u],otf[v][u]);
					}
		}
	}
	if (npt <= 1)
		return nrerror("Too few noise points in g4fit");
	var /= (npt - 1);
	if ((a = dvector(1, MA)) == (double *)0)
		return nrerror("Allocation error in g4fit-2");
	a[1] = 1.0;
	a[2] = *gam;
	a[3] = *a2;
	a[4] = *b2;
	a[5] = *pow;
	a[6] = 0.0;
	a[7] = 0.0;
	if ((lista = ivector(1, MA)) == (int *)0)
		return nrerror("Allocation error in g4fit-3");
	for (i = 1; i <= MA; i++)
		lista[i] = i;
	if ((covar = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g4fit-4");
	if ((alpha = dmatrix(1, MA, 1, MA)) == (double **)0)
		return nrerror("Allocation error in g4fit-5");

	mfit = 1;	/* fit amplitude, amp, of OTF */
	alamda = -1.0;	/* initialize mrq4min for a 1-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	alamda = 0.0;	/* no iterations, the problem is linear */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;

	mfit = 2;	/* fit amp and width, gam, of OTF */
	alamda = -1.0;	/* initialize mrq4min for a 2-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 2-parameter problem */
		ochisq = chisq;
		str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
			&npt, covar, alpha, &chisq, otf2d, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 2-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;

	mfit = 4;	/* fit amp, gam, and elongation, (a2, b2) of OTF */
	alamda = -1.0;	/* initialize mrq4min for a 4-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 4-parameter problem */
		ochisq = chisq;
		str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
			&npt, covar, alpha, &chisq, otf2d, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 4-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;

	mfit = 5;	/* fit amp, gam, a2, b2, and power, pow, of OTF */
	alamda = -1.0;	/* initialize mrq4min for a 5-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 5-parameter problem */
		ochisq = chisq;
		str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
			&npt, covar, alpha, &chisq, otf2d, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 5-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;

	mfit = 7;	/* fit amp, gam, a2, b2, pow, and (alf, bet) of OTF */
	alamda = -1.0;	/* initialize mrq4min for a 7-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	i = 0;	/* counts of iterations near convergence */
	j = 0;	/* counts of failures */
	while (i < 4 && j < 2) {	/* iterate the 7-parameter problem */
		ochisq = chisq;
		str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
			&npt, covar, alpha, &chisq, otf2d, &alamda);
		if (str != (char *)0) return str;
		if (chisq >= ochisq) {	/* failure */
			i = 0;
			j++;
		} else {	/* success */
			j = 0;
			if ((ochisq - chisq) < 0.1)
				i++;
		}
	}
	alamda = 0.0;	/* the final 7-parameter solution */
	str = mrq4min(otf, blank, var, uc, vc, r1, a, MA, lista, mfit,
		&npt, covar, alpha, &chisq, otf2d, &alamda);
	if (str != (char *)0) return str;
	if (npt <= mfit)
		return nrerror("Too few fitting points in g4fit");
	(*gam) = a[2];
	(*a2)  = a[3];
	(*b2)  = a[4];
	(*pow) = a[5];
	(*dgam) = sqrt(covar[2][2]*chisq/(npt - mfit));
	(*da2)  = sqrt(covar[3][3]*chisq/(npt - mfit));
	(*db2)  = sqrt(covar[4][4]*chisq/(npt - mfit));
	(*dpow) = sqrt(covar[5][5]*chisq/(npt - mfit));
	free_dmatrix(alpha, 1, 1);
	free_dmatrix(covar, 1, 1);
	free_ivector(lista, 1);
	free_dvector(a, 1);
	return (char *)0;
}
#undef PI
#undef MA
