
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "opcalc.h"
/*----------------------------------------------------------------------*/
#if	defined(ANSI)
void calc_uchar(void *buf1,
const void *buf2,
const int size,
const int op,
const double factor)
#else	/* old style */
void calc_uchar(buf1, buf2, size, op, factor)
void *buf1;
const void *buf2;
const int size;
const int op;
const double factor;
#endif
{
	int f = (int) factor;
	f = (f < 0)? -f: f;	/* abs(f) */
	f = (f == 0)? 1: f;	/* 0 -> 1 */
	switch (op) {
		case '+':
			{
				const int h = (int) UCHAR_Max;
				const UCHAR *q = (UCHAR *) buf2;
				UCHAR *p = (UCHAR *) buf1;
				int d, i;
				for (i = 0; i < size; ++i) {
					d = (int) p[i];
					d += q[i];
					p[i] = (UCHAR) ((d > h)? h: d);
				}
			}
			break;
		case '-':
			{
				const UCHAR *q = (UCHAR *) buf2;
				UCHAR *p = (UCHAR *) buf1;
				int d, i;
				for (i = 0; i < size; ++i) {
					d = (int) p[i];
					d -= q[i];
					p[i] = (UCHAR) ((d < 0)? 0: d);
				}
			}
			break;
		case '*':
			{
				const int h = (int) UCHAR_Max;
				const UCHAR *q = (UCHAR *) buf2;
				UCHAR *p = (UCHAR *) buf1;
				int d, i;
				for (i = 0; i < size; ++i) {
					d = (int) p[i];
					d *= q[i];
					d /= f;
					p[i] = (UCHAR) ((d > h)? h: d);
				}
			}
			break;
		case '/':
			{
				const int h = (int) UCHAR_Max;
				const UCHAR *q = (UCHAR *) buf2;
				UCHAR *p = (UCHAR *) buf1;
				int d, dn, i;
				for (i = 0; i < size; ++i) {
					dn = (int) q[i];
					d = (int) p[i];
					d *= f;
					d /= (dn == 0)? 1: dn;
					p[i] = (UCHAR) ((q[i] == 0 || d > h)? h: d);
				}
			}
			break;
		case '=':
			{
				const UCHAR *q = (UCHAR *) buf2;
				UCHAR *p = (UCHAR *) buf1;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = q[i];
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void calc_short(void *buf1,
const void *buf2,
const int size,
const int op,
const double factor,
const double blank)
#else	/* old style */
void calc_short(buf1, buf2, size, op, factor, blank)
void *buf1;
const void *buf2;
const int size;
const int op;
const double factor;
const double blank;
#endif
{
	const float z = FLT_Min;
	double f = factor;
	f = (f < 0)? -f: f;	/* abs(f) */
	f = (f < z)? 1.0: f;	/* 0 -> 1 */
	switch (op) {
		case '+':
			{
				const short b = (short) blank;
				const LONG h = (LONG) SHORT_Max;
				const LONG l = (LONG) SHORT_Min;
				const short *q = (short *) buf2;
				short *p = (short *) buf1;
				LONG d;
				short s;
				int i;
				for (i = 0; i < size; ++i) {
					d = (LONG) p[i];
					d += (LONG) q[i];
					s = (short) d;
					p[i] = (p[i]==b || q[i]==b || d>h || d<l)? b: s;
				}
			}
			break;
		case '-':
			{
				const short b = (short) blank;
				const LONG h = (LONG) SHORT_Max;
				const LONG l = (LONG) SHORT_Min;
				const short *q = (short *) buf2;
				short *p = (short *) buf1;
				LONG d;
				short s;
				int i;
				for (i = 0; i < size; ++i) {
					d = (LONG) p[i];
					d -= (LONG) q[i];
					s = (short) d;
					p[i] = (p[i]==b || q[i]==b || d<l || d>h)? b: s;
				}
			}
			break;
		case '*':
			{
				const short b = (short) blank;
				const float h = (float) SHORT_Max;
				const float l = (float) SHORT_Min;
				const short *q = (short *) buf2;
				short *p = (short *) buf1;
				double d;
				short s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= (double) q[i];
					d /= f;
					s = (short) (d + 0.5);
					t = (short) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || d>h || d<l)? b: s;
				}
			}
			break;
		case '/':
			{
				const short b = (short) blank;
				const float h = (float) SHORT_Max;
				const float l = (float) SHORT_Min;
				const short *q = (short *) buf2;
				short *p = (short *) buf1;
				double d;
				short s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= f;
					d /= (q[i] == 0)? 1.0: (double) q[i];
					s = (short) (d + 0.5);
					t = (short) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || q[i]==0 || d>h || d<l)? b: s;
				}
			}
			break;
		case '=':
			{
				const short b = (short) blank;
				const short *q = (short *) buf2;
				short *p = (short *) buf1;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] == b)? b: q[i];
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void calc_long(void *buf1,
const void *buf2,
const int size,
const int op,
const double factor,
const double blank)
#else	/* old style */
void calc_long(buf1, buf2, size, op, factor, blank)
void *buf1;
const void *buf2;
const int size;
const int op;
const double factor;
const double blank;
#endif
{
	const float z = FLT_Min;
	double f = factor;
	f = (f < 0)? -f: f;	/* abs(f) */
	f = (f < z)? 1.0: f;	/* 0 -> 1 */
	switch (op) {
		case '+':
			{
				const LONG b = (LONG) blank;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				const LONG *q = (LONG *) buf2;
				LONG *p = (LONG *) buf1;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d += (double) q[i];
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || d>h || d<l)? b: s;
				}
			}
			break;
		case '-':
			{
				const LONG b = (LONG) blank;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				const LONG *q = (LONG *) buf2;
				LONG *p = (LONG *) buf1;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d -= (double) q[i];
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || d<l || d>h)? b: s;
				}
			}
			break;
		case '*':
			{
				const LONG b = (LONG) blank;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				const LONG *q = (LONG *) buf2;
				LONG *p = (LONG *) buf1;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= (double) q[i];
					d /= f;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || d>h || d<l)? b: s;
				}
			}
			break;
		case '/':
			{
				const LONG b = (LONG) blank;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				const LONG *q = (LONG *) buf2;
				LONG *p = (LONG *) buf1;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= f;
					d /= (q[i] == 0)? 1.0: (double) q[i];
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || q[i]==b || q[i]==0 || d>h || d<l)? b: s;
				}
			}
			break;
		case '=':
			{
				const LONG  b = (LONG) blank;
				const LONG *q = (LONG *) buf2;
				LONG *p = (LONG *) buf1;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] == b)? b: q[i];
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void calc_float(void *buf1,
const void *buf2,
const int size,
const int op,
const double factor,
const double blank)
#else	/* old style */
void calc_float(buf1, buf2, size, op, factor, blank)
void *buf1;
const void *buf2;
const int size;
const int op;
const double factor;
const double blank;
#endif
{
	const float z = FLT_Min;
	double f = factor;
	f = (f < 0)? -f: f;
	f = (f < z)? 1.0: f;	/* 0 -> 1 */
	switch (op) {
		case '+':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float *q = (float *) buf2;
				float *p = (float *) buf1;
				float d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i] + q[i];
					p[i] = (p[i]>h || q[i]>h || d>h || d<-h)? b: d;
				}
			}
			break;
		case '-':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float *q = (float *) buf2;
				float *p = (float *) buf1;
				float d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i] - q[i];
					p[i] = (p[i]>h || q[i]>h || d>h || d<-h)? b: d;
				}
			}
			break;
		case '*':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float *q = (float *) buf2;
				float *p = (float *) buf1;
				double d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i];
					d *= q[i];
					d /= f;
					p[i] = (p[i]>h || q[i]>h || d>h || d<-h)? b: (float)d;
				}
			}
			break;
		case '/':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float *q = (float *) buf2;
				float *p = (float *) buf1;
				float dn;
				double d;
				int i;
				for (i = 0; i < size; ++i) {
					dn = q[i];
					d = p[i];
					d *= f;
					d /= (dn > -z && dn < z)? z: dn;
					p[i] = (p[i]>h || q[i]>h || d>h || d<-h)? b: (float)d;
				}
			}
			break;
		case '=':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float *q = (float *) buf2;
				float *p = (float *) buf1;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] > h)? b: q[i];
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void calc_complex(void *buf1,
const void *buf2,
const int size,
const int op,
const double factor,
const double blank)
#else	/* old style */
void calc_complex(buf1, buf2, size, op, factor, blank)
void *buf1;
const void *buf2;
const int size;
const int op;
const double factor;
const double blank;
#endif
{
	const double z = (double) FLT_Min;
	double f = factor;
	f = (f < 0)? -f: f;
	f = (f < z)? 1.0: f;	/* 0 -> 1 */
	switch (op) {
		case '+':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const COMPLEX *q = (COMPLEX *) buf2;
				COMPLEX *p = (COMPLEX *) buf1;
				float dx, dy;
				int i, ip, iq;
				for (i = 0; i < size; ++i) {
					dx = p[i].x + q[i].x;
					dy = p[i].y + q[i].y;
					ip = p[i].x > h || p[i].y > h;
					iq = q[i].x > h || q[i].y > h;
					p[i].x = (ip || iq || dx > h || dx < -h)? b: dx;
					p[i].y = (ip || iq || dy > h || dy < -h)? b: dy;
				}
			}
			break;
		case '-':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const COMPLEX *q = (COMPLEX *) buf2;
				COMPLEX *p = (COMPLEX *) buf1;
				float dx, dy;
				int i, ip, iq;
				for (i = 0; i < size; ++i) {
					dx = p[i].x - q[i].x;
					dy = p[i].y - q[i].y;
					ip = p[i].x > h || p[i].y > h;
					iq = q[i].x > h || q[i].y > h;
					p[i].x = (ip || iq || dx > h || dx < -h)? b: dx;
					p[i].y = (ip || iq || dy > h || dy < -h)? b: dy;
				}
			}
			break;
		case '*':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const COMPLEX *q = (COMPLEX *) buf2;
				COMPLEX *p = (COMPLEX *) buf1;
				double dx, dy;
				int i, ip, iq;
				for (i = 0; i < size; ++i) {
					dx = (double)p[i].x*q[i].x - (double)p[i].y*q[i].y;
					dx /= f;
					dy = (double)p[i].y*q[i].x + (double)p[i].x*q[i].y;
					dy /= f;
					ip = p[i].x > h || p[i].y > h;
					iq = q[i].x > h || q[i].y > h;
					p[i].x = (ip || iq || dx>h || dx<-h)? b: (float)dx;
					p[i].y = (ip || iq || dy>h || dy<-h)? b: (float)dy;
				}
			}
			break;
		case '/':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const COMPLEX *q = (COMPLEX *) buf2;
				COMPLEX *p = (COMPLEX *) buf1;
				double dn, dx, dy;
				int i, ip, iq;
				for (i = 0; i < size; ++i) {
					dn = (double)q[i].x*q[i].x + (double)q[i].y*q[i].y;
					dn = (dn < z)? z: dn;
					dx = (double)p[i].x*q[i].x + (double)p[i].y*q[i].y;
					dx *= f;
					dx /= dn;
					dy = (double)p[i].y*q[i].x - (double)p[i].x*q[i].y;
					dy *= f;
					dy /= dn;
					ip = p[i].x > h || p[i].y > h;
					iq = q[i].x > h || q[i].y > h;
					p[i].x = (ip || iq || dx>h || dx<-h)? b: (float)dx;
					p[i].y = (ip || iq || dy>h || dy<-h)? b: (float)dy;
				}
			}
			break;
		case '=':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const COMPLEX *q = (COMPLEX *) buf2;
				COMPLEX *p = (COMPLEX *) buf1;
				int i, ip, iq;
				for (i = 0; i < size; ++i) {
					ip = p[i].x > h || p[i].y > h;
					iq = q[i].x > h || q[i].y > h;
					p[i].x = (ip || iq)? b: q[i].x;
					p[i].y = (ip || iq)? b: q[i].y;
				}
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
