
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "opxtract.h"
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy11(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy11(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	UCHAR *p = (UCHAR *)bf1;
	const UCHAR *q = (UCHAR *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy12(void *bf1,
const void *bf2,
const int size,
const double blank2)
#else	/* old style */
static void Copy12(bf1, bf2, size, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank2;
#endif
{
	UCHAR *p = (UCHAR *)bf1;
	const short *q = (short *)bf2;
	const short b2 = (short)blank2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? 0: (UCHAR) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy13(void *bf1,
const void *bf2,
const int size,
const double blank2)
#else	/* old style */
static void Copy13(bf1, bf2, size, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank2;
#endif
{
	UCHAR *p = (UCHAR *)bf1;
	const float *q = (float *)bf2;
	const float h2 = (float) (0.9*blank2);
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] > h2)? 0: (UCHAR) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy14(void *bf1,
const void *bf2,
const int size,
const double blank2)
#else	/* old style */
static void Copy14(bf1, bf2, size, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank2;
#endif
{
	UCHAR *p = (UCHAR *)bf1;
	const LONG *q = (LONG *)bf2;
	const LONG b2 = (LONG)blank2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? 0: (UCHAR) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy15(void *bf1,
const void *bf2,
const int size,
const double blank2)
#else	/* old style */
static void Copy15(bf1, bf2, size, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank2;
#endif
{
	UCHAR *p = (UCHAR *)bf1;
	const COMPLEX *q = (COMPLEX *)bf2;
	const float h2 = (float) (0.9*blank2);
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i].x > h2 || q[i].y > h2)? 0: (UCHAR) q[i].x;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy21(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy21(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	short *p = (short *)bf1;
	const UCHAR *q = (UCHAR *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy22(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy22(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	short *p = (short *)bf1;
	const short *q = (short *)bf2;
	const short b2 = (short)blank2;
	const short b1 = (short)blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy23(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy23(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	short *p = (short *)bf1;
	const float *q = (float *)bf2;
	const float h2 = (float) (0.9*blank2);
	const short b1 = (short) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (short) ((q[i] > h2) ? b1 :
                      (q[i] < 0) ? q[i] - 0.5 : q[i] + 0.5);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy24(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy24(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	short *p = (short *)bf1;
	const LONG *q = (LONG *)bf2;
	const LONG b2 = (LONG)blank2;
	const short b1 = (short)blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: (short) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy25(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy25(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	short *p = (short *)bf1;
	const COMPLEX *q = (COMPLEX *)bf2;
	const float h2 = (float) (0.9*blank2);
	const short b1 = (short) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (short) ((q[i].x > h2 || q[i].y > h2) ? b1 :
                      (q[i].x < 0) ? q[i].x - 0.5 : q[i].x + 0.5);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy31(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy31(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	float *p = (float *)bf1;
	const UCHAR *q = (UCHAR *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (float) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy3U2(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy3U2(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	float *p = (float *)bf1;
	const USHORT *q = (USHORT *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (float) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy32(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy32(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	float *p = (float *)bf1;
	const short *q = (short *)bf2;
	const short b2 = (short) blank2;
	const float b1 = (float) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: (float) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy33(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy33(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	float *p = (float *)bf1;
	const float *q = (float *)bf2;
	const float h2 = (float) (0.9*blank2);
	const float b1 = (float) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] > h2)? b1: q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy34(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy34(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	float *p = (float *)bf1;
	const LONG *q = (LONG *)bf2;
	const LONG b2 = (LONG) blank2;
	const float b1 = (float) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: (float) q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy35(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy35(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	float *p = (float *)bf1;
	const COMPLEX *q = (COMPLEX *)bf2;
	const float h2 = (float) (0.9*blank2);
	const float b1 = (float) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i].x > h2 || q[i].y > h2)? b1: q[i].x;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy41(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy41(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	LONG *p = (LONG *)bf1;
	const UCHAR *q = (UCHAR *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy4U2(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy4U2(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	LONG *p = (LONG *)bf1;
	const USHORT *q = (USHORT *)bf2;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy42(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy42(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	LONG *p = (LONG *)bf1;
	const short *q = (short *)bf2;
	const short b2 = (short) blank2;
	const LONG b1 = (LONG) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy43(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy43(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	LONG *p = (LONG *)bf1;
	const float *q = (float *)bf2;
	const float h2 = (float) (0.9*blank2);
	const LONG b1 = (LONG) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (LONG) ((q[i] > h2) ? b1 :
                     (q[i] < 0) ? q[i] - 0.5 : q[i] + 0.5);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy44(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy44(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	LONG *p = (LONG *)bf1;
	const LONG *q = (LONG *)bf2;
	const LONG b2 = (LONG) blank2;
	const LONG b1 = (LONG) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (q[i] == b2)? b1: q[i];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy45(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy45(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	LONG *p = (LONG *)bf1;
	const COMPLEX *q = (COMPLEX *)bf2;
	const float h2 = (float) (0.9*blank2);
	const LONG b1 = (LONG) blank1;
	int i;
	for (i = 0; i < size; ++i)
		p[i] = (LONG) ((q[i].x > h2 || q[i].y > h2) ? b1 :
                     (q[i].x < 0) ? q[i].x - 0.5 : q[i].x + 0.5);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy51(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy51(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const UCHAR *q = (UCHAR *)bf2;
	int i;
	for (i = 0; i < size; ++i) {
		p[i].x = (float)q[i];
		p[i].y = (float)0;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy5U2(void *bf1,
const void *bf2,
const int size)
#else	/* old style */
static void Copy5U2(bf1, bf2, size)
void *bf1;
const void *bf2;
const int size;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const USHORT *q = (USHORT *)bf2;
	int i;
	for (i = 0; i < size; ++i) {
		p[i].x = (float)q[i];
		p[i].y = (float)0;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy52(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy52(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const short *q = (short *)bf2;
	const short b2 = (short)blank2;
	const float b1 = (float)blank1;
	int i, ip;
	for (i = 0; i < size; ++i) {
		ip = (q[i] == b2);
		p[i].x = (ip)? b1: (float)q[i];
		p[i].y = (ip)? b1: (float)0;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy53(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy53(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const float *q = (float *)bf2;
	const float h2 = (float) (0.9*blank2);
	const float b1 = (float)blank1;
	int i, ip;
	for (i = 0; i < size; ++i) {
		ip = (q[i] > h2);
		p[i].x = (ip)? b1: q[i];
		p[i].y = (ip)? b1: (float)0;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy54(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy54(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const LONG *q = (LONG *)bf2;
	const LONG b2 = (LONG)blank2;
	const float b1 = (float)blank1;
	int i, ip;
	for (i = 0; i < size; ++i) {
		ip = (q[i] == b2);
		p[i].x = (ip)? b1: (float)q[i];
		p[i].y = (ip)? b1: (float)0;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Copy55(void *bf1,
const void *bf2,
const int size,
const double blank1,
const double blank2)
#else	/* old style */
static void Copy55(bf1, bf2, size, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const double blank1;
const double blank2;
#endif
{
	COMPLEX *p = (COMPLEX *)bf1;
	const COMPLEX *q = (COMPLEX *)bf2;
	const float h2 = (float) (0.9*blank2);
	const float b1 = (float)blank1;
	int i, ip;
	for (i = 0; i < size; ++i) {
		ip = (q[i].x > h2 || q[i].y > h2);
		p[i].x = (ip)? b1: q[i].x;
		p[i].y = (ip)? b1: q[i].y;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void bufcpy(void *bf1,
const void *bf2,
const int size,
const int type1,
const int type2,
const double blank1,
const double blank2)
#else	/* old style */
void bufcpy(bf1, bf2, size, type1, type2, blank1, blank2)
void *bf1;
const void *bf2;
const int size;
const int type1;
const int type2;
const double blank1;
const double blank2;
#endif
{
	switch (type1) {
		case 1:
			switch (type2) {
				case  1:
					Copy11(bf1, bf2, size);
					break;
				case -2:
				case  2:
					Copy12(bf1, bf2, size, blank2);
					break;
				case  3:
					Copy13(bf1, bf2, size, blank2);
					break;
				case  4:
					Copy14(bf1, bf2, size, blank2);
					break;
				case  5:
					Copy15(bf1, bf2, size, blank2);
					break;
				default:
					break;
			}
			break;
		case 2:
			switch (type2) {
				case  1:
					Copy21(bf1, bf2, size);
					break;
				case -2:
				case  2:
					Copy22(bf1, bf2, size, blank1, blank2);
					break;
				case  3:
					Copy23(bf1, bf2, size, blank1, blank2);
					break;
				case  4:
					Copy24(bf1, bf2, size, blank1, blank2);
					break;
				case  5:
					Copy25(bf1, bf2, size, blank1, blank2);
					break;
				default:
					break;
			}
			break;
		case 3:
			switch (type2) {
				case  1:
					Copy31(bf1, bf2, size);
					break;
				case -2:
					Copy3U2(bf1, bf2, size); /* copy unsigned short */
					break;
				case  2:
					Copy32(bf1, bf2, size, blank1, blank2);
					break;
				case  3:
					Copy33(bf1, bf2, size, blank1, blank2);
					break;
				case  4:
					Copy34(bf1, bf2, size, blank1, blank2);
					break;
				case  5:
					Copy35(bf1, bf2, size, blank1, blank2);
					break;
				default:
					break;
			}
			break;
		case 4:
			switch (type2) {
				case  1:
					Copy41(bf1, bf2, size);
					break;
				case -2:
					Copy4U2(bf1, bf2, size); /* copy unsigned short */
					break;
				case  2:
					Copy42(bf1, bf2, size, blank1, blank2);
					break;
				case  3:
					Copy43(bf1, bf2, size, blank1, blank2);
					break;
				case  4:
					Copy44(bf1, bf2, size, blank1, blank2);
					break;
				case  5:
					Copy45(bf1, bf2, size, blank1, blank2);
					break;
				default:
					break;
			}
			break;
		case 5:
			switch (type2) {
				case  1:
					Copy51(bf1, bf2, size);
					break;
				case -2:
					Copy5U2(bf1, bf2, size); /* copy unsigned short */
					break;
				case  2:
					Copy52(bf1, bf2, size, blank1, blank2);
					break;
				case  3:
					Copy53(bf1, bf2, size, blank1, blank2);
					break;
				case  4:
					Copy54(bf1, bf2, size, blank1, blank2);
					break;
				case  5:
					Copy55(bf1, bf2, size, blank1, blank2);
					break;
				default:
					break;
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
