
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "pstring.h"
/*
	String handling support by Martin Weitzel.
	The macros ALLOC and FREE hide away some casts and take
	care of pre-ANSI-C free() which is not guaranteed to be
	safe for (void *)0 pointers.
*/
#define ALLOC(n,t)	(((t) *)malloc((n)*(sizeof(t))))
#if	defined(__STDC__)
#define FREE(p)	(free((void *)(p)))
#else
#define FREE(p)	(if ((p) != NULL) free((void *)(p)))
#endif
/*
	Concatenate variable number of strings in space allocated
	from the heap. If no more space is available, the program
	is aborted.
*/
char *pstr_x(const char *str, ...)
{
	const char *cp; /* tmpry to step over args */
	char *result;   /* tmpry for return value */
	va_list ap;     /* arglist iterator */
	size_t len = 1; /* total length of strs + 1 */
/* loop over args, sum length of all strs */
	va_start(ap, str);
	for (cp = str; cp != NULL; cp = va_arg(ap, const char *))
		len += strlen(cp);
	va_end(ap);
/* allocate space for all strs */
	if ((result = ALLOC(len, char)) == (char *)0)
		abort();
	result[0] = '\0';
/* loop over args, concatenate all strs */
	va_start(ap, str);
	for (cp = str; cp != NULL; cp = va_arg(ap, const char *))
		(void) strcat(result, cp);
	va_end(ap);
	return result;
}
/*
	Helper function for building temporary strings. It effectively
	"schedules a free" delayed by one call to this function, so the
	caller does not have to take care of saving the result of "strmake"
	in a variable and hand this to free later on.
	Note: With "(void) tmpstr((char *)0)" you may force to deallocate
	the last temporary string build with one of the "TMPSTR" macros.
*/
static char *last_str = (char *)0;
char *tmpstr(char *str)
{
	FREE(last_str);
	return (last_str = str);
}
