
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
#define	EPDN	2.6	/* Electrons per Digital Number */
#define	ILIM	2	/* number of iterations */
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int no_sky = 0;	/* no sky correction */
	double epdn = EPDN;	/* Electrons per Digital Number */
	int i, x, y;
	int xcen, ycen, radmin, radmax, skymin, skymax;
	int xc, yc, stx, sty, extx, exty;
	int id, type, npix, nlin, rw = 0;
	long y2, rmin2, rmax2, smin2, smax2, nobj, nsky;
	double Blank, objsum, objsig, skymean, skyvar, skysig, sum;
	void *buffer;
	char *image, *ptr;
	float **a;
	unsigned int err;
	int start[2], dim[2];
	const float FltBlank = (float)FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	float Init_Value = FltBlank;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "aphot: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "ne:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'n':	no_sky++; break;
			case 'e':	sscanf(optarg, "%lf", &epdn); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:	break;
		}
	}

	if (opt_err || n_arg < 6) {
		printf("aphot -n -e epdn image Xc Yc RadMax SkyMin SkyMax [RadMin]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	if (epdn <= 0.0) {
		fprintf(stderr,
			"aphot: Illegal number of electrons per DN = %g\n", epdn);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "aphot: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%d", &xcen);
	sscanf(arg[2], "%d", &ycen);
	sscanf(arg[3], "%d", &radmax);	/* outer aperture radius */
	sscanf(arg[4], "%d", &skymin);	/* inner rad of sky annulus */
	sscanf(arg[5], "%d", &skymax);	/* outer rad of sky annulus */

	if (n_arg > 6) {	/* inner aperture radius */
		sscanf(arg[6], "%d", &radmin);
	} else radmin = 0;

	if (radmax <= radmin) {
		fprintf(stderr, "aphot: radmax <= radmin\n");
		exit(EXIT_FAILURE);
	}

	if (skymin <= radmax) {
		fprintf(stderr, "aphot: skymin <= radmax\n");
		exit(EXIT_FAILURE);
	}

	if (skymax <= skymin) {
		fprintf(stderr, "aphot: skymax <= skymin\n");
		exit(EXIT_FAILURE);
	}

	rmin2 = radmin;
	rmin2 *= rmin2;	/* inner aperture radius squared */
	rmax2 = radmax;
	rmax2 *= rmax2;	/* outer aperture radius squared */
	smin2 = skymin;
	smin2 *= smin2;	/* inner sky radius squared */
	smax2 = skymax;
	smax2 *= smax2;	/* outer sky radius squared */

	stx = xcen - skymax;
	sty = ycen - skymax;
	if (stx < 0 || sty < 0) {
		fprintf(stderr, "aphot: sky annulus extends outside image\n");
		exit(EXIT_FAILURE);
	}
	xc = xcen - stx;
	yc = ycen - sty;

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "aphot: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "aphot: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr, "aphot: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	extx = 2 * skymax + 1;
	exty = extx;
	if (stx + extx > npix || sty + exty > nlin) {
		fprintf(stderr, "aphot: sky annulus extends outside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "aphot: object window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((buffer = (void *)calloc(extx, bytepix(type))) == (void *)0) {
		fprintf(stderr, "aphot: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	dim[0] = exty;	/* y dimension of object window */
	dim[1] = extx; /* x dimension of object window */
	start[0] = 0;	/* first index of y dimension */
	start[1] = 0;	/* first index of x dimension */

	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "aphot: daa: error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < exty; y++) {	/* read object window into array */
		getline(id, y, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					for (x = 0; x < extx; ++x)
						a[y][x] = (float)q[x];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x].x>h || q[x].y>h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "aphot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	nsky = 0L;	/* number of sky pixels */
	sum = 0.0;
	for (y = 0; y < exty; y++) {
		long r2;
		y2 = y - yc;
		y2 *= y2;
		for (x = 0; x < extx; x++) {
			r2 = x - xc;
			r2 *= r2;
			r2 += y2;
			if (smin2 <= r2 && r2 <= smax2)
				if (a[y][x] <= FltMax) {
					sum += a[y][x];
					nsky++;
				}
		}
	}

	skymean = sum/(double)nsky;	/* average sky level */

	nsky = 0L;
	sum = 0.0;
	for (y = 0; y < exty; y++) {
		long r2;
		double data;
		y2 = y - yc;
		y2 *= y2;
		for (x = 0; x < extx; x++) {
			r2 = x - xc;
			r2 *= r2;
			r2 += y2;
			if (smin2 <= r2 && r2 <= smax2)
				if (a[y][x] <= FltMax) {
					data = a[y][x] - skymean;
					sum += fabs(data);
					nsky++;
				}
		}
	}

	skyvar = (PI/2.0)*(sum/(double)(nsky-1L))*(sum/(double)nsky);
	skysig = sqrt(skyvar);	/* first estimate of sky sigma */

	printf("nsky  = %8ld    skymean = %14.8g    skysig = %14.8g\n",
			nsky, skymean, skysig);

	for (i = 0; i < ILIM; i++) {	/* iterate sky determination */
		double limit = 2.0 * skysig;	/* rejection limit */

		for (y = 0; y < exty; y++) {
			long r2;
			double data;
			y2 = y - yc;
			y2 *= y2;
			for (x = 0; x < extx; x++) {
				r2 = x - xc;
				r2 *= r2;
				r2 += y2;
				if (smin2 <= r2 && r2 <= smax2)
					if (a[y][x] <= FltMax) {
						data = a[y][x] - skymean;
						if (fabs(data) > limit)
							a[y][x] = FltBlank;
					}
			}
		}

		nsky = 0L;	/* new number of sky pixels */
		sum = 0.0;
		for (y = 0; y < exty; y++) {
			long r2;
			y2 = y - yc;
			y2 *= y2;
			for (x = 0; x < extx; x++) {
				r2 = x - xc;
				r2 *= r2;
				r2 += y2;
				if (smin2 <= r2 && r2 <= smax2)
					if (a[y][x] <= FltMax) {
						sum += a[y][x];
						nsky++;
					}
			}
		}

		skymean = sum/(double)nsky;	/* new sky average */

		nsky = 0L;
		sum = 0.0;
		for (y = 0; y < exty; y++) {
			long r2;
			double data;
			y2 = y - yc;
			y2 *= y2;
			for (x = 0; x < extx; x++) {
				r2 = x - xc;
				r2 *= r2;
				r2 += y2;
				if (smin2 <= r2 && r2 <= smax2)
					if (a[y][x] <= FltMax) {
						data = a[y][x] - skymean;
						sum += fabs(data);
						nsky++;
					}
			}
		}

		skysig = (PI/2.0)*(sum/(double)(nsky-1L))*(sum/(double)nsky);
		skysig = sqrt(skysig);	/* new estimate of sky sigma */

		printf("nsky  = %8ld    skymean = %14.8g    skysig = %14.8g\n",
				nsky, skymean, skysig);
	}

	nobj = 0L;	/* number of pixels in aperture */
	objsum = 0.0;
	for (y = 0; y < exty; y++) {
		long r2;
		y2 = y - yc;
		y2 *= y2;
		for (x = 0; x < extx; x++) {
			r2 = x - xc;
			r2 *= r2;
			r2 += y2;
			if (rmin2 <= r2 && r2 <= rmax2)
				if (a[y][x] <= FltMax) {
					if (no_sky != 0)
						objsum += a[y][x];
					else
						objsum += a[y][x] - skymean;
					nobj++;
				}
		}
	}

	if (no_sky != 0)
		objsig = sqrt(objsum/epdn);
	else {
		skyvar = (double)nobj*(1.0 + (double)nobj/(double)nsky)*skyvar;
		objsig = sqrt(objsum/epdn + skyvar);
	}
	printf("nobj  = %8ld    objsum  = %14.8g    objsig = %14.8g\n",
			nobj, objsum, objsig);

	free(ptr);
	free(buffer);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
