
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image, *ptr;
	short **a, Blank;
	int xmin1, xmax1, xmin2, xmax2, xmin3, xmax3;
	int i, n, x, extx, xcen, dx1, dx2, dx3;
	int y, ymin, ymax, exty;
	int id, type, npix, nlin;
	double fa, fb, fx, xa, xb;
	unsigned int err;
	int start[2], dim[2];

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 6) {
		printf("Usage: badcol image Xc dX1 dX2 dX3 [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "badcol: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(argv[3], "%d", &dx1);
	sscanf(argv[4], "%d", &dx2);
	sscanf(argv[5], "%d", &dx3);

	if (dx1 < 0) {
		fprintf(stderr, "badcol: dx1 = %d\n", dx1);
		exit(EXIT_FAILURE);
	} else if (dx2 <= dx1) {
		fprintf(stderr, "badcol: dx2 = %d <= dx1 = %d\n", dx2, dx1);
		exit(EXIT_FAILURE);
	} else if (dx3 < dx2) {
		fprintf(stderr, "badcol: dx3 = %d < dx2 = %d\n", dx3, dx2);
		exit(EXIT_FAILURE);
	}

	sscanf(argv[2], "%d", &xcen);

	if ((xmin3 = xcen - dx3) < 0) {
		fprintf(stderr, "badcol: xcen = %d < dx3 = %d\n", xcen, dx3);
		exit(EXIT_FAILURE);
	}

	xmin2 = xcen - dx2;
	xmax2 = xcen + dx2;
	xmin1 = xcen - dx1;
	xmax1 = xcen + dx1;

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "badcol: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open image for read-only */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "badcol: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((xmax3 = xcen + dx3) >= npix) {
		fprintf(stderr,
			"badcol: xcen + dx3 = %d >= npix = %d\n", xmax3, npix);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (argc > 6) sscanf(argv[6], "%d", &ymin); else ymin = 0;

	if (ymin < 0) {
		fprintf(stderr, "badcol: ymin = %d < 0\n", ymin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (argc > 7) sscanf(argv[7], "%d", &ymax); else ymax = nlin - 1;

	if (ymax >= nlin) {
		fprintf(stderr, "badcol: ymax = %d >= nlin = %d\n", ymax, nlin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (ymax < ymin) {
		fprintf(stderr, "badcol: ymax = %d < ymin = %d\n", ymax, ymin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	extx = xmax3 - xmin3 + 1;
	exty = ymax - ymin + 1;

	if (type != 2) {
		fprintf(stderr, "badcol: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = (short) blank(image);

	dim[0] = exty;	/* y dimension of object window */
	dim[1] = extx; /* x dimension of object window */
	start[0] = 0;	/* first index of y dimension */
	start[1] = 0;	/* first index of x dimension */

	if ((a = (short **)daa(sizeof(short), 2, dim, start,
			&err, &ptr, &Blank)) == (short **)0) {
		fprintf(stderr, "badcol: daa: error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (!window(id, xmin3, ymin, extx, exty)) {
		fprintf(stderr, "badcol: object window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < exty; ++y)	/* read window into buffer array */
		getline(id, y, &a[y][0]);
	xmin2 -= xmin3;	/* coordinates relative to (xmin3, 0) */
	xmin1 -= xmin3;
	xmax1 -= xmin3;
	xmax2 -= xmin3;
	xmax3 -= xmin3;
	for (y = 0; y < exty; ++y) {	/* process line */
		fa = 0.0;
		xa = 0.0;
		for (n = 0, x = 0; x <= xmin2; ++x)
			if (a[y][x] != Blank) {
				xa += (double)x;
				fa += (double)a[y][x];
				++n;
			}
		if (n == 0) continue;
		xa /= n;
		fa /= n;
		fb = 0.0;
		xb = 0.0;
		for (n = 0, x = xmax2; x <= xmax3; ++x)
			if (a[y][x] != Blank) {
				xb += (double)x;
				fb += (double)a[y][x];
				++n;
			}
		if (n == 0) continue;
		xb /= n;
		fb /= n;
		fa /= (xb - xa);
		fb /= (xb - xa);
		for (x = xmin1; x <= xmax1; ++x) {	/* interpolate */
			fx = (xb - x)*fa + (x - xa)*fb;
			a[y][x] = (short) ((fx < 0.0)? fx - 0.5: fx + 0.5);
		}
	}
	setbufsiz(4096);
	if (!chiomod(id, 2)) {	/* change IO mode to update */
		fprintf(stderr, "badcol: Can't change IO mode for %s\n", image);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < exty; ++y)
		putline(id, y, a[y]);
	free(ptr);
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "badcol: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
