
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "opcalc.h"	/* function declaration for opcalc.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image1, *image2, *buf, str[5];
	char *array1 = (char *)0;
	char *array2 = (char *)0;
	char *ptr = (char *)0;
	int k, l, m, n, bsize, narray, op;
	int id1, type1, npix1, nlin1, rw1 = 2;
	int id2, type2, npix2, nlin2, rw2 = 0;
	unsigned u, ioff, asize, nb, nblok, mblok, maxblok;
	long isize;
	double Blank1, Blank2;
	double factor = 1.0;	/* default value */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if(argc < 4) {
		fprintf(stderr, "Usage: calc image1 op image2 [factor]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(argv[1]);
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "calc: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	sscanf(argv[2], "%4s", str);
	op=(strncmp(str,   "+", 1) == 0)? '+':
		(strncmp(str, "add", 3) == 0)? '+':
		(strncmp(str, "ADD", 3) == 0)? '+':
		(strncmp(str,   "-", 1) == 0)? '-':
		(strncmp(str, "sub", 3) == 0)? '-':
		(strncmp(str, "SUB", 3) == 0)? '-':
		(strncmp(str,   "*", 1) == 0)? '*':
		(strncmp(str, "mul", 3) == 0)? '*':
		(strncmp(str, "MUL", 3) == 0)? '*':
		(strncmp(str,   "/", 1) == 0)? '/':
		(strncmp(str, "div", 3) == 0)? '/':
		(strncmp(str, "DIV", 3) == 0)? '/':
		(strncmp(str,   "=", 1) == 0)? '=':
		(strncmp(str,  "eq", 2) == 0)? '=':
		(strncmp(str,  "EQ", 2) == 0)? '=': 0;

	image2 = truncname(argv[3]);
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "calc: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}

	if(argc > 4) sscanf(argv[4], "%lf", &factor);

	/* read image names from header directory */
	if ((k = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "calc: prolog: Error = %d\n", k);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id1 = iopen(image1, rw1, &type1, &npix1, &nlin1)) < 0) {
		fprintf(stderr, "calc: can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type1 < 1 || type1 > 5) {
		fprintf(stderr, "calc: %s has illegal type = %d\n", image1, type1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank1 = blank(image1);
	if ((id2 = iopen(image2, rw2, &type2, &npix2, &nlin2)) < 0) {
		fprintf(stderr, "calc: can't open %s\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type1 != type2) {
		fprintf(stderr,
			"calc: %s and %s have incompatible types\n", image1, image2);
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (npix1 != npix2) {
		fprintf(stderr,
			"calc: %s and %s have incompatible x-dimensions\n",
				image1, image2);
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (nlin1 != nlin2) {
		fprintf(stderr,
			"calc: %s and %s have incompatible y-dimensions\n",
				image1, image2);
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank2 = blank(image2);
	if (Blank1 != 0.0 && fabs(Blank2/Blank1 - 1.0) > 1.0E-10 ||
		 Blank1 == 0.0 && Blank2 != 0.0) {
		fprintf(stderr,
			"calc: %s and %s have incompatible BLANKs\n", image1, image2);
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	isize = imagesize(image1);	/* image size in bytes */
	bsize = buffsize();	/* block size in bytes */
	maxblok = INT_Max/bsize;	/* max number of blocks available */
	mblok = (unsigned)((isize - 1L)/bsize + 1L);	/* blocks in image */
	if (mblok > maxblok)	/* find maximum block size of array */
		mblok = maxblok;
	asize = 0;	/* array size in bytes */
	for (nblok = mblok; nblok > 0; nblok--) {	/* find 2 big arrays */
		asize = nblok*bsize;	/* size of each buffer array */
		if ((array1 = (char *)malloc(asize)) != (char *)0)
			if ((array2 = (char *)malloc(asize)) != (char *)0)
				break;	/* space for 2 big arrays allocated */
			else
				free(array1);	/* bad luck with second array */
	}
	if (nblok <= 0) {	/* arrays have not been allocated */
		fprintf(stderr, "calc: malloc: Array allocation failed\n");
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (nblok < mblok) {	/* no space for maximum size */
		nb = nblok;
		nb++;
		nb >>= 1;
		asize = nb*bsize;	/* array size reduced by 50% */
		if ((ptr = (char *)realloc(array1, asize)) != (char *)0) {
			array1 = ptr;	/* first array successfully reallocated */
			if ((ptr = (char *)realloc(array2, asize)) != (char *)0)
				array2 = ptr;	/* second array reallocated */
			nblok = nb;
		} else
			asize = nblok*bsize;	/* reallocation failed */
	}
	narray = (int)((isize - 1L)/asize + 1L);	/* no. of buf arrays */
	for (ioff = 0, k = 0; k < narray; k++) {	/* step array index */
		n = 0;	/* initialize pixel counter for array1 */
		buf = array1;	/* initialize block pointer for array1 */
		for (u = 0; u < nblok; u++)	/* step block index */
			if ((l = bread(id1, ioff + u, buf)) == 0)	/* read 1 block */
				break;	/* no more blocks, exit loop */
			else {
				n += l;	/* update number of pixels read into array1 */
				buf += bsize;	/* increment block pointer */
			}
		m = 0;	/* initialize pixel counter for array2 */
		buf = array2;	/* initialize block pointer for array2 */
		for (u = 0; u < nblok; u++)	/* step block index */
			if ((l = bread(id2, ioff + u, buf)) == 0)	/* read 1 block */
				break;	/* no more blocks, exit loop */
			else {
				m += l;	/* update number of pixels read into array2 */
				buf += bsize;	/* increment block pointer */
			}
		if (n != m) {
			fprintf(stderr,
				"calc: Different number of pixels read from %s and %s\n",
				image1, image2);
			iclose(id2);
			iclose(id1);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		switch (type1) {	/* operate on full buffer arrays */
			case 1:
				calc_uchar(array1, array2, n, op, factor);
				break;
			case 2:
				calc_short(array1, array2, n, op, factor, Blank1);
				break;
			case 3:
				calc_float(array1, array2, n, op, factor, Blank1);
				break;
			case 4:
				calc_long(array1, array2, n, op, factor, Blank1);
				break;
			case 5:
				calc_complex(array1, array2, n, op, factor, Blank1);
				break;
		}
		buf = array1;	/* block pointer for array1 */
		for (u = 0; u < nblok; u++)	/* step block index */
			if (bwrite(id1, ioff + u, buf) == 0)	/* write 1 block */
				break;	/* no more blocks, exit loop */
			else
				buf += bsize;	/* increment block pointer */
		ioff += nblok;	/* increment block offset of array start */
	}
	free(array2);	/* remember to deallocate big arrays */
	free(array1);
	iclose(id2);
	iclose(id1);
	sethistory(image1, "calc");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "calc: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
