
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declaration for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
#define ABS(a)	(((a) < Zero) ? -(a) : (a))	/* Zero = 0 */
/*--------------------------------------------------------------------*/
#define NDAT	24	/* maximum number of data points included in fit */
#define NFIT	11	/* number of functions to be fitted */
/*--------------------------------------------------------------------*/
/* default mask indicating pixels included in fit */
/* the central pixel is excluded */
static int _Mask[5][5] = {{ 1, 1, 1, 1, 1},
                          { 1, 1, 1, 1, 1},
                          { 1, 1, 0, 1, 1},
                          { 1, 1, 1, 1, 1},
                          { 1, 1, 1, 1, 1}};
/*----------------------------------------------------------------------
	"solve" calculates the solution of linear system, A*x = b
	Do not use it if 'decomp' has detected singularity.
	Input:
      n      Order of matrix.
      *A[]   Triangular matrix A obtained from 'decomp'.
      b[]    Right hand side vector, b.
      ipvt[] Pivot vector obtained from 'decomp'.
	Output:
      b[]    Solution vector, x.
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 55.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 25-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void solve(int n, double * const A[], double b[], const int ipvt[])
#else	/* old style */
void solve(n, A, b, ipvt)
int n;
double * const A[];
double b[];
const int ipvt[];
#endif
{
	int k;

	if (n > 1) {

		/* forward elimination */
		for (k = 0; k < (n-1); ++k) {
			int i, m;
			double t;

			m = ipvt[k];
			t = b[m];
			b[m] = b[k];
			b[k] = t;

			for (i = (k+1); i < n; ++i)
				b[i] += A[k][i]*t;
		}

		/* back substitution */
		for (k = (n-1); k > 0; --k) {
			int i;
			double t;

			b[k] /= A[k][k];
			t = b[k];
			for (i = 0; i < k; ++i)
				b[i] -= A[k][i]*t;
		}
	}
	b[0] /= A[0][0];
}
/*----------------------------------------------------------------------
	"decomp" decomposes a square matrix A by Gaussian elimination and
	estimates the condition of the matrix.
	Use "solve" to compute solutions to linear systems.
	Input:
      n      Order of square matrix.
      *A[]   The matrix A to be triangularized.

	Output:
		A contains an upper triangular matrix U and a permuted
		version of a lower triangular matrix I-L so that
		(Permutation Matrix)*A = L*U.

	Return:
		An estimate of the 'condition' of A.

		For the linear system A*x = b, changes in A and b MAY
		cause changes 'condition' times as large in x.
		If (condition+1.0 == condition), A is singular to working
		precision. condition is set to 1.0e+32 if exact singularity
		is detected.

      ipvt[]    The pivot vector.
      ipvt[k]   The index of the k-th pivot row.
      ipvt[n-1] (-1)**(Number of Interchanges)

	Work Space:
		The wector work[] must be declared and included in the
		call. Its input contents are ignored. Its output
		contents are usually unimportant.

	The determinant of A can be obtained on output by
		Det(A) = ipvt[n-1]*A[0][0]*A[1][1]*......*A[n-1][n-1]
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 52.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 25-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double decomp(int n, double *A[], int ipvt[], double work[])
#else	/* old style */
double decomp(n, A, ipvt, work)
int n, ipvt[];
double *A[], work[];
#endif
{
	int k;
	double anorm, ynorm, znorm, cond;
	const double Big = 1.0e+32;
	const double One = 1.0;
	const double Zero = 0.0;

	ipvt[n-1] = 1;

	/* 1-by-1 */
	if (n == 1) {
		cond = (A[0][0] != Zero)? One: Big;
		return cond;
	}

	/* compute 1-norm of A */
	anorm = Zero;
	for (k = 0; k < n; ++k) {
		int i;
		double t;

		t = Zero;
		for (i = 0; i < n; ++i)
			t += ABS(A[k][i]);
		if (t > anorm) anorm = t;
	}

	/* Gaussian elimination with partial pivoting */
	for (k = 0; k < (n-1); ++k) {
		int i, j, m;
		double t;

		/* find pivot */
		m = k;
		for (i = (k+1); i < n; ++i)
			if (ABS(A[k][i]) > ABS(A[k][m])) m = i;
		ipvt[k] = m;
		if(m != k) ipvt[n-1] = -ipvt[n-1];
		t = A[k][m];
		A[k][m] = A[k][k];
		A[k][k] = t;

		/* skip step if pivot is zero */
		if (t != Zero) {

			/* compute multipliers */
			for (i = (k+1); i < n; ++i)
				A[k][i] /= -t;

			/* interchange and eliminate by columns */
			for (j = (k+1); j < n; ++j) {
				t = A[j][m];
				A[j][m] = A[j][k];
				A[j][k] = t;
				if (t != Zero)
					for (i = (k+1); i < n; ++i)
						A[j][i] += A[k][i]*t;
			}
		}
	}
	/*
	cond = (1-norm of A)*(an estimate of 1-norm of A-inverse)
	Estimate obtained by one step of inverse iteration for the
	small singular vector. This involves solving two systems
	of equations, (A-transpose)*Y = E and A*Z = Y where E is
	a vector of +1 or -1 chosen to cause growth in Y.
	Estimate = (1-norm of Z)/(1-norm of Y)
	*/

	/* solve (A-transpose)*Y = E */
	for (k = 0; k < n; ++k) {
		int i;
		double ek, t;

		t = Zero;
		if (k > 0)
			for (i = 0; i < k; ++i)
				t += A[k][i]*work[i];
		ek = (t < Zero)? -One: One;
		if (A[k][k] == Zero) {
			cond = Big;
			return cond;
		}
		work[k] = -(ek + t)/A[k][k];
	}
	for (k = (n-2); k >= 0; --k) {
		int i, m;
		double t;

		t = Zero;
		for (i = (k+1); i < n; ++i)
			t += A[k][i]*work[k];
		work[k] = t;
		m = ipvt[k];
		if (m != k) {
			t = work[m];
			work[m] = work[k];
			work[k] = t;
		}
	}
	ynorm = Zero;
	for (k = 0; k < n; ++k)
		ynorm += ABS(work[k]);

	/* solve A*Z = Y */
	solve(n, A, work, ipvt);

	znorm = Zero;
	for (k = 0; k < n; ++k)
		znorm += ABS(work[k]);

	/* estimate condition */
	cond = anorm*znorm/ynorm;
	if (cond < One) cond = One;
	return cond;
}
/*----------------------------------------------------------------------
   "variance" derives the variance from data using gain and read noise.
   Input:
      gain    CCD gain in electrons per ADU.
      rnoise  CCD rms read noise in electrons.
      serror  CCD scale error as a fraction.
      d[5][5] Data array in ADU units.
   Output:
      v[5][5] Variance in ADU squared.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 30-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void variance(
double gain,
double rnoise,
double serror,
float (d[])[5],
float (v[])[5])
#else	/* old style */
void variance(gain, rnoise, serror, d, v)
double gain;	/* gain in electrons per ADU */
double rnoise;	/* read noise in electrons */
double serror;	/* scale error as a fraction */
float (d[])[5];	/* data array in ADU units */
float (v[])[5];	/* variance in ADU squared */
#endif
{
	int i;
	float *dat = (float *)d;
	float *var = (float *)v;
	double s, v0 = rnoise/gain;
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);

	v0 *= v0;
	for (i = 0; i < 25; ++i)
		if (dat[i] <= FltMax) {
			s = serror*dat[i];
			s *= s;
			var[i] = (float) (dat[i]/gain + s + v0);
		} else
			var[i] = FltBlank;
}
/*--------------------------------------------------------------------*/
static float _F[NFIT][5][5];	/* the NFIT functions to be fitted */
static int _m[25];	/* mask indicating pixels to be included in fit */
/*----------------------------------------------------------------------
   "InitMask" initializes the pixel mask, and the function array.
   Globals:
      _F[NFIT][5][5] Basic set of fitting functions.
      _Mask[5][5]    Default pixel mask.
      _m[25]         Current pixel mask.
   Input:
      mode     4th order term in radius (0), in x (1), or in y (-1).
      d[5][5]  Data array in ADU units.
   Return:
      Number of valid pixels that should be included in fit.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 25-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int InitMask(int mode, float (d[])[5])
#else	/* old style */
static int InitMask(mode, d)
int mode;	/* indicates which 4th order term should be included */
float (d[])[5];	/* data array in ADU units */
#endif
{
	int i, x, y, ndat = 0;
	int *mask = (int *)_Mask;
	float *dat = (float *)d;
	const float FltMax = (float) (0.9*FLT_BLANK);

	/* initialize the pixel mask */
	for (i = 0; i < 25; ++i)
		_m[i] = (dat[i] > FltMax)? 0: mask[i];

	/* count the number of valid pixels */
	for (i = 0; i < 25; ++i)
		if (_m[i]) ndat++;

	/* initialize the function matrices */
	for (y = -2; y <= 2; ++y) {
		long yy = y*y - 2;
		for (x = -2; x <= 2; ++x) {
			long xx = x*x - 2;
			long xy = x*y;
			long rr = xx + yy;
			_F[0][y+2][x+2] = 1;
			_F[1][y+2][x+2] = x;
			_F[2][y+2][x+2] = y;
			_F[3][y+2][x+2] = rr;
			_F[4][y+2][x+2] = xx - yy;
			_F[5][y+2][x+2] = xy;
			_F[6][y+2][x+2] = ((5*xx - 7)*x)/6;
			_F[7][y+2][x+2] = ((5*yy - 7)*y)/6;
			_F[8][y+2][x+2] = yy*x;
			_F[9][y+2][x+2] = xx*y;
			if (mode > 0)	/* 4th order term in x */
				_F[10][y+2][x+2] = (5*(7*xx - 3)*xx - 98)/12;
			else if (mode < 0)	/* 4th order term in y */
				_F[10][y+2][x+2] = (5*(7*yy - 3)*yy - 98)/12;
			else	/* 4th order term in radius */
				_F[10][y+2][x+2] = (5*(7*rr - 3)*rr - 196)/2;
		}
	}
	return ndat;
}
/*----------------------------------------------------------------------
   "RejectOne" conditionally rejects the one pixel with largest residual.
   Globals:
      _m[25]  Mask indicating pixels which should be included.
   Input:
      kappa   Data point is rejected if abs(r[][]) > kappa*sigma.
      v[5][5] Variance array.
      r[5][5] Residual array.
   Return:
      1: if 1 point is rejected, 0: if none is rejected.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 21-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int RejectOne(double kappa, float (v[])[5], float (r[])[5])
#else	/* old style */
static int RejectOne(kappa, v, r)
double kappa;	/* rejection limit is kappa*sqrt(v[][]) */
float (v[])[5];	/* variance array */
float (r[])[5];	/* residual array */
#endif
{
	int i, imax = 0, reject = 0;
	float vmax;
	float *var = (float *)v;
	float *res = (float *)r;
	const float Zero = (float)0;

	vmax = Zero;
	for (i = 0; i < 25; ++i)
		if (_m[i]) {
			float rabs = ABS(res[i]);
			if (vmax < rabs) {
				imax = i;
				vmax = rabs;
			}
		}
	vmax /= kappa;
	vmax *= vmax;
	if (vmax > var[imax]) {
		_m[imax] = 0;
		reject = 1;
	}
	return reject;
}
/*----------------------------------------------------------------------
   "prod" calculates a weighted inner product of two arrays.
   Globals:
      _m[25]  Terms are only included in sum if _m[] != 0.
   Input:
      x[5][5] First array.
      y[5][5] Second array.
      v[5][5] Variance array.
   Return:
      The inner product.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 21-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double prod(float (x[])[5], float (y[])[5], float (v[])[5])
#else	/* old style */
double prod(x, y, v)
float (x[])[5];	/* first array */
float (y[])[5];	/* second array */
float (v[])[5];	/* weight = 1/v[][] */
#endif
{
	int i;
	float *p = (float *)x;
	float *q = (float *)y;
	float *var = (float *)v;
	double sum = 0.0;

	for (i = 0; i < 25; ++i)
		if (_m[i])
			sum += p[i]*q[i]/var[i];

	return sum;
}
/*----------------------------------------------------------------------
   "lincomb" calculates a linear combination of basic functions.
   Globals:
      _F[NFIT][5][5] Basic set of fitting functions.
   Input:
      p[NFIT]        The NFIT fitting parameters.
   Output:
      y[5][5]        Resulting linear combination.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 21-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void lincomb(double p[], float (y[])[5])
#else	/* old style */
void lincomb(p, y)
double p[];	/* fitted parameters */
float (y[])[5];	/* fitted data array */
#endif
{
	int i, k;
	float (*fun)[25] = (float (*)[25])_F;
	float *fit = (float *)y;

	for (i = 0; i < 25; ++i) {
		double sum = 0.0;

		for (k = 0; k < NFIT; ++k)
			sum += p[k] * fun[k][i];
		fit[i] = (float) sum;
	}
}
/*----------------------------------------------------------------------
   "matsub" calculates a the residual array.
   Input:
      d[5][5] Data array.
      y[5][5] Fitted array.
   Output:
      r[5][5] Residual array.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 21-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void matsub(float (d[])[5], float (y[])[5], float (r[])[5])
#else	/* old style */
void matsub(d, y, r)
float (d[])[5];	/* data array */
float (y[])[5];	/* fitted data array */
float (r[])[5];	/* residual array */
#endif
{
	int i;
	float *dat = (float *)d;
	float *fit = (float *)y;
	float *res = (float *)r;
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);

	for (i = 0; i < 25; ++i)
		res[i] = (dat[i] > FltMax)? FltBlank: dat[i] - fit[i];
}
/*----------------------------------------------------------------------
   "normal" calculates the Normal Equations.
   Globals:
      _F[NFIT][5][5] Basic set of fitting functions.
   Input:
      d[5][5]        Data array.
      v[5][5]        Variance array.
   Output:
      a[NFIT][NFIT]  Left hand side of Normal Equations.
      b[NFIT]        Right hand side of Normal Equations.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 25-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void normal(float (d[])[5], float (v[])[5], double (a[])[NFIT], double b[])
#else	/* old style */
void normal(d, v, a, b)
float (d[])[5];	/* data array */
float (v[])[5];	/* variance array */
double (a[])[NFIT];
double b[];
#endif
{
	int i, j;

	/* calculate the Normal Equations */
	/* assign lower triangular part of A */
	for (j = 0; j < NFIT; ++j) {
		b[j] = prod(_F[j], d, v);
		for (i = 0; i <= j; ++i)
			a[j][i] = prod(_F[j], _F[i], v);
	}

	/* assign upper triangular part of A */
	for (j = 1; j < NFIT; ++j)
		for (i = 0; i < j; ++i)
				a[i][j] = a[j][i];
}
/*----------------------------------------------------------------------
   "iscr" checks if the central pixel is a cosmic ray.
   Input:
      mode     Indicates which 4th order term should be included in fit.
               (0: 4th order in r, 1: 4th order in x, -1: 4th oder in y)
      gain     CCD gain in electrons per ADU.
      rnoise   CCD rms read noise in electrons.
      serror   CCD scale error as a fraction.
      kappa    The criterion is (d[2][2] - y[2][2]) > kappa*sigma[2][2].
      npt      Max No of data points rejected from fit.
   Input/Output:
      d[5][5]  Data array. If yes: d[2][2] = y[2][2] (fitted value).
   Return:
      1: Yes, 0: No.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 30-Aug-93
----------------------------------------------------------------------*/
#define KAPPA	2.0	/* Rejection factor for data excluded from fit */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int iscr(
int mode,
double gain,
double rnoise,
double serror,
double kappa,
int npt,
float (d[])[5])
#else	/* old style */
int iscr(mode, gain, rnoise, serror, kappa, npt, d)
int mode;	/* which 4th order term should be included? */
double gain;	/* gain in electrons per ADU */
double rnoise;	/* read noise in electrons */
double serror;	/* scale error as a fraction */
double kappa;	/* rejection factor */
int npt;	/* max number of points rejected */
float (d[])[5];	/* data array */
#endif
{
	int ipvt[NFIT], ndat, k;
	double a[NFIT][NFIT], *A[NFIT], p[NFIT], work[NFIT], cond;
	float v[5][5], y[5][5], r[5][5];
	const float FltMax = (float) (0.9*FLT_BLANK);
	const double One = 1.0;

	/* check if the central pixel is marked as BLANK? */
	if (d[2][2] > FltMax)
		return 0;

	/* assign pointer array */
	for (k = 0; k < NFIT; ++k)
		A[k] = a[k];

	/* initialize the pixel mask, and the function matrices */
	ndat = InitMask(mode, d);	/* No. of data points in fit */

	/* check if there are too few valid pixels */
	if (ndat < (NDAT-npt))
		return 0;

	/* calculate the variance from data, gain, read-noise, and scale error */
	variance(gain, rnoise, serror, d, v);

	/* calculate the Normal Equations */
	normal(d, v, a, p);

	/* decompose the A matrix */
	cond = decomp(NFIT, A, ipvt, work);

	/* check if A is singular to working precision */
	if ((cond + One) == cond)	/* solution is impossible */
		return 0;

	/* solve the Normal Equations */
	solve(NFIT, A, p, ipvt);

	/* calculate the fitted array */
	lincomb(p, y);

	/* calculate the residual array */
	matsub(d, y, r);

	for (k = 0; k < npt; ++k) {
		if (RejectOne(KAPPA, v, r) == 0)
			break;

		/* re-calculate the Normal Equations */
		normal(d, v, a, p);

		/* decompose the new A matrix */
		cond = decomp(NFIT, A, ipvt, work);

		/* check if A is singular to working precision */
		if ((cond + One) == cond)	/* solution is impossible */
			break;

		/* solve the new Normal Equations */
		solve(NFIT, A, p, ipvt);

		/* calculate the fitted array */
		lincomb(p, y);

		/* calculate the residual array */
		matsub(d, y, r);
	}

	/* is the central data point d[2][2] likely to be a cosmic ray? */
	if (r[2][2] < kappa*sqrt(v[2][2]))
		return 0;	/* No! */

	/* replace the central data point with the fitted value */
	d[2][2] = y[2][2];
	return 1;	/* Yes */
}
/*----------------------------------------------------------------------
   "isspike" checks if the central pixel is a spike.
   Input:
      gain     CCD gain in electrons per ADU.
      rnoise   CCD rms read noise in electrons.
      serror   CCD scale error as a fraction.
      d[5][5]  Data array.
   Return:
      1: Yes, 0: No.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 30-Aug-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int isspike(double gain, double rnoise, double serror, float (d[])[5])
#else	/* old style */
int isspike(gain, rnoise, serror, d)
double gain;	/* gain in electrons per ADU */
double rnoise;	/* read noise in electrons */
double serror;	/* scale error as a fraction */
float (d[])[5];	/* data array */
#endif
{
	double s, v0 = rnoise/gain;
	float dat, x = (float) (-FLT_BLANK);
	const float FltMax = (float) (0.9*FLT_BLANK);

	/* check if the central pixel is marked as BLANK? */
	if (d[2][2] > FltMax)
		return 0;

	/* find the maximum of the 4 sums of opposite pixels */
	if (d[2][1] <= FltMax && d[2][3] <= FltMax) {
		dat = d[2][1] + d[2][3];
		x = MAX(dat, x);
	}
	if (d[1][2] <= FltMax && d[3][2] <= FltMax) {
		dat = d[1][2] + d[3][2];
		x = MAX(dat, x);
	}
	if (d[1][1] <= FltMax && d[3][3] <= FltMax) {
		dat = d[1][1] + d[3][3];
		x = MAX(dat, x);
	}
	if (d[1][3] <= FltMax && d[3][1] <= FltMax) {
		dat = d[1][3] + d[3][1];
		x = MAX(dat, x);
	}

	/* all 8 neighbour pixels are marked as BLANK */
	if (x < -FltMax)
		return 0;

	v0 *= v0;
	s = serror*d[2][2];	/* scale error */
	s *= s;
	dat = (float) (sqrt(d[2][2]/gain + s + v0) + x/2);

	return ((d[2][2] > dat)? 1: 0);
}
/*----------------------------------------------------------------------
	"cosmic" detects and replaces cosmic ray events.

	A pixel is suspected to be a spike if it is greater than the maximum
	of the 4 means of opposite pairs of pixels plus the standard deviation.
	All polynomial terms in (x, y) up to the 3rd order plus a 4th order
	(in r, x, or y) term are fitted to a 5x5 square region centred on any
	suspected spike (the central pixel is omitted from the fit).
	The fit is repeated after the one pixel having the largest residual
	has been omitted. This pixel rejection procedure can be repeated
	several times. If the residual of the central pixel is larger than
	kappa*sigma the pixel value is replaced with the fitted value.
	NOTE, the image data should be over sampled (PSF FWHM > 5 pixels).
   The sample spacing can be chosen as 1 pixel (default), 1.41 pixel
   (-m2), or 2 pixels (-m3).
   The -m2 option should be chosen when the PSF is well sampled.
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 25-Aug-93, 07-Dec-93
----------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv, **arg;

	int option, opt_err = 0, n_arg = 0;
	int mode = 0;	/* rotationally symmetric 4th order term */
	double gain = 3.0;	/* gain in electrons per ADU */
	double rnoise = 7.0;	/* read noise in electrons */
	double serror = 0.01;	/* scale error as a fraction */
	double kappa = 4.0;	/* rejection limit in sigma units */
	int npt = 1;	/* max number of rejected points per fit */
	int nit = 10;	/*	number of iterations */
	int method = 1;	/* method: 1, 2, 3: spacing = 1, 1.41, 2 pix */
	int step = 1;

	char *image;	/* input image */
	char *image1;	/* output image */

	int i, y, n;
	int id, type, ncols, nrows;
	double Blank;

	void *row;	/* row buffer */
	float **a;	/* pixel array a[][] allocated by daa */
	char *aptr;	/* base pointer of array space for a[][] */
	float **b;	/* pixel array b[][] allocated by daa */
	char *bptr;	/* base pointer of array space for b[][] */
	unsigned int err;	/* error code returned by daa */
	int start[2], dim[2];	/* starting indices and dimensions for daa */
	float d[5][5];	/* temporary data array */

	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Half = (float) 0.5;
	const float Zero = (float) 0;
	float Init_Value = FltBlank;	/* initial value for a[][] */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "cosmic: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	/* process command line options */
	while (((option = getopt(argc, argv, "xyg:n:s:k:r:i:m:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'x':	mode++; break;	/* 4th order term along x-axis */
			case 'y':	mode--; break;	/* 4th order term along y-axis */
			case 'g':	sscanf(optarg, "%lf", &gain); break;
			case 'n':	sscanf(optarg, "%lf", &rnoise); break;
			case 's':	sscanf(optarg, "%lf", &serror); break;
			case 'k':	sscanf(optarg, "%lf", &kappa); break;
			case 'r':	sscanf(optarg, "%d", &npt); npt=MIN(npt,4); break;
			case 'i':	sscanf(optarg, "%d", &nit); break;
			case 'm':	sscanf(optarg, "%d", &method);
							method = MAX(1, method); method = MIN(3, method);
							step = (method > 1)? 2: 1; break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	/* illegal options */
	if (opt_err || n_arg < 2) {
		printf(
"cosmic -x -y -g gain -n noise -s err -k kap -r npt -i nit -m met img img1\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);
	/* is image name legal? */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "cosmic: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[1]);
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "cosmic: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "cosmic: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open image for read only */
	if ((id = iopen(image, 0, &type, &ncols, &nrows)) < 0) {
		fprintf(stderr, "cosmic: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr, "cosmic: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	if (xist(image1)) {	/* does output image exist? */
		fprintf(stderr, "cosmic: %s: Existing image\n", image1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	dim[0] = nrows;	/* y-dimension */
	dim[1] = ncols;	/* x-dimension */
	start[0] = 0;	/* first index of y-dimension */
	start[1] = 0;	/* first index of x-dimension */

	/* allocate float array a[][] */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &aptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "cosmic: daa: Error code = %d\n", err);
		fprintf(stderr,
			"cosmic: Can't allocate a[%d][%d]\n", dim[0], dim[1]);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	dim[0] = 4*step+1;	/* number of rows in buffer */

	/* allocate float array buffer b[][] */
	if ((b = (float **)daa(sizeof(float), 2, dim, start,
			&err, &bptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "cosmic: daa: Error code = %d\n", err);
		fprintf(stderr,
			"cosmic: Can't allocate b[%d][%d]\n", dim[0], dim[1]);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate row buffer */
	if ((row = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "cosmic: Can't allocate row[%d]\n", ncols);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nrows; ++y) {	/* read image into array a[][] */
		getline(id, y, row);	/* read next row */
		switch (type) {
			case 1:
				{
					const UCHAR *q = (UCHAR *)row;
					float *p = a[y];
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (float)q[x];
				}
				break;
			case 2:
				{
					const short *q = (short *)row;
					float *p = a[y];
					const short b = (short)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:
				{
					const float *q = (float *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:
				{
					const LONG *q = (LONG *)row;
					float *p = a[y];
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:
				{
					const COMPLEX *q = (COMPLEX *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x].x > h || q[x].y > h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);	/* close image file */

	/* start iterating the search for cosmic ray events */
	for (n = 0; n < nit; ++n) {	/* next pass through image */
		int step2 = 2*step;
		int step4 = 4*step;

		/* write iteration number to screen */
		printf("Pass %d\r", n+1);
		fflush(stdout);

		/* copy the first (4*step) lines into the (4*step+1)-row buffer */
		for (y = 0; y < step4; ++y) {
			int x;
			for (x = 0; x < ncols; ++x)
				b[y+1][x] = a[y][x];
		}

		/* loop through rows */
		for (y = step4; y < nrows; ++y) {	/* process next row */
			int x, j;

			/* shift 1 row back in buffer */
			for (j = 0; j < step4; ++j)
				for (x = 0; x < ncols; ++x)
					b[j][x] = b[j+1][x];

			/* copy next row from image into buffer */
			for (x = 0; x < ncols; ++x)
				b[step4][x] = a[y][x];

			/* loop througt pixels */
			for (x = step2; x < ncols-step2; ++x) {	/* process next pixel */
				int i, j;

				/* copy next 5x5 data matrix from buffer */
				switch (method) {
					case 1:
						for (i = 0; i < 5; ++i) {
							int k, l = x - 2;
							for (k = 0; k < 5; ++k)
								d[i][k] = b[i][l+k];
						}
						break;
					case 2:
						for (i = 0; i < 5; ++i) {
							int k, l = x - i;
							for (k = 0; k < 5; ++k)
								d[i][k] = b[i+k][l+k];
						}
						break;
					case 3:
						j = 0;
						for (i = 0; i < 5; ++i) {
							int k, l = x - 4;
							for (k = 0; k < 5; ++k) {
								d[i][k] = b[j][l];
								l += 2;
							}
							j += 2;
						}
						break;
				}

				/* is the central pixel a spike? is it a cosmic ray? */
				if (isspike(gain, rnoise, serror, d))
					if (iscr(mode, gain, rnoise, serror, kappa, npt, d))
						a[y-step2][x] = d[2][2];	/* replace with fitted value */
			}
		}
	}

	if (cpi(image, image1) < 1) {	/* dublicate image header */
		fprintf(stderr,
			"cosmic: %s.hdr -> %s.hdr copy failed\n", image, image1);
		goto Finish;
	} else if (setimagefile(image1, "") == 0) {	/* reset FILE key */
		fprintf(stderr,
			"cosmic: Can't erase old FILE value for %s\n", image1);
		rmi(image1);	/* remove image header */
		goto Finish;
	}

	if ((id = icreat(image1, 1, type, ncols, nrows)) < 0) {
		fprintf(stderr, "cosmic: Can't create %s\n", image1);
		rmi(image1);	/* remove image header */
		goto Finish;
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "cosmic: update: Disk header not created\n");

	Blank = blank(image1);

	for (y = 0; y < nrows; ++y) {	/* write image to disk */
		switch (type) {
			case 1:
				{
					UCHAR *q = (UCHAR *)row;
					const float *p = a[y];
					int x;
					for (x = 0; x < ncols; ++x)
						q[x] = (UCHAR)(p[x] + Half);
				}
				break;
			case 2:
				{
					short *q = (short *)row;
					const float *p = a[y];
					const short b = (short)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b:
								(short)((p[x]<Zero)? p[x]-Half: p[x]+Half);
				}
				break;
			case 3:
				{
					float *q = (float *)row;
					const float *p = a[y];
					const float b = (float)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b: p[x];
				}
				break;
			case 4:
				{
					LONG *q = (LONG *)row;
					const float *p = a[y];
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b:
								(LONG)((p[x]<Zero)? p[x]-Half: p[x]+Half);
				}
				break;
			case 5:
				{
					COMPLEX *q = (COMPLEX *)row;
					const float *p = a[y];
					const float b = (float)Blank;
					int x, ip;
					for (x = 0; x < ncols; ++x) {
						ip = p[x] > FltMax;
						q[x].x = (ip)? b: p[x];
						q[x].y = (ip)? b: Zero;
					}
				}
				break;
		}
		putline(id, y, row);	/* write row to disk */
	}

	iclose(id);	/* close output image */
	setiname(image1);
	sethistory(image1, "cosmic");
	sethistory(image1, image1);

Finish:

	/* free row[] */
	free(row);
	/* free buffer array */
	free(bptr);
	/* free pixel array */
	free(aptr);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
