
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "wild.h"	/* function prototype for wild.c */
#include "daa.h"	/* function declaration for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define IMAX	 20 /* maximum number of input images */
#define MAXLEN  16 /* maximum length of output file name */
/*--------------------------------------------------------------------*/
#define F_OK    0       /* pixel is thought to be OK */
#define F_DNB   1       /* diagonal neighbor of a CR */
#define F_NNB   2       /* nearest neighbor of a CR */
#define F_DCR   3       /* diagonal neighbor decided to be CR */
#define F_NCR   4       /* nearest neighbor decided to be CR */
#define F_CR    5       /* pixel decided to be a CR (on its own) */
#define F_UND   6       /* pixel that started out undefined */
#define F_SAT   7       /* pixel that started out saturated */
/*--------------------------------------------------------------------*/
#define RO      0       /* read only */
#define WO      1       /* write only */
/*--------------------------------------------------------------------*/
typedef struct {     /* input parameter structure */
   double rnoise;    /* read noise in electrons */
   double gain;      /* gain in electrons per ADU */
   double frac;      /* fractional error */
   double dark;      /* dark rate in ADU/s */
   double blank;     /* blank value */
   double etime;     /* exposure time */
   double dtime;     /* dark exposure time */
   double sat;       /* saturation level */
   double rg2;       /* read noise over gain squared */
   double ig;        /* inverse gain */
   double dc;        /* dark current in electrons */
   double rlev;      /* relative signal level */
   double rlev2;     /* relative signal level squared */
   int nuntest;      /* number of untested pixels */
   int nnuntest;     /* number of untested nearest neighbors of CRs */
   int nduntest;     /* number of untested diagonal neighbors of CRs */
   int nok;          /* number of OK pixels */
   int nsat;         /* number of pixels that started out saturated */
   int nundef;       /* number of pixels that started out undefined */
   int ncr;          /* number of cosmic rays from main test */
   int nncr;         /* number of cosmic rays from nearest neigh. test */
   int ndcr;         /* number of cosmic rays from diagonal neigh test */
} IPARAM;
/*--------------------------------------------------------------------*/
typedef struct {     /* output parameter structure */
   double effread;   /* effective read noise */
   double effgain;   /* effective gain */
   double ogain;     /* output gain */
   double otime;     /* output time */
   int naundef;      /* number of undefined pixels in output */
   int naok;         /* number of OK pixels in output */
} OPARAM;
/*--------------------------------------------------------------------*/
typedef struct {     /* control flag structure */
   int above;        /* line above NN row */
   int nnrow;        /* nearest neighbor row */
   int below;        /* line below NN row */
   int do_top;       /* look in line above */
   int do_bottom;    /* look in line below */
   int do_two_sided; /* do a two sided search */
} DOFLAG;
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void cptoflt(int ncols,
int type,
double blank,
void const *row,
float frow[])
#else	/* old style */
void cptoflt(ncols, type, blank, row, frow)
int ncols, type;
double blank;
void const *row;
float frow[];
#endif
{
   const float FltBlank = (float) FLT_BLANK;
   switch (type) {
      case 1:
             {
                const UCHAR *q = (UCHAR *)row;
                float *p = frow;
                int x;
                for (x = 0; x < ncols; ++x)
                   p[x] = (float)q[x];
             }
             break;
      case 2:
             {
                const short *q = (short *)row;
                float *p = frow;
                const short b = (short)blank;
                int x;
                for (x = 0; x < ncols; ++x)
                   p[x] = (q[x] == b)? FltBlank: (float)q[x];
             }
             break;
      case 3:
             {
                const float *q = (float *)row;
                float *p = frow;
                const float h = (float) (0.9*blank);
                int x;
                for (x = 0; x < ncols; ++x)
                   p[x] = (q[x] > h)? FltBlank: q[x];
             }
             break;
      case 4:
             {
                const LONG *q = (LONG *)row;
                float *p = frow;
                const LONG b = (LONG)blank;
                int x;
                for (x = 0; x < ncols; ++x)
                   p[x] = (q[x] == b)? FltBlank: (float)q[x];
             }
             break;
      case 5:
             {
                const COMPLEX *q = (COMPLEX *)row;
                float *p = frow;
                const float h = (float) (0.9*blank);
                int x;
                for (x = 0; x < ncols; ++x)
                   p[x] = (q[x].x > h || q[x].y > h)? FltBlank: q[x].x;
             }
             break;
   }
}
/*--------------------------------------------------------------------
  This routine is used by the cosmic ray split program to
  process a "main" line.  (As opposed to a nearest neighbor line)
  Lines for all images to be compared are read in and converted to
  internal data representation.  Variance arrays are filled and the
  flag array is filled with indications of undefined.
  Then we go through all points and look for cosmic rays.
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void main_row(DOFLAG flag,
double thresh,
int nimg,
int ncols,
IPARAM ip[],
float **a[],
float **w[],
char **f[])
#else	/* old style */
void main_row(flag, thresh, nimg, ncols, ip, a, w, f)
DOFLAG flag;
double thresh;
int nimg, ncols;
IPARAM ip[];
float **a[], **w[];
char **f[];
#endif
{
   const float FltBlank = (float) FLT_BLANK;
   const float FltMax = (float) (0.9*FLT_BLANK);
   double thresh2 = thresh*thresh; /* threshold squared */
   int x, j;

   /* Loop over the input images */
   for (j = 0; j < nimg; j++) {
      float s = (float) ip[j].sat;     /* saturation level */
      float h = FltMax;                /* undefined above this */
      float *d = &a[flag.above][j][0]; /* pixel values on input line */
      float *v = &w[flag.above][j][0]; /* pixel weights */
      char  *c = &f[flag.above][j][0]; /* pixel quality flags */ 
      int x;

      /* Now we need to go through the arrays checking for */
      /* saturated/undefined values */
      for (x = 0; x < ncols; x++) {

         /* We may have undefined data */
         if (d[x] > h) {
            /* Here we have undefined values */
            c[x] = F_UND;
         } else if (d[x] >= s) {

            /* Here we may have saturated values in the data */
            /* - set them undefined */
            d[x] = FltBlank;
            c[x] = F_SAT;
         } else
            c[x] = F_OK;
      }

      /* OK, now the data buffer for this line is filled with */
      /* the data in DN-units. And the flag buffer has been */
      /* filled with flags indicating if the data were OK, */
      /* undefined, or saturated. Fill the weight buffer with the */
      /* inverse variance in the data and then scale both the */
      /* data buffer and the weight buffer for the signal level */
      for (x = 0; x < ncols; x++) {
         if (c[x] == F_OK) {
            double dn, vv;
            dn = (double) d[x];
            dn = MAX(dn,1.0);
            vv = dn*ip[j].frac;
            vv = ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv;
            v[x] = (float) (1.0/(vv*ip[j].rlev2));
            d[x] = (float) (d[x]*ip[j].rlev);
         } else
            v[x] = FltBlank;
      }
   }

   /* OK, each line has been scaled and a weight has been calculated */
   /* Now we need to go through and determine which pixels are cosmic */
   /* ray hits. Loop over the pixels: */
   for (x = 0; x < ncols; x++) {
      int n_left = nimg;      /* Assume all lines contain good pixels */
      int first_time = 1;     /* Show it's the first time */

      /* For each pixel, we loop until there is only one left */
      /* or we don't throw out cosmic rays */
      while (n_left > 1) {

         /* Set the maximum/minimum value, average, variance */
         float v_big   =  FltBlank;
         float v_small =  FltBlank;
         int j_big     = -1;
         int j_small   = -1;
         float d_big   = -FltMax;
         float d_small =  FltMax;
         double sums   =  0.0;    /* Sum of the weights times the signal */
         double sumv   =  0.0;    /* Sum of the weights */
         n_left        =  0;      /* Reset number left */

         /* Loop over input images */
         for (j = 0; j < nimg; j++) {

            /* See if we have a good pixel */
            if (f[flag.above][j][x] == F_OK) {

               /* Get data value and weight */
               float dn = a[flag.above][j][x];
               float vv = w[flag.above][j][x];

               /* Add to sums */
               sums += vv*dn;
               sumv += vv;

               /* Increment the number we have left */
               n_left++;

               /* Is this one the largest one we have so far? */
               if (dn > d_big) {

                  /* If so, remember it. */
                  d_big = dn;
                  v_big = vv;
                  j_big = j;
               }

               /* Is this one the smallest one we have so far? */
               if (dn < d_small) {

                  /* If so, remember. */
                  d_small = dn;
                  v_small = vv;
                  j_small = j;
               }
            }
         }

         /* OK, now how many do we have left? */
         if (n_left > 1) {
            double dn, vv, r_big, r_small;
            double sumsb2 = 0.0;
            double sumvb2 = 0.0;

            /* Here, more than 1 is left - might the biggest be a */
            /* cosmic ray? Remove its contributions from the sums */
            double sumsb = sums - v_big*d_big;
            double sumvb = sumv - v_big;

            /* Compute variance and mean of all but the biggest */
            sumvb = 1.0/sumvb;   /* Variance */
            sumsb = sumvb*sumsb; /* Mean */
 
            /* Now SUMSB is the mean signal (corresponding to a signal */
            /* level of 1). But it's been calculated with weights that */
            /* are biased. That is a smaller signal, gives a smaller */
            /* variance and hence a larger weight. Thus the mean in SUMS */
            /* is biased towards a signal less than the true signal. */
            /* So we assume SUMS is the true signal, recalculate weights */
            /* and recalculate the mean and variance. */
            for (j = 0; j < nimg; j++) {
               if (f[flag.above][j][x] == F_OK && j != j_big) {

                  /* OK, here we have a pixel which is still OK and */
                  /* which is not the biggest (ie, the one under test). */
                  /* So compute its weight based on SUMS and then compute */
                  /* its contribution to the sum and the sum of the weights */
                  /* Scale the average to the signal level of this pixel */
                  dn = sumsb/ip[j].rlev;
                  dn = MAX(dn,1.0);
                  vv = dn*ip[j].frac;
                  vv = ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv;
                  vv = 1.0/(vv*ip[j].rlev2);
                  sumsb2 += vv*a[flag.above][j][x];
                  sumvb2 += vv;
               }
            }
            sumvb2 = 1.0/sumvb2;     /* New variance */
            sumsb2 = sumvb2*sumsb2;  /* New mean */

            /* Now SUMSB2 is the mean signal level estimated from all but */
            /* the largest pixel. This corresponds to the output signal */
            /* level. SUMVB2 is the estimated variance for this mean. */
            /* Now we need to estimate the variance we would have in */
            /* the largest one if its signal were SUMSB2 */
            dn = sumsb2/ip[j_big].rlev;
            dn = MAX(dn,1.0);
            vv = dn*ip[j_big].frac;
            vv = ip[j_big].rlev2 *
                 (ip[j_big].rg2 + dn*ip[j_big].ig + ip[j_big].dc + vv*vv);

            /* If the difference between the actual signal level and the */
            /* mean of the other pixels is greater than the threshold */
            /* times the square root of the total variance, we decide */
            /* it's a cosmic ray. */
            r_big = d_big - sumsb2;

            /* Compute the ratio of the squared difference to the variance */
            r_big = (r_big*r_big)/(sumvb2 + vv);

            /* See if we are doing two sided */
            if (flag.do_two_sided > 0) {
               double sumss2 = 0.0;
               double sumvs2 = 0.0;

               /* Here, we're doing two sided, so look at the smallest */
               /* pixel. Remove its contributions from the sums. */
               double sumss = sums - v_small*d_small;
               double sumvs = sumv - v_small;

               /* Compute variance and mean of all but the biggest */
               sumvs = 1.0/sumvs;    /* Variance */
               sumss = sumvs*sumss;  /* Mean */

               /* Now SUMSS is the mean signal (corresponding to the output */
               /* signal level). But it's been calculated with weights that */
               /* are biased. That is a smaller signal, gives a smaller */
               /* variance and hence a larger weight. Thus the mean in */
               /* SUMSS is biased towards a signal less than the true */
               /* signal. So we assume SUMSS is the true signal, */
               /* recalculate weights and recalculate the mean and variance */
               for (j = 0; j < nimg; j++) {
                  if (f[flag.above][j][x] == F_OK && j != j_small) {

                     /* OK, here we have a pixel which is still OK and */
                     /* which is not the smallest (ie, the one under test) */
                     /* So compute its weight based on SUMSS and then */
                     /* compute its contribution to the sum and the sum of */
                     /* the weights. Scale the average to the signal level */
                     /* of this pixel. */
                     dn = sumss/ip[j].rlev;
                     dn = MAX(dn,1.0);
                     vv = dn*ip[j].frac;
                     vv = ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv;
                     vv = 1.0/(vv*ip[j].rlev2);
                     sumss2 += vv*a[flag.above][j][x];
                     sumvs2 += vv;
                  }
               }
               sumvs2 = 1.0/sumvs2;	/* New variance */
               sumss2 = sumvs2*sumss2;	/* New mean */

               /* Now SUMSS2 is the mean signal level estimated from all */
               /* but the smallest pixel. This corresponds to the output */
               /* signal level. SUMVS2 is the estimated variance for this */
               /* mean. Now we need to estimate the variance we would have */
               /* in the smallest one if its signal were SUMSS2 */
               dn = sumss2/ip[j_small].rlev;
               dn = MAX(dn,1.0);
               vv = dn*ip[j_small].frac;
               vv =  ip[j_small].rlev2 *
                    (ip[j_small].rg2 + dn*ip[j_small].ig +
                     ip[j_small].dc + vv*vv);
               r_small = d_small - sumss2;

               /* Compute the ratio of the squared difference to the variance */
               r_small = (r_small*r_small)/(sumvs2 + vv);

               /* Now we've computed the ratio of the squared difference */
               /* between the biggest and smallest pixels and the means */
               /* of the others to the estimated variance.  We will throw */
               /* out whichever is larger, if it's also above the threshold */
               if (r_big >= r_small) {
                  if (r_big > thresh2) {

                     /* Here the biggest pixel is the farthest from the */
                     /* mean and is above threshold */
                     f[flag.above][j_big][x] = F_CR;

                     /* Make undefined */
                     a[flag.above][j_big][x] = FltBlank;

                     /* And decrement NLEFT */
                     n_left--;
                  } else

                     /* Here the one farthest from the mean is not above */
                     /* threshold, so set NLEFT=0 to terminate loop */
                     n_left = 0;
               } else {
                  if (r_small > thresh2) {

                     /* Here the smallest pixel is the farthest from the */
                     /* mean and is above threshold */
                     f[flag.above][j_small][x] = F_CR;

                     /* Make undefined */
                     a[flag.above][j_small][x] = FltBlank;

                     /* And decrement NLEFT */
                     n_left--;
                  } else

                     /* Here the one farthest from the mean is not above */
                     /* threshold, so set NLEFT=0 to terminate loop */
                     n_left = 0;
               }
            } else {

               /* Here we are doing only one sided testing, so look only */
               /* at the biggest */
               if (r_big > thresh2) {

                  /* Here the biggest pixel is the farthest from the */
                  /* mean and is above threshold */
                  f[flag.above][j_big][x] = F_CR;

                  /* Make undefined */
                  a[flag.above][j_big][x] = FltBlank;

                  /* And decrement NLEFT */
                  n_left--;
               } else

                  /* Here the one farthest from the mean is not above */
                  /* threshold, so set NLEFT=0 to terminate loop */
                  n_left = 0;
            }
         } else if (n_left == 1) {

            /* Here there was only one good pixel left. If this is the */
            /* first time through the loop, increment the count of */
            /* untested pixels for this file. */
            if (first_time != 0)
               (ip[j_big].nuntest)++;
         }

         /* Reset first time flag for next pass through the loop */
         first_time = 0;
      }
   }
}
/*--------------------------------------------------------------------
  This routine is used by the cosmic ray split program to process
  a line for cosmic rays that are neighbors of cosmic rays.
----------------------------------------------------------------------*/
#if     defined(ANSI)	/* ANSI-C */
void neigh_row(DOFLAG flag,
double nthresh,
double dthresh,
int nimg,
int ncols,
IPARAM ip[],
OPARAM *op,
float **a[],
float **w[],
char **f[])
#else	/* old style */
void neigh_row(flag, nthresh, dthresh, nimg, ncols, ip, op, a, w, f)
DOFLAG flag;
double nthresh, dthresh;
int nimg, ncols;
IPARAM ip[];
OPARAM *op;
float **a[], **w[];
char **f[];
#endif
{
   const float FltBlank = (float) FLT_BLANK;
   const double nthresh2 = nthresh*nthresh;
   const double dthresh2 = dthresh*dthresh;
   float *frow = &a[flag.below][0][0];  /* used as float buffer */
   char  *crow = &f[flag.below][0][0];  /* used as count buffer */
   int x, j;

   /* Loop over the images to set flags provided we really */
   /* are going to do neighbor tests */
   if (nthresh > 0.0 || dthresh > 0.0) {
      int x, j;

      for (j = 0; j < nimg; j++) {

         /* First do nearest neighbor tests if required */
         if (nthresh > 0.0) {
            char *c = &f[flag.nnrow][j][0];  /* pixel flags for NN line */
            int x;

            /* Loop over the pixels and look for left and right neighbors */
            for (x = 0; x < ncols; x++) {

               /* We don't need to do anything if we're not dealing */
               /* with a pixel that's still flagged as OK */
               if (c[x] == F_OK) {

                  /* Do neighbor on left side */
                  if (x > 0)
                     if (c[x-1] == F_CR) c[x] = F_NNB;

                  /* Do neighbor on right side */
                  if ((x+1) < ncols)
                     if (c[x+1] == F_CR) c[x] = F_NNB;
               }
            }

            /* Now look for nearest neighbor cosmic rays in */
            /* the line above, if we have a line above */
            if (flag.do_top) {
               char *c = &f[flag.nnrow][j][0];  /* flags for NN line */
               char *t = &f[flag.above][j][0];  /* flags above NN line */
               int x;

               /* Loop over the pixels and look for top neighbors */
               for (x = 0; x < ncols; x++) {

                  /* We don't need to do anything if we're not dealing */
                  /* with a pixel that's still flagged as OK */
                  if (c[x] == F_OK)
                     if (t[x] == F_CR) c[x] = F_NNB;
               }
            }

            /* Now look for nearest neighbor cosmic rays in the */
            /* line below if we have a line below */
            if (flag.do_bottom) {
               char *c = &f[flag.nnrow][j][0];  /* flags for NN line */
               char *b = &f[flag.below][j][0];  /* flags below NN line */
               int x;

               /* Loop over the pixels and look for bottom neighbors */
               for (x = 0; x < ncols; x++) {

                  /* We don't need to do anything if we're not dealing */
                  /* with a pixel that's still flagged as OK */
                  if (c[x] == F_OK)
                     if (b[x] == F_CR) c[x] = F_NNB;
               }
            }
         }

         /* Now look for diagonal neighbors if necessary */
         if (dthresh > 0.0) {

            /* Look for them in the line above if we have a line above */
            if (flag.do_top) {
               char *c = &f[flag.nnrow][j][0];  /* flags for NN line */
               char *t = &f[flag.above][j][0];  /* flags above NN line */
               int x;

               /* Loop over the pixels and look for neighbors */
               for (x = 0; x < ncols; x++) {

                  /* We don't need to do anything if we're not dealing */
                  /* with a pixel that's still flagged as OK */
                  if (c[x] == F_OK) {

                     /* Do neighbor on top left */
                     if (x > 0)
                        if (t[x-1] == F_CR) c[x] = F_DNB;

                     /* Do neighbor on top right */
                     if ((x+1) < ncols)
                        if (t[x+1] == F_CR) c[x] = F_DNB;
                  }
               }
            }

            /* Look for them in the line below if we have a line below */
            if (flag.do_bottom) {
               char *c = &f[flag.nnrow][j][0];  /* flags for NN line */
               char *b = &f[flag.below][j][0];  /* flags below NN line */
               int x;

               /* Loop over the pixels and look neighbors */
               for (x = 0; x < ncols; x++) {

                  /* We don't need to do anything if we're not dealing */
                  /* with a pixel that's still flagged as OK */
                  if (c[x] == F_OK) {

                     /* Do neighbor on bottom left */
                     if (x > 0)
                        if (b[x-1] == F_CR) c[x] = F_DNB;

                     /* Do neighbor on bottom right */
                     if ((x+1) < ncols)
                        if (b[x+1] == F_CR) c[x] = F_DNB;
                  }
               }
            }
         }
      }

      /* OK, now we have all the pixels in this line flagged - now loop */
      /* through the line and see which we have to perform tests on. */
      for (x = 0; x < ncols; x++) {

         /* Find out how many neighbor and how many good pixels there */
         /* are at this location */
         int n_neighbor = 0;
         int n_good = 0;
         int j;

         for (j = 0; j < nimg; j++) {
            int c = f[flag.nnrow][j][x];  /* flags for NN line */

            if (c == F_OK)
               n_good++;
            else if (c == F_NNB || c == F_DNB)
               n_neighbor++;
         }

         /* We only need to proceed if there were some neighbor flags set */
         if (n_neighbor > 0) {

            /* OK, we have some neighbor flags set - see if we have */
            /* anything to do a test with */
            if (n_good > 0) {

               /* OK, we have at least one good pixel at this spot - */
               /* compute the weighted mean of all the good pixels */
               double sums = 0.0;  /* Sum of weights times signals */
               double sumv = 0.0;  /* Sum of weights */
               double sums2 = 0.0;
               double sumv2 = 0.0;
               int j;

               for (j = 0; j < nimg; j++) {
                  if (f[flag.nnrow][j][x] == F_OK) {
                     sums += w[flag.nnrow][j][x]*a[flag.nnrow][j][x];
                     sumv += w[flag.nnrow][j][x];
                  }
               }

               /* OK, calculate the mean and the variance */
               sumv = 1.0/sumv;
               sums = sumv*sums;

               /* Now SUMS is the mean, but weighted with weights based */
               /* on the individual signal levels.  Recompute the mean */
               /* with weights based on the signal level from SUMS scaled */
               /* to the proper exposure time */
               for (j = 0; j < nimg; j++) {
                  if (f[flag.nnrow][j][x] == F_OK) {
                     double dn, vv;

                     /* OK, here we have a pixel which is OK and not a */
                     /* neighbor. So compute its weight based on SUMS and */
                     /* then compute its contribution to the sum and the */
                     /* sum of the weights. Scale the average to the */
                     /* signal level of this pixel */
                     dn = sums/ip[j].rlev;
                     dn = MAX(dn,1.0);
                     vv = dn*ip[j].frac;
                     vv = ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv;
                     vv = 1.0/(vv*ip[j].rlev2);
                     sums2 += vv*a[flag.nnrow][j][x];
                     sumv2 += vv;
                  }
               }
               sumv2 = 1.0/sumv2;    /* New variance */
               sums2 = sumv2*sums2;  /* New mean */

               /* Now loop over the images to find the neighbors and test */
               for (j = 0; j < nimg; j++) {
                  int c = f[flag.nnrow][j][x];
                  if (c == F_NNB || c == F_DNB) {
                     double test, dn, vv;

                     /* OK, we found a neighbor */
                     /* Compute difference between neighbor and mean */
                     test = a[flag.nnrow][j][x] - sums2;

                     /* Compute the variance for this pixel scaled */
                     /* to the appropriate signal level */
                     dn = sums2/ip[j].rlev;
                     dn = MAX(dn,1.0);
                     vv = dn*ip[j].frac;
                     vv = ip[j].rlev2 *
                         (ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv);

                     /* Now test */
                     if (test > 0.0) {
                        if (c == F_NNB) {
                           if (test*test > nthresh2*(sumv2 + vv)) {

                              /* Here it's a nearest neighbor cosmic ray */
                              f[flag.nnrow][j][x] = F_NCR;
                              a[flag.nnrow][j][x] = FltBlank;
                           }
                        } else {
                           if (test*test > dthresh2*(sumv2 + vv)) {

                              /* Here it's a diagonal neighbor cosmic ray */
                              f[flag.nnrow][j][x] = F_DCR;
                              a[flag.nnrow][j][x] = FltBlank;
                           }
                        }
                     } else if (test < 0.0 && flag.do_two_sided > 0) {

                        /* Here we are testing for negative going pixels */
                        if (c == F_NNB) {
                           if (test*test > nthresh2*(sumv2 + vv)) {

                              /* Here it's a nearest neighbor cosmic ray */
                              f[flag.nnrow][j][x] = F_NCR;
                              a[flag.nnrow][j][x] = FltBlank;
                           }
                        } else {
                           if (test*test > dthresh2*(sumv2 + vv)) {

                              /* Here it's a diagonal neighbor cosmic ray */
                              f[flag.nnrow][j][x] = F_DCR;
                              a[flag.nnrow][j][x] = FltBlank;
                           }
                        }
                     }
                  }
               }
            } else {
               int j;

               /* Here we had some neighbors of cosmic rays, but no good */
               /* pixels to test them against. So go through and increment */
               /* the counters for these neighbors */
               for (j = 0; j < nimg; j++) {
                  int c = f[flag.nnrow][j][x];
                  if (c == F_NNB)
                     (ip[j].nnuntest)++;
                  else if (c == F_DNB)
                     (ip[j].nduntest)++;
               }
            }
         }
      }
   }

   /* Now we have tested the pixels in this line to see if they */
   /* are neighbors of cosmic ray pixels and if so, we have tested */
   /* them to see if they are cosmic rays themselves.  So we are */
   /* ready to output the lines */

   /* Loop over the files, and update counters */
   for (j = 0; j < nimg; j++) {
      char *c = &f[flag.nnrow][j][0];  /* flags for NN line */
      int x;

      /* update counters */
      for (x = 0; x < ncols; x++) {
         if (c[x] == F_SAT)
            (ip[j].nsat)++;
         else if (c[x] == F_UND)
            (ip[j].nundef)++;
         else if (c[x] == F_CR)
            (ip[j].ncr)++;
         else if (c[x] == F_NCR)
            (ip[j].nncr)++;
         else if (c[x] == F_DCR)
            (ip[j].ndcr)++;
         else
            (ip[j].nok)++;
      }
   }

   /* Now, compute the weighted average and the count of good frames */
   for (x = 0; x < ncols; x++) {
      double sums = 0.0;
      double sumv = 0.0;
      int j;
      for (j = 0; j < nimg; j++) {
         int c = f[flag.nnrow][j][x];  /* pixel flags for NN line */

         if (c == F_OK || c == F_NNB || c == F_DNB) {
            sums += w[flag.nnrow][j][x]*a[flag.nnrow][j][x];
            sumv += w[flag.nnrow][j][x];
         }
      }
      if (sumv <= FLT_Min) {

         /* undefined output pixel */
         frow[x] = FltBlank;
         crow[x] = 0;
         (op->naundef)++;
      } else {
         double sums2 = 0.0;
         double sumv2 = 0.0;
         int sumn = 0;
         int j;

         sums /= sumv;

         /* Here we've computed a weighted average with weights based */
         /* on the signal at each pixel - recompute with weights */
         /* based on the average signal */
         for (j = 0; j < nimg; j++) {
            int c = f[flag.nnrow][j][x];  /* flags for NN line */

            if (c == F_OK || c == F_NNB || c == F_DNB) {
               double dn, vv;
               dn = sums/ip[j].rlev;
               dn = MAX(dn,1.0);
               vv = dn*ip[j].frac;
               vv = ip[j].rg2 + dn*ip[j].ig + ip[j].dc + vv*vv;
               vv = 1.0/(vv*ip[j].rlev2);
               sums2 += vv*a[flag.nnrow][j][x];
               sumv2 += vv;
               sumn++;
            }
         }
         if (sumv2 <= FLT_Min) {

            /* undefined output pixel */
            frow[x] = FltBlank;
            crow[x] = 0;
            (op->naundef)++;
         } else {

            /* good average output pixel */
            frow[x] = sums2/sumv2;
            (op->naok)++;
            crow[x] = sumn;
         }
      }
   }
}
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------
   This program is used to analyze cosmic ray split exposures
   or other sets of data where the values are positive and it
   is desired to compare them and throw out high values.

   crsplit.c is used to handle Hubble Space Telescope Wide Field
   Planetary Camera "Cosmic Ray Split" exposures. It may also be
   useful for filtering spurious data points in other kinds of data.

   "crsplit.c" is a translation of Edward J. Groth's Fortran program
   PICCRS to C. However, it only includes the calculation of the
   average output image, but this can be changed in the future.

   Bjarne B. Thomsen      Frbruary 1996
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
   int argc;
   char **argv, **arg;
   int option, opt_err = 0, n_arg = 0;

   double rnoise = 5.5;    /* WFPC2 read noise in electrons */
   int change_gain = 0;    /* force a change of gain ? */
   double gain = 7.0;      /* WFPC2 gain in electrons per ADU */
   double frac = 0.03;     /* WFPC2 fractional error */
   double thresh = 3.5;    /* sigma threshold for cosmic rays */
   double nthresh = 2.5;   /* sigma threshold for near neighbors */
   double dthresh = 2.5;   /* sigma threshold for diag neighbors */
   double dark = 0.0;      /* global dark rate in ADU/s */
   double sat = FLT_BLANK; /* global saturation level in ADU */
   int both = 0;           /* check both negative and positve residuals */
   int type_o = 3;         /* output average image is float */
   int type_c = 1;         /* output count image is UCHAR */

   IPARAM ip[IMAX];   /* structure with parameters for each input image */
   OPARAM op;         /* structure with parameters for the output image */
   DOFLAG flag;       /* structure with control flags */
   int Id[IMAX], Type[IMAX], Ncols[IMAX], Nrows[IMAX];
   char **image, **imgtab, *name;

   int id_o, id_c, ncols, nrows;
   char *image_o, file_o[MAXLEN];
   char *image_c = (char *)0; /* output count image */
   FILE *fp;

   void  *row;             /* void row buffer */
   float ***a;             /* 3-row pixel buffer a[nimg][3][ncols] */
   char  *aptr;            /* base pointer of array space for a[][][] */
   float ***w;             /* 3-row weight buffer w[nimg][3][ncols] */
   char  *wptr;            /* base pointer of array space for w[][][] */
   char  ***f;             /* 3-row flag buffer f[nimg][3][ncols] */
   char  *fptr;            /* base pointer of array space for f[][][] */
   unsigned int err;       /* error code returned by daa */
   int start[3], dim[3];   /* starting indices and dimensions for daa */

   int nimg, i, j, n, y, ntab;

   const float FltBlank = (float) FLT_BLANK;
   float Init_a = FltBlank; /* initial value for a[3][nimg][ncols] */
   char  Init_f = F_OK;     /* initial value for f[3][nimg][ncols] */
   float Init_w = FltBlank; /* initial value for w[3][nimg][ncols] */

#if defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
   argv = args(Argc, Argv, &argc);
#else	/* no substitution */
   argc = Argc;
   argv = Argv;
#endif

   if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
      fprintf(stderr, "crsplit: calloc: Can't allocate argument table\n");
      exit(EXIT_FAILURE);
   }

	/* process command line options */
   while (((option = getopt(argc, argv, "br:g:f:t:n:d:a:s:c:")) != NONOPT)
         || (optarg != NULL)) {
      switch(option) {
         case 'b': both++; break;
         case 'r': sscanf(optarg, "%lf", &rnoise); break;
         case 'g': sscanf(optarg, "%lf", &gain); change_gain++; break;
         case 'f': sscanf(optarg, "%lf", &frac); break;
         case 't': sscanf(optarg, "%lf", &thresh); break;
         case 'n': sscanf(optarg, "%lf", &nthresh); break;
         case 'd': sscanf(optarg, "%lf", &dthresh); break;
         case 'a': sscanf(optarg, "%lf", &dark); break;
         case 's': sscanf(optarg, "%lf", &sat); break;
         case 'c': image_c = optarg; break;	/* count image */
         case '?': opt_err++; break;	/* unknown option */
         case NONOPT: arg[n_arg++] = optarg; break;
         default: break;
      }
   }

   nimg = n_arg - 1;	/* number of input images */

   if (opt_err || nimg < 1) {
      printf("crsplit -b -r<rnoise> -g<gain> -f<frac> -t<thresh>\n");
      printf("        -n<nthresh> -d<dthresh> -a<dark> -s<saturation>\n");
      printf("        -c <image_cnt> image image1 [image2] ....\n");
      free(arg);
		fflush(stdout);
      exit(EXIT_FAILURE);
   }

   image_o = truncname(arg[0]);	/* output average image */

   /* is output image name legal? */
   if (strpbrk(image_o, punct) != (char *)0) {
      fprintf(stderr, "crsplit: Illegal image name: %s\n", image_o);
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* does output image exist? */
   if (xist(image_o)) {
      fprintf(stderr, "crsplit: %s: Existing image\n", image_o);
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* is output count image given */
   if (image_c != (char *)0) {

      image_c = truncname(image_c);

      /* is image name legal? */
      if (strpbrk(image_c, punct) != (char *)0) {
         fprintf(stderr, "crsplit: Illegal image name: %s\n", image_c);
         free(arg);
         exit(EXIT_FAILURE);
      }

      /* does count image name exist? */
      if (xist(image_c)) {
         fprintf(stderr, "crsplit: %s: Existing image\n", image_c);
         free(arg);
         exit(EXIT_FAILURE);
      }
   }

   if (thresh <= 0.0) {
      fprintf(stderr, "crsplit: The threshold must be greater than 0\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* The nearest neighbor and diagonal neighbor thresholds can be */
   /* 0 or less - this signifies the tests are not to be done */
   nthresh = MAX(nthresh,0.0);
   dthresh = MAX(dthresh,0.0);

   /* read image names from header directory */
   if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
      fprintf(stderr, "crsplit: prolog: Error = %d\n", i);
      free(arg);
      exit(EXIT_FAILURE);
   }

   if (nimg > 1) {	/* input images given on command line */
      if ((image = (char **)calloc(nimg, sizeof(char *))) == (char **)0) {
         fprintf(stderr, "crsplit: calloc: Allocation failed-1\n");
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }
      for (n = 0, i = 0; i < nimg; i++) {
         name = truncname(arg[i+1]);	/* next input image */
         if (strpbrk(name, punct) != (char *)0)
            printf("%s is not a legal image name\n", name);
         else if (xist(name))	/* check if image exists */
            image[n++] = name;	/* include in input list */
         else
            printf("%s does not exist\n", name);
      }
   } else {	/* wild card possible, if only one input image is given */
      imgtab = getitab(&ntab);	/* table containing all image names */
      image = (char **)0;	/* initialize the image table pointer */
      if (ntab > 0)	/* directory is not empty */
         if ((image = (char **)calloc(ntab, sizeof(char *))) == (char **)0) {
            fprintf(stderr, "crsplit: calloc: Allocation failed-2\n");
            if (epilog() < 1)	/* update disk headers */
               fprintf(stderr,
                  "crsplit: epilog: Disk header(s) not updated\n");
            free(arg);
            exit(EXIT_FAILURE);
         }
      name = arg[1];	/* input image, which may contain wild card */
      for (n = 0, i = 0; i < ntab; i++)	/* is name in image table */
         if (wildcmp(imgtab[i], name) == 0)
            image[n++] = imgtab[i];
   }

   nimg = n;	/* number of input images to process */
   if(nimg == 0) {
      fprintf(stderr, "crsplit: No input images found\n");
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   if(nimg > IMAX) {	/* too many input images */
      fprintf(stderr, "crsplit: Too many input images: %d\n", nimg);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   setipath(0);	/* select first token in IMGPATH */

   op.effread = 0.0;    /* effective read noise sum */
   op.effgain = 0.0;    /* effective gain sum */

   /* Now start going through the images and checking for errors */
   for (i = 0; i < nimg; i++) {	/* open input images */
      if ((Id[i]=iopen(image[i], RO, &Type[i], &Ncols[i], &Nrows[i])) < 0) {
         fprintf(stderr, "crsplit: Can't open %s\n", image[i]);
         for (j = (i-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }
      if (Type[i] < 1 || Type[i] > 5) {
         fprintf(stderr, "crsplit: %s has illegal type = %d\n",
            image[i], Type[i]);
         for (j = i; j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }

      /* The option -r <rnoise> is always used */
      ip[i].rnoise = MAX(rnoise,0.0);

      /* The obtion -g <gain> indicates that the gain specified */
      /* in the header of an input image should not be used */
      ip[i].gain = atodgain(image[i]);
      if (change_gain != 0 || ip[i].gain == 0.0)
         ip[i].gain = MAX(gain,0.0);

      /* The option -f <frac> is always used */
      ip[i].frac = MAX(frac,0.0);

      /* The option -a <dark> is always used */
      ip[i].dark = MAX(dark,0.0);

      /* Undefined value */
      ip[i].blank = blank(image[i]);

      /* 1 sec is used when an exposure time is not */
      /* specified for an individual input image */
      ip[i].etime = exptime(image[i]);
      if (ip[i].etime == 0.0) ip[i].etime = 1.0;

      /* dark time */
      ip[i].dtime = darktime(image[i]);
      if (ip[i].dtime == 0.0) ip[i].dtime = ip[i].etime;

      /* saturation level */
      ip[i].sat = MAX(sat,0.0);

      /* Now compute some derived parameters, so we don't have to */
      /* compute them during the row loop */

      /* RG2 is the read noise divided by the gain squared */
      ip[i].rg2 = ip[i].rnoise/ip[i].gain;
      ip[i].rg2 *= ip[i].rg2;

      /* IG is the inverse gain (DN per electron) */
      ip[i].ig = 1.0/ip[i].gain;

      /* DC is the dark current in electrons */
      ip[i].dc = ip[i].dark*ip[i].dtime/ip[i].gain;

      /* RLEV is the relative level of image I relative to the output */
      /* image (i.e. image I should be scaled by this quantity to make */
      /* the same DN as the output image */
      ip[i].rlev = (ip[i].gain * ip[0].etime)/(ip[0].gain * ip[i].etime);

      /* RLEV2 is RLEV^2 */
      ip[i].rlev2 = ip[i].rlev * ip[i].rlev;

      /* Set pixel quality counters to 0 */
      ip[i].nuntest = 0;
      ip[i].nnuntest = 0;
      ip[i].nduntest = 0;
      ip[i].nok = 0;
      ip[i].nsat = 0;
      ip[i].nundef = 0;
      ip[i].ncr = 0;
      ip[i].nncr = 0;
      ip[i].ndcr = 0;

      /* Compute sums */
      op.effgain += ip[i].etime;
      op.effread += (ip[i].etime/ip[i].rnoise)*(ip[i].etime/ip[i].rnoise);   
   }

   /* Now fix up the effective gain and effective read noise. */
   /* The output image will be normalized to the exposure time */
   /* and gain of the first image. This normalization is for the */
   /* purpose of photometry - i.e. converting DN to DN/sec to magnitudes. */
   /* The effective gain and effective read noise are for the purpose of */
   /* noise estimation - i.e. the noise in the output is estimated using */
   /* these quantities for the read noise and gain with the DN in the */
   /* output. */
   op.ogain = ip[0].gain;
   op.otime = ip[0].etime;
   op.effread = op.effgain/sqrt(op.effread);
   op.effgain = op.ogain * op.effgain / op.otime;

   /* check image types and sizes */
   for (i = 1; i < nimg; i++) {
      if (Ncols[i] != Ncols[0]) {
         fprintf(stderr,
               "crsplit: %s and %s have incompatible x-dimensions\n",
               image[i], image[0]);
         for (j = (nimg-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }
      if (Nrows[i] != Nrows[0]) {
         fprintf(stderr,
               "crsplit: %s and %s have incompatible y-dimensions\n",
               image[i], image[0]);
         for (j = (nimg-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }
   }

   ncols = Ncols[0];
   nrows = Nrows[0];

   /* copy header of first image to header of average image */
   if (cpi(image[0], image_o) < 1) {
      fprintf(stderr,
         "crsplit: %s.hdr -> %s.hdr copy failed\n", image[0], image_o);
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   } else if (setimagefile(image_o, "") == 0) {	/* reset FILE key */
      fprintf(stderr,
         "crsplit: Can't erase old FILE value for %s\n", image_o);
      rmi(image_o);	/* remove image header */
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   if ((id_o = icreat(image_o, WO, type_o, ncols, nrows)) < 0) {
      fprintf(stderr, "crsplit: Can't create %s\n", image_o);
      rmi(image_o);	/* remove image header */
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* create a count image, if necessary */
   if (image_c != (char *)0) {

      /* copy header of first image to header of count image */
      if (cpi(image[0], image_c) < 1) {
         fprintf(stderr,
            "crsplit: %s.hdr -> %s.hdr copy failed\n", image[0], image_c);
         iclose(id_o);
         for (j = (nimg-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr,
               "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      } else if (setimagefile(image_c, "") == 0) {	/* reset FILE key */
         fprintf(stderr,
            "crsplit: Can't erase old FILE value for %s\n", image_c);
         rmi(image_c);	/* remove image header */
         iclose(id_o);
         for (j = (nimg-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }

      if ((id_c = icreat(image_c, WO, type_c, ncols, nrows)) < 0) {
         fprintf(stderr, "crsplit: Can't create %s\n", image_c);
         rmi(image_c);	/* remove image header */
         iclose(id_o);
         for (j = (nimg-1); j >= 0; j--)
            iclose(Id[j]);
         free(image);
         if (epilog() < 1)	/* update disk headers */
            fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
         free(arg);
         exit(EXIT_FAILURE);
      }
   }   

   if (update() == 0)	/* update disk headers */
      fprintf(stderr, "crsplit: update: Disk header not created\n");

   dim[0] = 3;          /* 3 rows in pixel buffer */
   dim[1] = nimg;       /* number of input images */
   dim[2] = ncols;      /* number of pixels in a row */
   start[0] = 0;        /* index of first input image */
   start[1] = 0;        /* index of first row in buffer */
   start[2] = 0;        /* index of first pixel in a row */

   /* allocate COMPLEX row buffer as void pointer */
   if ((row = (void *)calloc(ncols, bytepix(5))) == (void *)0) {
      fprintf(stderr, "crsplit: Can't allocate row[%d]\n", ncols);
      if (image_c != (char *)0)
         iclose(id_c);
      iclose(id_o);
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* allocate pixel float array a[3][nimg][ncols] */
   if ((a = (float ***)daa(sizeof(float), 3, dim, start,
         &err, &aptr, &Init_a)) == (float ***)0) {
      fprintf(stderr, "crsplit: daa: Error code = %d\n", err);
      fprintf(stderr,
         "crsplit: Can't allocate a[%d][%d][%d]\n", dim[0], dim[1], dim[2]);
      free(row);
      if (image_c != (char *)0)
         iclose(id_c);
      iclose(id_o);
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* allocate flag char array f[3][nimg][ncols] */
   if ((f = (char ***)daa(sizeof(char), 3, dim, start,
         &err, &fptr, &Init_f)) == (char ***)0) {
      fprintf(stderr, "crsplit: daa: Error code = %d\n", err);
      fprintf(stderr,
         "crsplit: Can't allocate f[%d][%d][%d]\n", dim[0], dim[1], dim[2]);
      free(aptr);
      free(row);
      if (image_c != (char *)0)
         iclose(id_c);
      iclose(id_o);
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }

   /* allocate weight float array w[3][nimg][ncols] */
   if ((w = (float ***)daa(sizeof(float), 3, dim, start,
         &err, &wptr, &Init_w)) == (float ***)0) {
      fprintf(stderr, "crsplit: daa: Error code = %d\n", err);
      fprintf(stderr,
         "crsplit: Can't allocate w[%d][%d][%d]\n", dim[0], dim[1], dim[2]);
      free(fptr);
      free(aptr);
      free(row);
      if (image_c != (char *)0)
         iclose(id_c);
      iclose(id_o);
      for (j = (nimg-1); j >= 0; j--)
         iclose(Id[j]);
      free(image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
      free(arg);
      exit(EXIT_FAILURE);
   }
/*
   Now enter a loop over the lines. In general we input a line one row
   further along than we output. This allows us to implement the nearest
   neighbor and diagonal neighbor tests. For example, suppose we are reading
   line 102. Then lines 100 and 101 are still in memory. After line 102 is
   read in, we process it to find cosmic rays in it. Then we can process
   line 101 to see if the nearest neighbors or diagonal neighbors of cosmic
   rays are present. Then we can output line 101. This means that we want to
   loop for one more line than the number of lines in the file. The first
   pas through the loop we input a line and process, but we don't output a
   line. In the second pass, we input a line and process, and we process a
   line for neighbors (except there is is no line below it) and output.
   In the third pass we input a line and process and we process a line for
   neighbors and output. All subsequent passes are like the third pass except
   the last. In the last pass we don't input and process anything. Instead we
   process a line for neighbors (except there is no line above it) and
   output.
*/

   flag.below = 0;  /* row below nearest neighbor line */
   flag.nnrow = 1;  /* nearest neighbor line */
   flag.above = 2;  /* row above nearest neighbor line */

   /* The first time we process a neighbor line there will be */
   /* a neighbor line on the top, but not the bottom */
   flag.do_top = 1;
   flag.do_bottom = 0;

   /* do we wish to do a two sided search ? */
   flag.do_two_sided = both;

   /* Start the row loop */
   for (y = 0; y <= nrows; y++) {

      /* Input and process the line (if we're not off the top end yet) */
      if (y < nrows) {
         for (j = 0; j < nimg; j++) {
            getline(Id[j], y, row);
            cptoflt(ncols, Type[j], ip[j].blank, row, &a[flag.above][j][0]);
         }
         main_row(flag, thresh, nimg, ncols, ip, a, w, f);
      }

      /* Process and output the nearest neighbor line if above the first. */
      if (y > 0) {

         /* Reset doing the top one if the last line */
         if (y >= nrows) flag.do_top = 0;
         neigh_row(flag, nthresh, dthresh, nimg, ncols, ip, &op, a, w, f);
         putline(id_o, y-1, &a[flag.below][0][0]); /* average image */
         if (image_c != 0)        /* write count image, if needed */
            putline(id_c, y-1, &f[flag.below][0][0]);

         /* Turn on doing the bottom line */
         flag.do_bottom = 1;
      }

      /* Update the row numbers - for each of the three buffer lines. */
      /* This is just a permutation */
      n = flag.below;
      flag.below = flag.nnrow;
      flag.nnrow = flag.above;
      flag.above = n;
   }

   free(wptr);
   free(fptr);
   free(aptr);
   free(row);
   if (image_c != (char *)0)
      iclose(id_c);
   iclose(id_o);
   for (j = (nimg-1); j >= 0; j--)
      iclose(Id[j]);

   /* create output file name with extension */ 
   strcat(strcpy(file_o, image_o), ".crs");

   /* open output file */
   if ((fp = fopen(file_o, "w")) != NULL) {
      fprintf(fp, "Output gain (electrons/DN)..........: %f\n", op.ogain);
      fprintf(fp, "Output exposure time (input units)..: %f\n", op.otime);
      fprintf(fp, "Effective read noise (electrons)....: %f\n", op.effread);
      fprintf(fp, "Effective gain (electrons/DN).......: %f\n", op.effgain);
      fprintf(fp, "This times error is CR threshold....: %f\n", thresh);
      fprintf(fp, "Threshold for nearest neighbors.....: %f\n", nthresh);
      fprintf(fp, "Threshold for diagonal neighbors....: %f\n", dthresh);
      fprintf(fp, "Number of good pixels in output.....: %d\n", op.naok);
      fprintf(fp, "Number of undefined pixels in output: %d\n", op.naundef);
      for (j = 0; j < nimg; j++) {
         fprintf(fp, "Input image....................: %s\n", image[j]);
         fprintf(fp, "Saturation level...............: %g\n", ip[j].sat);
         fprintf(fp, "Read noise (electrons).........: %f\n", ip[j].rnoise);
         fprintf(fp, "Gain (electrons per DN)........: %f\n", ip[j].gain);
         fprintf(fp, "Dark current (DN/time units)...: %f\n", ip[j].dark);
         fprintf(fp, "Fractional error...............: %f\n", ip[j].frac);
         fprintf(fp, "Exposure time (arbitrary units): %f\n", ip[j].etime);
         fprintf(fp, "%8d Number of input undefined pixels\n", ip[j].nundef);
         fprintf(fp, "%8d Number of input saturated pixels\n", ip[j].nsat);
         fprintf(fp, "%8d Cosmic ray pixels - primary test\n", ip[j].ncr);
         fprintf(fp, "%8d Cosmic ray pixels - nearest neighbor test\n",
                      ip[j].nncr);
         fprintf(fp, "%8d Cosmic ray pixels - diagonal neighbor test\n",
                      ip[j].ndcr);
         fprintf(fp, "%8d Good pixels, not undefined/saturated/cosmic ray\n",
                      ip[j].nok);
         fprintf(fp, "%8d Pixels not tested in primary test\n",
                      ip[j].nuntest);
         fprintf(fp, "%8d Nearest neighbor pixels not tested\n",
                      ip[j].nnuntest);
         fprintf(fp, "%8d Diagonal neighbor pixels not tested\n",
                      ip[j].nduntest);
      }
      fclose(fp);
   }   
   free(image);

   setiname(image_o);	/* set IMAGE = Image Name */
   sethistory(image_o, "crsplit");
   sethistory(image_o, image_o);

   if (epilog() < 1)	/* update disk headers */
      fprintf(stderr, "crsplit: epilog: Disk header(s) not updated\n");
   free(arg);

	fflush(stdout);
   exit(EXIT_SUCCESS);
}
