
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "nr.h"	/* function declarations for nr.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int isign = 1;	/* forward FFT */
	float FltSub = (float)0;	/* substitute for blank input values */
	int n_arg = 0;
	char **arg;
	void *buffer;
	char *image1, *image2;
	COMPLEX **a;
	float *data;
	int ndim[3];	/* ndim[0] is not used */
	int i, x, y;
	int id1, id2, npix, nlin, type;
	double Blank1;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "fft: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "ib:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'i':	isign = -1; break;	/* backwards FFT */
			case 'b':	sscanf(optarg, "%f", &FltSub); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 2) {
		printf("Usage: fft -i -b FltSub image1 image2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[0]);	/* image name */
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "fft: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	image2 = truncname(arg[1]);	/* image name */
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "fft: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "fft: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */
	if ((id1 = iopen(image1, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "fft: Can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (i = 1; i < npix; i <<= 1)
		;
	if (i > npix) {
		fprintf(stderr,
			"fft: %s has illegal npix = %d\n", image1, npix);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (i = 1; i < nlin; i <<= 1)
		;
	if (i > nlin) {
		fprintf(stderr,
			"fft: %s has illegal nlin = %d\n", image1, nlin);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	ndim[0] = 0;	/* not used */
	ndim[1] = nlin;	/* number of rows */
	ndim[2] = npix;	/* number of columns */
	if (type < 1 || type > 5) {
		fprintf(stderr,
			"fft: %s has illegal type = %d\n", image1, type);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank1 = blank(image1);

	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "fft: Allocation of buffer failed\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(image2)) {
		fprintf(stderr, "fft: %s: Existing image\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image1, image2) < 1) {	/* copy image header */
		fprintf(stderr,
			"fft: %s.hdr -> %s.hdr copy failed\n", image1, image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image2, "") == 0) {	/* erase FILE value */
		fprintf(stderr,
			"fft: Can't erase old FILE value for %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((id2 = icreat(image2, 1, 5, npix, nlin)) < 0) {
		fprintf(stderr, "fft: Can't create %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "fft: update: Disk header not created\n");

	if ((a = cmatrix(1, nlin, 1, npix)) == (COMPLEX **)0) {
		fprintf(stderr, "fft: Allocation failed in cmatrix()\n");
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	data = (float *)&a[1][1];
	--data;	/* data[1] is the first element */

	for (y = 1; y <= nlin; ++y) {
		getline(id1, y-1, buffer);	/* fetch new row */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					for (x = 1, i = 0; i < npix; ++x, ++i) {
						a[y][x].y = Zero;
						a[y][x].x = (float)q[i];
					}
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank1;
					for (x = 1, i = 0; i < npix; ++x, ++i) {
						a[y][x].y = Zero;
						a[y][x].x = (q[i] == b)? FltSub: (float)q[i];
					}
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank1);
					for (x = 1, i = 0; i < npix; ++x, ++i) {
						a[y][x].y = Zero;
						a[y][x].x = (q[i] > h)? FltSub: q[i];
					}
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank1;
					for (x = 1, i = 0; i < npix; ++x, ++i) {
						a[y][x].y = Zero;
						a[y][x].x = (q[i] == b)? FltSub: (float)q[i];
					}
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank1);
					for (x = 1, i = 0; i < npix; ++x, ++i)
						if (q[i].x > h || q[i].y > h) {
							a[y][x].x = FltSub;
							a[y][x].y = FltSub;
						} else
							a[y][x] = q[i];
				}
				break;
		}
	}

	fourn(data, ndim, 2, isign);	/* 2-dim complex FFT */

	for (y = 1; y <= nlin; ++y)
		putline(id2, y-1, &a[y][1]);	/* save next row */

	free_cmatrix(a, 1, 1);
	iclose(id2);
	free(buffer);
	iclose(id1);

	sethistory(image2, "fft");

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "fft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
