
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "nr.h"	/* function declarations for nr.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static double FullWidth(double a1, double r1, double a2, double r2)
#else	/* old style */
static double FullWidth(a1, r1, a2, r2)
double a1, r1, a2, r2;
#endif
{
	double f, r, dr, rtb;
	const double x1 = r1*r1;
	const double x2 = r2*r2;
	const double A1 = a1/x1/(a1/x1 + a2/x2);
	const double A2 = a2/x2/(a1/x1 + a2/x2);
	const double rmin = MIN(r1,r2);
	const double eps = 1.0E-6 * rmin;

	r = rmin;
	r *= 0.01;
	do {
		r *= 2.0;
		f = A1*exp(-PI*r*r/x1) + A2*exp(-PI*r*r/x2);
	} while (f > 0.49);
	dr = r;
	rtb = r;
	do	{
		r = rtb - (dr *= 0.5);
		f = A1*exp(-PI*r*r/x1) + A2*exp(-PI*r*r/x2);
		if (f <= 0.5) rtb = r;
	} while (dr > eps && f != 0.5);
	return (2.0*rtb);
}
/*--------------------------------------------------------------------*/
#define	FWHM	5.0	/* guess full width at half maximum */
#define	RAD1	5.0	/* guess width of first Gaussian component */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int mode = 0;	/* Fit in (0): x-y, (1): x, (-1): y */
	int two_gauss = 0;	/* Fit two Gausians in 2-dimensions */
	/* Electrons per Digital Number */
	double epdn = -1.0;	/* if edpn < 0, equal weights are used */
	int i, x, y;
	int xcen, ycen, radmax, skymin, skymax;
	int stx, sty, ext;
	int id, type, npix, nlin;
	double Blank;
	double a1, a2, r1, r2, xc, dxc, yc, dyc, fwhm, dfwhm;
	void *buffer;
	char *image, *str;
	float **a;
	const float FltBlank = (float)FLT_BLANK;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "gauss: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "xyme:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'x':	mode++; break;	/* fit along x-axis */
			case 'y':	mode--; break;	/* fit along y-axis */
			case 'm':	two_gauss++; break;	/* fit of two Gaussian */
			case 'e':	sscanf(optarg, "%lf", &epdn); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 6) {
		printf(
		"Usage: gauss -x -y -m -e epdn image Xc Yc RadMax SkyMin SkyMax\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "gauss: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%d", &xcen);
	sscanf(arg[2], "%d", &ycen);
	sscanf(arg[3], "%d", &radmax);	/* aperture radius */
	sscanf(arg[4], "%d", &skymin);	/* inner rad of sky annulus */
	sscanf(arg[5], "%d", &skymax);	/* outer rad of sky annulus */

	if (radmax <= 0) {
		fprintf(stderr, "gauss: radmax <= 0\n");
		exit(EXIT_FAILURE);
	}
	if (skymin < radmax) {
		fprintf(stderr, "gauss: skymin < radmax\n");
		exit(EXIT_FAILURE);
	}
	if (skymax <= skymin) {
		fprintf(stderr, "gauss: skymax <= skymin\n");
		exit(EXIT_FAILURE);
	}
	stx = xcen - skymax;
	sty = ycen - skymax;
	if (stx < 0 || sty < 0) {
		fprintf(stderr, "gauss: Sky annulus extends outside image\n");
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "gauss: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "gauss: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr,
			"gauss: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	ext = 2*skymax + 1;
	if (stx + ext > npix || sty + ext > nlin) {
		fprintf(stderr, "gauss: Sky annulus extends outside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (!window(id, stx, sty, ext, ext)) {
		fprintf(stderr, "gauss: Object window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(ext, bytepix(type))) == (void *)0) {
		fprintf(stderr, "gauss: Allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((a = matrix(sty, sty+ext-1, stx, stx+ext-1)) == (float **)0) {
		fprintf(stderr, "gauss: Allocation failed in matrix()-2\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = sty; y < (sty+ext); ++y) {
		getline(id, y-sty, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					for (x = stx, i = 0; i < ext; ++x, ++i)
						a[y][x] = (float)q[i];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					for (x = stx, i = 0; i < ext; ++x, ++i)
						a[y][x] = (q[i] == b)? FltBlank: (float)q[i];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = stx, i = 0; i < ext; ++x, ++i)
						a[y][x] = (q[i] > h)? FltBlank: q[i];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					for (x = stx, i = 0; i < ext; ++x, ++i)
						a[y][x] = (q[i] == b)? FltBlank: (float)q[i];
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = stx, i = 0; i < ext; ++x, ++i)
						a[y][x] = (q[i].x>h || q[i].y>h)? FltBlank: q[i].x;
				}
				break;
		}
	}
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "gauss: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	a1 = 1.0;	/* guess amplitude of first Gaussian component */
	a2 = 0.0;	/* guess amplitude of second Gaussian component */
	r1 = RAD1;	/* guess width of first Gaussian component */
	r2 = 2.0*r1;
	fwhm = FWHM;	/* guess FWHM of single Gaussian */
	xc = (double) xcen;	/* guess of location */
	yc = (double) ycen;
	if (mode < 0) {	/* 1-dim fit along the y-axis */
		str = g1yfit(a, FltBlank, epdn, radmax, skymin, skymax,
				&xc, &yc, &fwhm, &dyc, &dfwhm);
		if (str != (char *)0) {
			fprintf(stderr, "gauss: %s", str);
			exit(EXIT_FAILURE);
		}
		free_matrix(a, sty, stx);
		printf("xc = %7.3f  yc = %7.3f+-%5.3f  fwhm = %5.3f+-%5.3f\n",
			xc, yc, dyc, fwhm, dfwhm);
	} else if (mode > 0) {	/* 1-dim fit along the x-axis */
		str = g1xfit(a, FltBlank, epdn, radmax, skymin, skymax,
				&xc, &yc, &fwhm, &dxc, &dfwhm);
		if (str != (char *)0) {
			fprintf(stderr, "gauss: %s", str);
			exit(EXIT_FAILURE);
		}
		free_matrix(a, sty, stx);
		printf("xc = %7.3f+-%5.3f  yc = %7.3f  fwhm = %5.3f+-%5.3f\n",
			xc, dxc, yc, fwhm, dfwhm);
	} else if (two_gauss > 0) {	/* fit of two Gaussians */
		str = g3fit(a, FltBlank, epdn, radmax, skymin, skymax,
				&a1, &r1, &a2, &r2, &xc, &yc);
		if (str != (char *)0) {
			fprintf(stderr, "gauss: %s", str);
			exit(EXIT_FAILURE);
		}
		free_matrix(a, sty, stx);
		printf("\n");
		printf("PSF(x,y) = A1/R1^2*exp(-Pi*((x-xc)^2+(y-yc)^2)/R1^2)\n");
		printf("         + A2/R2^2*exp(-Pi*((x-xc)^2+(y-yc)^2)/R2^2)\n\n");
		printf("A1: Relative Volume of 1'st Gaussian Component\n");
		printf("A2: Relative Volume of 2'nd Gaussian Component\n\n");
		printf("A1   = %7.4f  a1 = %13.7g\n", a1/(a1+a2), a1);
		printf("A2   = %7.4f  a2 = %13.7g\n", a2/(a1+a2), a2);
		printf("R1   = %7.3f\n", r1);
		printf("R2   = %7.3f\n", r2);
		printf("xc   = %7.3f\n", xc);
		printf("yc   = %7.3f\n", yc);
		printf("fwhm = %7.3f\n\n", FullWidth(a1, r1, a2, r2));
	} else {	/* 2-dimensional single Gaussian fit */
		str = g2fit(a, FltBlank, epdn, radmax, skymin, skymax,
				&xc, &yc, &fwhm, &dxc, &dyc, &dfwhm);
		if (str != (char *)0) {
			fprintf(stderr, "gauss: %s", str);
			exit(EXIT_FAILURE);
		}
		free_matrix(a, sty, stx);
		printf("xc = %7.3f+-%5.3f  yc = %7.3f+-%5.3f  fwhm = %5.3f+-%5.3f\n",
			xc, dxc, yc, dyc, fwhm, dfwhm);
	}
	free(buffer);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
