
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* For ANSI compatibility, errno declared */
#include <string.h>	/* required only for function declarations */
#include <graph.h>	/* Standard Video Modes defined in graph.h */
#include <conio.h>	/* Keyboard I/O */
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "fbuf.h"	/* constants and prototypes for fbuf.c */
#include "vga256.h"	/* function prototypes for vga256.c */
#include "hcconio.h"	/* function prototypes for hcconio.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define NPIX	32		/* number of pixels in picture */
#define NLIN	16		/* number of lines in picture */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define LF	10
#define CR	13
/*--------------------------------------------------------------------*/
#define CURSIZ	5	/* cursor size */
#define NAMSIZ	40	/* full file name size */
/*--------------------------------------------------------------------*/
#define NHUE	360	/* number of hues */
static UCHAR red[NHUE], green[NHUE], blue[NHUE];
/*--------------------------------------------------------------------*/
static void ReadRGB(void)	/* Read R, G, and B tables */
{
	char **path, fname[NAMSIZ];
	int count;
	FILE *stream;
	if ((path = getpath("IMSYS", &count)) == (char **)0)
		strcpy(fname, ".");	/* current directory */
	else
		strcpy(fname, path[0]);	/* pathname from environment */
	strcat(fname, "/rgbtab.dat");
	if ((stream = fopen(fname, "rb")) == (FILE *)0) {
		fprintf(stderr, "can't open %s\n", fname);
		exit(EXIT_FAILURE);
	}
	fread(red, sizeof (char), NHUE, stream);
	fread(green, sizeof (char), NHUE, stream);
	fread(blue, sizeof (char), NHUE, stream);
	fclose(stream);
}
/*--------------------------------------------------------------------*/
static UCHAR rgb[768];	/* RGB Lookup Table */
/*--------------------------------------------------------------------*/
static void Lookup(int hue)	/* Calculate Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	while (hue < 0)
		hue += NHUE;
	while (hue >= NHUE)
		hue -= NHUE;
	for (i = 0; i < 256; i++) {
		j = (hue + i) % NHUE;
		*p++ = red[j];
		*p++ = green[j];
		*p++ = blue[j];
	}
}
/*--------------------------------------------------------------------*/
static void GrayLookup(void)	/* Calculate Gray Scale Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	for (i = 0; i < 256; i++) {
		j = i >> 2;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
	}
}
/*--------------------------------------------------------------------*/
static void DrawCursor(	/* Draw cursor at (x,y) */
int x, int y, UCHAR ucv)
{
	int i;
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	for (i = 1; i <= CURSIZ; ++i) {
		if ((x-i) >= 0)
			draw_pixel(y, x - i, (UCHAR) (read_pixel(y, x - i) + ucv));
		if ((x+i) < row_siz)
			draw_pixel(y, x + i, (UCHAR) (read_pixel(y, x + i) + ucv));
		if ((y-i) >= 0)
			draw_pixel(y - i, x, (UCHAR) (read_pixel(y - i, x) + ucv));
		if ((y+i) < col_siz)
			draw_pixel(y + i, x, (UCHAR) (read_pixel(y + i, x) + ucv));
	}
}
/*--------------------------------------------------------------------*/
static UCHAR DisplayValue(	/* Reads display value at (x,y) */
int x, int y)
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	return read_pixel(y, x);
}
/*--------------------------------------------------------------------*/
static short **Array;	/* pointer to array of pointers */
static char *_Buffer;	/* pointer to buffer array */
static int _BufSize;	/* image buffer size */
static int _Frame = 1;	/* default frame buffer number */
static int _Page  = 0;	/* default window page number */
static int _Ncols = NPIX;	/* number of columns in Array[][] */
static int _Nrows = NLIN;	/* number of rows in Array[][] */
static int _Col0 = 0;	/* first column displayed */
static int _Row0 = 0;	/* first row displayed */
static int _Rotate = 0;	/* should image be rotated? */
static int _Bias = 0;	/* value subtracted before a shift */
static int _Shift = 0;	/* bits shifted right */
/*--------------------------------------------------------------------*/
static int PixelValue(	/* Reads pixel value at (x,y) */
int x, int y)
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	if (_Col0 < 0) _Col0 = 0;
	if (_Row0 < 0) _Row0 = 0;
	if (_Col0 >= _Ncols) _Col0 = _Ncols - 1;
	if (_Row0 >= _Nrows) _Row0 = _Nrows - 1;
	if (_Rotate) {
		x += _Row0;
		y += _Col0;
		if (x >= _Nrows || y >= _Ncols)
			return 0;
		else
			return _Array[x][y];
	} else {
		x += _Col0;
		y += _Row0;
		if (x >= _Ncols || y >= _Nrows)
			return 0;
		else
			return _Array[y][x];
	}
}
/*--------------------------------------------------------------------*/
char _String[50] = {'\0'};	/* global string containing output text */
/*--------------------------------------------------------------------*/
static void CursorControl(int hue,	/* Cursor control */
const int scr_line,	/* string written on this line */
const int ncols,	/* number of columns in Array[][] */
const int nrows,	/* number of rows in Array[][] */
const int col0,	/* first column displayed */
const int row0,	/* first row displayed */
const int rotate)	/* has image been rotated? */
{
	int xcen, ycen;
	int row_siz = rowsize();
	int col_siz = colsize();
	int c = 'm';	/* move cursor */
	UCHAR astep = 1;
	UCHAR bstep = 10;
	UCHAR color = 128;
	UCHAR ucv = color;
	if (rotate) {
		xcen = MIN(row_siz, nrows - row0)/2;
		ycen = MIN(col_siz, ncols - col0)/2;
	} else {
		xcen = MIN(row_siz, ncols - col0)/2;
		ycen = MIN(col_siz, nrows - row0)/2;
	}
	DrawCursor(xcen, ycen, ucv);
	while (c != 'e') {
		switch(c) {
			case 'c':	/* change cursor color */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						color += 8;
					if (c == 'P' || c == 'K')
						color -= 8;
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					ucv = color;
					sprintf(_String,
						"cursor color: %6d                    ", color);
					write_str(_String, scr_line, 0, 255);
					DrawCursor(xcen, ycen, ucv);
				}
			break;
			case 's':	/* change cursor step size */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						astep++;
					if (c == 'P' || c == 'K')
						astep--;
					if (c == 'Q' || c == 'G')
						bstep++;
					if (c == 'S' || c == 'O')
						bstep--;
					sprintf(_String,
						"steps: %6d %6d                    ", astep, bstep);
					write_str(_String, scr_line, 0, 255);
				}
			break;
			case 'h':	/* change starting hue for color lookup table */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M') {
						hue += 4;
						while (hue >= NHUE)
							hue -= NHUE;
					}
					if (c == 'P' || c == 'K') {
						hue -= 4;
						while (hue < 0)
							hue += NHUE;
					}
					sprintf(_String,
						"starting hue: %6d                    ", hue);
					write_str(_String, scr_line, 0, 255);
					Lookup(hue);
					set_color(rgb);
				}
			break;
			case 'g':	/* use gray scale lookup table */
				GrayLookup();
				set_color(rgb);
				c = 'm';	/* move cursor next time */
			break;
			case 'm':	/* move cursor */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					switch(c) {
						case 'H':
							ycen -= astep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'P':
							ycen += astep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'K':
							xcen -= astep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'M':
							xcen += astep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						case 'G':
							ycen -= bstep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'O':
							ycen += bstep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'S':
							xcen -= bstep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'Q':
							xcen += bstep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						default:
						break;
					}
					DrawCursor(xcen, ycen, ucv);
					sprintf(_String, "x=%4d y=%4d display=%4d pixel=%7d",
						xcen, ycen, DisplayValue(xcen, ycen),
						PixelValue(xcen, ycen, ncols, nrows, col0, row0, rotate));
					write_str(_String, scr_line, 0, 255);
				}
			break;
			default:
				while ((c = get_ch()) == 0)	/* get next non-arrow */
					c = get_ch();	/* skip second byte */
			break;
		}
	}
}
/*--------------------------------------------------------------------*/
static void Display(
int ncols,	/* number of columns in Array[][] */
int nrows,	/* number of rows in Array[][] */
int col0,	/* first column displayed */
int row0,	/* first row displayed */
int rotate,	/* should image be rotated? */
int bias,	/* value subtracted before a shift */
int shift)	/* bits shifted right */
{
	UCHAR Row[2048];	/* <------ REMEMBER THIS BUFFER */
	int row_siz = rowsize();
	int col_siz = colsize();
	if (col0 < 0) col0 = 0;
	if (row0 < 0) row0 = 0;
	if (col0 >= ncols) col0 = ncols - 1;
	if (row0 >= nrows) row0 = nrows - 1;
	if (rotate) {
		int npix = MIN(row_siz, nrows - row0);
		int nlin = MIN(col_siz, ncols - col0);
		int x, y, z;
		for (x = 0; x < nlin; ++x) {
			for (y = 0; y < npix; ++y) {
				z = Array[y+row0][x+col0] - bias;
				if (z < 0) z = 0;
				Row[y] = (UCHAR) (z >> shift);
			}
			draw_line(x, 0, Row, npix);
		}
	} else {
		int npix = MIN(row_siz, ncols - col0);
		int nlin = MIN(col_siz, nrows - row0);
		int x, y, z;
		for (y = 0; y < nlin; ++y) {
			for (x = 0; x < npix; ++x) {
				z = Array[y+row0][x+col0] - bias;
				if (z < 0) z = 0;
				Row[x] = (UCHAR) (z >> shift);
			}
			draw_line(y, 0, Row, npix);
		}
	}
}
/*--------------------------------------------------------------------*/
static void ReadImage(void)
{
	int i;
	switch (_Frame) {	/* read image from frame buffer */
		case 1:
			for (i = 0; i < _BufSize; i++)
				_Buffer[i] = fb1_win[0][i];
			break;
		case 2:
			for (i = 0; i < _BufSize; i++)
				_Buffer[i] = fb2_win[0][i];
			break;
	}
}
/*--------------------------------------------------------------------*/
#define	STRSIZ	80
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *command;
	int scr_line, vga_mode, mode;
	int i, hue, bias;
	int ncol = NPIX;
	int nrow = NLIN;
	int col0 = 0;
	int row0 = 0;
	int rotate = 1;
	int shift = 0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf("Usage: gshow command [ncol] [nrow] [frame] [page]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}
	command = argv[1];
	if (argc > 2) sscanf(argv[2], "%d", &ncol);	/* x-dimension */
	if (argc > 3) sscanf(argv[3], "%d", &nrow);	/* y-dimension */
	if (argc > 4) {
		sscanf(argv[4], "%d", &_Frame);	/* frame number */
		if (_Frame < 1 || _Frame > 2) {
			fprintf(stderr, "gshow: Frame number = %d\n", _Frame);
			exit(EXIT_FAILURE);
		}
	}
	if (argc > 5) {
		sscanf(argv[5], "%d", &_Page);	/* page number */
		if (_Page < 0 || _Page >= RAM_PAG) {
			fprintf(stderr, "gshow: Page number = %d\n", _Page);
			exit(EXIT_FAILURE);
		}
	}

	if ((_BufSize = 2*nrow*ncol) > PAG_SIZ) {
		fprintf(stderr, "gshow: image is bigger than one page\n");
		exit(EXIT_FAILURE);
	}

	if ((_Buffer = (char *)malloc(_BufSize)) == (char *)0) {
		fprintf(stderr, "gshow: Can't allocate buffer\n");
		exit(EXIT_FAILURE);
	}

	if ((Array = (short **)malloc(nrow*sizeof(short *))) == (short **)0){
		fprintf(stderr, "gshow: Can't allocate row table\n");
		exit(EXIT_FAILURE);
	}

	Array[0] = (short *)_Buffer;
	for (i = 1; i < nrow; i++)
		Array[i] = Array[i-1] + ncol;

	map_fram(1);	/* map frame buffer RAM */
	set_fwin(_Frame, _Page);	/* select frame and page */

	ReadImage();

	ReadRGB();	/* Read R, G, and B tables */
	hue = 200;
	Lookup(hue);	/* Calculate Lookup Table */
#if	defined(ET4000)	/* Tseng Labs Super VGA */
	vga_mode = _ET640x480;
	scr_line = 29;
	mode = '4';
#else		/* Paradise Super VGA */
	vga_mode = _PA640x480;
	scr_line = 29;
	mode = '3';
#endif
	while (1) {
		printf("Start display at column  (%d): ", col0);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &col0);
		printf("Start display at row     (%d): ", row0);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &row0);
		printf("Subtract offset value    (%d): ", bias);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &bias);
		printf("Divide by 2 to the power (%d): ", shift);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &shift);
		printf("Rotated image? (0|1)     (%d): ", rotate);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &rotate);
		set_mode(vga_mode);
		set_color(rgb);
		Display(extx, exty, col0, row0, rotate, bias, shift);
		CursorControl(hue, scr_line, extx, exty, col0, row0, rotate);
		set_mode(_TEXTC80);
		printf("%s\n", _String);
		printf("Repeat with other VGA mode?\n");
		printf("Standard    320x200: 1\n");
#if	defined(ET4000)	/* Tseng Labs Super VGA */
		printf("Tseng Labs  640x350: 2\n");
		printf("Tseng Labs  640x400: 3\n");
		printf("Tseng Labs  640x480: 4\n");
		printf("Tseng Labs  800x600: 5\n");
		printf("Tseng Labs 1024x768: 6\n");
#else		/* Paradise Super VGA */
		printf("Paradise    640x400: 2\n");
		printf("Paradise    640x480: 3\n");
#endif
		printf("Exit...............: 0\n");
		printf("Enter mode (%c): ", mode);
		fflush(stdout);
		i = getche();
		printf("\n");
		if (i != CR && i != LF)
			mode = i;
		if (mode < '1' || mode > '6')
			break;
		switch (mode) {
			case '1':
				vga_mode = _MRES256COLOR;
				scr_line = 23;
				break;
#if	defined(ET4000)	/* Tseng Labs Super VGA */
			case '2':
				vga_mode = _ET640x350;
				scr_line = 24;
				break;
			case '3':
				vga_mode = _ET640x400;
				scr_line = 24;
				break;
			case '4':
				vga_mode = _ET640x480;
				scr_line = 29;
				break;
			case '5':
				vga_mode = _ET800x600;
				scr_line = 36;
				break;
			case '6':
				vga_mode = _ET1024x768;
				scr_line = 46;
				break;
#else		/* Paradise Super VGA */
			case '2':
				vga_mode = _PA640x400;
				scr_line = 24;
				break;
			case '3':
				vga_mode = _PA640x480;
				scr_line = 29;
				break;
#endif
		}
	}
	free(Array);
	free(_Buffer);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
