
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <ctype.h>
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "wild.h"	/* function prototype for wild.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	LINSIZ	80	/* maximum length of input line */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0, n_arg = 0;
	int interactive = 0;
	char **arg, **itab, **imgtab, *name;
	char str[LINSIZ];
	int i, n, ntab;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "irm: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "i")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'i':	interactive++; break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 1) {
		printf("Usage: irm -i image1 [image2] [image3]...\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "irm: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if (n_arg > 1) {	/* remove images, if 2 or more are specified */

		if ((itab = (char **)calloc(n_arg, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "irm: calloc: Can't allocate image table\n");
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "irm: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}

		for (n = 0, i = 0; i < n_arg; ++i) {
			name = truncname(arg[i]);
			if (strpbrk(name, punct) != (char *)0)
				printf("%s is not a legal image name\n", name);
			else if (xist(name))	/* check if image exists */
				itab[n++] = name;
			else
				printf("%s does not exist\n", name);
		}

	} else {	/* wild card possible, if only first image is given */

		imgtab = getitab(&ntab);	/* table containing all image names */

		if (ntab > 0)
			if ((itab = (char **)calloc(ntab,
					sizeof(char *))) == (char **)0) {
				fprintf(stderr,
					"irm: calloc: Can't allocate image table\n");
				if (epilog() < 1)	/* update disk headers */
					fprintf(stderr,
						"irm: epilog: Disk header(s) not updated\n");
				exit(EXIT_FAILURE);
			}

		name = arg[0];

		for (n = 0, i = 0; i < ntab; ++i)	/* find name in image table */
			if (wildcmp(imgtab[i], name) == 0)
				itab[n++] = imgtab[i];
		if (n == 0)
			printf("%s does not exist\n", name);
	}

	for (i = 0; i < n; i++) {
		name = itab[i];	/* image name */
		if (interactive != 0) {
			printf("%s: ? (y/n/e) ", name);
			fgets(str, LINSIZ, stdin);
			switch (str[0]) {
				case 'y':
				case 'Y':
						if (iremove(name) < 0)	/* remove image */
							printf("Can't remove %s\n", name);
						break;
				case 'e':
				case 'E':
						goto End;
				case 'n':
				case 'N':
				default:
						break;
			}
		} else if (iremove(name) < 0)	/* remove image */
			printf("Can't remove %s\n", name);
	}

End:

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "irm: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
