
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
   int argc;
   char **argv;
   int option, opt_err = 0;
   int n_arg = 0;
   char **arg;
   int x_flip = 0;	/* mirror x-dimension */
   int y_flip = 0;	/* mirror y-dimension */
   char *image, *ptr;
   void **a, *buffer;
   int i, x, y, x0, y0;
   int id, ncols, nrows, type;
   int start[2], dim[2];
   unsigned int err;

#if	defined(ARGS)
   /* substitute some arguments (: or !) from previous command line */
   argv = args(Argc, Argv, &argc);
#else	/* no substitution */
   argc = Argc;
   argv = Argv;
#endif

   if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
      fprintf(stderr, "ishift: calloc: Can't allocate argument table\n");
      exit(EXIT_FAILURE);
   }

   while (((option = getopt(argc, argv, "xy")) != NONOPT)
         || (optarg != NULL)) {
      switch(option) {
         case 'x':	x_flip++; break;
         case 'y':	y_flip++; break;
         case '?':	opt_err++; break;
         case NONOPT:	arg[n_arg++] = optarg; break;
         default:	break;
      }
   }

   if (opt_err || n_arg < 3) {
      printf("Usage: ishift -x -y image X0 Y0\n");
		fflush(stdout);
      exit(EXIT_FAILURE);
   }

   image = truncname(arg[0]);	/* image name */
   if (strpbrk(image, punct) != (char *)0) {
      fprintf(stderr, "ishift: Illegal image name: %s\n", image);
      exit(EXIT_FAILURE);
   }

   sscanf(arg[1], "%d", &x0);
   sscanf(arg[2], "%d", &y0);

   /* read image names from header directory */
   if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
      fprintf(stderr, "ishift: prolog: Error = %d\n", i);
      exit(EXIT_FAILURE);
   }
   setipath(0);	/* select first token in IMGPATH */
   if ((id = iopen(image, 0, &type, &ncols, &nrows)) < 0) {
      fprintf(stderr, "ishift: Can't open %s\n", image);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr,
            "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
   if (type < 1 || type > 5) {
      fprintf(stderr,
         "ishift: %s has illegal type = %d\n", image, type);
      iclose(id);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr,
            "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
   if ((buffer = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
      fprintf(stderr, "ishift: Allocation of row buffer failed\n");
      iclose(id);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr,
            "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
   dim[0] = nrows;	/* number of rows: y-dimension */
   dim[1] = ncols;	/* number of columns: x-dimension */
   start[0] = 0;	/* first index of y-dimension */
   start[1] = 0;	/* first index of x-dimension */
   if ((a = (void **)daa(bytepix(type), 2, dim, start,
         &err, &ptr, NULL)) == (void **)0) {
      fprintf(stderr, "ishift: daa: Error code = %d\n", err);
      iclose(id);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr,
            "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
   for (y = 0; y < nrows; ++y) {	/* read image and shift rows */
      getline(id, y, buffer);
      i = ncols - x0;
      while (i < 0) i += ncols;
      while (i >= ncols) i -= ncols;
      switch (type) {
         case 1:	/* UCHAR */
            if (x_flip != 0) { /* mirror x-dimension */
               for (x = (ncols-1); x >= 0; ++i, --x) {
                  if (i == ncols) i = 0;
                  ((UCHAR **)a)[y][x] = ((UCHAR *)buffer)[i];
               }
            } else {
               for (x = 0; x < ncols; ++i, ++x) {
                  if (i == ncols) i = 0;
                  ((UCHAR **)a)[y][x] = ((UCHAR *)buffer)[i];
               }
            }
            break;
         case 2:	/* short */
            if (x_flip != 0) { /* mirror x-dimension */
               for (x = (ncols-1); x >= 0; ++i, --x) {
                  if (i == ncols) i = 0;
                  ((short **)a)[y][x] = ((short *)buffer)[i];
               }
            } else {
               for (x = 0; x < ncols; ++i, ++x) {
                  if (i == ncols) i = 0;
                  ((short **)a)[y][x] = ((short *)buffer)[i];
               }
            }
            break;
         case 3:	/* float */
            if (x_flip != 0) { /* mirror x-dimension */
               for (x = (ncols-1); x >= 0; ++i, --x) {
                  if (i == ncols) i = 0;
                  ((float **)a)[y][x] = ((float *)buffer)[i];
               }
            } else {
               for (x = 0; x < ncols; ++i, ++x) {
                  if (i == ncols) i = 0;
                  ((float **)a)[y][x] = ((float *)buffer)[i];
               }
            }
            break;
         case 4:	/* LONG int */
            if (x_flip != 0) { /* mirror x-dimension */
               for (x = (ncols-1); x >= 0; ++i, --x) {
                  if (i == ncols) i = 0;
                  ((LONG **)a)[y][x] = ((LONG *)buffer)[i];
               }
            } else {
               for (x = 0; x < ncols; ++i, ++x) {
                  if (i == ncols) i = 0;
                  ((LONG **)a)[y][x] = ((LONG *)buffer)[i];
               }
            }
            break;
         case 5:	/* COMPLEX */
            if (x_flip != 0) { /* mirror x-dimension */
               for (x = (ncols-1); x >= 0; ++i, --x) {
                  if (i == ncols) i = 0;
                  ((COMPLEX **)a)[y][x] = ((COMPLEX *)buffer)[i];
               }
            } else {
               for (x = 0; x < ncols; ++i, ++x) {
                  if (i == ncols) i = 0;
                  ((COMPLEX **)a)[y][x] = ((COMPLEX *)buffer)[i];
               }
            }
            break;
      }
   }
   if (!chiomod(id, 2)) {	/* change IO mode to update */
      fprintf(stderr, "ishift: Can't change IO mode for %s\n", image);
      iclose(id);
      if (epilog() < 1)	/* update disk headers */
         fprintf(stderr, "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
   i = nrows - y0;
   while (i < 0) i += nrows;
   while (i >= nrows) i -= nrows;
   if (y_flip != 0) { /* mirror y-dimension */
      for (y = (nrows-1); y >= 0; ++i, --y) { /* shift rows and write back */
         if (i == nrows) i = 0;
         switch (type) {
            case 1:	putline(id, y, ((UCHAR **)a)[i]); break;
            case 2:	putline(id, y, ((short **)a)[i]); break;
            case 3:	putline(id, y, ((float **)a)[i]); break;
            case 4:	putline(id, y, ((LONG **)a)[i]); break;
            case 5:	putline(id, y, ((COMPLEX **)a)[i]); break;
         }
      }
   } else {
      for (y = 0; y < nrows; ++i, ++y) { /* shift rows and write back */
         if (i == nrows) i = 0;
         switch (type) {
            case 1:	putline(id, y, ((UCHAR **)a)[i]); break;
            case 2:	putline(id, y, ((short **)a)[i]); break;
            case 3:	putline(id, y, ((float **)a)[i]); break;
            case 4:	putline(id, y, ((LONG **)a)[i]); break;
            case 5:	putline(id, y, ((COMPLEX **)a)[i]); break;
         }
      }
   }
   free(ptr);
   free(buffer);
   iclose(id);
   if (epilog() < 1) {	/* update disk headers */
      fprintf(stderr, "ishift: epilog: Disk header(s) not updated\n");
      exit(EXIT_FAILURE);
   }
	fflush(stdout);
   exit(EXIT_SUCCESS);
}
