
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define ABS(x)		(((x) < (0)) ? -(x) : (x))
/*--------------------------------------------------------------------*/
#define	MAXDIM	10
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static float Amotry(float *p[], float y[], float psum[], int ndim,
		float (*funk)(float x[]), int ihi, int *nfunk, double fac)
#else	/* old style */
static float Amotry(p, y, psum, ndim, funk, ihi, nfunk, fac)
int ndim, ihi, *nfunk;
float *p[], y[], psum[], (*funk)();
double fac;
#endif
{
	int j;
	float fac1, fac2, ytry, ptry[MAXDIM];

	if (ndim > MAXDIM) {
		fprintf(stderr, "Amotry: Too large dimension: ndim = %d\n", ndim);
		exit(EXIT_FAILURE);
	}

	fac1 = (float) ((1.0 - fac)/ndim);
	fac2 = (float) (fac1 - fac);
	for (j = 0; j < ndim; j++)
		ptry[j] = psum[j]*fac1 - p[ihi][j]*fac2;
	ytry = (*funk)(ptry);
	++(*nfunk);
	if (ytry < y[ihi]) {
		y[ihi] = ytry;
		for (j = 0; j < ndim; j++) {
			psum[j] += ptry[j] - p[ihi][j];
			p[ihi][j] = ptry[j];
		}
	}
	return ytry;
}
/*--------------------------------------------------------------------*/
#define	ALPHA	1.0
#define	BETA	0.5
#define	GAMMA	2.0
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int Amoeba(float *p[], float y[], int ndim, double ftol,
			float (*funk)(float x[]), int nmax)
#else	/* old style */
static int Amoeba(p, y, ndim, ftol, funk, nmax)
int ndim, nmax;
float *p[], y[], (*funk)();
double ftol;
#endif
{
	int i, j, ilo, ihi, inhi, mpts = ndim + 1, nfunk = 0;
	float ytry, ysave, psum[MAXDIM];
	const float Half = (float) 0.5;
	double rtol;

	if (ndim > MAXDIM) {
		fprintf(stderr, "Amoeba: Too large dimension: ndim = %d\n", ndim);
		exit(EXIT_FAILURE);
	}

	for (j = 0; j < ndim; j++) {
		float sum = (float)0;
		for (i = 0; i < mpts; i++)
			sum += p[i][j];
		psum[j] = sum;
	}
	for (;;) {
		ilo = 0;
		ihi = (y[0] > y[1]) ? (inhi = 1, 0) : (inhi = 0, 1);
		for (i = 0; i < mpts; i++) {
			if (y[i] < y[ilo]) ilo = i;
			if (y[i] > y[ihi]) {
				inhi = ihi;
				ihi = i;
			} else if (y[i] > y[inhi])
				if (i != ihi) inhi = i;
		}
		rtol = 2.0*fabs(y[ihi] - y[ilo])/(fabs(y[ihi]) + fabs(y[ilo]));

		if (rtol < ftol || nfunk > nmax) break;

		ytry = Amotry(p, y, psum, ndim, funk, ihi, &nfunk, -ALPHA);
		if (ytry <= y[ilo])
			ytry = Amotry(p, y, psum, ndim, funk, ihi, &nfunk, GAMMA);
		else if (ytry >= y[inhi]) {
			ysave = y[ihi];
			ytry = Amotry(p, y, psum, ndim, funk, ihi, &nfunk, BETA);
			if (ytry >= ysave) {
				for (i = 0; i < mpts; i++) {
					if (i != ilo) {
						for (j = 0; j < ndim; j++) {
							psum[j] = Half*(p[i][j] + p[ilo][j]);
							p[i][j] = psum[j];
						}
						y[i] = (*funk)(psum);
					}
				}
				nfunk += ndim;
				for (j = 0; j < ndim; j++) {
					float sum = (float)0;
					for (i = 0; i < mpts; i++)
						sum += p[i][j];
					psum[j] = sum;
				}
			}
		}
	}
	return nfunk;
}
/*--------------------------------------------------------------------*/
#undef	ALPHA
#undef	BETA
#undef	GAMMA
#undef	MAXDIM
/*--------------------------------------------------------------------*/
static float *Array = (float *)0;	/* pointer to image array */
static int Ncols = 0;	/* number of columns in array */
static int Nrows = 0;	/* number of rows in array */
static float Rmax = (float)0;	/* radius of fitting area */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static float Mad(float par[])
#else	/* old style */
static float Mad(par)	/* calculates the Mean Absolute Devition */
float par[];
#endif
{
	int k, y, npoints = 0;
	float rc2 = par[2]*par[2];	/* Rc^2 */
	float amp = rc2*par[1];	/* Ic*Rc^2 */
	float r2max = Rmax*Rmax;
	float mean_abs_dev = (float)0;
	const float FltMax = (float) (0.9*FLT_BLANK);

	for (k = 0, y = 0; y < Nrows; k += Ncols, ++y) {
		float y2 = (y - par[4]);	/* (y - yc) */
		int x;
		y2 *= y2;
		for (x = 0; x < Ncols; ++x) {
			float r2 = (x - par[3]);	/* (x - xc) */
			r2 *= r2;
			r2 += y2;
			if (r2 < r2max)
				if (Array[k+x] <= FltMax) {
					float d = Array[k+x] - par[0] - amp/(rc2 + r2);
					mean_abs_dev += ABS(d);
					++npoints;
				}
		}
	}
	if (npoints > 0)
		mean_abs_dev /= npoints;

	return mean_abs_dev;
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Smooth(float par[])
#else	/* old style */
static void Smooth(par)	/* calculates a synthetic image */
float par[];
#endif
{
	int k, y;
	float rc2 = par[2]*par[2];	/* Rc^2 */
	float amp = rc2*par[1];	/* Ic*Rc^2 */
	float r2max = Rmax*Rmax;

	for (k = 0, y = 0; y < Nrows; k += Ncols, ++y) {
		float y2 = (y - par[4]);	/* (y - yc) */
		int x;
		y2 *= y2;
		for (x = 0; x < Ncols; ++x) {
			float r2 = (x - par[3]);	/* (x - xc) */
			r2 *= r2;
			r2 += y2;
			if (r2 < r2max)
				Array[k+x] = par[0] + amp/(rc2 + r2);
		}
	}
}
/*----------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 13-Aug-92
----------------------------------------------------------------------*/
#define	MP	6	/* number of points in simplex */
#define	NP	5	/* number of parameters */
#define	FTOL	1.0e-6
#define	NMAX	5000	/* maximum number of function calls */
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int nmax = NMAX;	/* maximum number of function calls */
	char *image, *image1 = (char *)0;	/* output image */
	void *buffer;
	int i, k, y, id, type, nmad;
	float par[NP], mad[MP], aux[MP][NP], *simplex[MP];
	float mad_min;
	double Blank;
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Zero = (float)0;
	const float Half = (float)0.5;

	for (k = 0; k < MP; ++k)
		simplex[k] = &aux[k][0];

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "imake: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "n:o:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'n':	sscanf(optarg, "%d", &nmax); break;
			case 'o':	image1 = optarg; break;	/* output image */
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	/* illegal options */
	if (opt_err || n_arg < 7) {
		printf(
		"Usage: kingfit -n nmax -o image1 image Sky Ic Rc Xc Yc Rmax\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* input image */
	/* is image name legal? */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "kingfit: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	/* is output image legal? */
	if (image1 != (char *)0) {
		image1 = truncname(image1);
		if (strpbrk(image1, punct) != (char *)0) {
			fprintf(stderr,
				"kingfit: Illegal image name: %s\n", image1);
			exit(EXIT_FAILURE);
		}
	}

	sscanf(arg[1], "%f", &par[0]);	/* sky value */
	sscanf(arg[2], "%f", &par[1]);	/* central surface brightness Ic */
	sscanf(arg[3], "%f", &par[2]);	/* core radius Rc */
	sscanf(arg[4], "%f", &par[3]);	/* X-coordinate of center xc */
	sscanf(arg[5], "%f", &par[4]);	/* Y-coordinate of center yc */
	sscanf(arg[6], "%f", &Rmax);	/* max fitting radius Rmax */

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "kingfit: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open image for read-only */
	if ((id = iopen(image, 0, &type, &Ncols, &Nrows)) < 0) {
		fprintf(stderr, "kingfit: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "kingfit: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr,
			"kingfit: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "kingfit: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	if (image1 != (char *)0)
		if (xist(image1)) {	/* does output image exist? */
			fprintf(stderr, "kingfit: %s: Existing image\n", image1);
			iclose(id);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"kingfit: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}

	/* allocate float array Array[] */
	if ((Array = (float *)calloc(Nrows,
			Ncols * sizeof(float))) == (float *)0) {
		fprintf(stderr, "kingfit: Allocation of image array failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "kingfit: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate row buffer */
	if ((buffer = (void *)calloc(Ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "kingfit: Allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "kingfit: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < Nrows; ++y) {	/* read image into Array[] */
		getline(id, y, buffer);	/* read next row */
		switch (type) {
			case 1:	/* unsigned char */
				{
					float *p = &Array[y*Ncols];
					const UCHAR *q = (UCHAR *) buffer;
					int x;
					for (x = 0; x < Ncols; ++x)
						p[x] = (float) q[x];
				}
				break;
			case 2:	/* short int */
				{
					float *p = &Array[y*Ncols];
					const short *q = (short *) buffer;
					const short b = (short) Blank;
					int x;
					for (x = 0; x < Ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					float *p = &Array[y*Ncols];
					const float *q = (float *) buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < Ncols; ++x)
						p[x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					float *p = &Array[y*Ncols];
					const LONG *q = (LONG *) buffer;
					const LONG b = (LONG) Blank;
					int x;
					for (x = 0; x < Ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					float *p = &Array[y*Ncols];
					const COMPLEX *q = (COMPLEX *) buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < Ncols; ++x)
						p[x] = (q[x].x > h || q[x].y > h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);	/* close image file */

	/* calculate the starting simplex */
	for (k = 0; k < MP; ++k)
		for (i = 0; i < NP; ++i)
			simplex[k][i] = par[i];
	for (i = 0; i < NP; ++i)
		simplex[i+1][i] *= (float)1.2;

	/* calculate the functional values */
	for (k = 0; k < MP; ++k)
		mad[k] = Mad(simplex[k]);

	nmad = Amoeba(simplex, mad, NP, FTOL, Mad, nmax);

	printf("\nNo. of function evaluations: %4d\n", nmad);
	printf("Vertices of final 5-d simplex and\n");
	printf("function values at the vertices:\n\n");
	printf("%3s %8s %8s %8s %8s %8s %10s\n\n",
		"No", "Sky", "Ic", "Rc", "Xc", "Yc", "Function");
	for (k = 0; k < MP; k++) {
		printf("%3d ", k);
		for (i = 0; i < NP; ++i)
			printf("%8.2f ", simplex[k][i]);
		printf("%10.6f\n", mad[k]);
	}

	mad_min = mad[0];
	for (k = 0, i = 1; i < NP; ++i)
		if (mad[i] < mad_min) {
			mad_min = mad[i];
			k = i;
		}

	for (i = 0; i < NP; ++i)
		par[i] = simplex[k][i];

	/* create a smooth image */
	if (image1 != (char *)0) {

		printf("\nStart calculating the synthetic image\n");

		/* calculate a smooth image in Array[] */
		Smooth(par);

		if (cpi(image, image1) < 1) {	/* dublicate image header */
			fprintf(stderr,
				"kingfit: %s.hdr -> %s.hdr copy failed\n", image, image1);
			goto Finish;
		} else if (setimagefile(image1, "") == 0) {	/* reset FILE key */
			fprintf(stderr,
				"kingfit: Can't erase old FILE value for %s\n", image1);
			rmi(image1);	/* remove image header */
			goto Finish;
		}
		if ((id = icreat(image1, 1, type, Ncols, Nrows)) < 0) {
			fprintf(stderr, "kingfit: Can't create %s\n", image1);
			rmi(image1);	/* remove image header */
			goto Finish;
		}

		if (update() == 0)	/* update disk headers */
			fprintf(stderr, "kingfit: update: Disk header not created\n");

		Blank = blank(image1);

		printf("Write output image to disk\n");

		for (y = 0; y < Nrows; ++y) {	/* write image to disk */
			switch (type) {
				case 1:
					{
						UCHAR *q = (UCHAR *) buffer;
						const float *p = &Array[y*Ncols];
						int x;
						for (x = 0; x < Ncols; ++x)
							q[x] = (UCHAR)(p[x] + Half);
					}
					break;
				case 2:
					{
						short *q = (short *) buffer;
						const float *p = &Array[y*Ncols];
						const short b = (short) Blank;
						int x;
						for (x = 0; x < Ncols; ++x)
							q[x] = (p[x] > FltMax)? b:
									(short)((p[x]<0)? p[x]-Half: p[x]+Half);
					}
					break;
				case 3:
					{
						float *q = (float *) buffer;
						const float *p = &Array[y*Ncols];
						const float b = (float) Blank;
						int x;
						for (x = 0; x < Ncols; ++x)
							q[x] = (p[x] > FltMax)? b: p[x];
					}
					break;
				case 4:
					{
						LONG *q = (LONG *) buffer;
						const float *p = &Array[y*Ncols];
						const LONG b = (LONG) Blank;
						int x;
						for (x = 0; x < Ncols; ++x)
							q[x] = (p[x] > FltMax)? b:
									(LONG)((p[x]<0)? p[x]-Half: p[x]+Half);
					}
					break;
				case 5:
					{
						COMPLEX *q = (COMPLEX *) buffer;
						const float *p = &Array[y*Ncols];
						const float b = (float) Blank;
						int x, ip;
						for (x = 0; x < Ncols; ++x) {
							ip = p[x] > FltMax;
							q[x].x = (ip)? b: p[x];
							q[x].y = (ip)? b: Zero;
						}
					}
					break;
			}
			putline(id, y, buffer);	/* write row to disk */
		}

		iclose(id);	/* close output image */

		setiname(image1);
		sethistory(image1, "kingfit");
		sethistory(image1, image1);
	}

Finish:
	free(buffer);
	free(Array);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "kingfit: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
