
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define  FWHM  5.0
#define  FACT  1.5
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	double fwhm = FWHM;
	int i, j, x, y;
	int x_max, y_max;
	int id, type, npix, nlin, rw = 0;
	double x1, y1, x2, y2, xd, yd, yp, x0, y0, r12, co, si;
   double xc1, xc2, xc3, xc4, yc1, yc2, yc3, yc4;
   double xmin, xmax, ymin, ymax, wsum, dsum, weight;
	double Blank, data;
	void *buffer;
	char *image, *ptr;
	float **a;
	unsigned int err;
	int start[2], dim[2];
	const float FltBlank = (float)FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	float Init_Value = FltBlank;
   double ln2 = log(2.0);

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "addstar: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "w:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'w':	sscanf(optarg, "%lf", &fwhm); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 5) {
		printf("Usage: line -w fwhm image X1 Y1 X2 Y2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	if (fwhm <= 0.0) {
		fprintf(stderr,
			"line: Illegal full width at half maximum = %g\n", fwhm);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "line: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &x1);
	sscanf(arg[2], "%lf", &y1);
	sscanf(arg[3], "%lf", &x2);
	sscanf(arg[4], "%lf", &y2);
   r12 = sqrt((x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1));
	if (r12 == 0.0) {
		fprintf(stderr, "line: Length == 0.0\n");
		exit(EXIT_FAILURE);
	}
   co = (x2 - x1)/r12;
   si = (y2 - y1)/r12;
   x_max = (int) (r12 + 0.5);
   y_max = (int) (FACT*fwhm + 0.5);

   x0 = x1;
   y0 = y1;
   xc1 = x0 + y_max*si;
   yc1 = y0 - y_max*co;
   xc2 = x0 - y_max*si;
   yc2 = y0 + y_max*co;

   x0 = x1 + x_max*co;
   y0 = y1 + x_max*si;
   xc3 = x0 + y_max*si;
   yc3 = y0 - y_max*co;
   xc4 = x0 - y_max*si;
   yc4 = y0 + y_max*co;

   xmin = xc1;
   xmin = MIN(xmin,xc2);
   xmin = MIN(xmin,xc3);
   xmin = MIN(xmin,xc4);

   xmax = xc1;
   xmax = MAX(xmax,xc2);
   xmax = MAX(xmax,xc3);
   xmax = MAX(xmax,xc4);

   ymin = yc1;
   ymin = MIN(ymin,yc2);
   ymin = MIN(ymin,yc3);
   ymin = MIN(ymin,yc4);

   ymax = xc1;
   ymax = MAX(ymax,yc2);
   ymax = MAX(ymax,yc3);
   ymax = MAX(ymax,yc4);

	if (xmin < 0.0) {
		fprintf(stderr, "line: xmin < 0.0\n");
		exit(EXIT_FAILURE);
	}
	if (ymin < 0.0) {
		fprintf(stderr, "line: ymin < 0.0\n");
		exit(EXIT_FAILURE);
	}

   /* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "line: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "line: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr, "line: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xmax > npix-1) {
		fprintf(stderr, "line: xmax > %d\n", npix-1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (ymax > nlin-1) {
		fprintf(stderr, "line: ymax > %d\n", nlin-1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "line: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	dim[0] = nlin;	/* y dimension */
	dim[1] = npix; /* x dimension */
	start[0] = 0;	/* first index of y dimension */
	start[1] = 0;	/* first index of x dimension */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "line: daa: error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nlin; ++y) {	/* read image into a[][] */
		getline(id, y, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* convert unsigned char to float */
				{
					const UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (float)q[x];
				}
				break;
			case 2:	/* convert short int to float */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* convert LONG int to float */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x].x>h || q[x].y>h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "line: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

   /* step through pixels along the line */
   for (i = 0; i <= x_max; ++i) {
      x0 = x1 + i*co;
      y0 = y1 + i*si;
      dsum = 0.0;
      wsum = 0.0;
      /* step through pixels across to the line */
      for (j = -y_max; j <= y_max; ++j) {
         xd = x0 - j*si;
         yd = y0 + j*co;
         /* find the nearest pixel */
         x = (int) (xd + 0.5);
         y = (int) (yd + 0.5);
         /* actual distance of pixel from line */
         yp = (y - y0)*co - (x - x0)*si;
         yp = 2.0*yp/fwhm;
         weight = exp(-ln2*yp*yp);
         /* sum only good pixel data */
         if (a[y][x] <= FltMax) {
            dsum += weight*a[y][x];
            wsum += weight;
         }
      };
      data = (wsum == 0.0) ? FLT_BLANK : dsum/wsum;
      printf("%5d %12g\n", i, data);
   }

	free(ptr);
	free(buffer);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
