
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "wild.h"	/* function prototype for wild.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
char *type[] = {"undef", "uchar", "short", "float", "long", "cmplx"};
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int lev_dist = -1, long_list = 0;
	char *name, *dt, *ob, **imgtab, **itab = (char **)0;
	int i, m, n, ntab, nx, ny;
	long sz;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "list: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "lsd:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
         case 'l':   long_list++; break;
         case 's':   long_list--; break;
         case 'd':   sscanf(optarg, "%d", &lev_dist); break;
         case '?':   opt_err++; break;
         case NONOPT:   arg[n_arg++] = optarg; break;
         default:    break;
		}
	}

	if (opt_err) {
		printf("Usage: list -l -s -d levdist [image1] [image2] ...\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "list: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	if (n_arg > 1) {	/* list images, if 2 or more are specified */

		if ((itab = (char **)calloc(n_arg, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "list: calloc: Can't allocate image table\n");
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "list: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}

		for (n = 0, i = 0; i < n_arg; ++i) {
			name = truncname(arg[i]);
			if (strpbrk(name, punct) != (char *)0)
				printf("%s is not a legal image name\n", name);
			else if (xist(name))	/* check if image exists */
				itab[n++] = name;
			else
				printf("%s does not exist\n", name);
		}

	} else {	/* wild card possible, if only first image is given */

		imgtab = getitab(&ntab);	/* table containing all image names */

		if (ntab > 0)
			if ((itab = (char **)calloc(ntab,
					sizeof(char *))) == (char **)0) {
				fprintf(stderr,
					"list: calloc: Can't allocate image table\n");
				if (epilog() < 1)	/* update disk headers */
					fprintf(stderr,
						"list: epilog: Disk header(s) not updated\n");
				exit(EXIT_FAILURE);
			}

		if (n_arg == 0) name = "*"; else name = arg[0];

		if (lev_dist >= 0) {	/* use Levenstein-distance selection */
			for (n = 0, i = 0; i < ntab; ++i)	/* find names in table */
				if (levdist(name, imgtab[i]) <= lev_dist)
					itab[n++] = imgtab[i];
		} else {	/* use wildcard selection */
			for (n = 0, i = 0; i < ntab; ++i)	/* find names in table */
				if (wildcmp(imgtab[i], name) == 0)
					itab[n++] = imgtab[i];
		}

		if (n == 0)
			printf("%s does not exist\n", name);
	}

   if (long_list > 0)	/* long list option */
		for (i = 0; i < n; i++) {
			name = itab[i];	/* next image name */
			m = pixtype(name);	/* pixel type */
			if (m < 1 || m > 5)	/* type is not defined */
				m = pixsize(name);	/* use bytes per pixel */
			sz = imagesize(name);	/* image size in bytes */
			nx = naxis1(name);	/* x-dimension */
			ny = naxis2(name);	/* y-dimension */
			printf("%-8s%6s%8ld%6d%6d", name, type[m], sz, nx, ny);
			ob = object(name);	/* object name */
			if (ob != NULL) printf("\t%s", ob);
			dt = dateobs(name);	/* observation date */
			if (dt != NULL) printf("\t%s", dt);
			printf("\n");
		}
   else if (long_list < 0)	/* short list option */
		for (i = 0; i < n; i++) {
			name = itab[i];	/* next image name */
			printf("%s\n", name);
		}
   else {	/* normal list option */
		for (i = 1; i <= n; i++) {
			name = itab[i-1];	/* next image name */
			if ((i%8) != 0)
				printf("%-8s ", name);
			else
				printf("%-8s\n", name);
		}
		if (n%8 > 0)
			printf("\n");
   }

   if (long_list >= 0) {
      if (n == 1)
         printf("%d image\n", n);
      else if (n > 1)
         printf("%d images\n", n);
   }

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "list: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
