
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "wild.h"	/* function prototype for wild.c */
#include "daa.h"	/* function declarations for daa.c */
#include "medi.h"	/* function declarations for medi.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	IMAX	64	/* maximum number of input images */
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv, **arg;
	int option, opt_err = 0, n_arg = 0;
	double limit = 0.0;     /* clipping limit */
	char *file = (char *)0;	/* optional input file with weights */
	int i, j, n, nimg, ntab, n_weight;
	int id0, rw0 = 1;
	int id[IMAX], type[IMAX], npix[IMAX], nlin[IMAX], rw = 0;
	float weight[IMAX];
	double Blank[IMAX];
	char *ptr, *name, *image0;
	char **image, **imgtab;
	void **arr, *buf;
	unsigned int err;
	int start[2], dim[2];

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "median: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	/* process command line options */
	while (((option = getopt(argc, argv, "l:w:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'l': sscanf(optarg, "%lf", &limit); break;
			case 'w': file = optarg; break; /* input weights file */
			case '?': opt_err++; break;
			case NONOPT: arg[n_arg++] = optarg; break;
			default: break;
		}
	}

	nimg = n_arg - 1;	/* number of input images */

	if (opt_err || nimg < 1) {
		printf(
	"Usage: median -l limit -w file image image1 [image2] [image3] ...\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	/* read optional weights for the final mean, or use unit weights */
	if (file != (char *)0) { /* read weights from a given file */
		FILE *fp;
		if ((fp = fopen(file,"r")) == NULL) {
			fprintf(stderr, "median: Cannot open input file: %s\n", file);
			exit(EXIT_FAILURE);
		}
		for (n_weight = 0; n_weight < IMAX ; n_weight++)
			if (fscanf(fp, "%f", &weight[n_weight]) == EOF)
				break;
	} else /* use unit weights */
		for (n_weight = 0; n_weight < IMAX; n_weight++)
			weight[n_weight] = (float)1;

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "median: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	if (nimg > 1) {	/* input images given on command line */
		if ((image = (char **)calloc(nimg, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "median: calloc: Allocation failed-1\n");
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		for (n = 0, i = 0; i < nimg; i++) {
			name = truncname(arg[i+1]);	/* next input image */
			if (strpbrk(name, punct) != (char *)0)
				printf("%s is not a legal image name\n", name);
			else if (xist(name))	/* check if image exists */
				image[n++] = name;	/* include in input list */
			else
				printf("%s does not exist\n", name);
		}
	} else {	/* wild card possible, if only one input image is given */
		imgtab = getitab(&ntab);	/* table containing all image names */
		image = (char **)0;	/* initialize the image table pointer */
		if (ntab > 0)	/* directory is not empty */
			if ((image = (char **)calloc(ntab,
					sizeof(char *))) == (char **)0) {
				fprintf(stderr, "median: calloc: Allocation failed-2\n");
				if (epilog() < 1)	/* update disk headers */
					fprintf(stderr,
						"median: epilog: Disk header(s) not updated\n");
				exit(EXIT_FAILURE);
			}
		name = arg[1];	/* input image, which may contain wild card */
		for (n = 0, i = 0; i < ntab; i++)	/* is name in image table */
			if (wildcmp(imgtab[i], name) == 0)
				image[n++] = imgtab[i];
	}

	nimg = n;	/* number of input images to process */
	if(nimg == 0) {
		fprintf(stderr, "median: No input images found\n");
		exit(EXIT_FAILURE);
	}

	if(nimg > IMAX) {	/* too many input images */
		fprintf(stderr, "median: Too many input images: %d\n", nimg);
		exit(EXIT_FAILURE);
	}

	if(n_weight < nimg) {	/* too few weights */
		fprintf(stderr,
		"median: #weights (%d)  <  #images (%d)\n", n_weight, nimg);
		exit(EXIT_FAILURE);
	}

	image0 = truncname(arg[0]);	/* output image */
	if (strpbrk(image0, punct) != (char *)0) {
		fprintf(stderr, "median: Illegal image name: %s\n", image0);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	for (i = 0; i < nimg; i++) {	/* open input images */
		if ((id[i] = iopen(image[i], rw, &type[i], &npix[i], &nlin[i])) < 0) {
			fprintf(stderr, "median: Can't open %s\n", image[i]);
			for (j = (i-1); j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		if (type[i] < 1 || type[i] > 4) {
			fprintf(stderr, "median: %s has illegal type = %d\n",
				image[i], type[i]);
			for (j = i; j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		Blank[i] = blank(image[i]);
	}

	for (i = 1; i < nimg; i++) {	/* check image types and sizes */
		if (type[i] != type[0]) {
			fprintf(stderr,
				"median: %s and %s have incompatible types\n",
					image[i], image[0]);
			for (j = (nimg-1); j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		if (npix[i] != npix[0]) {
			fprintf(stderr,
				"median: %s and %s have incompatible x-dimensions\n",
					image[i], image[0]);
			for (j = (nimg-1); j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		if (nlin[i] != nlin[0]) {
			fprintf(stderr,
				"median: %s and %s have incompatible y-dimensions\n",
					image[i], image[0]);
			for (j = (nimg-1); j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		if (fabs(Blank[i]/Blank[0] - 1.0) > 1.0E-10) {
			fprintf(stderr,
				"median: %s and %s have incompatible BLANKs\n",
					image[i], image[0]);
			for (j = (nimg-1); j >= 0; j--)
				iclose(id[j]);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
	}

	if (xist(image0)) {	/* does output image exist? */
		fprintf(stderr, "median: %s: Existing image\n", image0);
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image[0], image0) < 1) {	/* dublicate image header */
		fprintf(stderr,
			"median: %s.hdr -> %s.hdr copy failed\n", image[0], image0);
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image0, "") == 0) {	/* reset FILE key */
		fprintf(stderr,
			"median: Can't erase old FILE value for %s\n", image0);
		rmi(image0);	/* remove image header */
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((id0 = icreat(image0, rw0, type[0], npix[0], nlin[0])) < 0) {
		fprintf(stderr, "median: Can't create %s\n", image0);
		rmi(image0);	/* remove image header */
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "median: update: Disk header not created\n");

	dim[0] = nimg;	/* number of input images */
	dim[1] = bpixels(image0);	/* pixels in one image block */
	start[0] = 0;	/* first image index */
	start[1] = 0;	/* first pixel index */

	if ((buf = (void *)calloc(dim[1], bytepix(type[0]))) == (void *)0) {
		fprintf(stderr, "median: Can't allocate output buffer\n");
		iclose(id0);
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((arr = (void **)daa(bytepix(type[0]), 2, dim, start,
			&err, &ptr, NULL)) == (void **)0) {
		fprintf(stderr, "median: daa: Error code = %d\n", err);
		iclose(id0);
		for (j = (nimg-1); j >= 0; j--)
			iclose(id[j]);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (i = 0; (n = bread(id[0], i, arr[0])) > 0; i++) {
		for (j = 1; j < nimg; j++)
			bread(id[j], i, arr[j]);
		switch (type[0]) {	/* process images */
			case 1:
				medi_uchar(buf, (const UCHAR **)arr,
				(const float *)weight, nimg, n, limit);
				break;
			case 2:
				medi_short(buf, (const short **)arr,
				(const float *)weight, nimg, n, limit, Blank[0]);
				break;
			case 3:
				medi_float(buf, (const float **)arr,
				(const float *)weight, nimg, n, limit, Blank[0]);
				break;
			case 4:
				medi_long(buf, (const LONG **)arr,
				(const float *)weight, nimg, n, limit, Blank[0]);
				break;
		}
		bwrite(id0, i, buf);
	}

	free(ptr);
	free(buf);
	iclose(id0);
	for (j = (nimg-1); j >= 0; j--)
		iclose(id[j]);
	free(image);

	setiname(image0);	/* set IMAGE = Image Name */
	sethistory(image0, "median");
	sethistory(image0, image0);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "median: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
