
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
#include <math.h>
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declaration for daa.c */
#include "nr.h"	/* function declarations for nr.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define ORDER	10	/* maximum order of harmonic fit */
#define MD	((ORDER+1)*2)	/* maximum number of parameters */
/*----------------------------------------------------------------------
	'CosSin' calculates (cos(i*x), sin(i*x)), i = 2,...,n, from
	given (cos(x), sin(x)).
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void CosSin(double *co, double *si, int n)
#else	/* old style */
void CosSin(co, si, n)
double co[], si[];
int n;
#endif
{
	int i;
	for (i = 1; i < n; ++i) {
		co[i] = co[0]*co[i-1] - si[0]*si[i-1];
		si[i] = si[0]*co[i-1] + co[0]*si[i-1];
	}
}
/*----------------------------------------------------------------------
	'solve' calculates the solution of linear system, A*x = b
	Do not use it if 'decomp' has detected singularity
	Input:
      n      Order of matrix.
      *A[]   Triangular matrix A obtained from 'decomp'.
      b[]    Right hand side vector, b.
      ipvt[] Pivot vector obtained from 'decomp'.
	Output:
      b[]    Solution vector, x.
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 55.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void solve(int n, double * const A[], double b[], const int ipvt[])
#else	/* old style */
void solve(n, A, b, ipvt)
int n;
double * const A[];
double b[];
const int ipvt[];
#endif
{
	int i, k, m;
	double t;

	if (n > 1) {

		/* forward elimination */
		for (k = 0; k < (n-1); ++k) {
			m = ipvt[k];
			t = b[m];
			b[m] = b[k];
			b[k] = t;
			for (i = (k+1); i < n; ++i)
				b[i] += A[k][i]*t;
		}

		/* back substitution */
		for (k = (n-1); k > 0; --k) {
			b[k] /= A[k][k];
			t = -b[k];
			for (i = 0; i < k; ++i)
				b[i] += A[k][i]*t;
		}
	}
	b[0] /= A[0][0];
}
/*----------------------------------------------------------------------
	'decomp' decomposes a square matrix A by Gaussian elimination and
	estimates the condition of the matrix.
	Use 'solve' to compute solutions to linear systems.

	Input:
      n      Order of square matrix.
      *A[]   The matrix A to be triangularized.

	Output:
		A contains an upper triangular matrix U and a permuted
		version of a lower triangular matrix I-L so that
		(Permutation Matrix)*A = L*U.

	Return:
		An estimate of the 'condition' of A.

		For the linear system A*x = b, changes in A and b MAY
		cause changes 'condition' times as large in x.
		If (condition+1.0 == condition), A is singular to working
		precision. condition is set to 1.0e+32 if exact singularity
		is detected.

      ipvt[]    The pivot vector.
      ipvt[k]   The index of the k-th pivot row.
      ipvt[n-1] (-1)**(Number of Interchanges)

	Work Space:
		The wector work[] must be declared and included in the
		call. Its input contents are ignored. Its output
		contents are usually unimportant.

	The determinant of A can be obtained on output by
		Det(A) = ipvt[n-1]*A[0][0]*A[1][1]*......*A[n-1][n-1]
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 52.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double decomp(int n, double *A[], int ipvt[], double work[])
#else	/* old style */
double decomp(n, A, ipvt, work)
int n, ipvt[];
double *A[], work[];
#endif
{
	int i, j, k, m;
	double ek, t, anorm, ynorm, znorm, cond;

	ipvt[n-1] = 1;

	/* 1-by-1 */
	if (n == 1) {
		if (A[0][0] != 0.0)
			cond = 1.0;
		else
			cond = 1.0e+32;
		return cond;
	}

	/* compute 1-norm of A */
	anorm = 0.0;
	for (j = 0; j < n; ++j) {
		t = 0.0;
		for (i = 0; i < n; ++i)
			t += fabs(A[j][i]);
		if (t > anorm)
			anorm = t;
	}

	/* Gaussian elimination with partial pivoting */
	for (k = 0; k < (n-1); ++k) {
		/* find pivot */
		m = k;
		for (i = (k+1); i < n; ++i)
			if (fabs(A[k][i]) > fabs(A[k][m]))
				m = i;
		ipvt[k] = m;
		if(m != k)
			ipvt[n-1] = -ipvt[n-1];
		t = A[k][m];
		A[k][m] = A[k][k];
		A[k][k] = t;

		/* skip step if pivot is zero */
		if (t != 0.0) {

			/* compute multipliers */
			for (i = (k+1); i < n; ++i)
				A[k][i] /= -t;

			/* interchange and eliminate by columns */
			for (j = (k+1); j < n; ++j) {
				t = A[j][m];
				A[j][m] = A[j][k];
				A[j][k] = t;
				if (t != 0.0)
					for (i = (k+1); i < n; ++i)
						A[j][i] += A[k][i]*t;
			}
		}
	}
	/*
	cond = (1-norm of A)*(an estimate of 1-norm of A-inverse)
	Estimate obtained by one step of inverse iteration for the
	small singular vector. This involves solving two systems
	of equations, (A-transpose)*Y = E and A*Z = Y where E is
	a vector of +1 or -1 chosen to cause growth in Y.
	Estimate = (1-norm of Z)/(1-norm of Y)
	*/

	/* solve (A-transpose)*Y = E */
	for (k = 0; k < n; ++k) {
		t = 0.0;
		if (k > 0)
			for (i = 0; i < k; ++i)
				t += A[k][i]*work[i];
		ek = 1.0;
		if (t < 0.0)
			ek = -1.0;
		if (A[k][k] == 0.0) {
			cond = 1.0e+32;
			return cond;
		}
		work[k] = -(ek + t)/A[k][k];
	}
	for (k = (n-2); k >= 0; --k) {
		t = 0.0;
		for (i = (k+1); i < n; ++i)
			t += A[k][i]*work[k];
		work[k] = t;
		m = ipvt[k];
		if (m != k) {
			t = work[m];
			work[m] = work[k];
			work[k] = t;
		}
	}
	ynorm = 0.0;
	for (i = 0; i < n; ++i)
		ynorm += fabs(work[i]);

	/* solve A*Z = Y */
	solve(n, A, work, ipvt);

	znorm = 0.0;
	for (i = 0; i < n; ++i)
		znorm += fabs(work[i]);

	/* estimate condition */
	cond = anorm*znorm/ynorm;
	if (cond < 1.0)
		cond = 1.0;
	return cond;
}
/*----------------------------------------------------------------------
	Finds the Median, or any other Quantile.
	The median of n float items a[i] is defined as the item which is
	less than (or equal to) half of the n items and which is larger
	than (or equal to) the other half of the n items. The method to
	be displayed easily generalizes to the problem of finding the
	kth smallest of n items. Finding the median represents the
	special case k = n/2.
	The algorithm invented by C. A. R. Hoare functions as
	follows. First, the partitioning operation of Quicksort (p. 77)
	is applied with l = 0 and r = (n-1), and with a[k] selected as
	splitting value x. The resulting index values i and j are such that
		1. a[h] <= x for all h < i
		2. a[h] >= x for all h > j
		3. i > j
	There are three possible cases that may arise:
		1. The splitting value x was too small; as a result,
		   the limit between the two partitions is below the
		   desired value k. The partitioning process has to be
		   repeated upon the elements a[i]...a[r].
		2. The chosen bound x was too large. The splitting
		   operation has to be repeated on the partition
		   a[l]...a[j].
		3. j < k < i: the element a[k] splits the array into two
		   partitions in the specified proportions and
		   therefore is the desired quantile.
	The splitting process has to be repeated until case 3. arises.
------------------------------------------------------------------------
	Niklaus Wirth:	Algorithms + Data Structure = Programs, p. 82.
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void find(int n, float a[], int k)
#else	/* old style */
void find(n, a, k)
int n, k;
float a[];
#endif
{
	int l, r, i, j;
	float w, x;

	l = 0;
	r = n - 1;

	while (l < r) {
		x = a[k];
		i = l;
		j = r;
		/* begin split */
		while (i <= j) {
			while (a[i] < x)
				++i;
			while (x < a[j])
				--j;
			if (i <= j) {
				w = a[i];
				a[i] = a[j];
				a[j] = w;
				++i;
				--j;
			}
		}
		/* end split */
		if (j < k)
			l = i;
		if (k < i)
			r = j;
	}
}
/*----------------------------------------------------------------------
	Define the functions rho(t), wgt(t), and phi(t) for various cases.
		wgt(t) = rho'(t)/t; phi(t) = rho''(t).
------------------------------------------------------------------------
	Gaussian Distribution of r[i]/scale: Ordinary Least Squares
		rho0(t) = 0.5*t^2
		wgt0(t) = 1
		phi0(t) = 1
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho0(double t)
#else	/* old style */
double rho0(t)
double t;
#endif
{
	return (0.5*t*t);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double one(double t)
#else	/* old style */
double one(t)
double t;
#endif
{
	return (1.0);
}
/*----------------------------------------------------------------------
	Hyperbolic Distribution of r[i]/scale:
		rho1(t) = sqrt(1 + t^2) - 1 ~ 0.5*t^2
		wgt1(t) = 1/sqrt(1 + t^2)
		phi1(t) = 1/sqrt(1 + t^2)^3
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho1(double t)
#else	/* old style */
double rho1(t)
double t;
#endif
{
	return (sqrt(t*t + 1.0) - 1.0);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt1(double t)
#else	/* old style */
double wgt1(t)
double t;
#endif
{
	return (1.0/sqrt(t*t + 1.0));
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double phi1(double t)
#else	/* old style */
double phi1(t)
double t;
#endif
{	double z = t*t + 1.0;
	return (1.0/(sqrt(z)*z));
}
/*----------------------------------------------------------------------
	Cauchy or Lorentzian Distribution of r[i]/scale:
		rho2(t) = log(1 + 0.5*t^2) ~ 0.5*t^2
		wgt2(t) = 1/(1 + 0.5*t^2)
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho2(double t)
#else	/* old style */
double rho2(t)
double t;
#endif
{
	return (log(0.5*t*t + 1.0));
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt2(double t)
#else	/* old style */
double wgt2(t)
double t;
#endif
{
	return (1.0/(0.5*t*t + 1.0));
}
/*----------------------------------------------------------------------
	Hoaglin, Mosteller and Tukey: Table 8-2, page 293.
	Tukey's biweight estimator (with B = 3.6):
		rho3(t) = B^2*(1 - (1 - (t/B)^2)^3)/6 ~ 0.5*t^2,  |t| < B
		wgt3(t) = (1 - (t/B)^2)^2,  |t| < B
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 15-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho3(double t)
#else	/* old style */
double rho3(t)
double t;
#endif
{
	double B = 3.6;
	double x = B*B/6.0;
	if (fabs(t) < B) {
		t /= B;
		t = (1.0 - t)*(1.0 + t);
		x *= (1.0 - t*t*t);
	}
	return x;
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt3(double t)
#else	/* old style */
double wgt3(t)
double t;
#endif
{
	double B = 3.6;
	double x = 0.0;
	if (fabs(t) < B) {
		t /= B;
		x = (1.0 - t)*(1.0 + t);
		x *= x;
	}
	return x;
}
/*----------------------------------------------------------------------
	The functions rho0(t), rho1(t), rho2(t), and rho3(t) are assigned
	to an array rho[] of pointers to function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*rho[])(double) = { rho0, rho1, rho2, rho3 };
#else	/* old style */
double (*rho[])()       = { rho0, rho1, rho2, rho3 };
#endif
/*----------------------------------------------------------------------
	The functions one(t), wgt1(t), wgt2(t), and wgt3(t) are assigned
	to an array wgt[] of pointers to function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*wgt[])(double) = { one, wgt1, wgt2, wgt3 };
#else	/* old style */
double (*wgt[])()       = { one, wgt1, wgt2, wgt3 };
#endif
/*----------------------------------------------------------------------
	The functions one(t), phi1(t), wgt2(t), and wgt3(t) are assigned
	to an array phi[] of pointers to function.
	The phi1(t) is an M-estimator, wgt2(t) and wgt3(t) are W-estimators.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*phi[])(double) = { one, phi1, wgt2, wgt3 };
#else	/* old style */
double (*phi[])()       = { one, phi1, wgt2, wgt3 };
#endif
/*----------------------------------------------------------------------
	'normal' calculates a set of general 'Normal Equations', Ap = b,
	which are used to solve an over-determined system of linear equations,
	y = Xp + r, by stepwise minimizing the general merit function
	chi(scale; p) = Sum of rho(r[i]/scale).
	Exploring Data Tables, Trends, and Shapes, Chp 8.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      r[]   One dimensional float array containing the n
            residuals calculated from the current regression
            coefficients.
      scale Scale estimate, which is kept fixed during the
            optimization procedure.
      wgt() wgt(z) = psi(z)/z = rho'(z)/z.
      phi() phi(z) = psi'(z) = rho''(z).
	Output:
      *A[]  Two dimensional double array containing the A
            matrix in the first m rows and m columns.
      b[]   One dimensional double array of size m containing
            the right hand side of the 'Normal Equations'.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void normal(int n, int m, float * const X[], const float r[],
	double scale, double (*wgt)(double), double (*phi)(double),
	double *A[], double b[])
#else	/* old style */
void normal(n, m, X, r, scale, wgt, phi, A, b)
int n, m;
float * const X[];
const float r[];
double scale, (*wgt)(), (*phi)(), *A[], b[];
#endif
{
	int i, j, k;
	double s, t, u, v;

	/* initialize b and lower triangular part of A */
	for (j = 0; j < m; ++j) {
		b[j] = 0.0;
		for (k = 0; k <= j; ++k)
			A[j][k] = 0.0;
	}

	/* calculate b and lower triangular part of A */
	for (i = 0; i < n; ++i) {
		t = r[i]/scale;
		u = phi(t);
		v = wgt(t)*r[i];
		for (j = 0; j < m; ++j) {
			b[j] += v*X[j][i];
			s = u*X[j][i];
			for (k = 0; k <= j; ++k)
				A[j][k] += s*X[k][i];
		}
	}

	/* assign upper triangular part of A */
	for (j = 1; j < m; ++j)
		for (k = 0; k < j; ++k)
				A[k][j] = A[j][k];

}
/*----------------------------------------------------------------------
	'chi' calculates the residuals r[i], and the merit function:
	chi(scale; p) = Sum of rho(r[i]/scale).
	Exploring Data Tables, Trends, and Shapes, Chp 8.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      y[]   One dimensional float array containing the n
            observations.
      scale Scale estimate, which is kept fixed during the
            optimization procedure.
      rho() chi(scale; p) = Sum of rho(r[i]/scale).
      p[]   One dimensional double array containing the m
            regression coefficients.
	Output:
      r[]   One dimensional float array containing the n
            residuals.
	Return:
            The merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double chi(int n, int m, float * const X[], const float y[],
	double scale, double (*rho)(double), const double p[], float r[])
#else	/* old style */
double chi(n, m, X, y, scale, rho, p, r)
int n, m;
float * const X[];
const float y[];
double scale, (*rho)();
const double p[];
float r[];
#endif
{
	int i, j;
	double dif, sum = 0.0;

	for (i = 0; i < n; ++i) {
		dif = y[i];
		for (j = 0; j < m; ++j)
			dif -= X[j][i]*p[j];
		r[i] = (float) dif;
		sum += rho(dif/scale);
	}
	return sum;
}
/*----------------------------------------------------------------------
	'ols', solves the Normal Equations, Ap = b, of the linear
	regression model, y = Xp + r.
	Exploring Data Tables, Trends, and Shapes, Chp 8, p283.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      y[]   One dimensional float array containing the n
            observations.
	Output:
      p[]   One dimensional double array of size m containing
            the OLS estimate of the regression coefficients.
      r[]   One dimensional float array containing the n
            residuals.
	Returns:
            The current value of the merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double ols(int n, int m, float * const X[], const float y[],
	double p[], float r[])
#else	/* old style */
double ols(n, m, X, y, p, r)
int n, m;
float * const X[];
const float y[];
double p[];
float r[];
#endif
{
	int ipvt[MD], k;
	double a[MD][MD], *A[MD], work[MD], cond;

	if (m > MD) {
		fprintf(stderr, "ols: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* assign pointer array */
	for (k = 0; k < MD; ++k)
		A[k] = a[k];

	/* calculate the normal equations */
	normal(n, m, X, y, 1.0, wgt[0], phi[0], A, p);

	/* decompose the A matrix */
	cond = decomp(m, A, ipvt, work);

	/* Check if A is singular to working precision */
	if ((cond + 1.0) == cond) {
		fprintf(stderr, "ols: A is singular to working precision\n");
		exit(EXIT_FAILURE);
	}

	/* solve the normal equations */
	solve(m, A, p, ipvt);

	/* calculate the residuals and return the merit function */
	return chi(n, m, X, y, 1.0, rho[0], p, r);
}
/*----------------------------------------------------------------------
	'mad' calculates the Median Absolute Deviation (MAD) of the
	residuals r[i].
	Exploring Data Tables, Trends, and Shapes, Chp 8, p302.
	Input:
      n     Number of observations.
      r[]   One dimensional float array containing the n
            current residuals.
	Return:
            The Median Absolute Deviation of r[i].
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 29-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double mad(int n, const float r[])
#else	/* old style */
double mad(n, r)
int n;
const float r[];
#endif
{
	int i, k1, k2;
	float *work;
	double rmad, rmedi;

	/* allocate the work[] vector */
	if ((work = (float *)calloc(n, sizeof(float))) == (float *)0) {
		fprintf(stderr, "mad: Can't allocate work[%d]\n", n);
		exit(EXIT_FAILURE);
	}

	/* make a working copy of the residuals */
	for (i = 0; i < n; ++i)
		work[i] = r[i];

	/* find the median */
	k1 = (n - 1)/2;
	k2 = n/2;
	find(n, work, k1);
	rmedi = work[k1];
	if (k2 != k1) {
		find(n, work, k2);
		rmedi = (rmedi + work[k2])/2.0;
	}

	/* calculate the absolute deviations */
	for (i = 0; i < n; ++i)
		work[i] = (float) fabs(work[i] - rmedi);

	/* find the Median of the Absolute Deviations */
	find(n, work, k1);
	rmad = work[k1];
	if(k2 != k1) {
		find(n, work, k2);
		rmad = (rmad + work[k2])/2.0;
	}

	/* free the work[] array */
	free(work);

	/* The MAD of a Normal Distribution is 0.6745 */
	/* times the usual Standard Deviation */

	return rmad;
}
/*----------------------------------------------------------------------
	'chistep' uses the Levenberg-Marquardt method to reduce the merit
	function 'chi' of the linear regression model, y = Xp + r, by
	using a robust estimator, possibly derived by maximum-likelihood
	arguments.
	Exploring Data Tables, Trends, and Shapes, Chp 8, p305.
	Numerical Recipes in C, Chp. 14.4, p540.
	Input:
      n      Number of observations.
      m      Number of regression coefficients.
      *X[]   Two dimensional float array containing the
             design matrix in the first m rows and n columns.
      y[]    One dimensional float array containing the n
             observations.
      scale  Scale estimate, which is kept fixed during the
             optimization procedure.
      rho()  chi(scale; p) = Sum of rho(r[i]/scale).
      wgt()  wgt(z) = psi(z)/z = rho'(z)/z.
      phi()  phi(z) = psi'(z) = rho''(z).
      p[]    One dimensional double array containing the m
             regression coefficients obtained from the
             previous iteration step.
      *lamda Marquardt's parameter.
	Output:
      p[]    One dimensional double array containing the m
             updated regression coefficients.
      r[]    One dimensional float array containing the n
             updated residuals.
      *lamda Marquardt's lamda parameter.
	Returns:
             The updated value of the merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 29-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double chistep(int n, int m, float * const X[], const float y[],
	double scale, double (*rho)(double), double (*wgt)(double),
	double (*phi)(double), double p[], float r[], double *lamda)
#else	/* old style */
double chistep(n, m, X, y, scale, rho, wgt, phi, p, r, lamda)
int n, m;
float * const X[];
const float y[];
float r[];
double scale, (*rho)(), (*wgt)(), (*phi)(), p[], *lamda;
#endif
{
	int ipvt[MD], j, k;
	double a[MD][MD], q[MD][MD];
	double *A[MD], *Q[MD], b[MD], dp[MD], work[MD];
	double oldchi, newchi, cond;

	if (m > MD) {
		fprintf(stderr, "chistep: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* assign pointer arrays */
	for (k = 0; k < MD; ++k) {
		A[k] = a[k];
		Q[k] = q[k];
	}

	/* calculate the residuals and the merit function */
	oldchi = chi(n, m, X, y, scale, rho, p, r);

	/* calculate the general 'Normal Equations' */
	normal(n, m, X, r, scale, wgt, phi, A, b);

	/* increase lamda until the merit function declines */
	do {

		/* make new working copies of A and b */
		for (j = 0; j < m; ++j) {
			for (k = 0; k < m; ++k)
				Q[j][k] = A[j][k];
			dp[j] = b[j];
		}

		/* increase the diagonal part of Q by Marquardt's prescription */
		if (lamda[0] > 0.0)
			for (k = 0; k < m; ++k)
				Q[k][k] *= (1.0 + lamda[0]);

		/* decompose the Q matrix */
		cond = decomp(m, Q, ipvt, work);

		/* Check if Q is singular to working precision */
		if ((cond + 1.0) == cond) {
			fprintf(stderr, "chistep: Q is singular to working precision\n");
			exit(EXIT_FAILURE);
		}

		/* solve for the increments to the regression coefficients */
		solve(m, Q, dp, ipvt);

		/* make a working copy of the updated coefficients */
		for (k = 0; k < m; ++k)
			work[k] = p[k] + dp[k];

		/* calculate the new residuals, and the new merit function */
		newchi = chi(n, m, X, y, scale, rho, work, r);

		/* update Marquardt's parameter */
		if (lamda[0] > 0.0) {
			if (newchi < oldchi)
				lamda[0] *= 0.1;
			else
				lamda[0] *= 10.0;
		}
	} while (lamda[0] > 0.0 && newchi >= oldchi);

	/* update the regression coefficient */
	for (k = 0; k < m; ++k)
		p[k] += dp[k];

	return newchi;
}
/*----------------------------------------------------------------------
	'robust' does a robust regression to a linear model, y = Xp + r.
	The methods used are based on:
		Chapter 8, Robust Regression, by Guoying Li in
		Exploring Data Tables, Trends, and Shapes, edited by
		D. C. Hoaglin, F. Mosteller, and J. W. Tukey
		John Wiley & Sons
------------------------------------------------------------------------
	The following steps are done					
		0. An ordinary least squares solution (OLS)
		1. An iterative M1-estimate with
			psi(t) = t*w(t), where w(t) = 1/sqrt(1 + t^2)
		2. An iterative W2-estimate with
			psi(t) = t*w(t), where w(t) = 1/(1 + 0.5*t^2)
		3. An iterative W3-estimate with
			psi(t) = t*w(t), where w(t) = (1 - (t/B)^2)^2,  |t| < B
------------------------------------------------------------------------
	Input:
      method 0: Least Squares, 1: M1-estimate,
             2: W2-estimate, and 3: W3-estimate.
      n      Number of observations.
      m      Number of regression coefficients.
      *X[]   Two dimensional float array containing the
             design matrix in the first m rows and n columns.
      y[]    One dimensional float array containing the n
             observations.
	Output:
      par[]  One dimensional double array containing the m
             regression coefficients.
      r[]    One dimensional float array containing the n
             final residuals.
	Return:
             The Median Absolute Deviation (MAD) divided by
             the MAD of a Standard Normal Distribution (StMAD).
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 15-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double robust(int method, int n, int m, float * const X[],
	const float y[], double par[], float r[])
#else	/* old style */
double robust(method, n, m, X, y, par, r)
int method, n, m;
float * const X[];
const float y[];
double par[];
float r[];
#endif
{
	int i, k, nit[3];
	double oldchi, newchi, lamda, scale;
	const double StMAD = 0.6745;

	/* iteration limits */
	nit[0] = 20;
	nit[1] = 20;
	nit[2] = 20;

	if (m > MD) {
		fprintf(stderr, "robust: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* begin with an Ordinary Least Squares solution */
	newchi = ols(n, m, X, y, par, r);

	/* find scale factor from MAD of the OLS-residuals */
	scale = mad(n, r)/StMAD;

	/* first obtain M1-estimate, then obtain W2-estimate */
	/* finally obtain W3-estimate */
	/* by using Marquardt's method, and a constant scale factor */
	for (k = 1; k <= method; ++k) {	/* step through methods */

		/* initialize Marquardt's parameter */
		lamda = 0.001;

		/* do the first step */
		newchi = chistep(n, m, X, y, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);

		/* do at most 'nit' steps */
		for (i = 0; i < nit[k-1]; ++i) {
			oldchi = newchi;
			newchi = chistep(n, m, X, y, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);
			if (fabs(oldchi - newchi) < 0.1)
				break;
		}

		/* no solution found after 'nit' steps */
		/* but continue anyway */
		if (i == nit[k-1]) {
			fprintf(stderr,
				"robust: No solution after %d %d-steps\n", nit[k-1], k);
		}

		/* do the final step */
		lamda = 0.0;
		newchi = chistep(n, m, X, y, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);

		/* find scale factor from MAD of the residuals */
		scale = mad(n, r)/StMAD;
	}

	return scale;
}
/*----------------------------------------------------------------------
	'harmonic' fits a Fourier series to pixel data between two ellipses
	centered at (xc, yc), and with effective radii [rmin, rmax], where
	the radius is defined as r = sqrt(b*a).
	The coefficients are found by fitting the series:
	pixel(r, v) = a0 + b0*(r - r0) + a1*cos(v) + b1*sin(v) + ....
------------------------------------------------------------------------
	Input:
      grad     Gradient term is included (grad != 0).
      olsq     Ordinary Least Squares, only (olsq != 0).
      ncols    Number of columns in image.
      nrows    Number of rows in image.
      **pixel  Two dimensional pixel array (float).
      xc, yc   Position of the origin of the polar coordinates.
      theta    Position angle of the major axis of the ellipses.
      q        Ratio of the minor to the major axis.
      rmin     Equivalent radius of inner ellipse.
      rmax     Equivalent radius of outer ellipse.
      order    Order of the harmonic series.
	Output:
      *r0      Mean de Vaucouleurs radius of the pixels.
      ac[0]    Mean of data values between the ellipses.
		bc[0]    Radial gradient, if grad != 0.
      ac[i]    Cos(i*theta) coefficient, for i = 1, 2,..., order
      bc[i]    Sin(i*theta) coefficient, for i = 1, 2,..., order
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 28-Feb-92, 10-Jul-93
----------------------------------------------------------------------*/
#define	OLS	0	/* Ordinary Least Squares */
#define	M1ES	1	/* M1-estimate */
#define	W2ES	2	/* W2-estimate */
#define	W3ES	3	/* W3-estimate */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int harmonic(
int grad,
int olsq,
int nrows,
int ncols,
float * const *pixel,
double xc,
double yc,
double theta,
double q,
double rmin,
double rmax,
int order,
double *r0,
double *ac,
double *bc)
#else	/* old style */
int harmonic(grad, olsq, nrows, ncols, pixel,
	xc, yc, theta, q, rmin, rmax, order, r0, ac, bc)
int grad, olsq, nrows, ncols, order;
float * const *pixel;
double xc, yc, theta, q, rmin, rmax, *r0, ac[], bc[];
#endif
{
	int xmin, xmax, ymin, ymax, j, k, l, n, nd;
	int m = 2*order + 1;
	double par[MD];
	double cos_theta, sin_theta, sqrt_q, co[ORDER], si[ORDER];
	double lrad, deVrad, xxmax, yymax, rmin2, rmax2, area;
	float **X, *y, *r;
	char *ptr;
	unsigned int err;
	int start[2], dim[2];
	float Init_Value = (float) 0;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const double RM1 = 7.5, RW2 = 21.5, RW3 = 63.5;

	if (grad != 0)	/* include a gradient parameter */
		m++;

	if (m > MD) {
		fprintf(stderr, "harmonic: Too many parameters: %d\n", m);
		return 0;
	}

	/* initialize the parameter array */
	for (k = 0; k < MD; ++k) {
		par[k] = 0.0;
	}

	/* take care of projection effect */
	cos_theta = cos(theta);
	sin_theta = sin(theta);
	sqrt_q = sqrt(q);

	xxmax = sqrt(cos_theta*cos_theta/q + sin_theta*sin_theta*q)*rmax;
	yymax = sqrt(sin_theta*sin_theta/q + cos_theta*cos_theta*q)*rmax;

	xmin = MAX((int)(xc - xxmax), 0);
	ymin = MAX((int)(yc - yymax), 0);

	xmax = MIN((int)(xc + xxmax) + 1, ncols - 1);
	ymax = MIN((int)(yc + yymax) + 1, nrows - 1);

	rmin2 = rmin * rmin;
	rmax2 = rmax * rmax;

	area = 4.0*(rmax2 - rmin2)*atan(1.0);	/* area of annulus in pixels */

	nd = (int)(1.05*area) + 50;

	/* allocate r[nd] vector */
	if ((r = (float *)calloc(nd, sizeof(float))) == (float *)0) {
		fprintf(stderr, "harmonic: Can't allocate r[%d]\n", nd);
		return 0;
	}

	/* allocate y[nd] vector */
	if ((y = (float *)calloc(nd, sizeof(float))) == (float *)0) {
		fprintf(stderr, "harmonic: Can't allocate y[%d]\n", nd);
		free(r);
		return 0;
	}

	dim[0] = m;
	dim[1] = nd;
	start[0] = 0;	/* first index */
	start[1] = 0;	/* first index */

	/* allocate X[m][nd] matrix */
	if ((X = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "harmonic: daa: Error code = %d\n", err);
		fprintf(stderr, "harmonic: Can't allocate X[%d][%d]\n", m, nd);
		free(y);
		free(r);
		return 0;
	}

/*
	Calculate the mean de Vaucouleurs radius, *r0,
	for pixels between the given ellipses.
*/
	deVrad = 0.0;
	n = 0;
	for (j = ymin; j <= ymax; ++j) {
		double dy_cos = (j - yc)*cos_theta;
		double dy_sin = (j - yc)*sin_theta;
		double dx, dy, rad2;
		int i;
		for (i = xmin; i <= xmax; ++i) {
			dx = ((i - xc)*cos_theta + dy_sin)*sqrt_q;
			dy = (dy_cos - (i - xc)*sin_theta)/sqrt_q;
			rad2 = dx*dx + dy*dy;
			if (rad2 > rmin2 && rad2 <= rmax2)
				if (pixel[j][i] <= FltMax) {
					deVrad += sqrt(sqrt(sqrt(rad2)));
					n++;
				}
		}
	}

	deVrad /= n;	/* the mean de Vaucouleurs radius */

	*r0 = deVrad;

	lrad = deVrad;
	lrad *= lrad;
	lrad *= lrad;	/* linear radius */

	if (n < m || n > nd) {
		fprintf(stderr, "harmonic: Too few/many equations: %d\n", n);
		free(ptr);
		free(y);
		free(r);
		return 0;
	}

	/* Calculate the design matrix */
	n = 0;
	for (j = ymin; j <= ymax; ++j) {
		double dy_cos = (j - yc)*cos_theta;
		double dy_sin = (j - yc)*sin_theta;
		double dx, dy, rad2;
		int i;
		for (i = xmin; i <= xmax; ++i) {
			dx = ((i - xc)*cos_theta + dy_sin)*sqrt_q;
			dy = (dy_cos - (i - xc)*sin_theta)/sqrt_q;
			rad2 = dx*dx + dy*dy;
			if (rad2 > rmin2 && rad2 <= rmax2)
				if (pixel[j][i] <= FltMax) {
					double rad = sqrt(rad2);
					int l, k = 0;

					/* fill data vector */
					y[n] = pixel[j][i];

					/* calculate cosines and sines */
					if (order > 0) {
						co[0] = dx / rad;
						si[0] = dy / rad;
						CosSin(co, si, order);
					}

					/* fill design matrix */
					X[k++][n]  = (float) 1.0;

					/* include gradient term */
					if (grad != 0)
						X[k++][n]  = (float) (sqrt(sqrt(rad)) - deVrad);

					for (l = 0; l < order; ++l) {
						X[k++][n] = (float)co[l];
						X[k++][n] = (float)si[l];
					}
					n++;
				}
		}
	}

	/* Solve equations */
	if (olsq != 0)	/* Ordinary Least Squares only */
		robust(OLS, n, m, X, y, par, r);
	else if (lrad < RM1)	/* Ordinary Least Squares only */
		robust(OLS, n, m, X, y, par, r);
	else if (lrad < RW2)	/* M1-estimate */
		robust(M1ES, n, m, X, y, par, r);
	else if (lrad < RW3)	/* W2-estimate */
		robust(W2ES, n, m, X, y, par, r);
	else	/* W3-estimate */
		robust(W3ES, n, m, X, y, par, r);

	/* Extract the Fourier coefficients ac[], bc[] */
	k = 0;
	ac[0] = par[k++];

	/* gradient parameter, if included */
	if (grad != 0)
		bc[0] = par[k++];

	/* cosine and sine coefficients */
	for (l = 1; l <= order; ++l) {
		ac[l] = par[k++];
		bc[l] = par[k++];
	}

	/* free arrays */
	free(ptr);
	free(y);
	free(r);

	return 1;
}
/*----------------------------------------------------------------------
   Logprof transform the mean radial profile to log(intensity) if the
   mean intensity profile is monotonic decreasing, if all points of
   profile are greater than 1. The value returned indicates if the
   output profile is on a log scale (1), or on a linear scale (0).
------------------------------------------------------------------------
	Input:
      npts     Number of points in intensity profile.
      linp[]   Array containing the linear intensities.
	Output:
      logp[]   Array containing log or (linear) intensities.
   Return:     1: log intensity; 0: linear intensity.
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 05-June-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int logprof(const double linp[], int npts, double logp[])
#else	/* old style */
int logprof(linp, npts, logp)
const double linp[];
int npts;
double logp[];
#endif
{
	int i;
	for (i = 1; i <= npts; ++i)
		logp[i] = linp[i];
	for (i = 1; i < npts; ++i)
		if (linp[i] <= linp[i+1])
			return 0;
	if (linp[npts] < 1.0)
		return 0;
	for (i = 1; i <= npts; ++i)
		logp[i] = log(linp[i]);
	return 1;
}
/*----------------------------------------------------------------------
	Smooth calculates a synthetic image for pixels between two ellipses
	centered at (xc, yc) from the fitted Fourier coefficients.
	The pixel values are calculated from the series:
	pixel(r, v) = a0(r) + a1(r)*cos(v) + b1(r)*sin(v) + .......
	where the radial tables r0[], an[i][], bn[i][]
	are interpolated by means of a Cubic Spline or Akima's method.
------------------------------------------------------------------------
	Input:
      method   == 0: Cubic Spline, != 0: Akima's method.
      ncols    Number of columns in image.
      nrows    Number of rows in image.
      **pixel  Two dimensional pixel array (float).
      xc, yc   Position of the origin of the polar coordinates.
      theta    Position angle of the major axis.
      q        Ratio of the minor to the major axis.
      rmin     Effective radius of the inner ellipse.
      rmax     Effective radius of the outer ellipse.
      order    Order of the harmonic series.
      npts     Number of points in the following radial tables.
      r0[]     Table containing de Vaucouleurs radii.
      an[0][]  Table containing the mean intensity values.
      an[i][]  Table containing Cos(i*theta) coefficients.
      bn[i][]  Table containing Sin(i*theta) coefficients.
	Output:
      **pixel  Two dimensional pixel array (float).
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 31-May-93
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int smooth(
int method,
int nrows,
int ncols,
float **pixel,
double xc,
double yc,
double theta,
double q,
double rmin,
double rmax,
int order,
int npts,
const double *r0,
double * const *an,
double * const *bn)
#else	/* old style */
int smooth(method, nrows, ncols, pixel, xc, yc, theta, q,
	rmin, rmax, order, npts, r0, an, bn)
int method, nrows, ncols, order, npts;
float **pixel;
double xc, yc, theta, q, rmin, rmax;
const double *r0;
double * const an[], * const bn[];
#endif
{
	char *err;
	int xmin, xmax, ymin, ymax, j, k, is_log;
	double xxmax, yymax, rmin2, rmax2;
	double cos_theta, sin_theta, sqrt_q, co[ORDER], si[ORDER];
	double **anpc, **bnpc = (double **)0;	/* pointers to 2 matrices */
	double *loga0;	/* pointer to log(intensity) profile */
	int size = npts;	/* cubic spline */

	if (method != 0) size = 3*npts - 1;	/* Akima's method */

	/* Allocate the Polynomial Coefficients of the Cosine Profiles */
	if ((anpc = dmatrix(0, order, 1, size)) == (double **)0) {
		fprintf(stderr,
			"smooth: Can't allocate anpc[%d][%d]\n", order, size);
		return 0;
	}

	/* Allocate the Polynomial Coefficients of the Sine Profiles */
	if (order > 0)
		if ((bnpc = dmatrix(1, order, 1, size)) == (double **)0) {
			fprintf(stderr,
				"smooth: Can't allocate bnpc[%d][%d]\n", order, size);
			free_dmatrix(anpc, 0, 1);
			return 0;
		}

	/* Allocate vector for log(intensity) profile */
	if ((loga0 = dvector(1, npts)) == (double *)0) {
		fprintf(stderr, "smooth: Can't allocate loga0[%d]\n", npts);
		if (order > 0) free_dmatrix(bnpc, 1, 1);
		free_dmatrix(anpc, 0, 1);
		return 0;
	}

	/* If possible, take the log of the mean intensity profile? */
	is_log = logprof(an[0], npts, loga0);

	/* call NAKSPL or AKIMA to obtain Polynomial Coefficients */
	err = (method != 0)	/* Not-A-Knode Spline or Akima's Method */
			?  akima(r0, loga0, npts, anpc[0])
			: nakspl(r0, loga0, npts, anpc[0]);
	if (err != (char *)0) {
		fprintf(stderr, "smooth: loga0[]: %s\n", err);
		free_dvector(loga0, 1);
		if (order > 0) free_dmatrix(bnpc, 1, 1);
		free_dmatrix(anpc, 0, 1);
		return 0;
	}
	for (k = 1; k <= order; ++k) {
		err = (method != 0)	/* Spline or Akima */
				?  akima(r0, an[k], npts, anpc[k])
				: nakspl(r0, an[k], npts, anpc[k]);
		if (err != (char *)0) {
			fprintf(stderr, "smooth: an[%d][]: %s\n", k, err);
			free_dvector(loga0, 1);
			if (order > 0) free_dmatrix(bnpc, 1, 1);
			free_dmatrix(anpc, 0, 1);
			return 0;
		}
		err = (method != 0)	/* Spline or Akima */
				?  akima(r0, bn[k], npts, bnpc[k])
				: nakspl(r0, bn[k], npts, bnpc[k]);
		if (err != (char *)0) {
			fprintf(stderr, "smooth: bn[%d][]: %s\n", k, err);
			free_dvector(loga0, 1);
			if (order > 0) free_dmatrix(bnpc, 1, 1);
			free_dmatrix(anpc, 0, 1);
			return 0;
		}
	}

	cos_theta = cos(theta);
	sin_theta = sin(theta);
	sqrt_q = sqrt(q);

	xxmax = sqrt(cos_theta*cos_theta/q + sin_theta*sin_theta*q)*rmax;
	yymax = sqrt(sin_theta*sin_theta/q + cos_theta*cos_theta*q)*rmax;

	xmin = MAX((int)(xc - xxmax), 0);
	ymin = MAX((int)(yc - yymax), 0);

	xmax = MIN((int)(xc + xxmax) + 1, ncols - 1);
	ymax = MIN((int)(yc + yymax) + 1, nrows - 1);

	rmin2 = rmin * rmin;
	rmax2 = rmax * rmax;

	for (j = ymin; j <= ymax; ++j) {
		double dy_cos = (j - yc)*cos_theta;
		double dy_sin = (j - yc)*sin_theta;
		double dx, dy, rad2;
		int i;
		for (i = xmin; i <= xmax; ++i) {
			dx = ((i - xc)*cos_theta + dy_sin)*sqrt_q;
			dy = (dy_cos - (i - xc)*sin_theta)/sqrt_q;
			rad2 = dx*dx + dy*dy;
			if (rad2 > rmin2 && rad2 <= rmax2) {
				double prof, anr, bnr;
				double rad = sqrt(rad2);
				double deVrad = sqrt(sqrt(rad));
				int k;

				prof = (method != 0)	/* Spline or Akima of log(intensity) */
						?  akint(r0, loga0, anpc[0], npts, deVrad)
						: splint(r0, loga0, anpc[0], npts, deVrad);

				if (is_log != 0) prof = exp(prof);

				/* calculate cosines and sines */
				if (order > 0) {
					co[0] = dx / rad;
					si[0] = dy / rad;
					CosSin(co, si, order);
				}

				for (k = 1; k <= order; ++k) {
					anr = (method != 0)	/* Spline or Akima */
							?  akint(r0, an[k], anpc[k], npts, deVrad)
							: splint(r0, an[k], anpc[k], npts, deVrad);
					bnr = (method != 0)	/* Spline or Akima */
							?  akint(r0, bn[k], bnpc[k], npts, deVrad)
							: splint(r0, bn[k], bnpc[k], npts, deVrad);
					prof += anr*co[k-1] + bnr*si[k-1];
				}
				pixel[j][i] = (float) prof;
			}
		}
	}

	free_dvector(loga0, 1);
	if (order > 0) free_dmatrix(bnpc, 1, 1);
	free_dmatrix(anpc, 0, 1);

	return 1;
}
/*----------------------------------------------------------------------
	This program calculates harmonic radial profiles along ellipses
	around a selected origin (centered on an extended object), and
	optionally produces a smooth image of the object, reconstructed
	from the radial profiles by a cubic spline interpolation.
	The methods used are based on:
		Chapter 8, Robust Regression, by Guoying Li in
		Exploring Data Tables, Trends, and Shapes, edited by
		D. C. Hoaglin, F. Mosteller, and J. W. Tukey
		John Wiley & Sons
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 31-May-93
------------------------------------------------------------------------
	Calculate radial harmonic profiles between rmin and rmax
	y = a0(r) + a1(r)*cos(theta) + b1(r)*sin(theta) + ..........
----------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv, **arg;
	int option, opt_err = 0, n_arg = 0;
	int method = 0;	/* cubic spline (default) or Akima's method */ 
	int olsq = 0;	/* robust regression is default */
	int grad = 1;	/* gradient term is default */
	int order = ORDER;	/* maximum order of harmonic fit */
	int npts = 15;	/* default number of radial points */
	double pa = 0.0;	/* position angle (in deg) of major axis */
	double eps = 0.0;	/* default ellipticity */
	char *file = (char *)0;	/* listing file */
	char *image1 = (char *)0;	/* output image */
	char *image;	/* input image */
	FILE *fp = (FILE *)0;	/* listing file pointer */
	void *row;	/* row buffer */
	int i, y, n, m;
	int id, type, ncols, nrows;
	double xc, yc, rmin, rmax, Blank, theta, q;
	double rinner, router, sqrtrad, sqrtmin, sqrtmax, sqrtstep;
	double ac[ORDER+1], bc[ORDER+1];	/* Fourier coefficients */
	double *r0, **an, **bn = (double **)0;	/* harmonic profiles */
	float **a;	/* pixel array a[][] allocated by daa */
	char *ptr;	/* base pointer of array space allocated by daa */
	unsigned int err;	/* error code returned by daa */
	int start[2], dim[2];	/* starting indices and dimensions for daa */
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Half = (float)0.5;
	const float Zero = (float)0;
	float Init_Value = FltBlank;	/* initial value for a[][] */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "profile: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	/* process command line options */
	while (((option = getopt(argc, argv, "asgp:e:n:f:o:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'a':	method++; break;	/* use Akima Interpolation */
			case 's':	olsq++; break;	/* Ordinary Least Squares only */
			case 'g':	grad = 0; break; /* no gradient term in fit */
			case 'p':	sscanf(optarg, "%lf", &pa); break;	/* position angle */
			case 'e':	sscanf(optarg, "%lf", &eps); break;	/* ellipticity */
			case 'n':	sscanf(optarg, "%d", &order); break;
			case 'f':	file   = optarg; break;	/* list file */
			case 'o':	image1 = optarg; break;	/* output image */
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	/* illegal options */
	if (opt_err || n_arg < 5) {
		printf(
"profile -a -s -g -p pa -e eps -n ord -f file -o im1 im Xc Yc Rmin Rmax [npt]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	theta = atan(1.0)*pa/45.0;	/* position angle in radians */

	if (eps < 0.0 || eps >= 1.0) {
		fprintf(stderr, "profile: Ellipticity = %f is illegal\n", eps);
		exit(EXIT_FAILURE);
	}
	q = 1.0 - eps;	/* ratio of the minor to the major axis */

	if (order > ORDER) {
		fprintf(stderr, "profile: Harmonic order = %d is too big\n", order);
		exit(EXIT_FAILURE);
	}

	if (order < 0) {
		fprintf(stderr, "profile: Harmonic order = %d is too small\n", order);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);
	/* is image name legal? */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "profile: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	if (image1 != (char *)0) {	/* is output image legal? */
		image1 = truncname(image1);
		if (strpbrk(image1, punct) != (char *)0) {
			fprintf(stderr,
				"profile: Illegal image name: %s\n", image1);
			exit(EXIT_FAILURE);
		}
	}

	sscanf(arg[1], "%lf", &xc);
	sscanf(arg[2], "%lf", &yc);
	sscanf(arg[3], "%lf", &rmin);
	sscanf(arg[4], "%lf", &rmax);
	if (n_arg > 5) sscanf(arg[5], "%d", &npts);

	if (rmax <= rmin) {
		fprintf(stderr, "profile: rmax <= rmin\n");
		exit(EXIT_FAILURE);
	}

	if (npts < 7) {
		fprintf(stderr, "profile: Number of radial points < 7\n");
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "profile: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open image for read only */
	if ((id = iopen(image, 0, &type, &ncols, &nrows)) < 0) {
		fprintf(stderr, "profile: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr,
			"profile: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	if (image1 != (char *)0)
		if (xist(image1)) {	/* does output image exist? */
			fprintf(stderr, "profile: %s: Existing image\n", image1);
			iclose(id);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"profile: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}

	dim[0] = nrows;	/* y-dimension */
	dim[1] = ncols;	/* x-dimension */
	start[0] = 0;	/* first index of y-dimension */
	start[1] = 0;	/* first index of x-dimension */

	/* allocate float array a[][] */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "profile: daa: Error code = %d\n", err);
		fprintf(stderr, "profile: Can't allocate a[%d][%d]\n", nrows, ncols);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate row buffer */
	if ((row = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "profile: Can't allocate row[%d]\n", ncols);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nrows; y++) {	/* read image into array a[][] */
		getline(id, y, row);	/* read next row */
		switch (type) {
			case 1:
				{
					const UCHAR *q = (UCHAR *)row;
					float *p = a[y];
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (float)q[x];
				}
				break;
			case 2:
				{
					const short *q = (short *)row;
					float *p = a[y];
					const short b = (short)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:
				{
					const float *q = (float *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:
				{
					const LONG *q = (LONG *)row;
					float *p = a[y];
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:
				{
					const COMPLEX *q = (COMPLEX *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x].x > h || q[x].y > h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);	/* close image file */

	/* use equal step size in the sqrt radius */
	sqrtmin = sqrt(rmin);
	sqrtmax = sqrt(rmax);
	sqrtstep = (sqrtmax - sqrtmin)/npts;

	/* allocate radius profile */
	if ((r0 = dvector(1, npts)) == (double *)0) {
		fprintf(stderr, "profile: Can't allocate r0[%d]\n", npts);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate Cosine profiles */
	if ((an = dmatrix(0, order, 1, npts)) == (double **)0) {
		fprintf(stderr,
			"profile: Can't allocate an[%d][%d]\n", order, npts);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate Sine profiles */
	if (order > 0)
		if ((bn = dmatrix(1, order, 1, npts)) == (double **)0) {
			fprintf(stderr,
				"profile: Can't allocate bn[%d][%d]\n", order, npts);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"profile: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}

	/* open listing file, if any */
	if (file != (char *)0)
		fp = fopen(file, "w");

	printf("\n");
	printf(
	"  r0[]    a0[]   PA    a1[]   b1[]   a2[]   b2[]  a3[]  b3[]  a4[]  b4[]\n");

	/* index starts at n = 1 */
	for (sqrtrad = sqrtmin, n = 1; n <= npts; sqrtrad += sqrtstep, n++) {
		rinner = sqrtrad;	/* define inner sqrt radius */
		router = sqrtrad + sqrtstep;	/* define outer sqrt radius */

		rinner *= rinner;	/* return to linear scale */
		router *= router;

		/* determine Fourier coefficients for the next annulus */
		if (harmonic(grad, olsq, nrows, ncols, a, xc, yc, theta, q,
				rinner, router, order, &r0[n], ac, bc) == 0) {
			fprintf(stderr, "profile: Error while fitting ring No. %d\n", n);
			goto Finish;
		}

		/* save the Fourier coefficients */
		an[0][n] = ac[0];
		for (i = 1; i <= order; ++i) {
			an[i][n] = ac[i];
			bn[i][n] = bc[i];
		}

		/* list the harmonic profiles on the terminal */
		printf("%6.4f%8.2f", r0[n], an[0][n]);
		if (order >= 2) {
			double pa2 = 22.5*atan2(bn[2][n], an[2][n])/atan(1.0);
			printf("%6.1f", pa2);
		} else
			printf("      ");
		m = MIN(2, order);
		for (i = 1; i <= m; ++i)
			printf("%7.1f%7.1f", an[i][n], bn[i][n]);
		m = MIN(4, order);
		for (i = 3; i <= m; ++i)
			printf("%6.1f%6.1f", an[i][n], bn[i][n]);
		printf("\n");

		/* write the harmonic profiles to the listing file */
		if (fp != (FILE *)0) {
			fprintf(fp, "%6.4f%9.2f", r0[n], an[0][n]);
			for (i = 1; i <= order; ++i)
				fprintf(fp, "%8.2f%8.2f", an[i][n], bn[i][n]);
			fprintf(fp, "\n");
		}
	}

	/* close listing file */
	if (fp != (FILE *)0)
		fclose(fp);

	/* create a smooth image */
	if (image1 != (char *)0) {

		printf("\nStart calculating the synthetic image\n");

		/* calculate a smooth annular image in array a[][] */
		if (smooth(method, nrows, ncols, a, xc, yc, theta, q,
				rmin, rmax, order, npts, r0, an, bn) == 0) {
			fprintf(stderr, "profile: Error while calculating image\n");
			goto Finish;
		}

		if (cpi(image, image1) < 1) {	/* dublicate image header */
			fprintf(stderr,
				"profile: %s.hdr -> %s.hdr copy failed\n", image, image1);
			goto Finish;
		} else if (setimagefile(image1, "") == 0) {	/* reset FILE key */
			fprintf(stderr,
				"profile: Can't erase old FILE value for %s\n", image1);
			rmi(image1);	/* remove image header */
			goto Finish;
		}

		/* create output image */
		if ((id = icreat(image1, 1, type, ncols, nrows)) < 0) {
			fprintf(stderr, "profile: Can't create %s\n", image1);
			rmi(image1);	/* remove image header */
			goto Finish;
		}

		if (update() == 0)	/* update disk headers */
			fprintf(stderr, "profile: update: Disk header not created\n");

		Blank = blank(image1);

		printf("Write output image to disk\n");

		for (y = 0; y < nrows; y++) {	/* write image to disk */
			switch (type) {
				case 1:
					{
						UCHAR *q = (UCHAR *)row;
						const float *p = a[y];
						int x;
						for (x = 0; x < ncols; ++x)
							q[x] = (UCHAR)(p[x] + Half);
					}
					break;
				case 2:
					{
						short *q = (short *)row;
						const float *p = a[y];
						const short b = (short)Blank;
						int x;
						for (x = 0; x < ncols; ++x)
							q[x] = (p[x] > FltMax)? b:
									(short)((p[x]<Zero)? p[x]-Half: p[x]+Half);
					}
					break;
				case 3:
					{
						float *q = (float *)row;
						const float *p = a[y];
						const float b = (float)Blank;
						int x;
						for (x = 0; x < ncols; ++x)
							q[x] = (p[x] > FltMax)? b: p[x];
					}
					break;
				case 4:
					{
						LONG *q = (LONG *)row;
						const float *p = a[y];
						const LONG b = (LONG)Blank;
						int x;
						for (x = 0; x < ncols; ++x)
							q[x] = (p[x] > FltMax)? b:
									(LONG)((p[x]<Zero)? p[x]-Half: p[x]+Half);
					}
					break;
				case 5:
					{
						COMPLEX *q = (COMPLEX *)row;
						const float *p = a[y];
						const float b = (float)Blank;
						int x, ip;
						for (x = 0; x < ncols; ++x) {
							ip = p[x] > FltMax;
							q[x].x = (ip)? b: p[x];
							q[x].y = (ip)? b: Zero;
						}
					}
					break;
			}
			putline(id, y, row);	/* write row to disk */
		}

		iclose(id);	/* close output image */

		setiname(image1);
		sethistory(image1, "profile");
		sethistory(image1, image1);
	}

Finish:

	/* free harmonic profiles */
	if (order > 0) free_dmatrix(bn, 1, 1);
	free_dmatrix(an, 0, 1);
	free_dvector(r0, 1);

	/* free row[] */
	free(row);
	/* free pixel array */
	free(ptr);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "profile: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
