
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int width = 1;	/* width of rings in pixels */
	int i, y;
	int xcen, ycen, radmax, promax, xmax, ymax;
	int xc, yc, stx, sty, extx, exty;
	int id, type, npix, nlin;
	double Blank;
	double *rad, *pro, umax, vmax;
	double a2 = 0.0;
	double b2 = 0.0;
	long *npt;
	void *buffer;
	char *image, *ptr;
	float **a;
	unsigned int err;
	int start[2], dim[2];
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	float Init_Value = FltBlank;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "addstar: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "w:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'w':	sscanf(optarg, "%d", &width); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (width <= 0) width = 1;

	if (opt_err || n_arg < 4) {
		printf("Usage: prophot -w width image Xc Yc RadMax [a2] [b2]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "prophot: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	sscanf(arg[1], "%d", &xcen);
	sscanf(arg[2], "%d", &ycen);
	sscanf(arg[3], "%d", &radmax);	/* max radius in pixels */
	if (n_arg > 4) sscanf(arg[4], "%lf", &a2);
	if (n_arg > 5) sscanf(arg[5], "%lf", &b2);

	promax = radmax/width;
	radmax = promax*width;
	if (promax <= 0) {
		fprintf(stderr, "prophot: promax <= 0\n");
		exit(EXIT_FAILURE);
	}

	if ((npt = (long *)calloc(promax, sizeof(long))) == (long *)0) {
		fprintf(stderr, "prophot: allocation of npt[] failed\n");
		exit(EXIT_FAILURE);
	}
	if ((rad = (double *)calloc(promax, sizeof(double))) == (double *)0) {
		fprintf(stderr, "prophot: allocation of rad[] failed\n");
		exit(EXIT_FAILURE);
	}
	if ((pro = (double *)calloc(promax, sizeof(double))) == (double *)0) {
		fprintf(stderr, "prophot: allocation of pro[] failed\n");
		exit(EXIT_FAILURE);
	}

	umax = sqrt((1.0 - a2)/(1.0 - a2*a2 - b2*b2));
	vmax = sqrt((1.0 + a2)/(1.0 - a2*a2 - b2*b2));

	xmax = (int) (radmax*umax + 1);
	ymax = (int) (radmax*vmax + 1);

	stx = xcen - xmax;
	sty = ycen - ymax;
	if (stx < 0 || sty < 0) {
		fprintf(stderr, "prophot: radius extends outside image\n");
		exit(EXIT_FAILURE);
	}
	xc = xcen - stx;
	yc = ycen - sty;
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "prophot: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "prophot: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "prophot: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	extx = 2*xmax + 1;
	exty = 2*ymax + 1;
	if (stx + extx > npix || sty + exty > nlin) {
		fprintf(stderr, "prophot: radius extends outside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "prophot: window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(extx, bytepix(type))) == (void *)0) {
		fprintf(stderr, "prophot: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	dim[0] = exty;	/* y dimension of object window */
	dim[1] = extx; /* x dimension of object window */
	start[0] = 0;	/* first index of y dimension */
	start[1] = 0;	/* first index of x dimension */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "prophot: daa: error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < exty; ++y) {	/* read object window into array */
		getline(id, y, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = 0; x < extx; ++x)
						a[y][x] = (float)q[x];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < extx; ++x)
						a[y][x] = (q[x].x>h || q[x].y>h)? FltBlank: q[x].x;
				}
				break;
		}
	}
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "prophot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (i = 0; i < promax; ++i) {
		pro[i] = 0.0;
		rad[i] = 0.0;
		npt[i] = 0;
	}
	for (y = 0; y < exty; ++y) {
		double v = y - yc;
		double v2 = v*v;
		double bv = 2.0*b2*v;
		int x;
		for (x = 0; x < extx; ++x) {
			double u = x - xc;
			double u2 = u*u;
			double r = sqrt(u2 + v2 + a2*(u2 - v2) + bv*u);
			int k = (int) (r/width);
			if (k < promax)
				if (a[y][x] <= FltMax) {
					rad[k] += r;
					pro[k] += a[y][x];
					npt[k]++;
				}
		}
	}
	for (i = 0; i < promax; ++i)
		if (npt[i] > 0) {
			rad[i] /= npt[i];
			pro[i] /= npt[i];
		}

	for (i = 0; i < promax; ++i)
		printf("%8.2f%16.8e%8d\n", rad[i], pro[i], npt[i]);

	free(ptr);
	free(buffer);
	free(pro);
	free(rad);
	free(npt);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
