
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* for ANSI compatibility */
#include <string.h>
#if	defined(SYSV) || defined(BSD)	/* Linux */
#include <unistd.h>
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "fbuf.h"	/* constants and prototypes for fbuf.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	NPIX	564
#define	NLIN	512
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image, *buffer;
	int id, block, i, j, k, pag_buf, bsize = buffsize();
	int npix = NPIX;
	int nlin = NLIN;
	int type = 2;	/* short */
	int rw = 1;	/* write only */
	int frame = 1;	/* default frame buffer number */
	int page = 0;	/* default window page number */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf("Usage: rdframe image [npix] [nlin] [frame] [page]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

#if   defined(SYSV) || defined(BSD)	/* Linux */
   if (iopl(3) != 0) {
		fprintf(stderr, "rdframe: Can't get I/O permission\n");
		exit(EXIT_FAILURE);
	}
#endif

	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "rdframe: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	if (argc > 2) sscanf(argv[2], "%d", &npix);	/* x-dimension */
	if (argc > 3) sscanf(argv[3], "%d", &nlin);	/* y-dimension */
	if (argc > 4) {
		sscanf(argv[4], "%d", &frame);	/* frame number */
		if (frame < 1 || frame > 2) {
			fprintf(stderr, "rdframe: Frame number = %d\n", frame);
			exit(EXIT_FAILURE);
		}
	}

	if (argc > 5) {
		sscanf(argv[5], "%d", &page);	/* page number */
		if (page < 0 || page >= RAM_PAG) {
			fprintf(stderr, "rdframe: Page number = %d\n", page);
			exit(EXIT_FAILURE);
		}
	}

	if ((buffer = (char *)malloc(bsize)) == (char *)0) {
		fprintf(stderr, "rdframe: Can't allocate output buffer\n");
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "rdframe: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id = icreat(image, rw, type, npix, nlin)) < 0) {
		fprintf(stderr, "rdframe: Can't create %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rdframe: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "rdframe: update: Disk header not created\n");

	setdateobs(image);	/* set DATE-OBS */
	setiname(image);	/* set IMAGE */

	pag_buf = (int)(PAG_SIZ/bsize);	/* number of blocks per page */
	map_fram(pag_buf);	/* map frame buffer RAM */

	block = 0;	/* image block number */
	for (k = page; k < RAM_PAG; k++) {	/* step through pages */
		set_fwin(frame, k);	/* select frame and page */
		for (j = 0; j < pag_buf; j++) {	/* step through blocks */
			switch (frame) {
				case 1:
					for (i = 0; i < bsize; i++)
						buffer[i] = fb1_win[j][i];
					break;
				case 2:
					for (i = 0; i < bsize; i++)
						buffer[i] = fb2_win[j][i];
					break;
			}
			if (bwrite(id, block, buffer) == 0)	/* write block */
				goto End;	/* exit after last block */
			block++;
		}
	}
	End:

	close_map();
	iclose(id);

	sethistory(image, "rdframe");
	sethistory(image, image);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "rdframe: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(buffer);

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
