
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "medi.h"	/* function declarations for medi.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int mx = 3;	/* x-dimension of running box */
	int my = 3;	/* y-dimension of running box */
	double limit = 0.0;	/* rejection limit is limit*MAD */
	int i, y;
	int id, id1, type, ncols, nrows, rw = 0, rw1 = 1;
	double Blank;
	char *ptr, *image, *image1;
	void **arr, *buf;
	unsigned int err;
	int start[2], dim[2];

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "runmed: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "x:y:l:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'x':	sscanf(optarg, "%d", &mx); mx /= 2; break;
			case 'y':	sscanf(optarg, "%d", &my); my /= 2; break;
			case 'l':	sscanf(optarg, "%lf", &limit); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 2) {
		printf("Usage: runmed -x mx -y my -l limit image image1\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* input image */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "runmed: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[1]);	/* output image */
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "runmed: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "runmed: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open input image */
	if ((id = iopen(image, rw, &type, &ncols, &nrows)) < 0) {
		fprintf(stderr, "runmed: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 4) {
		fprintf(stderr, "runmed: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(image1)) {	/* does output image exist? */
		fprintf(stderr, "runmed: %s: Existing image\n", image1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image, image1) < 1) {	/* dublicate image header */
		fprintf(stderr,
			"runmed: %s.hdr -> %s.hdr copy failed\n", image, image1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image1, "") == 0) {	/* reset FILE key */
		fprintf(stderr,
			"runmed: Can't erase old FILE value for %s\n", image1);
		rmi(image1);	/* remove image header */
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* create output image */
	if ((id1 = icreat(image1, rw1, type, ncols, nrows)) < 0) {
		fprintf(stderr, "runmed: Can't create %s\n", image1);
		rmi(image1);	/* remove image header */
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "runmed: update: Disk header not created\n");

	dim[0] = nrows;	/* number of rows */
	dim[1] = ncols;	/* number of columns */
	start[0] = 0;	/* first row index */
	start[1] = 0;	/* first column index */

	/* allocate image array arr[][] */
	if ((arr = (void **)daa(bytepix(type), 2, dim, start,
			&err, &ptr, NULL)) == (void **)0) {
		fprintf(stderr, "runmed: daa: Error code = %d\n", err);
		iclose(id1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nrows; ++y)	/* read image into arr[][] */
		getline(id, y, arr[y]);

	iclose(id);	/* close input image */

	/* allocate row buffer */
	if ((buf = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "runmed: Can't allocate row buffer\n");
		free(ptr);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	for (y = 0; y < nrows; ++y) {
		printf("Row No. %d\r", y);
		fflush(stdout);

		switch (type) {	/* process next row */
			case 1:
				i = rmed_uchar(buf, y, (const UCHAR **)arr,
						nrows, ncols, my, mx, limit);
				break;
			case 2:
				i = rmed_short(buf, y, (const short **)arr,
						nrows, ncols, my, mx, limit, Blank);
				break;
			case 3:
				i = rmed_float(buf, y, (const float **)arr,
						nrows, ncols, my, mx, limit, Blank);
				break;
			case 4:
				i = rmed_long(buf, y, (const LONG **)arr,
						nrows, ncols, my, mx, limit, Blank);
				break;
		}
		/* check if running box was too large */
		if (i != 0) {
			fprintf(stderr, "runmed: Can't process row No. %d\n", y);
			free(buf);
			free(ptr);
			iclose(id1);
			if (idel(image1) < 0)	/* delete image header and file */
				rmi(image1);	/* remove image header */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"runmed: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		/* write the processed row back */
		putline(id1, y, buf);
	}

	free(buf);
	free(ptr);
	iclose(id1);

	setiname(image1);	/* set IMAGE = Image Name */
	sethistory(image1, "runmed");
	sethistory(image1, image1);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}

