
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*----------------------------------------------------------------------*/
typedef struct {
	int x;
	int y;
} PEAK;
/*------------------------------------------------------------------------
	C  Implementation of modified Clean Algorithm
	William C. Keel, PASP 103, 723-729.
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static PEAK Peak(float array[], int nx, int ny)
#else	/* old style */
static PEAK Peak(array, nx, ny)
float array[];
int nx, ny;
#endif
{
	PEAK pk;
	int i, ipk = 0;
	float amax = (float)0;
	for (i = 0; i < nx*ny; ++i)
		if (array[i] > amax) {
			amax = array[i];
			ipk = i;
		} else if (array[i] < -amax) {
			amax = -array[i];
			ipk = i;
		}
	pk.x = ipk % nx;
	pk.y = ipk / nx;
	return pk;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Clean(int loops,
float res[], int nx, int ny,
float psf[], int mx, int my,
double gain, float cln[])
#else	/* old style */
static void Clean(loops, res, nx, ny, psf, mx, my, gain, cln)
int loops, nx, ny, mx, my;
float res[], psf[], cln[];
double gain;
#endif
{
	PEAK res_pk, psf_pk;
	int ydif, xdif;
	int i, j, x, y, y0, ym, x0, xm, nitr;
	float psf_sum, psf_max, res_max, ampl, flux, abs_flux;
	float flux_min = (float)1.0e8;
	const float Zero = (float)0;

	psf_pk = Peak(psf, mx, my);
	psf_max = psf[psf_pk.y * mx + psf_pk.x];

	psf_sum = Zero;
	for (i = 0; i < mx*my; ++i)
		psf_sum += psf[i];

	for (nitr = 0; nitr < loops; ++nitr) {

		res_pk = Peak(res, nx, ny);
		res_max = res[res_pk.y * nx + res_pk.x];

		ampl = (float)gain*res_max/psf_max;
		flux = ampl*psf_sum;
		cln[res_pk.y * nx + res_pk.x] += flux;
		abs_flux = (flux < Zero)? -flux: flux;
		if (abs_flux > flux_min+flux_min+flux_min) break;
		if (abs_flux < flux_min) flux_min = abs_flux;

		ydif = res_pk.y - psf_pk.y;
		y0 = MAX(-ydif, 0);
		ym = MIN(ny - ydif, my);
		xdif = res_pk.x - psf_pk.x;
		x0 = MAX(-xdif, 0);
		xm = MIN(nx - xdif, mx);
		for (y = y0; y < ym; ++y) {
			i = (ydif + y)*nx + xdif + x0;
			j = y*mx + x0;
			for (x = x0; x < xm; ++i, ++j, ++x) 
				res[i] -= ampl*psf[j];
		}
	}
}
/*--------------------------------------------------------------------*/
#define	LOOPS	1	/* number of iterations */
#define	GAIN	1.0
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int loops = LOOPS;	/* number of CLEAN iterations */
	double gain = GAIN;	/* loop gain */
	char *res_image, *psf_image, *cln_image;
	float *res, *psf, *cln;
	void *buffer;
	int i, y, nx, ny, mx, my, res_row, psf_row, buf_siz;
	int id_res, id_cln, id_psf, res_type, psf_type;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "imake: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "l:g:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'l':	sscanf(optarg, "%d", &loops); break;
			case 'g':	sscanf(optarg, "%lf", &gain); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 3) {
		printf("Usage: sclean -l loops -g gain resim psfim clnim\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	res_image = truncname(arg[0]);	/* residual image */
	if (strpbrk(res_image, punct) != (char *)0) {
		fprintf(stderr, "sclean: Illegal image name: %s\n", res_image);
		exit(EXIT_FAILURE);
	}

	psf_image = truncname(arg[1]);	/* PSF image */
	if (strpbrk(psf_image, punct) != (char *)0) {
		fprintf(stderr, "sclean: Illegal image name: %s\n", psf_image);
		exit(EXIT_FAILURE);
	}

	cln_image = truncname(arg[2]);	/* clean image */
	if (strpbrk(cln_image, punct) != (char *)0) {
		fprintf(stderr, "sclean: Illegal image name: %s\n", cln_image);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "sclean: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id_res = iopen(res_image, 0, &res_type, &nx, &ny)) < 0) {
		fprintf(stderr, "sclean: Can't open %s\n", res_image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (res_type < 1 || res_type > 5) {
		fprintf(stderr,
			"sclean: %s has illegal type = %d\n", res_image, res_type);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((id_psf = iopen(psf_image, 0, &psf_type, &mx, &my)) < 0) {
		fprintf(stderr, "sclean: Can't open %s\n", psf_image);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (psf_type < 1 || psf_type > 5) {
		fprintf(stderr,
			"sclean: %s has illegal type = %d\n", psf_image, psf_type);
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(cln_image)) {	/* does image exist? */
		fprintf(stderr, "sclean: %s: Existing image\n", cln_image);
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	res_row = nx * bytepix(res_type);
	psf_row = mx * bytepix(psf_type);
	buf_siz = MAX(res_row, psf_row);
	if ((buffer = (void *)malloc(buf_siz)) == (void *)0) {
		fprintf(stderr, "sclean: Allocation of line buffer failed\n");
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((res = (float *)calloc(ny, nx * sizeof(float))) == (float *)0) {
		fprintf(stderr, "sclean: Allocation of res array failed\n");
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((cln = (float *)calloc(ny, nx * sizeof(float))) == (float *)0) {
		fprintf(stderr, "sclean: Allocation of cln array failed\n");
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((psf = (float *)calloc(my, mx * sizeof(float))) == (float *)0) {
		fprintf(stderr, "sclean: Allocation of psf array failed\n");
		iclose(id_psf);
		iclose(id_res);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < ny; ++y) {
		getline(id_res, y, buffer);	/* fetch new line */
		switch (res_type) {
			case 1:	/* unsigned char */
				{
					float *p = &res[y*nx];
					const UCHAR *q = (UCHAR *) buffer;
					int x;
					for (x = 0; x < nx; ++x)
						p[x] = (float) q[x];
				}
				break;
			case 2:	/* short int */
				{
					float *p = &res[y*nx];
					const short *q = (short *) buffer;
					int x;
					for (x = 0; x < nx; ++x)
							p[x] = (float) q[x];
				}
				break;
			case 3:	/* float */
				{
					float *p = &res[y*nx];
					const float *q = (float *) buffer;
					int x;
					for (x = 0; x < nx; ++x)
							p[x] = q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					float *p = &res[y*nx];
					const LONG *q = (LONG *) buffer;
					int x;
					for (x = 0; x < nx; ++x)
							p[x] = (float) q[x];
				}
				break;
			case 5:	/* complex */
				{
					float *p = &res[y*nx];
					const COMPLEX *q = (COMPLEX *) buffer;
					int x;
					for (x = 0; x < nx; ++x)
						p[x] = q[x].x;
				}
				break;
		}
	}

	iclose(id_res);

	for (y = 0; y < my; ++y) {
		getline(id_psf, y, buffer);	/* fetch new line */
		switch (psf_type) {
			case 1:	/* unsigned char */
				{
					float *p = &psf[y*mx];
					const UCHAR *q = (UCHAR *) buffer;
					int x;
					for (x = 0; x < mx; ++x)
						p[x] = (float) q[x];
				}
				break;
			case 2:	/* short int */
				{
					float *p = &psf[y*mx];
					const short *q = (short *) buffer;
					int x;
					for (x = 0; x < mx; ++x)
							p[x] = (float) q[x];
				}
				break;
			case 3:	/* float */
				{
					float *p = &psf[y*mx];
					const float *q = (float *) buffer;
					int x;
					for (x = 0; x < mx; ++x)
							p[x] = q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					float *p = &psf[y*mx];
					const LONG *q = (LONG *) buffer;
					int x;
					for (x = 0; x < mx; ++x)
							p[x] = (float) q[x];
				}
				break;
			case 5:	/* complex */
				{
					float *p = &psf[y*mx];
					const COMPLEX *q = (COMPLEX *) buffer;
					int x;
					for (x = 0; x < mx; ++x)
						p[x] = q[x].x;
				}
				break;
		}
	}

	iclose(id_psf);

	for (i = 0; i < nx*ny; ++i)
		cln[i] = Zero;

	Clean(loops, res, nx, ny, psf, mx, my, gain, cln);

	if (cpi(res_image, cln_image) < 1) {	/* dublicate header */
		fprintf(stderr,
			"sclean: %s.hdr -> %s.hdr copy failed\n", res_image, cln_image);
		goto Finish;
	} else if (setimagefile(cln_image, "") == 0) {	/* reset FILE key */
		fprintf(stderr,
			"sclean: Can't erase old FILE value for %s\n", cln_image);
		rmi(cln_image);	/* remove image header */
		goto Finish;
	}

	if ((id_cln = icreat(cln_image, 1, 3, nx, ny)) < 0) {
		fprintf(stderr, "sclean: Can't create %s\n", cln_image);
		rmi(cln_image);	/* remove image header */
		goto Finish;
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "sclean: update: Disk header not created\n");

	for (y = 0; y < ny; ++y)
		putline(id_cln, y, &cln[y*nx]);
	iclose(id_cln);
Finish:

	free(psf);
	free(cln);
	free(res);
	free(buffer);

	sethistory(cln_image, "sclean");

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "sclean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
