
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* For ANSI compatibility, errno declared */
#include <string.h>	/* required only for function declarations */
#if	defined(MSC) || defined(HC2) || defined(WC)
/* MS-C, High-C V2.3, or WATCOM-C 9.0/386 */
#include <malloc.h>	/* _fmalloc and _ffree */
#endif
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
#include <graph.h>	/* Standard Video Modes defined in graph.h */
#endif
#include <conio.h>	/* Keyboard I/O */
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "imhdr.h"	/* function prototypes for imhdr.c */
#include "imkw.h"	/* function prototypes for imkw.c */
#include "imio.h"	/* function prototypes for imio.c */
#include "vga256.h"	/* function prototypes for vga256.c */
#include "hcconio.h"	/* function prototypes for hcconio.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define LF	10
#define CR	13
/*--------------------------------------------------------------------*/
#define CURSIZ	5	/* cursor size */
#define NAMSIZ	40	/* full file name size */
/*--------------------------------------------------------------------*/
#define NHUE	360	/* number of hues */
static UCHAR red[NHUE], green[NHUE], blue[NHUE];
/*--------------------------------------------------------------------*/
static void ReadRGB(void)	/* Read R, G, and B tables */
{
	char **path, fname[NAMSIZ];
	int count;
	FILE *stream;
	if ((path = getpath("IMSYS", &count)) == (char **)0)
		strcpy(fname, ".");	/* current directory */
	else
		strcpy(fname, path[0]);	/* pathname from environment */
	strcat(fname, "/rgbtab.dat");
	if ((stream = fopen(fname, "rb")) == (FILE *)0) {
		fprintf(stderr, "can't open %s\n", fname);
		exit(EXIT_FAILURE);
	}
	fread(red, sizeof (char), NHUE, stream);
	fread(green, sizeof (char), NHUE, stream);
	fread(blue, sizeof (char), NHUE, stream);
	fclose(stream);
}
/*--------------------------------------------------------------------*/
static UCHAR rgb[768];	/* RGB Lookup Table */
/*--------------------------------------------------------------------*/
static void Lookup(int hue)	/* Calculate Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	while (hue < 0)
		hue += NHUE;
	while (hue >= NHUE)
		hue -= NHUE;
	for (i = 0; i < 256; i++) {
		j = (hue + i) % NHUE;
		*p++ = red[j];
		*p++ = green[j];
		*p++ = blue[j];
	}
}
/*--------------------------------------------------------------------*/
static void GrayLookup(void)	/* Calculate Gray Scale Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	for (i = 0; i < 256; i++) {
		j = i >> 2;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
	}
}
/*--------------------------------------------------------------------*/
static void DrawCursor(	/* Draw cursor at (x,y) */
int x, int y, UCHAR ucv)
{
	int i;
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	for (i = 1; i <= CURSIZ; ++i) {
		if ((x-i) >= 0)
			draw_pixel(y, x - i, (UCHAR) (read_pixel(y, x - i) + ucv));
		if ((x+i) < row_siz)
			draw_pixel(y, x + i, (UCHAR) (read_pixel(y, x + i) + ucv));
		if ((y-i) >= 0)
			draw_pixel(y - i, x, (UCHAR) (read_pixel(y - i, x) + ucv));
		if ((y+i) < col_siz)
			draw_pixel(y + i, x, (UCHAR) (read_pixel(y + i, x) + ucv));
	}
}
/*--------------------------------------------------------------------*/
static UCHAR DisplayValue(	/* Reads display value at (x,y) */
int x, int y)
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	return read_pixel(y, x);
}
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
static short far **Array;	/* pointer to array of far pointers */
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2) || defined(WC)
/* High-C or WATCOM-C 9.0/386 */
static short **Array;	/* pointer to array of pointers */
#endif
/*--------------------------------------------------------------------*/
static int PixelValue(	/* Reads pixel value at (x,y) */
int x, int y,
int ncols,	/* number of columns in Array[][] */
int nrows,	/* number of rows in Array[][] */
int col0,	/* first column displayed */
int row0,	/* first row displayed */
int rotate)	/* has image been rotated? */
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	if (col0 < 0) col0 = 0;
	if (row0 < 0) row0 = 0;
	if (col0 >= ncols) col0 = ncols - 1;
	if (row0 >= nrows) row0 = nrows - 1;
	if (rotate) {
		x += row0;
		y += col0;
		if (x >= nrows || y >= ncols)
			return 0;
		else
			return Array[x][y];
	} else {
		x += col0;
		y += row0;
		if (x >= ncols || y >= nrows)
			return 0;
		else
			return Array[y][x];
	}
}
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
static int RowAlloc(	/* allocate array of pointers */
short far *array[],	/* table of far pointers to image rows */
const int nrow,	/* number of rows in window */
const int ncol)	/* number of pixels per row */
{
	int i;
	for (i = 0; i < nrow; i++)
		if ((array[i] = (short far *)_fmalloc(ncol*sizeof(short)))
				== (short far *)0) {
			if (--i >= 0) {
				_ffree(array[i]);
				array[i] = (short far *)0;
			}
			break;
		}
	return i;	/* actual number of rows allocated */
}
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2) || defined(WC)
/* High-C or WATCOM-C 9.0/386 */
static int RowAlloc(	/* allocate array of pointers */
short *array[],	/* table of pointers to image rows */
const int nrow,	/* number of rows in window */
const int ncol)	/* number of pixels per row */
{
	int i;
	for (i = 0; i < nrow; i++)
		if ((array[i] = (short *)malloc(ncol*sizeof(short)))
				== (short *)0) {
			if (--i >= 0) {
				free(array[i]);
				array[i] = (short *)0;
			}
			break;
		}
	return i;	/* actual number of rows allocated */
}
#endif
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
static void RowFree(	/* free array of row pointers */
short far *array[],	/* table of far pointers to image rows */
const int nrow)	/* number of rows in window */
{
	int i;
	for (i = 0; i < nrow; i++)
		if (array[i] == (short far *)0)
			break;
	for (--i; i >= 0; i--)
		_ffree(array[i]);
}
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2) || defined(WC)
/* High-C or WATCOM-C 9.0/386 */
static void RowFree(	/* free array of row pointers */
short *array[],	/* table of pointers to image rows */
const int nrow)	/* number of rows in window */
{
	int i;
	for (i = 0; i < nrow; i++)
		if (array[i] == (short *)0)
			break;
	for (--i; i >= 0; i--)
		free(array[i]);
}
#endif
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
static void RowCopy(
short far *drow,	/* far pointer to short row */
short *srow,	/* pointer to short row */
const int ncol,	/* number of pixels per row */
const short blank)	/* pixels with this value are set to zero */
{
	short far *dp = drow;
	short *p = srow;
	short *e = p + ncol;
	for (; p < e; p++)
		if (*p != blank)
			*dp++ = *p;
		else
			*dp++ = 0;
}
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2) || defined(WC)
/* High-C or WATCOM-C 9.0/386 */
static void RowCopy(
short *drow,	/* pointer to short row */
short *srow,	/* pointer to short row */
const int ncol,	/* number of pixels per row */
const short blank)	/* pixels with this value are set to zero */
{
	short *dp = drow;
	short *p = srow;
	short *e = p + ncol;
	for (; p < e; p++)
		if (*p != blank)
			*dp++ = *p;
		else
			*dp++ = (short)0;
}
#endif
/*--------------------------------------------------------------------*/
char _String[50] = {'\0'};	/* global string containing output text */
/*--------------------------------------------------------------------*/
static void CursorControl(int hue,	/* Cursor control */
const int scr_line,	/* string written on this line */
const int ncols,	/* number of columns in Array[][] */
const int nrows,	/* number of rows in Array[][] */
const int col0,	/* first column displayed */
const int row0,	/* first row displayed */
const int rotate)	/* has image been rotated? */
{
	int xcen, ycen;
	int row_siz = rowsize();
	int col_siz = colsize();
	int c = 'm';	/* move cursor */
	UCHAR astep = 1;
	UCHAR bstep = 10;
	UCHAR color = 128;
	UCHAR ucv = color;
	if (rotate) {
		xcen = MIN(row_siz, nrows - row0)/2;
		ycen = MIN(col_siz, ncols - col0)/2;
	} else {
		xcen = MIN(row_siz, ncols - col0)/2;
		ycen = MIN(col_siz, nrows - row0)/2;
	}
	DrawCursor(xcen, ycen, ucv);
	while (c != 'e') {
		switch(c) {
			case 'c':	/* change cursor color */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						color += 8;
					if (c == 'P' || c == 'K')
						color -= 8;
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					ucv = color;
					sprintf(_String,
						"cursor color: %6d                    ", color);
					write_str(_String, scr_line, 0, 255);
					DrawCursor(xcen, ycen, ucv);
				}
			break;
			case 's':	/* change cursor step size */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						astep++;
					if (c == 'P' || c == 'K')
						astep--;
					if (c == 'Q' || c == 'G')
						bstep++;
					if (c == 'S' || c == 'O')
						bstep--;
					sprintf(_String,
						"steps: %6d %6d                    ", astep, bstep);
					write_str(_String, scr_line, 0, 255);
				}
			break;
			case 'h':	/* change starting hue for color lookup table */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M') {
						hue += 4;
						while (hue >= NHUE)
							hue -= NHUE;
					}
					if (c == 'P' || c == 'K') {
						hue -= 4;
						while (hue < 0)
							hue += NHUE;
					}
					sprintf(_String,
						"starting hue: %6d                    ", hue);
					write_str(_String, scr_line, 0, 255);
					Lookup(hue);
					set_color(rgb);
				}
			break;
			case 'g':	/* use gray scale lookup table */
				GrayLookup();
				set_color(rgb);
				c = 'm';	/* move cursor next time */
			break;
			case 'm':	/* move cursor */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					switch(c) {
						case 'H':
							ycen -= astep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'P':
							ycen += astep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'K':
							xcen -= astep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'M':
							xcen += astep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						case 'G':
							ycen -= bstep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'O':
							ycen += bstep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'S':
							xcen -= bstep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'Q':
							xcen += bstep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						default:
						break;
					}
					DrawCursor(xcen, ycen, ucv);
					sprintf(_String, "x=%4d y=%4d display=%4d pixel=%7d",
						xcen, ycen, DisplayValue(xcen, ycen),
						PixelValue(xcen, ycen, ncols, nrows, col0, row0, rotate));
					write_str(_String, scr_line, 0, 255);
				}
			break;
			default:
				while ((c = get_ch()) == 0)	/* get next non-arrow */
					c = get_ch();	/* skip second byte */
			break;
		}
	}
}
/*--------------------------------------------------------------------*/
static void Display(
int ncols,	/* number of columns in Array[][] */
int nrows,	/* number of rows in Array[][] */
int col0,	/* first column displayed */
int row0,	/* first row displayed */
int rotate,	/* should image be rotated? */
int bias,	/* value subtracted before a shift */
int shift)	/* bits shifted right */
{
	UCHAR Row[2048];	/* <------ REMEMBER THIS BUFFER */
	int row_siz = rowsize();
	int col_siz = colsize();
	if (col0 < 0) col0 = 0;
	if (row0 < 0) row0 = 0;
	if (col0 >= ncols) col0 = ncols - 1;
	if (row0 >= nrows) row0 = nrows - 1;
	if (rotate) {
		int npix = MIN(row_siz, nrows - row0);
		int nlin = MIN(col_siz, ncols - col0);
		int x, y, z;
		for (x = 0; x < nlin; ++x) {
			for (y = 0; y < npix; ++y) {
				z = Array[y+row0][x+col0] - bias;
				if (z < 0) z = 0;
				Row[y] = (UCHAR) (z >> shift);
			}
			draw_line(x, 0, Row, npix);
		}
	} else {
		int npix = MIN(row_siz, ncols - col0);
		int nlin = MIN(col_siz, nrows - row0);
		int x, y, z;
		for (y = 0; y < nlin; ++y) {
			for (x = 0; x < npix; ++x) {
				z = Array[y+row0][x+col0] - bias;
				if (z < 0) z = 0;
				Row[x] = (UCHAR) (z >> shift);
			}
			draw_line(y, 0, Row, npix);
		}
	}
}
/*--------------------------------------------------------------------*/
#define	STRSIZ	80
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int hue, y, i;
	int xmin, xmax, ymin, ymax;
	int stx, sty, extx, exty;
	int id, type, ncol, nrow, rw = 0;
	int scr_line, vga_mode, mode;
	short *row, Blank;
	char *image, str[STRSIZ];
	int bias;
	int col0 = 0;
	int row0 = 0;
	int rotate = 0;
	int shift = 3;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf(
			"Usage: show image [Xmin] [Xmax] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}
	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "show: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "show: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, rw, &type, &ncol, &nrow)) < 0) {
		fprintf(stderr, "show: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type != 2) {
		fprintf(stderr, "show: %s: Image type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = (short) blank(image);

	if (argc > 2) sscanf(argv[2], "%d", &xmin); else xmin = 0;
	if (argc > 3) sscanf(argv[3], "%d", &xmax); else xmax = ncol - 1;
	if (argc > 4) sscanf(argv[4], "%d", &ymin); else ymin = 0;
	if (argc > 5) sscanf(argv[5], "%d", &ymax); else ymax = nrow - 1;

	stx = xmin;
	sty = ymin;
	extx = xmax - xmin + 1;
	exty = ymax - ymin + 1;
	if (stx < 0) {
		extx += stx;
		stx = 0;
	}
	if (sty < 0) {
		exty += sty;
		sty = 0;
	}
	if (extx <= 0 || exty <= 0) {
		fprintf(stderr, "show: window is placed above/left of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx >= ncol || sty >= nrow) {
		fprintf(stderr, "show: window is placed below/right of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx + extx > ncol)
		extx = ncol - stx;
	if (sty + exty > nrow)
		exty = nrow - sty;
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "show: window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((row = (short *)malloc(extx * sizeof(short))) == (short *)0) {
		fprintf(stderr, "show: Allocation of row buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
#if	defined(MSC)	/* MS-C */
	if ((Array = (short far * *)malloc(
			exty * sizeof(short far *))) == (short far * *)0){
#endif
#if	defined(HC1) || defined(HC2) || defined(WC)
	/* High-C or WATCOM-C 9.0/386 */
	if ((Array = (short **)malloc(exty*sizeof(short *))) == (short **)0){
#endif
		fprintf(stderr, "show: Allocation of row table failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	exty = RowAlloc(Array, exty, extx);
	printf("%d rows allocated\n", exty);
	for (y = 0; y < exty; y++) {
		getline(id, y, row);
		RowCopy(Array[y], row, extx, Blank);
	}
	bias = 32767;
	for (y = 3; y < exty; y += 7) {
		int x;
		for (x = 3; x < extx; x += 7)
			if (bias > Array[y][x])	/* find minimum */
				bias = Array[y][x];
	}
	bias -= 50;
	ReadRGB();	/* Read R, G, and B tables */
	hue = 200;
	Lookup(hue);	/* Calculate Lookup Table */
#if	defined(ET4000)	/* Tseng Labs Super VGA */
	vga_mode = _ET1024x768;
	scr_line = 46;
	mode = '6';
#else		/* Paradise Super VGA */
	vga_mode = _PA640x480;
	scr_line = 29;
	mode = '3';
#endif
	while (1) {
		printf("Start display at column  (%d): ", col0);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &col0);
		printf("Start display at row     (%d): ", row0);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &row0);
		printf("Subtract offset value    (%d): ", bias);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &bias);
		printf("Divide by 2 to the power (%d): ", shift);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &shift);
		printf("Rotated image? (0|1)     (%d): ", rotate);
		if (gets(str) != NULL && strlen(str) > 0)
			sscanf(str, "%d", &rotate);
		set_mode(vga_mode);
		set_color(rgb);
		Display(extx, exty, col0, row0, rotate, bias, shift);
		CursorControl(hue, scr_line, extx, exty, col0, row0, rotate);
		set_mode(_TEXTC80);
		printf("%s\n", _String);
		printf("Repeat with other VGA mode?\n");
		printf("Standard    320x200: 1\n");
#if	defined(ET4000)	/* Tseng Labs Super VGA */
		printf("Tseng Labs  640x350: 2\n");
		printf("Tseng Labs  640x400: 3\n");
		printf("Tseng Labs  640x480: 4\n");
		printf("Tseng Labs  800x600: 5\n");
		printf("Tseng Labs 1024x768: 6\n");
#else		/* Paradise Super VGA */
		printf("Paradise    640x400: 2\n");
		printf("Paradise    640x480: 3\n");
#endif
		printf("Exit...............: 0\n");
		printf("Enter mode (%c): ", mode);
		fflush(stdout);
		i = getche();
		printf("\n");
		if (i != CR && i != LF)
			mode = i;
		if (mode < '1' || mode > '6')
			break;
		switch (mode) {
			case '1':
				vga_mode = _MRES256COLOR;
				scr_line = 23;
				break;
#if	defined(ET4000)	/* Tseng Labs Super VGA */
			case '2':
				vga_mode = _ET640x350;
				scr_line = 24;
				break;
			case '3':
				vga_mode = _ET640x400;
				scr_line = 24;
				break;
			case '4':
				vga_mode = _ET640x480;
				scr_line = 29;
				break;
			case '5':
				vga_mode = _ET800x600;
				scr_line = 36;
				break;
			case '6':
				vga_mode = _ET1024x768;
				scr_line = 46;
				break;
#else		/* Paradise Super VGA */
			case '2':
				vga_mode = _PA640x400;
				scr_line = 24;
				break;
			case '3':
				vga_mode = _PA640x480;
				scr_line = 29;
				break;
#endif
		}
	}
	RowFree(Array, exty);
	free(Array);
	free(row);
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "show: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
