
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#define EXTX	11
#define EXTY	11
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image;
	int x, y, stx, sty, xcen, ycen, imag;
	int type, npix, nlin, id, flag;
	double buffer[EXTX];
	int extx = EXTX, exty = EXTY;
	int rw = 0;	/* read only */
	int skipx = 0, skipy = 0;
	void *buf = (void *)buffer;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf("Usage: spot image [X] [Y] [Imag]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}
	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "spot: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	if (argc > 2) sscanf(argv[2], "%d", &xcen); else xcen = 0;
	if (argc > 3) sscanf(argv[3], "%d", &ycen); else ycen = 0;
	if (argc > 4) sscanf(argv[4], "%d", &imag); else imag = 0;

	stx = xcen - extx/2;	/* start-x */
	if (stx < 0) {
		extx += stx;
		skipx -= stx;
		stx = 0;
	}
	sty = ycen - exty/2;
	if (sty < 0) {
		exty += sty;
		skipy -= sty;
		sty = 0;
	}
	if (extx <= 0 || exty <= 0) {
		fprintf(stderr, "spot: window is placed above/left of image\n");
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((flag = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "spot: prolog: Error = %d\n", flag);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "spot: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "spot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx >= npix || sty >= nlin) {
		fprintf(stderr, "spot: window is placed below/right of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "spot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx + extx > npix)
		extx = npix - stx;
	if (sty + exty > nlin)
		exty = nlin - sty;
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "spot: window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "spot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "spot: Type = %d is not implemented\n", type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "spot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < skipy; y++)
		printf("\n");
	switch(type) {
		case 1:
			{
				unsigned char *pix = (unsigned char *)buf;
				for (y = 0; y < exty; y++) {
					for (x = 0; x < skipx; x++)
						printf("%7s", "");
					getline(id, y, pix);
					for (x = 0; x < extx; x++)
						printf("%7d", pix[x]);
					printf("\n");
				}
			}
			break;
		case 2:
			{
				short *pix = (short *)buf;
				for (y = 0; y < exty; y++) {
					for (x = 0; x < skipx; x++)
						printf("%7s", "");
					getline(id, y, pix);
					for (x = 0; x < extx; x++)
						printf("%7d", pix[x]);
					printf("\n");
				}
			}
			break;
		case 3:
			{
				float *pix = (float *)buf;
				for (y = 0; y < exty; y++) {
					for (x = 0; x < skipx; x++)
						printf("%7s", "");
					getline(id, y, pix);
					for (x = 0; x < extx; x++)
						printf("%7.1f", pix[x]);
					printf("\n");
				}
			}
			break;
		case 4:
			{
				LONG *pix = (LONG *)buf;
				for (y = 0; y < exty; y++) {
					for (x = 0; x < skipx; x++)
						printf("%7s", "");
					getline(id, y, pix);
					for (x = 0; x < extx; x++)
						printf("%7ld", pix[x]);
					printf("\n");
				}
			}
			break;
		case 5:
			{
				COMPLEX *pix = (COMPLEX *)buf;
				for (y = 0; y < exty; y++) {
					for (x = 0; x < skipx; x++)
						printf("%7s", "");
					getline(id, y, pix);
					for (x = 0; x < extx; x++)
						if (imag)
							printf("%7.1f", pix[x].y);
						else
							printf("%7.1f", pix[x].x);
					printf("\n");
				}
			}
			break;
	}
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "spot: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
