
/**************************************************************************/
/*                                                                        */
/*   This program is free software; you can redistribute it and/or modify */
/*   it under the terms of the GNU General Public License as published by */
/*   the Free Software Foundation; either version 2 of the License, or    */
/*   (at your option) any later version.                                  */
/*                                                                        */
/*   This program is distributed in the hope that it will be useful,      */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*   GNU General Public License for more details.                         */
/*                                                                        */
/*   You should have received a copy of the GNU General Public License    */
/*   along with this program; if not, write to the                        */
/*       Free Software Foundation, Inc.,                                  */
/*       59 Temple Place, Suite 330,                                      */
/*       Boston, MA  02111-1307                                           */
/*       USA                                                              */
/*                                                                        */
/**************************************************************************/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <ctype.h>	/* ctype macros */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.h */
#include "wild.h"	/* function prototype for wild.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	FSIZE	256	/* maximum field length */
#define	CSIZE	256	/* maximum comment length */
#define	PSIZE	128	/* maximum path length */
#define	FNSIZE	51	/* maximum file name size */
#define	RSIZE	2880	/* number of bytes in FITS record */
#define	MAXREC	128	/* maximum number of FITS records in array */
/*--------------------------------------------------------------------*/
typedef union {	/* LONG or float word */
	LONG  l;
	float f;
} LONGWORD;
/*--------------------------------------------------------------------*/
typedef struct {	/* keyword structure */
	short type;
	char *key;
} KEY;
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Field(
char *field,
const char *text)
#else	/* old style */
static char *Field(field, text)	/* extract value field from text */
char *field;
const char *text;
#endif
{
	const char *p = text;
	char *q = field;
	for (; isspace(*p); p++)
		;	/* skip leading spaces */
	while (*p != '\0' && *p != '\\')	/* copy until '\0' or '\\' */
		*q++ = *p++;
	for (--q; (q >= field) && isspace(*q); q--)
		;
	*++q = '\0';	/* truncate trailing spaces */
	return(field);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Comment(
char *comm,
const char *text)
#else	/* old style */
static char *Comment(comm, text)	/* extract comment field from text */
char *comm;
const char *text;
#endif
{
	const char *p = text;
	for (; (*p != '\0') && (*p != '\\'); p++)
		;	/* skip until '\\' */
	if (*p == '\\') {
		p++;	/* skip '\\' */
		for (; isspace(*p); p++)
			;	/* skip leading spaces */
	}
	*comm = '\0';
	if (strlen(p) > 0) {	/* comment string not empty */
		strcat(comm, "/ ");
		strcat(comm, p);
	}
	return(comm);
}
/*--------------------------------------------------------------------*/
static char ifile[PSIZE] = {(char)0};	/* image file path */
static char ipath[PSIZE] = {(char)0};	/* image directory path */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void SetPath(
const int i)
#else	/* old style */
static void SetPath(i)	/* assign image path from environment */
const int i;	/* select i'th path name */
#endif
{
	char **p;
	int n;
	if((p = getpath("IMGPATH", &n)) == (char **)0)
		strcpy(ipath, ".");	/* copy current directory path */
	else if(i < n)
		strcpy(ipath, p[i]);	/* copy i'th environment path */
	else
		strcpy(ipath, ".");	/* copy current directory path */
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void AssignImageFile(
const char *file)
#else	/* old style */
static void AssignImageFile(file)	/* cat image file to path name */
const char *file;	/* image file */
#endif
{
	strcpy(ifile, ipath);	/* image directory path */
	if(strlen(ifile) > 0)
		strcat(ifile, "/");	/* append "/" */
	strcat(ifile, file);	/* append image file */
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void ReOrder(void *buffer,
const int nrec,
const int type,
const double scale,
const double zero,
const double Blank)
#else	/* old style */
static void ReOrder(buffer, nrec, type, scale, zero, Blank)
void *buffer;	/* DATA buffer */
const int nrec;	/* number of FITS records */
const int type;	/* pixel type */
const double scale;	/* BSCALE value */
const double zero;	/* BZERO value */
const double Blank;	/* BLANK value */
#endif
{
	int scaling = (scale == 0.0 || scale == 1.0)? 0: 1;
	switch (type) {
		case 2:
			{	/* no scaling of short data */
				short w, m = 0xFF;
				short b = (short) Blank;
				short z = (short) ((zero < 0.0)? zero-0.5: zero+0.5);
				short *p = (short *) buffer;
				short *e = p + 1440 * nrec;
				char  *q = (char *) buffer;
				for (; p < e; p++) {
					w = *p;
					w = (w != b)? w - z: w;
					*q++ = (char) ((w >> 8) & m);	/* high order byte */
					*q++ = (char) (w & m);	/* low order byte */
				}
			}
			break;
		case 3:
			if (scaling) {	/* scaling of float to LONG data */
				LONG w, m = 0xFF;
				float h = (float) (0.9*FLT_BLANK);
				float *p = (float *) buffer;
				float *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					if (*p <= h) {
						double d = *p;
						d -= zero;
						d /= scale;
						w = (LONG) ((d < 0.0)? d - 0.5: d + 0.5);
					} else w = LONG_Min;
					*q++ = (char) ((w >> 24) & m);	/* most significant byte */
					*q++ = (char) ((w >> 16) & m);
					*q++ = (char) ((w >> 8) & m);
					*q++ = (char) (w & m);	/* least significant byte */
				}
			} else {	/* no scaling of float data */
				LONGWORD w;
				LONG m = 0xFF;
				float *p = (float *) buffer;
				float *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w.f = *p;
					*q++ = (char) ((w.l >> 24) & m);	/* most significant byte */
					*q++ = (char) ((w.l >> 16) & m);
					*q++ = (char) ((w.l >> 8) & m);
					*q++ = (char) (w.l & m);	/* least significant byte */
				}
			}
			break;
		case 4:
			{	/* no scaling of LONG data */
				LONG w, m = 0xFF;
				LONG b = (LONG) Blank;
				LONG z = (LONG) ((zero < 0.0)? zero-0.5: zero+0.5);
				LONG *p = (LONG *) buffer;
				LONG *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w = *p;
					w = (w != b)? w - z: w;
					*q++ = (char) ((w >> 24) & m);	/* most significant byte */
					*q++ = (char) ((w >> 16) & m);
					*q++ = (char) ((w >> 8) & m);
					*q++ = (char) (w & m);	/* least significant byte */
				}
			}
			break;
		default:
			break;
	}
}
/*--------------------------------------------------------------------*/
#define	NKEYS	191	/* number of FITS keywords, including "HISTORY" */
#define	LOGICAL	1	/* logical variable */
#define	INTEGER	2	/* integer variable */
#define	REAL	3	/* real variable */
#define	CHARACTER	4	/* character variable, normally 8 characters */
#define	HISTORY	5	/* history field */
/*--------------------------------------------------------------------*/
static KEY  words[] = {	/* table containing FITS keywords */
	LOGICAL  , "SIMPLE",    /* +Standard FITS format */
	INTEGER  , "BITPIX",    /* +Number of bits storing pixel values */
	INTEGER  , "NAXIS",     /* +Number of axes in frame; 0 if no data follow */
	INTEGER  , "NAXIS1",    /* +Number of pixels/axis; 0 if no data follow */
	INTEGER  , "NAXIS2",    /* +Number of pixels/axis; 0 if no data follow */
	LOGICAL  , "EXTEND",    /* +Standard extensions are possible */
	INTEGER  , "BLANK",     /* +Specifies that the pixel value is undefined */
	REAL     , "BSCALE",    /* +PhysicalValue=ArrayValue*BSCALE+BZERO */
	REAL     , "BZERO",     /* +PhysicalValue=ArrayValue*BSCALE+BZERO */
	REAL     , "DATAMIN",   /* +Minimum valid physical value represented */
	REAL     , "DATAMAX",   /* +Maximum valid physical value represented */
	INTEGER  , "OPSIZE",    /* &PSIZE of original image */
	CHARACTER, "FITSDATE",  /* &Date FITS file was created */
	CHARACTER, "FILENAME",  /* &Original GEIS header file name with _cvt */
	CHARACTER, "ODATTYPE",  /* &Original datatype */
	INTEGER  , "SDASMGNU",  /* &GCOUNT of original image */
	CHARACTER, "DADSFILE",
	CHARACTER, "DADSCLAS",
	CHARACTER, "DADSDATE",
	REAL     , "CD1_1",
	REAL     , "CD1_2",
	REAL     , "CD2_1",
	REAL     , "CD2_2",
	LOGICAL  , "MIR_REVR",
	REAL     , "ORIENTAT",
	INTEGER  , "FILLCNT",
	INTEGER  , "ERRCNT",
	REAL     , "FPKTTIME",
	REAL     , "LPKTTIME",
	INTEGER  , "DETECTOR",
	REAL     , "DEZERO",
	REAL     , "BIASEVEN",
	REAL     , "BIASODD",
	REAL     , "GOODMIN",
	REAL     , "GOODMAX",
	REAL     , "DATAMEAN",
	INTEGER  , "GPIXELS",
	INTEGER  , "SOFTERRS",
	INTEGER  , "CALIBDEF",
	INTEGER  , "STATICD",
	INTEGER  , "ATODSAT",
	INTEGER  , "DATALOST",
	INTEGER  , "BADPIXEL",
	INTEGER  , "OVERLAP",
	CHARACTER, "PHOTMODE",
	REAL     , "PHOTFLAM",
	REAL     , "PHOTZPT",
	REAL     , "PHOTPLAM",
	REAL     , "PHOTBW",
	REAL     , "MEDIAN",
	REAL     , "MEDSHADO",
	REAL     , "HISTWIDE",
	REAL     , "SKEWNESS",
	REAL     , "MEANC10",
	REAL     , "MEANC25",
	REAL     , "MEANC50",
	REAL     , "MEANC100",
	REAL     , "MEANC200",
	REAL     , "MEANC300",
	REAL     , "BACKGRND",
	CHARACTER, "ROOTNAME",  /* &Rootname of the observation set */
	CHARACTER, "FILETYPE",  /* &shp, ext, edq, sdq, sci */
	CHARACTER, "MODE",      /* &instr. mode: FULL (full res.), AREA (area int.) */
	CHARACTER, "SERIALS",   /* &serial clocks: ON, OFF */
	CHARACTER, "IMAGETYP",  /* &DARK/BIAS/IFLAT/UFLAT/VFLAT/KSPOT/EXT/ECAL */
	CHARACTER, "CDBSFILE",  /* &GENERIC/BIAS/DARK/FLAT/MASK/NO */
	INTEGER  , "PKTFMT",    /* &packet format code */
	CHARACTER, "FILTNAM1",  /* &first filter name */
	CHARACTER, "FILTNAM2",  /* &second filter name */
	INTEGER  , "FILTER1",   /* &first filter number (0-48) */
	INTEGER  , "FILTER2",   /* &second filter number (0-48) */
	INTEGER  , "FILTROT",   /* &partial filter rotation angle (degrees) */
	INTEGER  , "LRFWAVE",   /* &linear ramp filter wavelength */
	REAL     , "UCH1CJTM",  /* &TEC cold junction #1 temperature (Celsius) */
	REAL     , "UCH2CJTM",  /* &TEC cold junction #2 temperature (Celsius) */
	REAL     , "UCH3CJTM",  /* &TEC cold junction #3 temperature (Celsius) */
	REAL     , "UCH4CJTM",  /* &TEC cold junction #4 temperature (Celsius) */
	REAL     , "UBAY3TMP",  /* &bay 3 A1 temperature (deg C) */
	CHARACTER, "KSPOTS",    /* &Status of Kelsall spot lamps: ON, OFF */
	CHARACTER, "SHUTTER",   /* &Shutter in place at the beginning of the exp */
	REAL     , "ATODGAIN",  /* &Analog to Digital Gain (Electrons/DN) */
	CHARACTER, "MASKCORR",  /* &Do mask correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "ATODCORR",  /* &Do A-to-D correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "BLEVCORR",  /* &Do bias level correction: PERFORM, OMIT, COMPLE */
	CHARACTER, "BIASCORR",  /* &Do bias correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "DARKCORR",  /* &Do dark correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "FLATCORR",  /* &Do flat field correction: PERFORM, OMIT, COMPLE */
	CHARACTER, "SHADCORR",  /* &Do shaded shutter correction: PERFORM, OMIT, CO */
	CHARACTER, "DOSATMAP",  /* &Output saturated pixel map: PERFORM, OMIT, COMP */
	CHARACTER, "DOPHOTOM",  /* &Fill photometry keywords: PERFORM, OMIT, COMPLE */
	CHARACTER, "DOHISTOS",  /* &Make histograms: PERFORM, OMIT, COMPLETE */
	CHARACTER, "OUTDTYPE",  /* &Output image datatype: REAL, LONG, SHORT */
	CHARACTER, "MASKFILE",  /* &name of the input DQF of known bad pixels */
	CHARACTER, "ATODFILE",  /* &name of the A-to-D conversion file */
	CHARACTER, "BLEVFILE",  /* &Engineering file with extended register data */
	CHARACTER, "BLEVDFIL",  /* &Engineering file DQF */
	CHARACTER, "BIASFILE",  /* &name of the bias frame reference file */
	CHARACTER, "BIASDFIL",  /* &name of the bias frame reference DQF */
	CHARACTER, "DARKFILE",  /* &name of the dark reference file */
	CHARACTER, "DARKDFIL",  /* &name of the dark reference DQF */
	CHARACTER, "FLATFILE",  /* &name of the flat field reference file */
	CHARACTER, "FLATDFIL",  /* &name of the flat field reference DQF */
	CHARACTER, "SHADFILE",  /* &name of the reference file for shutter shading */
	CHARACTER, "PHOTTAB",   /* &name of the photometry calibration table */
	CHARACTER, "GRAPHTAB",  /* &the HST graph table */
	CHARACTER, "COMPTAB",   /* &the HST components table */
 	INTEGER  , "SATURATE",  /* &Data value at which saturation occurs */
	REAL     , "USCALE",    /* &Scale factor for output image */
	REAL     , "UZERO",     /* &Zero point for output image */
	INTEGER  , "READTIME",  /* &Length of time for CCD readout in clock ticks */
	REAL     , "PA_V3",     /* &position angle of v3 axis of HST */
	REAL     , "RA_SUN",    /* &right ascension of the sun (deg) */
	REAL     , "DEC_SUN",   /* &declination of the sun (deg) */
	REAL     , "EQNX_SUN",  /* &equinox of the sun */
	LOGICAL  , "MTFLAG",    /* &moving target flag */
	REAL     , "EQRADTRG",  /* &equatorial radius of target */
	REAL     , "FLATNTRG",  /* &flattening of target */
	REAL     , "NPDECTRG",  /* &north pole declination of target */
	REAL     , "NPRATRG",   /* &north pole right ascension of target */
	REAL     , "ROTRTTRG",  /* &rotation rate of target */
	REAL     , "LONGPMER",  /* &longitude of prime meridian */
	REAL     , "EPLONGPM",  /* &epoch of longitude of prime meridian */
	REAL     , "SURFLATD",  /* &surface feature latitude */
	REAL     , "SURFLONG",  /* &surface feature longitude */
	REAL     , "SURFALTD",  /* &surface feature altitude */
	INTEGER  , "PODPSFF",   /* &0=(no podps fill), 1=(podps fill present) */
	INTEGER  , "STDCFFF",   /* &0=(no st dcf fill), 1=(st dcf fill present) */
	CHARACTER, "STDCFFP",   /* &st dcf fill pattern (hex) */
	INTEGER  , "RSDPFILL",  /* &bad data fill value for calibrated images */
	INTEGER  , "UEXPODUR",  /* &Commanded duration of exposure (seconds) */
	INTEGER  , "NSHUTA17",  /* &Number of AP17 shutter B closes */
	INTEGER  , "UEXPOTIM",  /* &Major frame pulse time preceding exposure start */
	CHARACTER, "PSTRTIME",  /* &Predicted obs. start time  (yyyy:ddd:hh:mm:ss) */
	CHARACTER, "PSTPTIME",  /* &Predicted obs. stop  time  (yyyy:ddd:hh:mm:ss) */
	REAL     , "SUNANGLE",  /* &angle between sun and V1 axis (deg) */
	REAL     , "MOONANGL",  /* &angle between moon and V1 axis (deg) */
	REAL     , "SUN_ALT",   /* &altitude of the sun above Earth's limb (deg) */
	CHARACTER, "FGSLOCK",   /* &commanded FGS lock (FINE,COARSE,GYROS,UNKNOWN) */
	CHARACTER, "TARGNAME",  /* &proposer's target name */
	REAL     , "RA_TARG",   /* &right ascension of the target (deg) (J2000) */
	REAL     , "DEC_TARG",  /* &declination of the target (deg) (J2000) */
	INTEGER  , "PROPOSID",  /* &PEP proposal identifier */
	CHARACTER, "PEP_EXPO",  /* &PEP exposure identifier including sequence */
	CHARACTER, "LINENUM",   /* &PEP proposal line number */
	CHARACTER, "SEQLINE",   /* &PEP line number of defined sequence */
	CHARACTER, "SEQNAME",   /* &PEP define/use sequence name */
	CHARACTER, "BUNIT",     /* +Unit after application of BSCALE and BZERO */
	REAL     , "CRVAL1",    /* +CTYPE1 coordinate at the reference point CRPIX1 */
	REAL     , "CRPIX1",    /* +Location of a reference point along axis-1 */
	REAL     , "CDELT1",    /* +Derivative of CTYPE1 wrt the pixel index at CRPIX1 */
	REAL     , "CROTA1",    /* +Rotation angle between axis-1 and direction of CTYPE1 */
	CHARACTER, "CTYPE1",    /* +Name of coordinate represented by axis-1 */
	REAL     , "CRVAL2",    /* +CTYPE2 coordinate at the reference point CRPIX2 */
	REAL     , "CRPIX2",    /* +Location of a reference point along axis-2 */
	REAL     , "CDELT2",    /* +Derivative of CTYPE2 wrt the pixel index at CRPIX2 */
	REAL     , "CROTA2",    /* +Rotation angle between axis-2 and direction of CTYPE2 */
	CHARACTER, "CTYPE2",    /* +Name of coordinate represented by axis-2 */
	REAL     , "EQUINOX",   /* +Equinox in years for the celestial coordinate system */
	REAL     , "EPOCH",     /* +Deprecated: Use EQUINOX instead */
	REAL     , "JD-START",
	REAL     , "JD-STOP",
	REAL     , "MJD-OBS",   /* ?MJD start */
	REAL     , "TM-START",
	REAL     , "EXPSTART",  /* &Exposure start time (Modified Julian Date) */
	REAL     , "EXPEND",    /* &Exposure end time (Modified Julian Date) */
	REAL     , "EXPTIME",   /* &Exposure duration (seconds)--calculated */
	CHARACTER, "EXPFLAG",   /* &Exposure interruption indicator */
	REAL     , "DARKTIME",  /* &Dark time (seconds) */
	REAL     , "GAIN",
	REAL     , "RDNOISE",
	REAL     , "RA",
	REAL     , "DEC",
	REAL     , "AIRMASS",
	CHARACTER, "OBJECT",    /* +Name of the object observed */
	CHARACTER, "FILTER",
	CHARACTER, "GRISM",
	CHARACTER, "TIMESYS",   /* +Principal time system; default is UTC */
	CHARACTER, "DATE",      /* +YYYY-MM-DDThh:mm:ss UTC creation date of HDU */
	CHARACTER, "DATE-OBS",  /* +YYYY-MM-DDThh:mm:ss UTC at start of observation */
	CHARACTER, "TIME-OBS",  /* &UT time of start of observation (hh:mm:ss) */
	CHARACTER, "RADECSYS",  /* &Coordinate reference frame */
	CHARACTER, "UTC",       /* ?UTC at start (sec) */
	CHARACTER, "LST",       /* ?LST at start (sec) */
	CHARACTER, "ORIGIN",    /* +Organization creating the FITS file */
	CHARACTER, "INSTRUME",  /* +Instrument used to acquire the data */
	CHARACTER, "DISPAXIS",
	CHARACTER, "TELESCOP",  /* +Telescope used to acquire the data */
	CHARACTER, "OBSERVER",  /* +Who acquired the data? */
	CHARACTER, "COMMENT",
	CHARACTER, "IMAGE",
	HISTORY  , "HISTORY"
};
/*--------------------------------------------------------------------*/
static char card[81];	/* card image plus terminating '\0' */
static char field[FSIZE];	/* value field plus terminating '\0' */
static char comm[CSIZE];	/* comment field plus terminating '\0' */
static char *tmp = "itemp";
static char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
static char ext[5] = {'.', 'f', 'i', 't', '\0'};
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;

	int option, opt_err = 0, n_arg = 0;
	double scale = 1.0;	/* BSCALE value */
	double zero = 0.0;	/* BZERO value */
	char **arg;

	char **itab, **imgtab, *charbuf, (*record)[80];
	char *image, *keyw, *text, *fitsfile, name[FNSIZE];
	char *fitspath = getenv("FITSPATH");
	void *rec = (void *)0, *ptr = (void *)0;
	FILE *istream, *ostream;
	int nrec, nhead, type, vtype, i, j, k, l, n, ntab;
	unsigned nread, nwrite, asize, mrec, mr, maxr, u, fitsfilelength;
	double Blank;

	if (fitspath == NULL || strlen(fitspath) == 0)
		fitsfilelength = FNSIZE;
	else
		fitsfilelength = strlen(fitspath) + 1 + FNSIZE;
	if ((fitsfile = (char *)calloc(fitsfilelength, sizeof(char))) == NULL) {
		fprintf(stderr, "wrfits: calloc: Can't allocate file name\n");
		exit(EXIT_FAILURE);
	}

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "wrfits: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "s:z:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 's':	sscanf(optarg, "%lf", &scale); break;	/* BSCALE */
			case 'z':	sscanf(optarg, "%lf", &zero); break;	/* BZERO */
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 1) {
		printf(
		"Usage: wrfits -s scale -z zero {image1|@file} [image2] [image3]...\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "wrfits: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	if (n_arg > 1) {	/* write FITS images, if 2 or more are specified */
		if ((itab = (char **)calloc(n_arg, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "wrfits: calloc: Allocation failed\n");
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "wrfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		for (n = 0, i = 0; i < n_arg; ++i) {
			image = truncname(arg[i]);
			if (strpbrk(image, punct) != (char *)0)
				printf("%s is not a legal image name\n", image);
			else if (xist(image))	/* check if image exists */
				itab[n++] = image;
			else
				printf("%s does not exist\n", image);
		}
	} else {	/* wild card possible, if only first image is given */
		imgtab = getitab(&ntab);	/* table containing all image names */
		itab = (char **)0;	/* initialize image table pointer */
		if (ntab > 0)
			if ((itab = (char **)calloc(ntab,
						sizeof(char *))) == (char **)0) {
				fprintf(stderr, "wrfits: calloc: Allocation failed\n");
				if (epilog() < 1)	/* update disk headers */
					fprintf(stderr,
						"wrfits: epilog: Disk header(s) not updated\n");
				exit(EXIT_FAILURE);
			}
		image = arg[0];
		for (n = 0, i = 0; i < ntab; ++i)	/* find name in image table */
			if (wildcmp(imgtab[i], image) == 0)
				itab[n++] = imgtab[i];
		if (n == 0)
			printf("%s does not exist\n", image);
	}

	for (ntab = 0, i = 0; i < n; ++i)
		if (simple(itab[i]))
			itab[ntab++] = itab[i];
		else
			printf("%s is not a simple image\n", itab[i]);

	maxr = UINT_Max/RSIZE;	/* maximum possible number of records */
	if (maxr > MAXREC)	/* but not more than MAXREC records */
		maxr = MAXREC;
	rec = (void *)0;	/* buffer array has not been allocated */
	asize = 0;	/* array size in bytes */
	for (mrec = maxr; mrec > 0; mrec--) {	/* find big array */
		asize = mrec*RSIZE;	/* try this size */
		if ((rec = (void *)malloc(asize)) != (void *)0)
			break;	/* big array allocated */
	}

	if (mrec <= 0) {	/* array was not allocated */
		fprintf(stderr, "wrfits: malloc: Buffer allocation failed\n");
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "wrfits: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (mrec < maxr) {	/* no space for maximum size */
		mr = mrec;
		mr++;
		mr >>= 1;
		asize = mr*RSIZE;	/* array size reduced by 50% */
		if ((ptr = (void *)realloc(rec, asize)) == (void *)0)
			asize = mrec*RSIZE;	/* reallocation failed */
		else {	/* successful reallocation */
			mrec = mr;	/* number of records reallocated */
			rec = ptr;
		}
	}

	record = (char (*)[80])rec;
	charbuf = (char *)rec;
	for (n = 0; n < ntab; ++n) {	/* begin converting images */
		if (xist(tmp))
			if (idel(tmp) < 0)	/* delete image tmp, if existing */
				rmi(tmp);	/* remove tmp header, if not deleted */
		image = itab[n];	/* next image name */
		cpi(image, tmp);	/* make working copy of image node */
		type = pixtype(tmp);	/* get pixel type */
		Blank = blank(tmp);	/* get BLANK value */
		if (!(scale == 0.0 || scale == 1.0))
			setbscale(tmp, scale);	/* set BSCALE */
		if (zero != 0.0)
			setbzero(tmp, zero);	/* set BZERO */
		if (type == 3)	/* 32-bit float */
			if (!(scale == 0.0 || scale == 1.0)) {
				setpixtype(tmp, 4);	/* convert float to LONG data */
				setbitpix(tmp, 32);	/* float is converted to LONG */
				setblank(tmp, (double) LONG_Min);	/* set BLANK value */
			}
		SetPath(0);	/* in order to get direct access to image file */
		AssignImageFile(imagefile(image));	/* assign private image file */
		if ((istream = fopen(ifile, O_RO)) == (FILE *)0) {
			printf("can't open image file %s for %s\n", ifile, image);
			goto Next_Image;
#if	BUF_SIZ	/* set internal buffer size */
		} else if (setvbuf(istream, NULL, _IOFBF, buffsize())) {
			printf("%s: setvbuf() failed for image file %s\n",
			image, ifile);
			fclose(istream);	/* close image file */
			goto Next_Image;
#endif
		}
		fitsfile[0] = '\0';
		if (fitspath != NULL && strlen(fitspath) > 0) {
			strcat(fitsfile, fitspath);
			strcat(fitsfile, "/");
		}
		strcat(fitsfile, image);
		strcat(fitsfile, ext);	/* append extension to image name */
		if ((ostream = fopen(fitsfile, C_WO)) == (FILE *)0) {
			printf("can't create FITS file %s\n", fitsfile);
			fclose(istream);	/* close image file */
			goto Next_Image;
#if	BUF_SIZ	/* set internal buffer size */
		} else if (setvbuf(ostream, NULL, _IOFBF, buffsize())) {
			printf("setvbuf() failed for %s\n", fitsfile);
			fclose(ostream);	/* close FITS file */
			fclose(istream);	/* close image file */
			goto Next_Image;
#endif
		}
		for (k = 0, i = 0; i < NKEYS; i++) {
			vtype = words[i].type;	/* value type */
			keyw = words[i].key;	/* keyword */
			text = getk(tmp, keyw);	/* get keyword text */
			if (text != (char *)0) {	/* check if keyword is found */
				Field(field, text);	/* extract value field */
				Comment(comm, text);	/* extract comment field */
				switch (vtype) {
					case LOGICAL:	/* logical variable */
					case INTEGER:	/* integer variable */
					case REAL:	/* real variable */
						sprintf(card, "%-8.8s= %20.20s%1c%-49.49s",
							keyw, field, ' ', comm);
						break;
					case CHARACTER:	/* character variable, normally 8 chars */
						if ((l = strlen(field)) > 8) {	/* not fixed format */
							l += 2;
							strcat(field, "'");	/* terminate string */
							for (j = 20; j > l; j--)
								strcat(field, " ");  /* fill with spaces */
							strcat(field, " ");	/* append a space */
							strcat(field, comm);	/* append comment field */
							sprintf(card, "%-8.8s= '%-69.69s",
								keyw, field);
						} else	/* fixed format string */
							sprintf(card, "%-8.8s= '%-8.8s'%11c%-49.49s",
								keyw, field, ' ', comm);
						break;
					case HISTORY:	/* history field */
						sprintf(card, "%-8.8s%-72.72s", keyw, field);
						break;
				}
				for (j = 0; j < 80; j++)
					record[k][j] = card[j];
				k++;
			}
		}
		if (getk(image, "IMAGE") == (char *)0) {
			strcpy(name, image);	/* make copy of image name */
			for (j = 0; name[j] != 0; j++)	/* convert to lowercase */
				name[j] = tolower(name[j]);
			sprintf(card, "%-8.8s= '%-8.8s'%62c", "IMAGE", name, ' ');
			for (j = 0; j < 80; j++)
				record[k][j] = card[j];
			k++;
		}
		sprintf(card, "%-8.8s%72c", "END", ' ');	/* terminate header */
		for (j = 0; j < 80; j++)
			record[k][j] = card[j];
		nhead = k/36 + 1;	/* total number of header records */
		k++;
		sprintf(card, "%80c", ' ');	/* fill with spaces */
		for (; k < 36*nhead; k++)
			for (j = 0; j < 80; j++)
				record[k][j] = card[j];
		if (fwrite(rec, RSIZE, nhead, ostream) != nhead) {	/* write header */
			printf("Error writing FITS header to %s\n", fitsfile);
			fclose(istream);	/* close image file */
			fclose(ostream);	/* close FITS file */
			goto Next_Image;
		}
		while ((nread = fread(rec, 1, asize, istream)) > 0) {	/* DATA */
			nrec = (nread < asize) ? (nread - 1)/RSIZE + 1 : mrec;
			nwrite = RSIZE * nrec;	/* bytes to be written */
			for (u = nread; u < nwrite; u++)	/* remainder padded with zeros */
				charbuf[u] = 0;
			ReOrder(rec, nrec, type, scale, zero, Blank);	/* reorder bytes */
			if (fwrite(rec, 1, nwrite, ostream) != nwrite) {	/* write DATA */
				printf("Error writing DATA buffer to %s\n", fitsfile);
				fclose(istream);	/* close image file */
				fclose(ostream);	/* close FITS file */
				goto Next_Image;
			}
		}
		if (ferror(istream) != 0) {	/* error, when reading image file */
			printf("Error reading image file %s for %s\n", ifile, image);
			fclose(istream);	/* close image file */
			fclose(ostream);	/* close FITS file */
			goto Next_Image;
		}
		fclose(istream);	/* close image file */
		fclose(ostream);	/* close FITS file */
		Next_Image: rmi(tmp);	/* remove temporary image node */
	}

	free(rec);	/* remember to deallocate big array */
	free(itab);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "wrfits: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
